/**
 * SECTION:st-applet-calendar
 * @short_description: 日期的显示
 *
 * 按照用户点击需求显示具体的年月日：
 *      当选项为月份时，点击显示该月份的天数；
 *      当选项为年时，显示该年的12个月份；
 *      当选项为年的区间时，显示该区间的年份。
 *
 */

#include <glib.h>
#include <glib/gi18n.h>
#include "st.h"

#include "applet-calendar.h"
#include "applet-calendar-item.h"

#define CALENDAR_PARAM_COMMON (G_PARAM_STATIC_NAME|G_PARAM_STATIC_NICK|G_PARAM_STATIC_BLURB)
#define CALENDAR_PARAM_READWRITE (G_PARAM_READWRITE|CALENDAR_PARAM_COMMON)
#define CALENDAR_PARAM_READABLE (G_PARAM_READABLE|CALENDAR_PARAM_COMMON)
#define CALENDAR_PARAM_WRITABLE (G_PARAM_WRITABLE|CALENDAR_PARAM_COMMON)

#define SEC_IN_A_MIN (60)
#define SEC_IN_AN_HOUR (60 * SEC_IN_A_MIN)
#define SEC_IN_A_DAY (24 * SEC_IN_AN_HOUR)

#define ACTOR_DEFAULT_WIDTH (168)
#define ACTOR_DEFAULT_HEIGHT (198)
#define TOP_BOX_DEFAULT_HEIGHT (30)
#define TOP_BOX_DEFAULT_WIDTH ACTOR_DEFAULT_WIDTH
#define MAIN_BOX_DEFAULT_WIDTH ACTOR_DEFAULT_WIDTH
#define MAIN_BOX_DEFAULT_HEIGHT (ACTOR_DEFAULT_HEIGHT - TOP_BOX_DEFAULT_HEIGHT)

#define ANIMATION_DURATION_TIME (600)
#define ANIMATION_DELAY_TIME (200)
#define ANIMATION_MODE_OUT (CLUTTER_EASE_OUT_SINE)
#define ANIMATION_MODE_IN (CLUTTER_EASE_IN_SINE)

struct _AppletCalendarPrivate {

    StWidget *topBox;
    StWidget *mainBox;
    StWidget *swipeBox[3];

    StWidget *viewTypeSwitch;

    gint viewType;

    guint selectedYear;
    guint selectedMonth;
    guint selectedDayOfMonth;

    gint day_month[6][7];
    gint day[6][7];

    gint focus_unit;

    guint year_before : 1;

    gint week_start;
    gboolean show_week_num;
};

static const guint month_length[2][13] = {
    {0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31},
    {0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
};

static const guint days_in_months[2][14] = {
    {0, 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365},
    {0, 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366}
};

static char* default_abbreviated_dayname[7] = {0};
static char* default_monthname[12] = {0};

static gboolean leap(guint year) {
    return ((((year % 4) == 0) && ((year % 100) != 0)) || ((year % 400) == 0));
}

static guint day_of_week(guint year, guint mm, guint dd) {
    GDateTime * dt;
    guint days;

    dt = g_date_time_new_local(year, mm, dd, 1, 1, 1);
    if (NULL == dt) {
        return 0;
    }

    days = g_date_time_get_day_of_week(dt);
    g_date_time_unref(dt);

    return days;
}

static guint week_of_year(guint year, guint mm, guint dd) {
    GDateTime * dt;
    guint week;

    dt = g_date_time_new_local(year, mm, dd, 1, 1, 1);
    if (NULL == dt) {
        return 0;
    }

    week = g_date_time_get_week_of_year(dt);
    g_date_time_unref(dt);

    return week;
}

static gboolean isWorkDay(guint dayofweek) {
    guint days = dayofweek % 7;
    return (0 != days && 6 != days);
}

enum {
    VIEWTYPE_DAY = 0,
    VIEWTYPE_MONTH,
    VIEWTYPE_YEAR
};

enum {
    MONTH_PREV,
    MONTH_CURRENT,
    MONTH_NEXT
};

enum {
    MONTH_CHANGED_SIGNAL,
    DAY_SELECTED_SIGNAL,
    DAY_SELECTED_DOUBLE_CLICK_SIGNAL,
    PREV_MONTH_SIGNAL,
    NEXT_MONTH_SIGNAL,
    LAST_SIGNAL
};

enum {
    PROP_0,
    PROP_YEAR,
    PROP_MONTH,
    PROP_DAY,
    PROP_SHOW_WEEK_NUMBERS,
};

G_DEFINE_TYPE_WITH_PRIVATE(AppletCalendar, applet_calendar, G_TYPE_OBJECT)

static guint applet_calendar_signals[LAST_SIGNAL] = {0};

static void applet_calendar_finalize(GObject * calendar);
//static void applet_calendar_set_property(GObject * obj,
//                                       guint prop_id,
//                                       const GValue * value);

//static void applet_calendar_get_property(GObject * obj,
//                                       guint prop_id,
//                                       GValue * value,
//                                       GParamSpec * pspec);

static void buildHeader(AppletCalendar *calendar);

static void calendar_compute_days(AppletCalendar* calendar);

static void setDate(AppletCalendar* calendar, GDateTime * date, gboolean isForceLoad, GTimeSpan changed);

static void onPreButtonClicked(StButton *button, int clickedBtn, gpointer user_data);
static void onNextButtonClicked(StButton *button, int clickedBtn, gpointer user_data);
static void onDateLabelClicked(StButton *button, int clickedBtn, gpointer user_data);
static void onCalendarItemClicked(StButton * button, int clickedBtn, gpointer user_data);

static void updateDayView(AppletCalendar* calendar, gboolean isForceLoad, GTimeSpan changed);
static void updateMonthView(AppletCalendar* calendar, gboolean isForceLoad, GTimeSpan changed);
static void updateYearView(AppletCalendar* calendar, gboolean isForceLoad, GTimeSpan changed);

static void setSensitive(StWidget* actor, gboolean sensitive) {
    g_return_if_fail(NULL != actor && ST_IS_WIDGET(actor));

    clutter_actor_set_reactive(CLUTTER_ACTOR(actor), sensitive);
    st_widget_set_can_focus(ST_WIDGET(actor), sensitive);
    if (sensitive) {
        st_widget_remove_style_pseudo_class(ST_WIDGET(actor), "insensitive");
    } else {
        st_widget_remove_style_pseudo_class(ST_WIDGET(actor), "hover");
        st_widget_remove_style_pseudo_class(ST_WIDGET(actor), "active");
        st_widget_add_style_pseudo_class(ST_WIDGET(actor), "insensitive");
    }
}

static void applet_calendar_class_init(AppletCalendarClass *klass) {
    //g_print("%s --> enter!\n", __func__);
    GObjectClass * objClass = G_OBJECT_CLASS(klass);

//    objClass->get_property = applet_calendar_get_property;
//    objClass->set_property = applet_calendar_set_property;
    objClass->finalize = applet_calendar_finalize;

    /**
     * CdosCalendar:year:
     *
     * The selected year.
     * This property gets initially set to the current year.
     */
//    g_object_class_install_property(objClass,
//                                    PROP_YEAR,
//                                    g_param_spec_int("year",
//                                                     "Year",
//                                                     "The selected year",
//                                                     0, G_MAXINT >> 9, 0,
//                                                     CALENDAR_PARAM_READWRITE));

    /**
     * CdosCalendar:month:
     *
     * The selected month (as a number between 1 and 12).
     * This property gets initially set to the current month.
     */
//    g_object_class_install_property(objClass,
//                                    PROP_MONTH,
//                                    g_param_spec_int("month",
//                                                     "Month",
//                                                     "The selected month (as a number between 1 and 12)",
//                                                     1, 12, 0,
//                                                     CALENDAR_PARAM_READWRITE));

    /**
     * CdosCalendar:day:
     *
     * The selected day (as a number between 1 and 31, or 0
     * to unselect the currently selected day).
     * This property gets initially set to the current day.
     */
//    g_object_class_install_property(objClass,
//                                    PROP_DAY,
//                                    g_param_spec_int("day",
//                                                     "Day",
//                                                     "The selected day (as a number between 1 and 31, or 0 to unselect the currently selected day)",
//                                                     0, 31, 0,
//                                                     CALENDAR_PARAM_READWRITE));

    /**
     * CdosCalendar:show-week-numbers:
     *
     * Determines whether week numbers are displayed.
     *
     */
//    g_object_class_install_property(objClass,
//                                    PROP_SHOW_WEEK_NUMBERS,
//                                    g_param_spec_boolean("show-week-numbers",
//                                                         "Show Week Numbers",
//                                                         "If TRUE, week numbers are displayed",
//                                                         FALSE,
//                                                         CALENDAR_PARAM_READWRITE));

    /**
     * CdosCalendar:month-changed:
     * @calendar: the object which received the signal.
     *
     * Emitted when the user clicks a button to change the selected month on a
     * calendar.
     */
    applet_calendar_signals[MONTH_CHANGED_SIGNAL] =
            g_signal_new(I_("month-changed"),
                         G_OBJECT_CLASS_TYPE(objClass),
                         G_SIGNAL_RUN_FIRST,
                         G_STRUCT_OFFSET(AppletCalendarClass, month_changed),
                         NULL, NULL,
                         g_cclosure_marshal_VOID__VOID,
                         G_TYPE_NONE, 0);

    /**
     * CdosCalendar:day-selected:
     * @calendar: the object which received the signal.
     *
     * Emitted when the user selects a day.
     */
    applet_calendar_signals[DAY_SELECTED_SIGNAL] =
            g_signal_new(I_("day-selected"),
                         G_OBJECT_CLASS_TYPE(objClass),
                         G_SIGNAL_RUN_FIRST,
                         G_STRUCT_OFFSET(AppletCalendarClass, day_selected),
                         NULL, NULL,
                         g_cclosure_marshal_VOID__VOID,
                         G_TYPE_NONE, 0);

    /**
     * CdosCalendar:day-selected-double-click:
     * @calendar: the object which received the signal.
     *
     * Emitted when the user double-clicks a day.
     */
    applet_calendar_signals[DAY_SELECTED_DOUBLE_CLICK_SIGNAL] =
            g_signal_new(I_("day-selected-double-click"),
                         G_OBJECT_CLASS_TYPE(objClass),
                         G_SIGNAL_RUN_FIRST,
                         G_STRUCT_OFFSET(AppletCalendarClass, day_selected_double_click),
                         NULL, NULL,
                         g_cclosure_marshal_VOID__VOID,
                         G_TYPE_NONE, 0);

    /**
     * CdosCalendar:prev-month:
     * @calendar: the object which received the signal.
     *
     * Emitted when the user switched to the previous month.
     */
    applet_calendar_signals[PREV_MONTH_SIGNAL] =
            g_signal_new(I_("prev-month"),
                         G_OBJECT_CLASS_TYPE(objClass),
                         G_SIGNAL_RUN_FIRST,
                         G_STRUCT_OFFSET(AppletCalendarClass, prev_month),
                         NULL, NULL,
                         g_cclosure_marshal_VOID__VOID,
                         G_TYPE_NONE, 0);

    /**
     * CdosCalendar:next-month:
     * @calendar: the object which received the signal.
     *
     * Emitted when the user switched to the next month.
     */
    applet_calendar_signals[NEXT_MONTH_SIGNAL] =
            g_signal_new(I_("next-month"),
                         G_OBJECT_CLASS_TYPE(objClass),
                         G_SIGNAL_RUN_FIRST,
                         G_STRUCT_OFFSET(AppletCalendarClass, next_month),
                         NULL, NULL,
                         g_cclosure_marshal_VOID__VOID,
                         G_TYPE_NONE, 0);
    //g_print("%s --> end!\n", __func__);
}

static void applet_calendar_init(AppletCalendar *self) {
    //g_print("%s --> enter!\n", __func__);
    AppletCalendarPrivate * priv;
    time_t tmp_time;
    time_t secs;
    struct tm * tm;
    char buffer[255];
    int i = 0;

    priv = self->priv = applet_calendar_get_instance_private(self);

    if (!default_abbreviated_dayname[0]) {
        for (i = 0; i < 7; ++i) {
            tmp_time = (i + 3) * SEC_IN_A_DAY;
            strftime(buffer, sizeof(buffer), "%a", localtime(&tmp_time));
            default_abbreviated_dayname[i] = g_locale_to_utf8(buffer, -1, NULL, NULL, NULL);
        }
    }
    if (!default_monthname[0]) {
        for (i = 0; i < 12; ++i) {
            tmp_time = i * 32 * SEC_IN_A_DAY;
            strftime(buffer, sizeof(buffer), "%B", localtime(&tmp_time));
            default_monthname[i] = g_locale_to_utf8(buffer, -1, NULL, NULL, NULL);
        }
    }

    secs = time(NULL);
    tm = localtime(&secs);

    priv->selectedYear = tm->tm_year + 1900;
    priv->selectedMonth = tm->tm_mon + 1;
    priv->selectedDayOfMonth = tm->tm_mday;

    priv->focus_unit = -1;
    priv->year_before = 1;

    priv->week_start = 0;

    priv->show_week_num = FALSE;

    calendar_compute_days(self);

    priv->viewType = VIEWTYPE_DAY;

    self->actor = st_box_layout_new();
    st_widget_set_style_class_name(self->actor, "calendar");
    st_box_layout_set_vertical(ST_BOX_LAYOUT(self->actor), TRUE);

    priv->topBox = st_box_layout_new();
    clutter_actor_set_x_align(CLUTTER_ACTOR(priv->topBox), CLUTTER_ACTOR_ALIGN_CENTER);
    st_widget_set_style_class_name(priv->topBox, "calendar-top");
    st_box_layout_set_vertical(ST_BOX_LAYOUT(priv->topBox), FALSE);

    priv->mainBox = st_box_layout_new();
    st_widget_set_style_class_name(priv->mainBox, "calendar-main-container");
    st_box_layout_set_vertical(ST_BOX_LAYOUT(priv->mainBox), TRUE);
    clutter_actor_set_x_align(CLUTTER_ACTOR(priv->mainBox), CLUTTER_ACTOR_ALIGN_CENTER);

    priv->swipeBox[VIEWTYPE_DAY] = NULL;
    priv->swipeBox[VIEWTYPE_MONTH] = NULL;
    priv->swipeBox[VIEWTYPE_YEAR] = NULL;

    st_box_layout_insert_actor(ST_BOX_LAYOUT(self->actor), CLUTTER_ACTOR(priv->topBox), -1);
    st_box_layout_insert_actor(ST_BOX_LAYOUT(self->actor), CLUTTER_ACTOR(priv->mainBox), -1);
    clutter_actor_set_reactive(CLUTTER_ACTOR(self->actor), TRUE);

    buildHeader(self);
    updateDayView(self, FALSE, 0);
    //g_print("%s --> end!\n", __func__);
}

/**
 * applet_calendar_new:
 *
 * 新建日历
 *
 * Returns: 新建日历对象
 */
AppletCalendar* applet_calendar_new() {
    //g_print("%s --> enter!\n", __func__);
    return g_object_new(APPLET_TYPE_CALENDAR, NULL);
}

static void calendar_compute_days(AppletCalendar *calendar) {
    //g_print("%s --> enter!\n", __func__);
    AppletCalendarPrivate *priv = calendar->priv;
    gint month = 1;
    gint year = 1970;
    gint ndays_in_month = 31;
    gint ndays_in_prevmonth = 31;
    gint first_day = 4;
    gint row = 0;
    gint col = 0;
    gint day = 1;

    year = priv->selectedYear;
    month = priv->selectedMonth;

    ndays_in_month = month_length[leap(year)][month];

    first_day = day_of_week(year, month, 1);
    first_day = (first_day + 7 - priv->week_start) % 7;
    if (0 == first_day) {
        first_day = 7;
    }

    if (1 < month) {
        ndays_in_prevmonth = month_length[leap(year)][month - 1];
    } else {
        ndays_in_prevmonth = month_length[leap(year - 1)][12];
    }

    day = ndays_in_prevmonth - first_day + 1;
    for (col = 0; col < first_day; ++col) {
        priv->day[0][col] = day;
        priv->day_month[0][col] = MONTH_PREV;
        ++day;
    }

    row = first_day / 7;
    col = first_day % 7;
    for (day = 1; day <= ndays_in_month; ++day) {
        priv->day[row][col] = day;
        priv->day_month[row][col] = MONTH_CURRENT;

        ++col;
        if (7 == col) {
            ++row;
            col = 0;
        }
    }

    day = 1;
    for (; row <= 5; ++row) {
        for (; col <= 6; ++col) {
            priv->day[row][col] = day;
            priv->day_month[row][col] = MONTH_NEXT;
            ++day;
        }
        col = 0;
    }
    //g_print("%s --> end!\n", __func__);
}

void applet_calendar_update(AppletCalendar* self) {
    g_return_if_fail(NULL != self && APPLET_IS_CALENDAR(self));
    GDateTime * newDate = g_date_time_new_now_local();
    setDate(self, newDate, FALSE, 0);
    g_date_time_unref(newDate);
}

static void buildHeader(AppletCalendar *calendar) {
    //g_print("%s --> enter!\n", __func__);
    AppletCalendarPrivate* priv = calendar->priv;
    StWidget* topBoxArr = st_box_layout_new();

    st_box_layout_set_vertical(ST_BOX_LAYOUT(topBoxArr), FALSE);

    clutter_actor_destroy_all_children(CLUTTER_ACTOR(priv->topBox));

    priv->viewTypeSwitch = st_button_new_with_label("");
    st_widget_set_style_class_name(priv->viewTypeSwitch, "calendar-month-label");
    st_box_layout_insert_actor(ST_BOX_LAYOUT(priv->topBox), CLUTTER_ACTOR(priv->viewTypeSwitch), -1);
    g_signal_connect(G_OBJECT(priv->viewTypeSwitch), "clicked", G_CALLBACK(onDateLabelClicked), calendar);
    clutter_container_child_set(CLUTTER_CONTAINER(priv->topBox), CLUTTER_ACTOR(priv->viewTypeSwitch), "expand", TRUE, "x_fill", FALSE, "y_fill", FALSE, "x_align", ST_ALIGN_START, "y_align", ST_ALIGN_MIDDLE, NULL);

    st_box_layout_insert_actor(ST_BOX_LAYOUT(priv->topBox), CLUTTER_ACTOR(topBoxArr), -1);
    clutter_container_child_set(CLUTTER_CONTAINER(priv->topBox), CLUTTER_ACTOR(topBoxArr), "expand", TRUE, "x_fill", FALSE, "y_fill", FALSE, "x_align", ST_ALIGN_END, "y_align", ST_ALIGN_MIDDLE, NULL);

    StWidget* back = st_button_new();
    st_widget_set_style_class_name(ST_WIDGET(back), "calendar-change-month-back");
    st_box_layout_insert_actor(ST_BOX_LAYOUT(topBoxArr), CLUTTER_ACTOR(back), -1);
    g_signal_connect(G_OBJECT(back), "clicked", G_CALLBACK(onPreButtonClicked), calendar);
    clutter_container_child_set(CLUTTER_CONTAINER(topBoxArr), CLUTTER_ACTOR(back), "expand", FALSE, "x_fill", FALSE, "y_fill", FALSE, "x_align", ST_ALIGN_END, "y_align", ST_ALIGN_MIDDLE, NULL);

    StWidget* forward = st_button_new();
    st_widget_set_style_class_name(ST_WIDGET(forward), "calendar-change-month-forward");
    st_box_layout_insert_actor(ST_BOX_LAYOUT(topBoxArr), CLUTTER_ACTOR(forward), -1);
    g_signal_connect(G_OBJECT(forward), "clicked", G_CALLBACK(onNextButtonClicked), calendar);
    clutter_container_child_set(CLUTTER_CONTAINER(topBoxArr), CLUTTER_ACTOR(forward), "expand", FALSE, "x_fill", FALSE, "y_fill", FALSE, "x_align", ST_ALIGN_END, "y_align", ST_ALIGN_MIDDLE, NULL);

    //g_print("%s --> end!\n", __func__);
}

static void applyDateBrowseAction(AppletCalendar* calendar, int yearChange, int monthChange) {
    AppletCalendarPrivate* priv = calendar->priv;
    StWidget* swipebox = NULL;
    switch (priv->viewType) {
    case VIEWTYPE_DAY:
        swipebox = priv->swipeBox[0];
        break;
    case VIEWTYPE_MONTH:
        swipebox = priv->swipeBox[1];
        break;
    case VIEWTYPE_YEAR:
        swipebox = priv->swipeBox[2];
        break;
    default:
        break;
    }

    g_return_if_fail(NULL != swipebox);

    gint oldYear = priv->selectedYear;
    gint oldMonth = priv->selectedMonth;
    gint oldDayOfMonth = priv->selectedDayOfMonth;
    gint newMonth = oldMonth + monthChange;
    if (12 < newMonth) {
        yearChange = yearChange + 1;
        newMonth = 1;
    } else if (1 > newMonth) {
        yearChange = yearChange - 1;
        newMonth = 12;
    }

    gint newYear = oldYear + yearChange;
    gint newDayOfMonth = oldDayOfMonth;
    int newMonthLength = month_length[leap(newYear)][newMonth];
    if (newMonthLength < newDayOfMonth) {
       newDayOfMonth = newMonthLength;
    }

    GDateTime * oldDate = g_date_time_new_local(oldYear, oldMonth, oldDayOfMonth, 0, 0, 0);
    GDateTime * newDate = g_date_time_new_local(newYear, newMonth, newDayOfMonth, 0, 0, 0);
    setDate(calendar, newDate, FALSE, g_date_time_difference(oldDate, newDate));
    g_date_time_unref(newDate);
    g_date_time_unref(oldDate);
    // TODO: need to add tweener animations here.
}

static void onPreButtonClicked(StButton *button, int clickedBtn, gpointer user_data) {
    g_return_if_fail(APPLET_IS_CALENDAR(user_data));
    AppletCalendarPrivate* priv = APPLET_CALENDAR(user_data)->priv;
    switch (priv->viewType) {
    case VIEWTYPE_DAY:
        applyDateBrowseAction(APPLET_CALENDAR(user_data), 0, -1);
        break;
    case VIEWTYPE_MONTH:
        applyDateBrowseAction(APPLET_CALENDAR(user_data), -1, 0);
        break;
    case VIEWTYPE_YEAR:
        applyDateBrowseAction(APPLET_CALENDAR(user_data), -1 * 10, 0);
        break;
    default:
        break;
    }
}

static void onNextButtonClicked(StButton *button, int clickedBtn, gpointer user_data) {
    g_return_if_fail(APPLET_IS_CALENDAR(user_data));
    AppletCalendarPrivate* priv = APPLET_CALENDAR(user_data)->priv;
    switch (priv->viewType) {
    case VIEWTYPE_DAY:
        applyDateBrowseAction(APPLET_CALENDAR(user_data), 0, +1);
        break;
    case VIEWTYPE_MONTH:
        applyDateBrowseAction(APPLET_CALENDAR(user_data), +1, 0);
        break;
    case VIEWTYPE_YEAR:
        applyDateBrowseAction(APPLET_CALENDAR(user_data), +1 * 10, 0);
        break;
    default:
        break;
    }
}

static void onDateLabelClicked(StButton *button, int clickedBtn, gpointer user_data) {
    //g_print("%s --> enter!", __func__);
    g_return_if_fail(APPLET_IS_CALENDAR(user_data));
    AppletCalendarPrivate *priv = APPLET_CALENDAR(user_data)->priv;
    switch (priv->viewType) {
    case VIEWTYPE_DAY:
        {
            gfloat x, y;
            clutter_actor_get_position(priv->swipeBox[VIEWTYPE_DAY], &x, &y);
            //g_print("%s:%d --> (x, y) is (%.03f, %.03f)!\n", __func__, __LINE__, x, y);
        }
        clutter_actor_save_easing_state(priv->swipeBox[VIEWTYPE_DAY]);
        clutter_actor_set_scale(priv->swipeBox[VIEWTYPE_DAY], 0.2, 0.2);
        clutter_actor_set_opacity(priv->swipeBox[VIEWTYPE_DAY], 0);
        clutter_actor_set_easing_duration(priv->swipeBox[VIEWTYPE_DAY], ANIMATION_DURATION_TIME);
        clutter_actor_set_easing_mode(priv->swipeBox[VIEWTYPE_DAY], ANIMATION_MODE_OUT);
        clutter_actor_restore_easing_state(priv->swipeBox[VIEWTYPE_DAY]);

        updateMonthView(APPLET_CALENDAR(user_data), TRUE, 0);
        clutter_actor_set_scale(priv->swipeBox[VIEWTYPE_MONTH], 2, 2);
        clutter_actor_set_opacity(priv->swipeBox[VIEWTYPE_MONTH], 0);
        clutter_actor_save_easing_state(priv->swipeBox[VIEWTYPE_MONTH]);
        clutter_actor_set_scale(priv->swipeBox[VIEWTYPE_MONTH], 1, 1);
        clutter_actor_set_opacity(priv->swipeBox[VIEWTYPE_MONTH], 255);
        clutter_actor_set_easing_duration(priv->swipeBox[VIEWTYPE_MONTH], ANIMATION_DURATION_TIME);
        clutter_actor_set_easing_delay(priv->swipeBox[VIEWTYPE_MONTH], ANIMATION_DELAY_TIME);
        clutter_actor_set_easing_mode(priv->swipeBox[VIEWTYPE_MONTH], ANIMATION_MODE_OUT);
        clutter_actor_restore_easing_state(priv->swipeBox[VIEWTYPE_MONTH]);
        break;
    case VIEWTYPE_MONTH:
        {
            gfloat x, y;
            clutter_actor_get_position(priv->swipeBox[VIEWTYPE_MONTH], &x, &y);
            //g_print("%s:%d --> (x, y) is (%.03f, %.03f)!\n", __func__, __LINE__, x, y);
        }
        clutter_actor_save_easing_state(priv->swipeBox[VIEWTYPE_MONTH]);
        clutter_actor_set_scale(priv->swipeBox[VIEWTYPE_MONTH], 0.2, 0.2);
        clutter_actor_set_opacity(priv->swipeBox[VIEWTYPE_MONTH], 0);
        clutter_actor_set_easing_duration(priv->swipeBox[VIEWTYPE_MONTH], ANIMATION_DURATION_TIME);
        clutter_actor_set_easing_mode(priv->swipeBox[VIEWTYPE_MONTH], ANIMATION_MODE_OUT);
        clutter_actor_restore_easing_state(priv->swipeBox[VIEWTYPE_MONTH]);

        updateYearView(APPLET_CALENDAR(user_data), TRUE, 0);
        clutter_actor_set_scale(priv->swipeBox[VIEWTYPE_YEAR], 2, 2);
        clutter_actor_set_opacity(priv->swipeBox[VIEWTYPE_YEAR], 0);
        clutter_actor_save_easing_state(priv->swipeBox[VIEWTYPE_YEAR]);
        clutter_actor_set_scale(priv->swipeBox[VIEWTYPE_YEAR], 1, 1);
        clutter_actor_set_opacity(priv->swipeBox[VIEWTYPE_YEAR], 255);
        clutter_actor_set_easing_duration(priv->swipeBox[VIEWTYPE_YEAR], ANIMATION_DURATION_TIME);
        clutter_actor_set_easing_delay(priv->swipeBox[VIEWTYPE_YEAR], ANIMATION_DELAY_TIME);
        clutter_actor_set_easing_mode(priv->swipeBox[VIEWTYPE_YEAR], ANIMATION_MODE_OUT);
        clutter_actor_restore_easing_state(priv->swipeBox[VIEWTYPE_YEAR]);

        setSensitive(ST_WIDGET(priv->viewTypeSwitch), FALSE);
        break;
    default:
        // default: do nothing!
        return;
    }
}

static gboolean isTheSameYear(GDateTime * date1, GDateTime * date2) {
    gint year1 = g_date_time_get_year(date1);
    gint year2 = g_date_time_get_year(date2);
    return (year1 == year2);
}

static gboolean isTheSameMonth(GDateTime * date1, GDateTime * date2) {
    gint month1 = g_date_time_get_month(date1);
    gint month2 = g_date_time_get_month(date2);
    return (isTheSameYear(date1, date2) && (month1 == month2));
}

static gboolean isTheSameDate(GDateTime * date1, GDateTime * date2) {
    gint day1 = g_date_time_get_day_of_month(date1);
    gint day2 = g_date_time_get_day_of_month(date2);
    return (isTheSameMonth(date1, date2) && (day1 == day2));
}

static void setDate(AppletCalendar *calendar, GDateTime *date, gboolean isForceLoad, GTimeSpan changed) {
    AppletCalendarPrivate *priv = calendar->priv;
    GDateTime* selectedDate = g_date_time_new_local(priv->selectedYear, priv->selectedMonth, priv->selectedDayOfMonth, 1, 1, 1);
    if (!isTheSameDate(date, selectedDate)) {
        g_date_time_get_ymd(date, &(priv->selectedYear), &(priv->selectedMonth), &(priv->selectedDayOfMonth));
        switch (priv->viewType) {
        case VIEWTYPE_DAY:
            updateDayView(calendar, isForceLoad, changed);
            break;
        case VIEWTYPE_MONTH:
            updateMonthView(calendar, isForceLoad, changed);
            break;
        case VIEWTYPE_YEAR:
            updateYearView(calendar, isForceLoad, changed);
            break;
        default:
            break;
        }
    } else {
        if (isForceLoad) {
            updateDayView(calendar, isForceLoad, changed);
        }
    }
    g_date_time_unref(selectedDate);
}

static void updateDayView(AppletCalendar *calendar, gboolean isForceLoad, GTimeSpan changed) {
    //g_print("%s --> enter!\n", __func__);
    calendar_compute_days(calendar);
    AppletCalendarPrivate* priv = calendar->priv;
    StWidget* swipebox;
    gint rowCount = 7;
    gint offsetCols = (priv->show_week_num ? 1 : 0);
    gint colCount = 7 + offsetCols;
    int row = 0;
    int col = 0;
    char buffer[32];
    int selectedYear = priv->selectedYear;
    int selectedMonth = priv->selectedMonth;
    GDateTime* nowDate = g_date_time_new_now_local();

    priv->viewType = VIEWTYPE_DAY;

    clutter_actor_destroy_all_children(CLUTTER_ACTOR(priv->mainBox));

    priv->swipeBox[priv->viewType] = st_box_layout_new();
    st_box_layout_set_vertical(ST_BOX_LAYOUT(priv->swipeBox[priv->viewType]), FALSE);
    swipebox = priv->swipeBox[priv->viewType];

    st_box_layout_insert_actor(CLUTTER_ACTOR(priv->mainBox), CLUTTER_ACTOR(priv->swipeBox[priv->viewType]), -1);
    clutter_actor_set_name(CLUTTER_ACTOR(swipebox), "calendar-table");

    GDateTime* selectedDate = g_date_time_new_local(priv->selectedYear, priv->selectedMonth, priv->selectedDayOfMonth, 1, 1, 1);
    gchar* dateLabelText = g_date_time_format(selectedDate, C_("calendar heading", "%Y %b"));
    st_button_set_label(ST_BUTTON(priv->viewTypeSwitch), dateLabelText);
    g_date_time_unref(selectedDate);
    g_free(dateLabelText);

    for (col = 0; col < colCount; ++col) {
        StWidget* daylayout = st_box_layout_new();
        st_box_layout_set_vertical(ST_BOX_LAYOUT(daylayout), TRUE);
        for (row = 0; row < rowCount; ++row) {
            if (priv->show_week_num && (0 == col)) {
                if (0 == row) {
                    StWidget* weeknumbtn = st_button_new_with_label("");
                    st_widget_add_style_class_name(CLUTTER_ACTOR(weeknumbtn), "calendar-day-base");
                    st_widget_add_style_class_name(ST_WIDGET(weeknumbtn), "calendar-day");
                    st_widget_add_style_class_name(CLUTTER_ACTOR(weeknumbtn), "calendar-week-number");
                    st_box_layout_insert_actor(CLUTTER_ACTOR(daylayout), CLUTTER_ACTOR(weeknumbtn), -1);
                } else {
                    int day = priv->day[row - 1][6];
                    int month = selectedMonth;
                    int year = selectedYear;
                    switch (priv->day_month[row - 1][6]) {
                    case MONTH_PREV:
                        --month;
                        if (1 > month) {
                            month = 12;
                            --year;
                        }
                        break;
                    case MONTH_CURRENT:
                        break;
                    case MONTH_NEXT:
                        ++month;
                        if (12 < month) {
                            month = 1;
                            ++year;
                        }
                        break;
                    default:
                        g_assert(FALSE);
                        break;
                    }

                    int week = week_of_year(year, month, day);
                    g_snprintf(buffer, sizeof(buffer), "%02d", week);
                    StWidget* weeknumbtn = st_button_new_with_label(buffer);
                    st_widget_add_style_class_name(CLUTTER_ACTOR(weeknumbtn), "calendar-day-base");
                    st_widget_add_style_class_name(ST_WIDGET(weeknumbtn), "calendar-day");
                    st_widget_add_style_class_name(CLUTTER_ACTOR(weeknumbtn), "calendar-week-number");
                    st_box_layout_insert_actor(CLUTTER_ACTOR(daylayout), CLUTTER_ACTOR(weeknumbtn), -1);
                } // end: if (0 == row)
            } else {
                if (0 == row) {
                    StWidget* weekTitleItem = st_button_new_with_label(default_abbreviated_dayname[(col - offsetCols + priv->week_start) % 7]);
                    st_widget_add_style_class_name(CLUTTER_ACTOR(weekTitleItem), "calendar-day-base");
                    st_widget_add_style_class_name(ST_WIDGET(weekTitleItem), "calendar-day");
                    st_widget_add_style_class_name(CLUTTER_ACTOR(weekTitleItem), "calendar-week-number");
                    st_box_layout_insert_actor(CLUTTER_ACTOR(daylayout), CLUTTER_ACTOR(weekTitleItem), -1);
                } else {
                    g_snprintf(buffer, sizeof(buffer), "%d", priv->day[row - 1][col - offsetCols]);
                    StWidget* daybtn = applet_calendar_item_new_with_label_day_of_month(buffer, priv->day[row - 1][col - offsetCols]);

                    st_widget_add_style_class_name(ST_WIDGET(daybtn), "calendar-day-base");
                    st_widget_add_style_class_name(ST_WIDGET(daybtn), "calendar-day");
                    if (isWorkDay(col - offsetCols + priv->week_start)) {
                        st_widget_add_style_class_name(ST_WIDGET(daybtn), "calendar-work-day");
                    } else {
                        st_widget_add_style_class_name(ST_WIDGET(daybtn), "calendar-nonwork-day");
                    }

                    switch (priv->day_month[row - 1][col - offsetCols]) {
                    case MONTH_CURRENT: {
                        GDateTime * currentDate = g_date_time_new_local(selectedYear, selectedMonth, priv->day[row - 1][col - offsetCols],
                                                                        0, 0, 0);
                        if (isTheSameDate(nowDate, currentDate)) {
                            st_widget_add_style_class_name(ST_WIDGET(daybtn), "calendar-today");
                        }
                        g_date_time_unref(currentDate);
                    }
                        break;
                    case MONTH_PREV:
                    case MONTH_NEXT:
                        st_widget_add_style_class_name(ST_WIDGET(daybtn), "calendar-other-month-day");
                        break;
                    default:
                        break;
                    }
                    st_box_layout_insert_actor(CLUTTER_ACTOR(daylayout), CLUTTER_ACTOR(daybtn), -1);
                }
            } // end: if (priv->show_week_num && (0 == col))
        } // end: for row
        st_box_layout_insert_actor(CLUTTER_ACTOR(swipebox), CLUTTER_ACTOR(daylayout), -1);
    } // end: for col
    g_date_time_unref(nowDate);
    //g_print("%s --> end!\n", __func__);
}

static void updateMonthView(AppletCalendar *calendar, gboolean isForceLoad, GTimeSpan changed) {
    AppletCalendarPrivate* priv = calendar->priv;
    StWidget* swipebox = NULL;
    int row = 0;
    int col = 0;
    int rowCount = 3;
    int colCount = 4;
    GDateTime* nowDate = g_date_time_new_now_local();
    int selectedYear = priv->selectedYear;

    priv->viewType = VIEWTYPE_MONTH;

    clutter_actor_destroy_all_children(CLUTTER_ACTOR(priv->mainBox));

    priv->swipeBox[priv->viewType] = st_box_layout_new();
    st_box_layout_set_vertical(ST_BOX_LAYOUT(priv->swipeBox[priv->viewType]), TRUE);
    swipebox = priv->swipeBox[priv->viewType];
    st_widget_set_style(swipebox,"padding-top:33px;padding-bottom:33px;");
    st_box_layout_insert_actor(ST_BOX_LAYOUT(priv->mainBox), CLUTTER_ACTOR(priv->swipeBox[priv->viewType]), -1);
    clutter_actor_set_name(CLUTTER_ACTOR(swipebox), "calendar-table");
    GDateTime* selectedDate = g_date_time_new_local(priv->selectedYear, priv->selectedMonth, priv->selectedDayOfMonth, 1, 1, 1);
    gchar* dateLabelText = g_date_time_format(selectedDate, C_("calendar heading", "%Y"));
    st_button_set_label(ST_BUTTON(priv->viewTypeSwitch), dateLabelText);
    g_date_time_unref(selectedDate);
    g_free(dateLabelText);

    for (row = 0; row < rowCount; ++row) {
        StWidget* monthlayout = st_box_layout_new();
        st_box_layout_set_vertical(ST_BOX_LAYOUT(monthlayout), FALSE);
        for (col = 0; col < colCount; ++col) {
            int index = row * colCount + col;
            StWidget* monthButton = applet_calendar_item_new_with_label_month(default_monthname[index], index + 1);
            g_signal_connect(G_OBJECT(monthButton), "clicked", G_CALLBACK(onCalendarItemClicked), calendar);

            st_widget_add_style_class_name(ST_WIDGET(monthButton), "calendar-month-base");
            st_widget_add_style_class_name(ST_WIDGET(monthButton), "calendar-month");
            GDateTime* currentDate = g_date_time_new_local(selectedYear, index + 1, 1, 0, 0, 0);
            if (isTheSameMonth(nowDate, currentDate)) {
                st_widget_add_style_class_name(ST_WIDGET(monthButton), "calendar-current-month");
            }
            g_date_time_unref(currentDate);
            st_box_layout_insert_actor(CLUTTER_ACTOR(monthlayout), CLUTTER_ACTOR(monthButton), -1);
        }
        st_box_layout_insert_actor(CLUTTER_ACTOR(swipebox), CLUTTER_ACTOR(monthlayout), -1);
    }
}

static void updateYearView(AppletCalendar* calendar, gboolean isForceLoad, GTimeSpan changed) {
    AppletCalendarPrivate* priv = calendar->priv;
    StWidget* swipebox = NULL;
    int row = 0;
    int col = 0;
    int rowCount = 3;
    int colCount = 4;
    char buffer[32];
    GDateTime* nowDate = g_date_time_new_now_local();
    int nowYear = g_date_time_get_year(nowDate);
    int selectedYear = priv->selectedYear;

    int beginYear = ((int)selectedYear / 10) * 10 - 1;
    int endYear = ((int)selectedYear / 10 + 1) * 10;
    int iterYear = beginYear;

    priv->viewType = VIEWTYPE_YEAR;

    clutter_actor_destroy_all_children(CLUTTER_ACTOR(priv->mainBox));

    priv->swipeBox[priv->viewType] = st_box_layout_new();
    st_box_layout_set_vertical(ST_BOX_LAYOUT(priv->swipeBox[priv->viewType]), TRUE);
    swipebox = priv->swipeBox[priv->viewType];
    st_widget_set_style(swipebox,"padding-top:33px;padding-bottom:33px;");

    st_box_layout_insert_actor(ST_BOX_LAYOUT(priv->mainBox), CLUTTER_ACTOR(priv->swipeBox[priv->viewType]), -1);
    clutter_actor_set_name(CLUTTER_ACTOR(swipebox), "calendar-table");

    g_snprintf(buffer, sizeof(buffer), "%d-%d", beginYear, endYear);
    st_button_set_label(ST_BUTTON(priv->viewTypeSwitch), buffer);

    for (row = 0; row < rowCount; ++row) {
        StWidget* yearLayout = st_box_layout_new();
        st_box_layout_set_vertical(ST_BOX_LAYOUT(yearLayout), FALSE);
        for (col = 0; col < colCount; ++col) {
            int index = row * colCount + col;
            iterYear = beginYear + index;
            g_snprintf(buffer, sizeof(buffer), "%d", iterYear);
            StWidget* yearButton = applet_calendar_item_new_with_label_year(buffer, iterYear);
            g_signal_connect(G_OBJECT(yearButton), "clicked", G_CALLBACK(onCalendarItemClicked), calendar);

            st_widget_add_style_class_name(ST_WIDGET(yearButton), "calendar-year-base");
            st_widget_add_style_class_name(ST_WIDGET(yearButton), "calendar-year");
            if (nowYear == iterYear) {
                st_widget_add_style_class_name(ST_WIDGET(yearButton), "calendar-current-year");
            } else if (((int)(iterYear / 10)) != ((int)(selectedYear / 10))) {
                st_widget_add_style_class_name(ST_WIDGET(yearButton), "calendar-other-x-year");
            }
            st_box_layout_insert_actor(CLUTTER_ACTOR(yearLayout), CLUTTER_ACTOR(yearButton), -1);
        }
        st_box_layout_insert_actor(CLUTTER_ACTOR(swipebox), CLUTTER_ACTOR(yearLayout), -1);
    }
}

static void onCalendarItemClicked(StButton * button, int clickedBtn, gpointer user_data) {
    g_return_if_fail(APPLET_IS_CALENDAR_ITEM(button));
    g_return_if_fail(APPLET_IS_CALENDAR(user_data));
    AppletCalendarPrivate* priv = APPLET_CALENDAR(user_data)->priv;
    int itemOpt = applet_calendar_item_get_option(APPLET_CALENDAR_ITEM(button));
    int itemData = applet_calendar_item_get_item_data(APPLET_CALENDAR_ITEM(button));

    int newYear = priv->selectedYear;
    int newMonth = priv->selectedMonth;
    int newDayOfMonth = priv->selectedDayOfMonth;

    setSensitive(ST_WIDGET(priv->viewTypeSwitch), TRUE);

    switch (priv->viewType) {
    case VIEWTYPE_DAY:
        break;
    case VIEWTYPE_MONTH:
        g_return_if_fail(OPT_MONTH == itemOpt);
        clutter_actor_save_easing_state(priv->swipeBox[VIEWTYPE_MONTH]);
        clutter_actor_set_scale(priv->swipeBox[VIEWTYPE_MONTH], 2, 2);
        clutter_actor_set_opacity(priv->swipeBox[VIEWTYPE_MONTH], 0);
        clutter_actor_set_easing_duration(priv->swipeBox[VIEWTYPE_MONTH], ANIMATION_DURATION_TIME);
        clutter_actor_set_easing_mode(priv->swipeBox[VIEWTYPE_MONTH], ANIMATION_MODE_OUT);
        clutter_actor_restore_easing_state(priv->swipeBox[VIEWTYPE_MONTH]);

        newMonth = itemData;
        updateDayView(APPLET_CALENDAR(user_data), FALSE, 0);
        break;
    case VIEWTYPE_YEAR:
        g_return_if_fail(OPT_YEAR == itemOpt);
        clutter_actor_save_easing_state(priv->swipeBox[VIEWTYPE_YEAR]);
        clutter_actor_set_scale(priv->swipeBox[VIEWTYPE_YEAR], 2, 2);
        clutter_actor_set_opacity(priv->swipeBox[VIEWTYPE_YEAR], 0);
        clutter_actor_set_easing_duration(priv->swipeBox[VIEWTYPE_YEAR], ANIMATION_DURATION_TIME);
        clutter_actor_set_easing_mode(priv->swipeBox[VIEWTYPE_YEAR], ANIMATION_MODE_OUT);
        clutter_actor_restore_easing_state(priv->swipeBox[VIEWTYPE_YEAR]);

        newYear = itemData;
        updateMonthView(APPLET_CALENDAR(user_data), FALSE, 0);
        break;
    default:
        break;
    }

    GDateTime * newDate = g_date_time_new_local(newYear, newMonth, newDayOfMonth, 0, 0, 0);
    setDate(APPLET_CALENDAR(user_data), newDate, FALSE, 0);
    g_date_time_unref(newDate);

    clutter_actor_set_scale(priv->swipeBox[priv->viewType], 0.2, 0.2);
    clutter_actor_set_opacity(priv->swipeBox[priv->viewType], 0);
    clutter_actor_save_easing_state(priv->swipeBox[priv->viewType]);
    clutter_actor_set_scale(priv->swipeBox[priv->viewType], 1, 1);
    clutter_actor_set_opacity(priv->swipeBox[priv->viewType], 255);
    clutter_actor_set_easing_duration(priv->swipeBox[priv->viewType], ANIMATION_DURATION_TIME);
//    clutter_actor_set_easing_delay(priv->swipeBox[priv->viewType], ANIMATION_DELAY_TIME);
    clutter_actor_set_easing_mode(priv->swipeBox[priv->viewType], ANIMATION_MODE_OUT);
    clutter_actor_restore_easing_state(priv->swipeBox[priv->viewType]);
}

static void applet_calendar_finalize(GObject * calendar) {
    g_return_if_fail(APPLET_IS_CALENDAR(calendar));
    AppletCalendarPrivate* priv = APPLET_CALENDAR(calendar)->priv;

    g_object_unref(priv->viewTypeSwitch);
    clutter_actor_destroy_all_children(CLUTTER_ACTOR(priv->topBox));
    clutter_actor_destroy_all_children(CLUTTER_ACTOR(priv->mainBox));
    clutter_actor_destroy_all_children(CLUTTER_ACTOR(APPLET_CALENDAR(calendar)->actor));

    int i = 0;
    for (i = 0; i < 3; ++i) {
        if (NULL != priv->swipeBox[i]) {
            clutter_actor_destroy_all_children(CLUTTER_ACTOR(priv->swipeBox[i]));
            g_object_unref(priv->swipeBox[i]);
            priv->swipeBox[i] = NULL;
        }
    }
}

//static void applet_calendar_set_property(GObject * obj,
//                                       guint prop_id,
//                                       const GValue * value) {

//}

//static void applet_calendar_get_property(GObject * obj,
//                                       guint prop_id,
//                                       GValue * value,
//                                       GParamSpec * pspec) {

//}
