#include <string.h>
#include "applet-nm.h"
#include "cdos-popup.h"
#include "cdos-popup-separator-menu-item.h"
#include "cdos.h"
#include "cdos-util.h"

#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include "applet-nm-connection.h"
#include "applet-nm-wired-section-title-menu-item.h"
#include "applet-nm-wireless-section-menu-item.h"
#include "applet-nm-menu-item.h"
#include "applet-nm-device-wired.h"
#include "applet-nm-device-wireless.h"

#include "applet-nm-private.h"

/**
 * SECTION:applet-nm
 * @title: AppletNM
 * @short_description: Network applet 
 *
 * 管理网络连接，启用或禁用网络设备
 */

#define DEFAULT_PERIODIC_UPDATE_FREQUENCY_SECONDS 10
#define FAST_PERIODIC_UPDATE_FREQUENCY_SECONDS    2

const AppletEnumValue applet_nm_device_info[DeviceCount] = {
    {DeviceWired, "wired"},
    {DeviceWireless, "wireless"},
    {DeviceWwan, "wwan"},
    {DeviceVPN, "vpn"}
};

static void
applet_nm_sync_section_title (AppletNM *self, AppletNMDeviceType type);

static void
applet_nm_connection_removed (NMConnection *connection, AppletNM *self);

static void
applet_nm_update_connection (NMConnection *connection, AppletNM *self);

static void
applet_nm_device_added (NMClient *client, NMDevice *device, AppletNM *self);

static void applet_nm_set_ap_icon (AppletNM *self, NMAccessPoint *ap);

static void applet_nm_update_icon (NMClient *NmClient, gpointer spec, AppletNM *self);

const AppletEnumValue ctypes[] = 
{
    {DeviceWireless, NM_SETTING_WIRELESS_SETTING_NAME},
    {DeviceWired, NM_SETTING_WIRED_SETTING_NAME},
    {DeviceWired, NM_SETTING_PPPOE_SETTING_NAME},
    {DeviceWired, NM_SETTING_PPP_SETTING_NAME},
    {DeviceWwan,  NM_SETTING_BLUETOOTH_SETTING_NAME},
    {DeviceWwan,  NM_SETTING_CDMA_SETTING_NAME},
    {DeviceWwan,  NM_SETTING_GSM_SETTING_NAME},
    {DeviceVPN,   NM_SETTING_VPN_SETTING_NAME}
};

gint 
nm_g_ptr_array_find (const GPtrArray *array, gpointer element)
{
    guint i;
    for (i = 0; array && (i < array->len); i++)
        if (array->pdata[i] == element)
            return i;
    return -1;
}

static AppletNMDeviceType
get_ctype_category (const gchar *type)
{
    if (type == NULL)
        return DeviceInvalid;
    int i;
    for (i = 0; i < G_N_ELEMENTS (ctypes); i++)
    {
        if (strcmp (type, ctypes[i].value) == 0)
            return ctypes[i].key;
    }
    return DeviceInvalid;
}

G_DEFINE_TYPE ( AppletNM, applet_nm, CDOS_TYPE_ICONAPPLET)

static void 
menu_item_enabled_changed (AppletNMWirelessSectionMenuItem *item,
                           gboolean enabled,
                           AppletNM *self)
{
    GList *devices = self->devices[item->device_type].devices;
    GList *iter = NULL;
    for (iter = devices; iter; iter = iter->next)
    {
        applet_nm_device_set_enabled (iter->data, enabled);
    }
    applet_nm_sync_section_title (self, item->device_type);
}
static AppletNMWirelessSectionMenuItem *
make_toggle_item (AppletNM *self, AppletNMDeviceType type, const char *title)
{
    AppletNMWirelessSectionMenuItem *item = applet_nm_wireless_section_menu_item_new (self->client, type, title);
    g_signal_connect (item, "enabled-changed", G_CALLBACK (menu_item_enabled_changed), self);
    return item;
}

static void
applet_nm_finalize(GObject *object)
{
    AppletNM *applet = APPLET_NM (object);
    g_object_unref (applet->settings);
    g_free (applet->base_class);
    g_free (applet->wifi_class);
    g_free (applet->ethernet_class);
    g_free (applet->gdm_class);
    g_free (applet->cdma_class);
    g_free (applet->wimax_class);

    g_slist_free (applet->connections);

    GList *dev;
    gint i;
    for (i = 0; i < DeviceCount; i++)
    {
        GList *devices = applet->devices[i].devices;
        for (dev = devices; dev; dev = dev->next)
        {
            applet_nm_device_destroy (dev->data);
        }
        g_list_free (devices);
    }
    g_free (applet->devices);

    G_OBJECT_CLASS(applet_nm_parent_class)->finalize(object);
}

static void 
applet_nm_class_init (AppletNMClass *klass) {

    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->finalize = applet_nm_finalize;
}

static void
applet_nm_add_action_callback (CdosPopupBaseMenuItem *item, 
                               ClutterEvent *event,
                               gboolean keep,
                               AppletNM *self)
{
    if (self->client)
        nm_client_networking_set_enabled (self->client, TRUE);
}

static void
cdos_add_action_util_callback ()
{
    cdos_util_spawn_command_line ("nm-connection-editor");
}

static void
applet_nm_set_icon (AppletNM *self, const gchar *name)
{
    if (g_strcmp0 (self->currentIconName, name))
    {
        cdos_iconapplet_set_icon_symbolic_name (CDOS_ICONAPPLET(self), name);
        g_free (self->currentIconName);
        self->currentIconName = g_strdup (name);
    }
}

static void
applet_nm_on_signal_destroy (AppletNMMessageTraySource *source, AppletNM *self)
{
    self->source = NULL;
}

static void
applet_nm_ensure_source (AppletNM *self)
{
    if( !(self->source) )
    {
        self->source = applet_nm_message_tray_source_new();
        g_signal_connect(self->source,\
                         "destroy",\
                         G_CALLBACK(applet_nm_on_signal_destroy),\
                         self);

        CdosMessageTray *messageTray = cdos_monkey_plugin_get_message_tray (monkey_plugin_get());
        if (messageTray)
        {
            cdos_message_tray_add (messageTray, CDOS_SOURCE( self->source ));
        }
    }
}

static void 
update_for_device (CdosPopupSwitchMenuItem *item, AppletNMDevice *device,
                   AppletNMDeviceType type)
{
    switch (type)
    {
        case DeviceWired:
            applet_nm_wired_section_title_menu_item_update_for_device (
                    item, device);
            break;
        case DeviceWireless:
            applet_nm_wireless_section_menu_item_update_for_device (
                    item, device);
            break;
        default:
            break;
    }
}

static void
applet_nm_sync_section_title (AppletNM *self, AppletNMDeviceType type)
{
    GList *devices = self->devices[type].devices;
    CdosPopupSwitchMenuItem *item = self->devices[type].item;
    CdosPopupMenuSection *section = self->devices[type].section;
    if (devices == NULL)
    {
        if (section)
            clutter_actor_hide ((ClutterActor *)section->actor);
    }
    else
    {
        clutter_actor_show ((ClutterActor *)section->actor);
        if (devices->next == NULL)
        {
            AppletNMDevice *dev = devices->data;
            clutter_actor_hide (CLUTTER_ACTOR (cdos_popup_base_menu_item_get_actor (CDOS_POPUP_BASE_MENU_ITEM (dev->status_item))));
            update_for_device (item, dev, type);
        }
        else
        {
            GList *iter = NULL;
            for (iter = devices; iter; iter = iter->next)
            {
                //CdosPopupSwitchMenuItem *status_item = ((AppletNMDevice *)iter->data)->status_item;
                //clutter_actor_show ((ClutterActor *)(CDOS_POPUP_BASE_MENU_ITEM(status_item)->actor));
                AppletNMDevice *dev_tmp = (AppletNMDevice *)iter->data;
                CdosPopupSwitchMenuItem *status_item = dev_tmp->status_item;
                if (dev_tmp->category == DeviceWired)
                {
                    clutter_actor_hide ((ClutterActor *)(CDOS_POPUP_BASE_MENU_ITEM(status_item)->actor));
                }
                else
                {
                    clutter_actor_show ((ClutterActor *)(CDOS_POPUP_BASE_MENU_ITEM(status_item)->actor));
                }
            }
            update_for_device (item, NULL, type);
        }
    }
}

//_readDevices:
static void
applet_nm_read_devices (AppletNM *self)
{
    const GPtrArray *devices = nm_client_get_devices (self->client);
    gint i;
    for (i = 0; devices && i < devices->len; i++)
    {
        NMDevice *a = devices->pdata[i];
        g_debug ("add device: %s", nm_device_get_description (a));
        applet_nm_device_added (self->client, a, self );
    }
}

void
applet_nm_state_changed (AppletNM *applet, AppletNMDevice *wrapper)
{
    g_return_if_fail (APPLET_IS_NM (applet));
    applet_nm_sync_section_title (applet, wrapper->category);
}

static void
applet_nm_destroy (AppletNMDeviceWired *wrapper, gpointer user_data)
{
    applet_nm_finalize ((GObject *)user_data);
}

static void
applet_nm_device_added (NMClient *client, NMDevice *device, AppletNM *self)
{
    AppletNMDevice *delegate = g_object_get_data ((GObject *)device, "delegate");
    if (delegate)
    {
        //already seen, not adding again
        return;
    }
    AppletNMDeviceType type;
    NMDeviceType t = nm_device_get_device_type (device);
    if (t == NM_DEVICE_TYPE_ETHERNET)
        type = DeviceWired;
    else if (t == NM_DEVICE_TYPE_WIFI)
        type = DeviceWireless;
    else 
        return;
    AppletNMDevice *wrapper;
    if (type == DeviceWired)
        wrapper = (AppletNMDevice *)applet_nm_device_wired_new (self, device, self->connections);
    else if (type == DeviceWireless)
        wrapper = (AppletNMDevice *)applet_nm_device_wireless_new (self, device, self->connections);

    g_object_set_data ((GObject *)device, "delegate", wrapper);

    cdos_popup_menu_base_add_actor (CDOS_POPUP_MENU_BASE (self->devices[type].section),
            (ClutterActor *)wrapper->section->actor);

    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->devices[type].section),
                                        CDOS_POPUP_BASE_MENU_ITEM (wrapper->status_item), 1);
    self->devices[type].devices = g_list_append (self->devices[type].devices, wrapper);

    applet_nm_sync_section_title (self, type);

}

//_deviceRemoved:
static void
applet_nm_device_removed (NMClient *client, NMDevice *device , AppletNM *self)
{
    //if (!(applet_nm_device_hash_get_delegate(device)))
    AppletNMDevice *delegate = g_object_get_data ((GObject *)device, "delegate");
    if (delegate == NULL)
    {
        g_warning ("Removing a network device that was not added");
        return;
    }
    applet_nm_device_destroy (delegate);
    GList *devices = self->devices[delegate->category].devices;
    devices = g_list_remove (devices, delegate);
    self->devices[delegate->category].devices = devices;
    applet_nm_sync_section_title (self, delegate->category);
}

/*
static void
applet_nm_active_connection_changed (NMActiveConnection *activeConnection, gpointer spec, AppletNM *self)
{
    if ((nm_active_connection_get_state(activeConnection) == NM_ACTIVE_CONNECTION_STATE_ACTIVATED)
            && (applet_nm_active_connection_get_primary_device (activeConnection))
            && (applet_nm_device_hash_get_notification (applet_nm_active_connection_get_primary_device (activeConnection))))
    {
        cdos_notification_destroy (applet_nm_device_hash_get_notification (applet_nm_active_connection_get_primary_device (activeConnection)), 1);//暂时为1...
        applet_nm_device_hash_set_notification (applet_nm_active_connection_get_primary_device (activeConnection), NULL);
    }
}
*/

static void
applet_nm_read_connections (AppletNM *self)
{

    if (self->connections)
    {
        g_slist_free (self->connections);
        self->connections = NULL;
    }
    GSList *connections = nm_remote_settings_list_connections (self->settings);
    GSList *iter = NULL;
    for (iter = connections; iter; iter=iter->next)
    {
        NMConnection *conn = NM_CONNECTION (iter->data);

        AppletNMConnection *table = applet_nm_connection_get_data (conn);
        if (table->uuid)
        {
            //connection was already seen (for example because NetworkManager was restarted)
            continue;
        }
        table->connection = conn;
        table->removed_id = g_signal_connect (conn, "removed", G_CALLBACK (applet_nm_connection_removed), self);
        table->updated_id = g_signal_connect (conn, "updated", G_CALLBACK (applet_nm_update_connection), self);

        applet_nm_update_connection (conn, self);
    }
    self->connections = connections;
}

static void
applet_nm_update_connection (NMConnection *connection,
                             AppletNM *self)
{
    NMSettingConnection *s_con = nm_connection_get_setting_connection (connection);
    const char *type = nm_setting_connection_get_connection_type (s_con);
    NMSetting *setting = nm_connection_get_setting_by_name (connection, type);
    /*
    if (NM_IS_SETTING_WIRELESS (setting))
    {
        setting = nm_connection_get_setting (connection, NM_TYPE_SETTING_WIRELESS_SECURITY);
    }else{
        setting = nm_connection_get_setting_by_name (connection, type);
    }
    */
    //NM_IS_SETTING_CONNECTION(connectionSettings);

    AppletNMConnection *table = applet_nm_connection_get_data (connection);
    table->type = nm_setting_get_name (setting);
    table->section = get_ctype_category (table->type);
    table->name = nm_setting_connection_get_id (s_con);
    table->uuid = nm_setting_connection_get_uuid (s_con);
    table->timestamp = nm_setting_connection_get_timestamp (s_con);

    g_debug ("update connection: type:%s, name:%s, uuid:%s", table->type, table->name, table->uuid);

    GList *devices = self->devices[DeviceWired].devices;
    GList *iter;
    for (iter = devices; iter; iter = iter->next)
        applet_nm_device_check_connection (iter->data, connection);

    devices = self->devices[DeviceWireless].devices;
    for (iter = devices; iter; iter = iter->next)
        applet_nm_device_check_connection (iter->data, connection);
}

static void
applet_nm_new_connection (NMRemoteSettings *settings, NMConnection *connection, AppletNM *self)
{
    AppletNMConnection *table = applet_nm_connection_get_data (connection);
    if (table->uuid)
    {
        //connection was already seen
        return;
    }
    table->removed_id = g_signal_connect (connection, "removed",
                                          G_CALLBACK (applet_nm_connection_removed), self);
    table->updated_id = g_signal_connect (connection, "updated",
                                          G_CALLBACK (applet_nm_update_connection), self);

    applet_nm_update_connection (connection, self);
    self->connections = g_slist_append (self->connections, connection);

    //applet_nm_update_icon (NULL, NULL, self);
}

static void
applet_nm_connection_removed (NMConnection *connection, AppletNM *self)
{
    AppletNMConnection *conn = applet_nm_connection_get_data (connection);

    gint pos = g_slist_index (self->connections, connection);
    if (pos != -1)
        self->connections = g_slist_remove (self->connections, connection);
    AppletNMDeviceType section = conn->section;
    if( (section != DeviceVPN) && (section != DeviceInvalid))
    {
        GList *devices = self->devices[section].devices;
        GList *iter = NULL;
        for (iter = devices; iter; iter = iter->next)
            applet_nm_device_remove_connection (iter->data, connection);
    }

    conn->uuid = 0;
    //g_signal_handler_disconnect (connection, conn->removed_id);
    //g_signal_handler_disconnect (connection, conn->updated_id);
    g_free (conn);
}

//_hideDevices:
static void
applet_nm_hide_devices (AppletNM *self)
{
    self->devicesHidden = TRUE;
    //if( self->_devices->wired == "wired" )
    /* FIXME */
    //if (self->devices[DeviceWired] != NULL)
    {
        clutter_actor_hide (CLUTTER_ACTOR (self->devices[DeviceWired].section->actor));
    }
}

//_showNormal:
static void
applet_nm_show_normal (AppletNM *self)
{
    if( !(self->devicesHidden) )//nothing to do
    {
        return;
    }
    self->devicesHidden = FALSE;
    clutter_actor_hide ((ClutterActor *)self->statusSection->actor);

    applet_nm_sync_section_title (self, DeviceWired);
    applet_nm_sync_section_title (self, DeviceWireless);
    applet_nm_sync_section_title (self, DeviceWwan);

    //vpn模块，暂时不考虑
    //if (!this._devices.vpn.device.empty)
    //            this._devices.vpn.section.actor.show();
    //if( !applet_nm_device_vpn_get_empty() )
    //{
        //TODO:
    //}
}

static void 
applet_nm_sync_connection (NMClient *client,
                           GParamSpec *pspec,
                           AppletNM *self)
{
    NMActiveConnection *connection;
    connection = nm_client_get_primary_connection (self->client);
    self->mainConnection = connection;
    if (!self->mainConnection)
    {
        for (AppletNMDeviceType type = DeviceWired; type < DeviceCount; type ++)
        {
            GList *devices = self->devices[type].devices;
            for (GList *iter = devices; iter; iter = iter->next)
            {
                AppletNMDevice *device = iter->data;
                NMDevice *dev = device->device;
                connection = nm_device_get_active_connection (dev);
                if (connection)
                {
                    self->mainConnection = connection;
                    applet_nm_update_icon (NULL, NULL, self);
                    return;
                }
            }
        }
    }
    applet_nm_update_icon (NULL, NULL, self);
}

static void 
applet_nm_active_connection_state_changed (NMActiveConnection *connection,
                                           GParamSpec *pspec,
                                           AppletNM *self)
{
    NMActiveConnectionState state = nm_active_connection_get_state (connection);
    if (state != NM_ACTIVE_CONNECTION_STATE_ACTIVATING)
    {
        //g_signal_handlers_disconnect_by_func (connection,
        //                                      applet_nm_active_connection_state_changed,
        //                                      self);
        applet_nm_sync_connection (NULL, NULL, self);
    }
}

/* 
 * 如果Networkmanager没用配置自动检查connectivity，
 * 那么在网络由 正在连接-->已连接 转变后，
 * NMClient 不会收到 connectivity 变化的信号。
 * 可以通过 activating-connection 和 active-connections 的变化检测网络状态变化
 */
static void
applet_nm_sync_activating_connection (NMClient *client,
                                      GParamSpec *pspec,
                                      AppletNM *self)
{
    NMActiveConnection *connection;
    connection = nm_client_get_activating_connection (self->client);
    if (!connection)
    {
        applet_nm_sync_connection (NULL, NULL, self);
        return;
    }
    self->mainConnection = connection;
    g_signal_connect (connection, "notify::state", G_CALLBACK (applet_nm_active_connection_state_changed), self);
    applet_nm_update_icon (NULL, NULL, self);
}

static gboolean
applet_nm_has_active_device (AppletNM *self)
{
    for (AppletNMDeviceType type = DeviceWired; type < DeviceCount; type ++)
    {
        GList *devices = self->devices[type].devices;
        for (GList *iter = devices; iter; iter = iter->next)
        {
            AppletNMDevice *device = iter->data;
            NMDevice *dev = device->device;
            NMDeviceState state = nm_device_get_state (dev);
            if (state > NM_DEVICE_STATE_DISCONNECTED ||
                state < NM_DEVICE_STATE_DEACTIVATING)
            {
                return TRUE;
            }
        }
    }
    return FALSE;
}

static void 
applet_nm_sync_active_connections (NMClient *client,
                                   GParamSpec *pspec,
                                   AppletNM *self)
{
    /*
     * FIXME
     * 这里好像是libnm-glib的问题，在所有网络设备都断开连接时，
     * 依然可以获取到active-connections，
     * 所以在这里对所有设备状态进行判断，
     * 注意不能使用 nm_client_get_state() 函数判断网络状态。
     */
    gboolean has_active_device = applet_nm_has_active_device (self);
    if (!has_active_device)
    {
        self->mainConnection = NULL;
        applet_nm_update_icon (NULL, NULL, self);
        return;
    }
    const GPtrArray *connections = nm_client_get_active_connections (self->client);
    if (!connections || (connections->len == 0))
    {
        self->mainConnection = NULL;
        applet_nm_update_icon (NULL, NULL, self);
        return;
    }
    /*
     * 在主网卡已连接时，NMClient不会收到activating-connection信号，
     * 也不能获取到 activating-connection，只能收到active-connetions信号。
     */
    NMActiveConnection *tmp = self->mainConnection;
    gint i;
    for (i = 0; i < connections->len; i++)
    {
        NMActiveConnection *active = connections->pdata[i];
        /*
        NMRemoteConnection *r_con;
        const char *path = nm_active_connection_get_connection (active);
        r_con = nm_remote_settings_get_connection_by_path (self->settings, path);
        const char *id = nm_connection_get_id ((NMConnection *)r_con);
        g_message ("sync_active_connections: %s", id);
        */
        if (nm_active_connection_get_state (active) == NM_ACTIVE_CONNECTION_STATE_ACTIVATING)
        {
            self->mainConnection = active;
            g_signal_connect (active, "notify::state", G_CALLBACK (applet_nm_active_connection_state_changed), self);
            break;
        }
    }
    //if (tmp != self->mainConnection)
        applet_nm_update_icon (NULL, NULL, self);
}

static void 
applet_nm_sync_connectivity (NMClient *client,
                             GParamSpec *pspec,
                             AppletNM *self)
{
    applet_nm_update_icon (NULL, NULL, self);
}

static void
applet_nm_sync_nm_state (NMClient *NmClient, gpointer spec, AppletNM *self)
{
    if (!(nm_client_get_manager_running (self->client)))
    {
        //g_print ("NetworkManager is not running,hiding...\n");
        clutter_actor_hide ((ClutterActor *)CDOS_APPLET(self)->actor);
        return;
    }
    else
    {
        clutter_actor_show ((ClutterActor *)CDOS_APPLET(self)->actor);
    }

    if (!(nm_client_networking_get_enabled (self->client)))
    {
        applet_nm_set_icon (self, "network-offline");
        applet_nm_hide_devices (self);
        cdos_popup_menu_item_set_label (self->statusItem, _("Networking is disabled"));
        clutter_actor_show ((ClutterActor *)self->statusSection->actor);
        return;
    }
    applet_nm_show_normal (self);
    applet_nm_sync_connection (NULL, NULL, self);
}

static void 
applet_nm_ap_strength_changed (NMAccessPoint *ap, GParamSpec *pspec, AppletNM *self)
{
    applet_nm_set_ap_icon (self, ap);
}

static void 
applet_nm_set_ap_icon (AppletNM *self, NMAccessPoint *ap)
{
    if (self->active_access_point && NM_IS_ACCESS_POINT (self->active_access_point))
    {
        g_signal_handlers_disconnect_by_func (self->active_access_point, applet_nm_ap_strength_changed, self);
    }
    self->active_access_point = ap;
    guint8 strength = nm_access_point_get_strength (ap);

    if (self->connectivity != 0x2)
        applet_nm_set_icon (self, "network-wireless-no-route");
    else {
        const gchar *icon = applet_nm_get_icon (ap, applet_nm_signal_to_icon (strength));
        applet_nm_set_icon (self, icon);
    }
    g_signal_connect (ap, "notify::strength", G_CALLBACK (applet_nm_ap_strength_changed), self);
    char *utf8_ssid = applet_nm_get_utf8_ssid (nm_access_point_get_ssid (ap));
    gchar *tooltip;
    if (self->connectivity == 0x1) {
        tooltip = g_strdup_printf ("%s: %s (%d%%)",
                                    _("Wireless connection, maybe need authenticate"),
                                    utf8_ssid,
                                    strength);
    }
    else if (self->connectivity == 0){
        tooltip = g_strdup_printf ("%s: %s (%d%%)",
                                   _("Wireless connection, limited access"),
                                   utf8_ssid,
                                   strength);

    }
    else /* (self->connectivity == 0x2) */
    {
        tooltip = g_strdup_printf ("%s: %s (%d%%)",
                                   _("Wireless connection"),
                                   utf8_ssid,
                                   strength);
    }
    cdos_applet_set_applet_tooltip (CDOS_APPLET (self), tooltip);
    g_free (tooltip);
    g_free (utf8_ssid);
}

static void
applet_nm_update_icon (NMClient *client, gpointer spec, AppletNM *self)
{
    NMActiveConnection *mc;
    gboolean has_active_device = applet_nm_has_active_device (self);
    if (has_active_device)
        mc = self->mainConnection;
    else
    {
        self->mainConnection = NULL;
        mc = NULL;
    }
    if (!mc)
    {
        applet_nm_set_icon(self, "network-offline");
        cdos_applet_set_applet_tooltip (CDOS_APPLET(self), _("No connection"));
        return;
    }
    if (!NM_IS_ACTIVE_CONNECTION (mc))
        return;

    NMActiveConnectionState state = nm_active_connection_get_state (mc);
    if (state > NM_ACTIVE_CONNECTION_STATE_ACTIVATED)
        return;

    const GPtrArray *devices = nm_active_connection_get_devices (mc);
    if (!devices || !devices->len)
    {
        applet_nm_sync_connection (NULL, NULL, self);
        return;
    }
    NMDevice *device = devices->pdata[0];
    NMDeviceType type = nm_device_get_device_type (device);

    if (state == NM_ACTIVE_CONNECTION_STATE_ACTIVATING)
    {
        switch (type)
        {
            case NM_DEVICE_TYPE_ETHERNET:
                applet_nm_set_icon (self, "network-wired-acquiring");
                cdos_applet_set_applet_tooltip (CDOS_APPLET(self), _("Connecting to the wired network..."));
                break;
            case NM_DEVICE_TYPE_MODEM:
                applet_nm_set_icon (self, "network-celluar-acquiring");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connecting to the celluar network..."));
                break;
            case NM_DEVICE_TYPE_WIFI:
                applet_nm_set_icon (self, "network-wireless-acquiring");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connecting to the wireless network..."));
                break;
            /*
            case DeviceVPN:
                applet_nm_set_icon (self, "network-vpn-acquiring");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connecting to the VPN network..."));
                break;
            */
            default:
                applet_nm_set_icon (self, "network-wired-acquiring");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connecting to the wired network..."));
                break;
        }
    }
    else if (state == NM_ACTIVE_CONNECTION_STATE_ACTIVATED)
    {
        NMConnectivityState connectivity;
        //connectivity = nm_client_get_connectivity (self->client);
        connectivity = nm_client_check_connectivity (self->client, NULL, NULL);
        gboolean is_portal = FALSE;
        gboolean is_global = FALSE;
        /*
         * portal: 因特网连接被网关劫持，用户需要在浏览器中向网关进行身份验证。
         * limited: 主机已经连接到网络，但是没用完全的因特网访问权限，而且也
         *          没用检测到中间网关
         * global: 已连接到因特网
         */
        if (connectivity == NM_CONNECTIVITY_PORTAL)
            is_portal = TRUE;
        else if (connectivity == NM_CONNECTIVITY_LIMITED)
            is_global = FALSE;
        else if (connectivity == NM_CONNECTIVITY_FULL ||
                 connectivity == NM_CONNECTIVITY_UNKNOWN)
            is_global = TRUE;
        self->connectivity = (is_global << 1) | is_portal;
        if (type != NM_DEVICE_TYPE_WIFI)
        {
            if (self->active_access_point && NM_IS_ACCESS_POINT (self->active_access_point))
            {
                g_signal_handlers_disconnect_by_func (self->active_access_point, applet_nm_ap_strength_changed, self);
                self->active_access_point = NULL;
            }
        }
        switch (type)
        {
            case NM_DEVICE_TYPE_ETHERNET:
                if (is_global)
                    applet_nm_set_icon (self, "network-wired");
                else if (is_portal || !is_global)
                {
                    applet_nm_set_icon (self, "network-wired-no-route");
                }
                /*
                else
                    applet_nm_set_icon (self, "network-error");
                */
                cdos_applet_set_applet_tooltip (CDOS_APPLET(self), _("Connected to the wired network"));
                break;
            case NM_DEVICE_TYPE_MODEM:
                applet_nm_set_icon (self, "network-celluar-connected");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connected to the celluar network"));
                break;
            case NM_DEVICE_TYPE_WIFI:
            {
                const char *path = nm_active_connection_get_connection (mc);
                if (path)
                {
                    NMRemoteConnection *r_con = nm_remote_settings_get_connection_by_path (self->settings, path);
                    if (r_con)
                    {
                        NMSettingIP4Config *ip4_config = nm_connection_get_setting_ip4_config ((NMConnection *)r_con);
                        const char *method = nm_setting_ip4_config_get_method (ip4_config);
                        if (g_strcmp0 (method, NM_SETTING_IP4_CONFIG_METHOD_SHARED) == 0)
                        {
                            applet_nm_set_icon (self, "network-wireless-hotspot");
                            cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Hotspot Active"));
                            return;
                        }
                    }
                }
                //if (device)
                {
                    NMDeviceWifi *wdev = NM_DEVICE_WIFI (device);
                    NMAccessPoint *ap = nm_device_wifi_get_active_access_point (wdev);
                    NM80211Mode mode = nm_device_wifi_get_mode (wdev);
                    if (!ap)
                    {
                        if (mode != NM_802_11_MODE_ADHOC)
                        {
                            g_warning (_("An active wireless connection, in infrastructure mode, involves no access point?"));
                            //break;
                        }
                        if (is_global)
                            applet_nm_set_icon (self, "network-wireless-connected");
                        else
                            applet_nm_set_icon (self, "network-wireless-no-route");
                        cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connected to the wireless network"));
                    } else 
                        applet_nm_set_ap_icon (self, ap);
                }
                break;
            }
            /*
            case DeviceVPN:
                applet_nm_set_icon (self, "network-vpn");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connected to the VPN network"));
                break;
            */
            default:
                applet_nm_set_icon (self, "network-wired");
                cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Connected to the wired network"));
                break;
        }
    }
    else {
        applet_nm_set_icon(self, "network-offline");
        cdos_applet_set_applet_tooltip (CDOS_APPLET(self), _("No connection"));
    }
}

static void
cdos_nm_my_applet_on_applet_removed_from_panel()
{
    //TODO:...
    //外部函数，暂时未实现
}

static void
applet_nm_on_connections_read (NMRemoteSettings* settings, AppletNM *self)
{
    applet_nm_read_connections (self);
    applet_nm_read_devices (self);
    applet_nm_sync_nm_state (NULL, NULL, self);

    if (!self->inited)
    {
        self->inited = TRUE;
        g_signal_connect (self->client,
                         "notify::manager-running",
                         G_CALLBACK (applet_nm_sync_nm_state),
                         self);
        g_signal_connect (self->client,
                         "notify::networking-enabled",
                         G_CALLBACK (applet_nm_sync_nm_state),
                         self);
        g_signal_connect (self->client,
                         "notify::state",
                         G_CALLBACK (applet_nm_sync_nm_state),
                         self);
        g_signal_connect (self->client,
                          "notify::active-connections",
                          /* G_CALLBACK (applet_nm_update_vpn_active_connections), */
                          G_CALLBACK (applet_nm_sync_active_connections),
                          self);
        g_signal_connect (self->client,
                          "notify::connectivity",
                          G_CALLBACK (applet_nm_sync_connectivity),
                          self);
        g_signal_connect (self->client,
                          "notify::primary-connection",
                          G_CALLBACK (applet_nm_sync_connection),
                          self);
        g_signal_connect (self->client,
                          "notify::activating-connection",
                          G_CALLBACK (applet_nm_sync_activating_connection),
                          self);
        g_signal_connect (self->client,
                         "device-added",
                         G_CALLBACK (applet_nm_device_added),
                         self);
        g_signal_connect (self->client,
                         "device-removed",
                         G_CALLBACK (applet_nm_device_removed),
                         self);
        g_signal_connect (self->settings,
                         "new-connection",
                         G_CALLBACK (applet_nm_new_connection),
                         self);
    }
}

static void
applet_nm_init (AppletNM *self)
{
    self->devices = g_new0(AppletNMDeviceObj, DeviceCount);
    self->base_class = applet_nm_device_class_new ();
    self->ethernet_class = applet_nm_device_wired_class_new (self);
    self->wifi_class = applet_nm_device_wireless_class_new (self);
    self->client = NULL;
    self->settings = NULL;
    self->currentIconName = NULL;
}

static void 
applet_nm_try_init (AppletNM *self)
{
    if (!self->client || !self->settings)
        return;

    /* Wired Device */
    self->devices[DeviceWired].section = cdos_popup_menu_section_new ();
    self->devices[DeviceWired].devices = NULL;
    self->devices[DeviceWired].item = (CdosPopupSwitchMenuItem *)applet_nm_wired_section_title_menu_item_new(_("Wired"));
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->devices[DeviceWired].section),
            CDOS_POPUP_BASE_MENU_ITEM (self->devices[DeviceWired].item), -1);

    //clutter_actor_hide (CLUTTER_ACTOR (cdos_popup_menu_base_get_box (CDOS_POPUP_MENU_BASE (self->devices[DeviceWired].section))));
    cdos_popup_menu_base_add_actor (CDOS_POPUP_MENU_BASE (self->menu),
            (ClutterActor *)self->devices[DeviceWired].section->actor);

    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->menu),
                                       CDOS_POPUP_BASE_MENU_ITEM (cdos_popup_separator_menu_item_new ()), -1);

    /* Wireless Device */
    self->devices[DeviceWireless].section = cdos_popup_menu_section_new ();
    self->devices[DeviceWireless].item = (CdosPopupSwitchMenuItem *)make_toggle_item (self, DeviceWireless, _("Wireless"));
    cdos_popup_menu_base_add_menu_item ((CdosPopupMenuBase *)self->devices[DeviceWireless].section,
            (CdosPopupBaseMenuItem *)self->devices[DeviceWireless].item, -1);
    cdos_popup_menu_base_add_actor ((CdosPopupMenuBase *)self->menu,
                (ClutterActor *)self->devices[DeviceWireless].section->actor);
    cdos_popup_menu_base_add_actor ((CdosPopupMenuBase *)self->menu,
            (ClutterActor *)self->devices[DeviceWireless].section->actor);


    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->menu),
                                       CDOS_POPUP_BASE_MENU_ITEM (cdos_popup_separator_menu_item_new ()), -1);

    //open the network settings in the panel
    CdosPopupMenuItem *network_settings =
            CDOS_POPUP_MENU_ITEM (cdos_popup_menu_base_add_settings_action (CDOS_POPUP_MENU_BASE (self->menu),
                        _("Network Settings"), "network"));
    cdos_popup_menu_item_set_label_align(network_settings,ST_ALIGN_MIDDLE,0);
    StLabel *network_settings_label = cdos_popup_menu_item_get_label(network_settings);
    st_widget_add_style_class_name (ST_WIDGET(network_settings_label),"popup-menu-item-label-link");
    /*
    CdosPopupMenuItem *network_connections = 
            cdos_popup_menu_base_add_action (CDOS_POPUP_MENU_BASE (self->menu), _("Network Connections"),
                        cdos_add_action_util_callback, NULL);

    clutter_actor_set_margin_left (CLUTTER_ACTOR (cdos_popup_menu_item_get_label (network_connections)), (gfloat)20);
    */
    applet_nm_on_connections_read (self->settings, self);
}

static void 
new_client_finish (NMClient *client,
                   GAsyncResult *res,
                   AppletNM *self)
{
    self->client = nm_client_new_finish (res, NULL);
    applet_nm_try_init (self);
}

static void 
new_remote_settings_finish (NMRemoteSettings *settings,
                            GAsyncResult *res,
                            AppletNM *self)
{
    self->settings = nm_remote_settings_new_finish (res, NULL);
    applet_nm_try_init (self);
    /*
    self->connectionsReadId = g_signal_connect (self->settings, 
                                                "connections-read",
                                                G_CALLBACK (applet_nm_on_connections_read),
                                                self);
    */
}

static void
applet_nm_init_after (AppletNM *self,
                      MetaData *metadata,
                      guint orientation,
                      guint panel_height)
{

    //StBoxLayout *left = cdos_panel_get_leftbox ((CdosPanel *)cdos_panel);
    //st_container_add_actor ((ClutterContainer *)left, (ClutterActor *)CDOS_APPLET(self)->actor);

    self->metadata = metadata;
    CdosSystrayManager *systray = cdos_monkey_plugin_get_systray_manager (monkey_plugin_get ());
    cdos_systray_manager_register_role (systray, "network", "network@cdos.org");

    self->menuManager = cdos_applet_get_menu_manager((CdosApplet *)self);
    self->menu = cdos_applet_get_popup_menu((CdosApplet *)self);

    st_widget_add_style_class_name(ST_WIDGET(self->menu->actor),"nm-popup-menu");
    //st_container_add_actor ((ClutterContainer *)ui_group, (ClutterActor *)self->menu->actor);
    //StContainer *menu = (StContainer *)((CdosPopupMenuBase *)self->menu)->box;
    cdos_popup_menu_manager_add_menu (self->menuManager, CDOS_POPUP_MENU_BASE (self->menu), -1);

    applet_nm_set_icon (self, "network-offline");
    self->statusSection = cdos_popup_menu_section_new ();
    PopupBaseMenuItemInit *params = g_new0 (PopupBaseMenuItemInit, 1);
    params->style_class = "popup-inactive-menu-item";
    params->reactive    = 0;
    self->statusItem    = cdos_popup_menu_item_new();
	cdos_popup_menu_item_set_label(self->statusItem, _("Status"));

	cdos_popup_base_menu_item_set_params((CdosPopupBaseMenuItem *)self->statusItem, params);
    g_free(params);
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->statusSection),
                                       (CdosPopupBaseMenuItem *)self->statusItem, -1);

    cdos_popup_menu_base_add_action (CDOS_POPUP_MENU_BASE (self->statusSection),
                                     _("Enable networking"),
                                     (GCallback)applet_nm_add_action_callback,
                                     self);
    clutter_actor_hide ((ClutterActor *)cdos_popup_menu_base_get_box (CDOS_POPUP_MENU_BASE (self->statusSection)));

    cdos_popup_menu_base_add_actor ((CdosPopupMenuBase *)self->menu, 
            (ClutterActor *)CDOS_POPUP_MENU_BASE(self->statusSection)->box);

    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (self->menu),
                                       CDOS_POPUP_BASE_MENU_ITEM (cdos_popup_separator_menu_item_new()),
                                       -1);

    nm_client_new_async (NULL, (GAsyncReadyCallback)new_client_finish, self);
    self->bus = dbus_g_bus_get (DBUS_BUS_SYSTEM, NULL);
    if (!self->bus)
        return;
    DBusConnection *bus_conn = dbus_g_connection_get_connection (self->bus);
    dbus_connection_set_exit_on_disconnect (bus_conn, 0);
    nm_remote_settings_new_async (self->bus,
                                  NULL,
                                  (GAsyncReadyCallback)new_remote_settings_finish,
                                  self);
}

static AppletNM *
applet_nm_new (MetaData *metadata, guint orientation, guint panel_height )
{
    AppletNM *self;
    self = g_object_new (APPLET_TYPE_NM, NULL);
    applet_nm_init_after (self, metadata, orientation, 55);

    return self;
}

/**
 * applet_network_main:
 *
 * @orientation: 文本的方向
 * @panel_height：panel的方向
 * @mata: 新的Meta结构体
 *
 * 创建一个新的applet
 * Returns:新创建的 #CdosApplet
 *
 * Since:1.0.0
 */

CdosApplet *
applet_network_main (MetaData *meta, guint orientation,  guint panel_height)
{
   AppletNM *appletnm;
   appletnm = applet_nm_new (meta, orientation, panel_height);
    cdos_iconapplet_reset_for_systray(CDOS_ICONAPPLET(appletnm));
   return CDOS_APPLET(appletnm);
}

GSList *
applet_nm_get_all_connections (AppletNM *applet)
{
    GSList *connections = nm_remote_settings_list_connections (applet->settings);
    GSList *iter, *next;
    for (iter = connections; iter; iter = next)
    {
        NMConnection *conn = iter->data;
        next = iter->next;
        NMSettingConnection *s_con = nm_connection_get_setting_connection (conn);
        if (s_con && nm_setting_connection_get_master (s_con))
            connections = g_slist_delete_link (connections, iter);
    }
    return connections;
}


//_notifyForDevice:
static void
applet_nm_notify_for_device_signal_destroy (CdosNotification *notification, guint reason, AppletNMDevice *device)
{
    cdos_notification_destroy (device->notification, 0);
    device->notification = NULL;
}
static void
applet_nm_notify_for_device (AppletNMDevice *device,
                                  gchar *iconName, gchar *title, gchar *text, gboolean urgency)
{

    if (device->notification)
    {
        cdos_notification_destroy (device->notification, EXPIRED);
    }
    AppletNM *applet = device->data;
    applet_nm_ensure_source (applet);
    StIcon * myAppletIcon = ST_ICON (st_icon_new());
    st_icon_set_icon_name (myAppletIcon, iconName);
    st_icon_set_icon_type (myAppletIcon, ST_ICON_SYMBOLIC);
    //icon_size: this._source.ICON_SIZE...
    st_icon_set_icon_size (myAppletIcon, SOURCE_ICON_SIZE);
    Params *myAppletParam = g_new0 (Params, 1);
    myAppletParam->icon = myAppletIcon;
    device->notification = cdos_notification_new (CDOS_SOURCE (applet->source), title, text, myAppletParam);
    cdos_notification_set_urgency (device->notification, urgency);
    cdos_notification_set_transient (device->notification, TRUE);
    g_signal_connect (device->notification, "destroy",
                      G_CALLBACK (applet_nm_notify_for_device_signal_destroy), device);
    cdos_source_notify (CDOS_SOURCE (applet->source), device->notification);
}

//_deviceAdded
void
applet_nm_activation_failed (AppletNMDevice *device, guint reason)//...
{
    applet_nm_notify_for_device (device, "network-error", _("Connection failed"),
                                 _("Activation of network connection failed"),2);

}
