/** 
 * Copyright (c) 2016-2018 CPU and Fundamental Software Research Center, CAS
 *  
 * This software is published by the license of CPU-OS License，you can use and 
 * distibuted this software under this License. See CPU-OS Liense for more detail.
 *
 * You should have receive a copy of CPU-OS License. If not, please contact us 
 * by email <support_os@cpu-os.ac.cn>  
 * 
 **/
#include <girepository.h>
#include <gio/gio.h>
#include <math.h>
#include <glib.h>
#include <glib/gprintf.h>
#include <upower.h>

#include "applet-power.h"
#include "cdos-monkey-plugin.h"
#include "cdos-systray.h"

#define POWER_CHARGING_EFFECT_LOOP 5

/**
 * SECTION: cdos-power
 * @Title: AppletPower
 * @Short_description: 电源管理器
 * 
 * AppletPower 通过 settings daemon 的 power 模块来监听电源的状态变化
 * 比如拔出电源、电量变化等信号会触发 GUI 部分，从而显示变化。
 */

#define 	CDOS_NAME	"org.freedesktop.UPower" 
#define 	CDOS_PATH	"/org/freedesktop/UPower/devices/DisplayDevice" 
#define 	CDOS_INTERFACE		"org.freedesktop.UPower.Device" 
#define     POWER_SETTINGS_SCHEMAS "org.cdos.settings-daemon.plugins.power"

#define BATTERY_LOW_EFFECT_TIMES 3
G_DEFINE_TYPE (AppletPower, applet_power, CDOS_TYPE_ICONAPPLET);

/*
 * settings daemon 中接口编译老是不起作用，干脆在 power 中不使用 settings daemon 的 dbus
 * 直接使用 org.freedesktop.UPower 的接口
 * d-feet 可以查询的 object path 有好几个
 * {'DaemonVersion': '0.99.4',
 *  'LidIsClosed': False,
 *  'LidIsPresent': True,
 *  'OnBattery': True} 这几个是 org.freedesktop.UPower 的属性。
 *  其中 OnBattery 可以用于判断当前是否使用电池
 *
 * {'Capacity': 0.0,
 *  'Energy': 34.78,
 *  'EnergyEmpty': 0.0,
 *  'EnergyFull': 39.46,
 *  'EnergyFullDesign': 0.0,
 *  'EnergyRate': 14.699,
 *  'HasHistory': False,
 *  'HasStatistics': False,
 *  'IconName': 'battery-full-symbolic',  图标名称
 *  'IsPresent': True,	
 *  'IsRechargeable': False,
 *  'Luminosity': 0.0,
 *  'Model': '',
 *  'NativePath': '',
 *  'Online': False,
 *  'Percentage': 88.0,		剩余电量的百分比
 *  'PowerSupply': True,	
 *  'Serial': '',
 *  'State': 2,			电池状态：0 未知 1 充电、2 放电、3 空liao、4 充满、 5 暂停充电、 6 暂停放电
 *  'Technology': 0,
 *  'Temperature': 0.0,
 *  'TimeToEmpty': 8518,	电池还可以坚持多久
 *  'TimeToFull': 0,		还需要多久电池充满
 *  'Type': 2,		 
 *  'UpdateTime': 1494825678,
 *  'Vendor': '',
 *  'Voltage': 0.0,
 *  'WarningLevel': 1}  这个是 org.freedesktop.UPower.Device 的属性
 *
 * */

struct _AppletPowerPrivate
{
    GDBusConnection *connection;
    GDBusProxy *proxy;
    CdosPopupMenu *menu;
    CdosPopupMenuItem *action;
    MetaData *metadata;
    GSettings *settings;
    const gchar *icon_name;
    gdouble battery_percentage;
    UpDeviceState state;
    UpDeviceKind kind;

    //低电量告警：
    gboolean battery_low;
    gboolean battery_low_effect_flag;
    guint flag_num;
    guint battery_low_timeout_id;
    gint battery_low_threshold;

    //充电动态过程：
    guint battery_charging_timeout_id;
    //循环次数
    guint battery_charging_loop_times;
    //当次循环标识，刷新图标
    guint battery_charging_loop_flag;
};

static void
power_devices_changed (GDBusProxy *proxy,
                       GVariant   *changed_properties,
                       GStrv       invalidated_properties,  
                       gpointer    user_data);
static gboolean
on_action_clicked (ClutterActor *actor, 
                   ClutterEvent *event, 
                   gboolean keep, 
                   gpointer user_data);
static void
applet_power_low_battery (AppletPower *self);
static gboolean
applet_power_low_battery_effect (AppletPower *self);
static gboolean
applet_power_charging_effect (AppletPower *self);
static void
applet_power_settings_on_changed (GSettings *settings, gchar *key, gpointer user_data);
static const gchar *
power_get_icon_name (gdouble percent, UpDeviceState state);
static void
power_update_applet_icon (AppletPower *self);

static void
applet_power_battery_missing (AppletPower *self)
{
    AppletPowerPrivate *priv = self->priv;
    priv->icon_name = "battery-missing-symbolic";
    power_update_applet_icon (self);
    cdos_applet_set_applet_tooltip (CDOS_APPLET (self), _("Something wrong happened!"));
}

static void
applet_power_connection_dbus (AppletPower *self)
{
    AppletPowerPrivate *priv = self->priv;
    GError *error = NULL;
    priv->connection = g_bus_get_sync (G_BUS_TYPE_SYSTEM, NULL, &error);
    if (error)
    {
        goto onerror;
    }
    priv->proxy = g_dbus_proxy_new_sync (priv->connection,
                              G_DBUS_PROXY_FLAGS_NONE,
                              NULL,
                              CDOS_NAME,
                              CDOS_PATH,
                              CDOS_INTERFACE,
                              NULL,
                              &error);
    if (error)
    {
        goto onerror;
    }

    g_signal_connect (priv->proxy, 
                      "g-properties-changed", 
                      G_CALLBACK (power_devices_changed), 
                      self);
    //防止重启后，无法完成初始化；
    gchar **property_names = g_dbus_proxy_get_cached_property_names (priv->proxy);
    if (property_names)
    {
        power_devices_changed (priv->proxy, NULL, NULL, self);
        g_strfreev (property_names);
    }

onerror:
    if (error)
    {
        g_warning ("\n%s (%d):%s;\n", __func__, __LINE__, error->message);
        g_error_free (error);
        error = NULL;
        applet_power_battery_missing (self);
    }
}

static void
applet_power_dispose (GObject *object)
{
    AppletPower *self = APPLET_POWER (object);
    AppletPowerPrivate *priv = self->priv;
    if (priv->battery_low_timeout_id > 0)
    {
        g_source_remove (priv->battery_low_timeout_id);
        priv->battery_low_timeout_id = 0;
    }

    if (priv->battery_charging_timeout_id)
    {
        g_source_remove (priv->battery_charging_timeout_id);
        priv->battery_charging_timeout_id = 0;
    }
    
    g_clear_object (&priv->proxy);
    g_clear_object (&priv->connection);
    g_clear_object (&priv->action);
    g_clear_object (&priv->settings);

    G_OBJECT_CLASS (applet_power_parent_class)->dispose (object);
}

static void
applet_power_finalize (GObject *object)
{
    AppletPower *self = APPLET_POWER (object);
    AppletPowerPrivate *priv = self->priv;

    G_OBJECT_CLASS (applet_power_parent_class)->finalize (object);
}

static void
applet_power_register_role (AppletPower *self, gchar *role_name)
{
    AppletPowerPrivate *priv = self->priv;
    MonkeyPlugin *plugin = monkey_plugin_get ();
    CdosSystrayManager *systray_m = cdos_monkey_plugin_get_systray_manager (plugin);
    cdos_systray_manager_register_role (systray_m, role_name, priv->metadata->name);
}

static gboolean
applet_power_on_applet_clicked (gpointer data)
{
    AppletPower *self = APPLET_POWER (data);
    AppletPowerPrivate *priv = self->priv;

    cdos_popup_menu_base_toggle (CDOS_POPUP_MENU_BASE (priv->menu));
    return FALSE;
}

static void
applet_power_class_init (AppletPowerClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
    CdosAppletClass *parent_class = CDOS_APPLET_CLASS (klass);

    object_class->dispose = applet_power_dispose;
    object_class->finalize = applet_power_finalize;

    g_type_class_add_private (klass, sizeof (AppletPowerPrivate ));
}

static void
applet_power_init (AppletPower *self)
{
    AppletPowerPrivate *priv;
    priv = self->priv = G_TYPE_INSTANCE_GET_PRIVATE ((self), APPLET_TYPE_POWER, AppletPowerPrivate);

    priv->battery_low_timeout_id = 0;

    priv->battery_charging_timeout_id = 0;
    priv->battery_charging_loop_times = 0;
    priv->battery_charging_loop_flag = 0;

    priv->state = UP_DEVICE_STATE_UNKNOWN;
    priv->kind = UP_DEVICE_KIND_UNKNOWN;

    priv->settings = g_settings_new (POWER_SETTINGS_SCHEMAS);
    g_signal_connect (priv->settings,
                      "changed::percentage-low",
                      G_CALLBACK (applet_power_settings_on_changed),
                      self);
    priv->battery_low_threshold = g_settings_get_int (priv->settings, "percentage-low");

    priv->menu = cdos_applet_get_popup_menu (CDOS_APPLET (self));
    priv->action = cdos_popup_menu_item_new ();
    cdos_popup_menu_item_set_label_align (priv->action, ST_ALIGN_MIDDLE, 0);
    cdos_popup_menu_item_set_label (priv->action, _("Power Manager"));
    StLabel *label = cdos_popup_menu_item_get_label (priv->action);
    st_widget_add_style_class_name (ST_WIDGET (label), "popup-menu-item-label-link");
    g_signal_connect (priv->action, 
                      "activate", 
                      G_CALLBACK (on_action_clicked), 
                      self);
    cdos_popup_menu_base_add_menu_item (CDOS_POPUP_MENU_BASE (priv->menu), 
                                        CDOS_POPUP_BASE_MENU_ITEM (priv->action), 
                                        -1);
    cdos_popup_menu_close (CDOS_POPUP_MENU_BASE (priv->menu), FALSE);
    clutter_actor_hide (CLUTTER_ACTOR (CDOS_APPLET (self)->actor));

    priv->battery_low = FALSE;
    applet_power_connection_dbus (self);
}

static void
applet_power_settings_on_changed (GSettings *settings, gchar *key, gpointer user_data)
{
    AppletPower *self = APPLET_POWER (user_data);
    AppletPowerPrivate *priv = self->priv;
    gint percentage_low = g_settings_get_int (settings, "percentage-low");
    priv->battery_low_threshold = percentage_low;

    if (priv->battery_low == FALSE && 
        priv->battery_percentage <= priv->battery_low_threshold)
    {
        priv->battery_low = TRUE;
    } 
    else if (priv->battery_low == TRUE && 
             priv->battery_percentage > priv->battery_low_threshold) 
    {
        priv->battery_low = FALSE;

    }
    power_devices_changed (priv->proxy, NULL, NULL, self);
}

static gboolean
applet_power_low_battery_effect (AppletPower *self)
{
    AppletPowerPrivate *priv = self->priv;
    ClutterActor *actor = cdos_iconapplet_get_applet_icon (CDOS_ICONAPPLET (self));
    if (priv->battery_low && priv->flag_num <= BATTERY_LOW_EFFECT_TIMES)
    {
        if (priv->battery_low_effect_flag == TRUE)
        {
            clutter_actor_hide (actor);
            priv->battery_low_effect_flag = FALSE;
        } else {
            priv->flag_num++;
            clutter_actor_show (actor);
            priv->battery_low_effect_flag = TRUE;
        }
        return TRUE;
    } else {
        clutter_actor_show (actor);
        priv->battery_low_effect_flag = FALSE;
    }
    priv->battery_low_timeout_id = 0;
    return FALSE;
}

static void
applet_power_low_battery (AppletPower *self)
{
    AppletPowerPrivate *priv = self->priv;
    if (priv->battery_charging_timeout_id > 0)
    {
        g_source_remove (priv->battery_charging_timeout_id);
        priv->battery_charging_timeout_id = 0;
    }

    priv->flag_num = 0;
    if (priv->battery_low_timeout_id == 0)
    {
        priv->battery_low_timeout_id = \
            g_timeout_add (500, 
                           (GSourceFunc)applet_power_low_battery_effect, 
                           self);
    }
}

static gboolean
on_action_clicked (ClutterActor *actor, ClutterEvent *event, gboolean keep, gpointer user_data)
{
    cdos_util_spawn_command_line ("cdos-settings power");
    return FALSE;
}

static const gchar *
power_get_icon_name (gdouble percent, UpDeviceState state)
{
    if (state == UP_DEVICE_STATE_EMPTY) 
    {
        return "battery-empty-symbolic";
    } 
    else if (state == UP_DEVICE_STATE_FULLY_CHARGED) 
    {
        if (percent == 100.0)
        {
            return "battery-full-charged-symbolic";
        } else {
            return "battery-full-charging-symbolic";
        }
    }

    if (percent > 0 && percent <= 10.0)
    {
        if (state == UP_DEVICE_STATE_DISCHARGING)
            return "battery-caution-symbolic";
        else if (state == UP_DEVICE_STATE_CHARGING)
            return "battery-low-charging-symbolic";
    }
    else if (percent > 10.0 && percent < 20.0)
    {
        if (state == UP_DEVICE_STATE_DISCHARGING)
            return "battery-low-symbolic";
        else if (state == UP_DEVICE_STATE_CHARGING)
            return "battery-low-charging-symbolic";
    }
    else if (percent >= 20.0 && percent < 40.0)
    {
        if (state == UP_DEVICE_STATE_DISCHARGING)
            return "battery-poor-symbolic";
        else if (state == UP_DEVICE_STATE_CHARGING)
            return "battery-poor-charging-symbolic";
    }
    else if (percent >= 40.0 && percent < 60.0)
    {
        if (state == UP_DEVICE_STATE_DISCHARGING)
            return "battery-good-symbolic";
        else if (state == UP_DEVICE_STATE_CHARGING)
            return "battery-good-charging-symbolic";
    }
    else if (percent >= 60.0 && percent < 80.0)
    {
        if (state == UP_DEVICE_STATE_DISCHARGING)
            return "battery-better-symbolic";
        else if (state == UP_DEVICE_STATE_CHARGING)
            return "battery-better-charging-symbolic";
    }
    else if (percent >= 80.0 && percent < 90.0)
    {
        if (state == UP_DEVICE_STATE_DISCHARGING)
            return "battery-excellent-symbolic";
        else if (state == UP_DEVICE_STATE_CHARGING)
            return "battery-excellent-charging-symbolic";
    }
    else if (percent >= 90.0 && percent < 95.0)
    {
        if (state == UP_DEVICE_STATE_DISCHARGING)
            return "battery-almost-full-symbolic";
        else if (state == UP_DEVICE_STATE_CHARGING)
            return "battery-almost-full-charging-symbolic";
    } 
    else if (percent >= 95.0)
    {
        if (state == UP_DEVICE_STATE_DISCHARGING)
            return "battery-full-symbolic";
        else if (state == UP_DEVICE_STATE_CHARGING)
            return "battery-full-charging-symbolic";
    }
    return "battery-missing-symbolic";
}

static void
power_update_applet_icon (AppletPower *self)
{
    AppletPowerPrivate *priv = self->priv;
    if (priv->icon_name)
    {
        cdos_iconapplet_set_icon_symbolic_name (CDOS_ICONAPPLET (self), priv->icon_name);
        clutter_actor_show (CLUTTER_ACTOR (CDOS_APPLET (self)->actor));
        ClutterActor *applet_icon = cdos_iconapplet_get_applet_icon (CDOS_ICONAPPLET (self));
        st_widget_add_style_class_name (ST_WIDGET (applet_icon),"applet-systray-icon");
    } else {
        cdos_popup_menu_close (CDOS_POPUP_MENU_BASE (priv->menu), FALSE);
        clutter_actor_hide (CLUTTER_ACTOR (CDOS_APPLET (self)->actor));
    }

}

static gboolean
applet_power_charging_effect (AppletPower *self)
{
    AppletPowerPrivate *priv = self->priv;
    gboolean result = G_SOURCE_CONTINUE;
    

    gdouble percentage = priv->battery_percentage;
    guint flag = priv->battery_charging_loop_flag;
    gdouble new_percentage = percentage + flag * 20;

    //动画过程中可能出现插拔现象，造成紊乱
    if (percentage >= 95.0 || priv->state != UP_DEVICE_STATE_CHARGING)
    {
        priv->battery_charging_timeout_id = 0;
        result = G_SOURCE_REMOVE;
        new_percentage = percentage;
        goto updateicon;
    }

    priv->battery_charging_loop_flag++;
    //重置，并终止动态过程
    if (priv->battery_charging_loop_times >= POWER_CHARGING_EFFECT_LOOP)
    {
        new_percentage = percentage; 
        result = G_SOURCE_REMOVE;
        priv->battery_charging_timeout_id = 0;
    } else {
        new_percentage = new_percentage < 100 && new_percentage >= 90 ? 90 : new_percentage;
        if (new_percentage >= 100.0)
        {
            new_percentage = 100;
            priv->battery_charging_loop_flag = 0;
            priv->battery_charging_loop_times++;
        }
    }

updateicon:
    priv->icon_name = power_get_icon_name (new_percentage, priv->state);
    power_update_applet_icon (self);
    return result;
}

//我们希望插入电源后，电源指示图标会有一段动态效果。
static void
applet_power_switch_charging (AppletPower *self)
{
    AppletPowerPrivate *priv = self->priv;
    
    if (priv->battery_low_timeout_id > 0)
    {
        g_source_remove (priv->battery_low_timeout_id);
        priv->battery_low_timeout_id = 0;

        ClutterActor *actor = cdos_iconapplet_get_applet_icon (CDOS_ICONAPPLET (self));
        clutter_actor_show (actor);
    }

    //必须进行重置，防止动画过程中插拔造成的紊乱
    if (priv->battery_charging_timeout_id > 0)
    {
        g_source_remove (priv->battery_charging_timeout_id);
        priv->battery_charging_timeout_id = 0;
    }

    //重置循环次数和标识
    priv->battery_charging_loop_times = 0;
    priv->battery_charging_loop_flag = 0;
    priv->battery_charging_timeout_id = \
        g_timeout_add (500, 
                       (GSourceFunc)applet_power_charging_effect, 
                       self);
}

static void
power_devices_changed (GDBusProxy *proxy,
                       GVariant   *changed_properties,
                       GStrv       invalidated_properties,
                       gpointer    user_data)
{
    AppletPower *self = APPLET_POWER (user_data);
    AppletPowerPrivate *priv = self->priv;
    // 电源类型
    GVariant *kind = g_dbus_proxy_get_cached_property (proxy, "Type");
    if (kind)
    {
        priv->kind = g_variant_get_uint32 (kind);
        g_variant_unref (kind);
    }
    if (priv->kind != UP_DEVICE_KIND_BATTERY)
    {
        clutter_actor_hide (CLUTTER_ACTOR (CDOS_APPLET (self)->actor));
        return ;
    }
    // 电量百分比
    GVariant *percentage = g_dbus_proxy_get_cached_property (proxy, "Percentage");
    // 电源状态
    GVariant *power_state = g_dbus_proxy_get_cached_property (proxy, "State");
    if (!percentage || !power_state)
    {
        applet_power_battery_missing (self);
        return ;
    }
    // 以上都是 GVariant 类型的数据，需要转换类型使用
    gdouble percent;
    guint state;
    percent = g_variant_get_double (percentage);
    state = g_variant_get_uint32 (power_state);

    priv->battery_percentage = percent;

    gint64 seconds = 0;
    GVariant *time_to_end = NULL;
    gchar *state_label = NULL;

    switch (state) 
    {
        case UP_DEVICE_STATE_CHARGING:
            // 正在充电，获取 TimeToFull, INT64, 单位秒
            time_to_end = g_dbus_proxy_get_cached_property (proxy, "TimeToFull");
            state_label = _("Charging ");
            if (time_to_end)
            {
                seconds = g_variant_get_int64 (time_to_end);
            }
            break;
        case UP_DEVICE_STATE_DISCHARGING:
            // 正在耗电，获取 TimeToEmpty, INT64, 单位秒
            time_to_end = g_dbus_proxy_get_cached_property (proxy, "TimeToEmpty");
            state_label = _("Discharging ");
            if (time_to_end)
            {
                seconds = g_variant_get_int64 (time_to_end);
            }
            if (priv->battery_percentage <= priv->battery_low_threshold)
            {
                priv->battery_low = TRUE;
                applet_power_low_battery (self);
            } else {
                priv->battery_low = FALSE;
            }
            
            break;
        case UP_DEVICE_STATE_EMPTY:
            state_label = _("EmptyBat");
            break;
        case UP_DEVICE_STATE_FULLY_CHARGED:
            state_label = _("FullBat");
            break;
        default:
            break;
    }

    gchar *tooltip = NULL;
    if (state == UP_DEVICE_STATE_CHARGING || state == UP_DEVICE_STATE_DISCHARGING)
    {
        guint minutes = round (seconds / 60);
        guint hour = round (minutes / 60);
        if (percent < 100.0)
        {
            if (hour > 0)
            {
                minutes = minutes % 60;
                tooltip = g_strdup_printf (_("%s %.1f%: \n Time left %d hour %d minutes"), state_label, percent, hour, minutes);
            } 
            else if (minutes > 0)
            {
                tooltip = g_strdup_printf (_("%s %.1f%: \n Time left %d minutes"), state_label, percent, minutes);
            } else {
                tooltip = g_strdup_printf ("%s %.1f%%", state_label, percent); 
            } 
        } else {
            tooltip = g_strdup_printf ("%s %.1f%%: \n %s", state_label, percent, _("FullBat"));
        }
        
    } 
    else if (state == UP_DEVICE_STATE_EMPTY || state == UP_DEVICE_STATE_FULLY_CHARGED)
    {
        if (percent < 100.0)
        {
            tooltip = g_strdup_printf ("%s %.1f%%", state_label, percent); 
        } else {
            tooltip = g_strdup (state_label);
        }
    }
    cdos_applet_set_applet_tooltip (CDOS_APPLET (self), tooltip);
    //更新图标，并避开充电过程的动画
    if (priv->battery_charging_timeout_id == 0)
    {
        priv->icon_name = power_get_icon_name (percent, state);
        power_update_applet_icon (self);
    }

    if (state == UP_DEVICE_STATE_CHARGING &&
        priv->state != UP_DEVICE_STATE_UNKNOWN && 
        state != priv->state)
    {
        applet_power_switch_charging (self);
    }
    priv->state = state;
    g_free (tooltip);
    g_variant_unref (percentage);
    g_variant_unref (power_state);
    if (time_to_end)
    {
        g_variant_unref (time_to_end);
    }
}

static AppletPower *
applet_power_new (MetaData *meta, guint orientation, guint panel_height){

    AppletPower *self = NULL;
    self = g_object_new (APPLET_TYPE_POWER, NULL);
    AppletPowerPrivate *priv = self->priv;
    priv->metadata = meta; 

    applet_power_register_role (self, "power");
    applet_power_register_role (self, "battery");

    return self;
}

/**
 * applets_power_main:
 *
 * 新建对象，applet manager 新建 power 模块对象，将该 applet 的图标添加到任务栏右侧
 *
 * Return value: #AppletPower 对象
 */
AppletPower *
applets_power_main (MetaData *meta, guint orientation, guint panel_height, gchar *instance_id){
    static AppletPower *power;
    if (power == NULL)
    {
        power = applet_power_new (meta, orientation, panel_height);
        cdos_iconapplet_reset_for_systray (CDOS_ICONAPPLET (power));
    }
    return power;
}
