#include "applet-place-device-info.h"
#include <gio/gio.h>

/**
 * SECTION: applet-place-device-info.c
 * @Title: CdosPlaceDeviceInfo
 * @short_description: 继承自CdosPlaceInfo，增加了一个mount信息。
 *
 * CdosPlaceDeviceInfo保存mount的基本信息
 */

struct _CdosPlaceDeviceInfoPrivate
{
        GMount *mount;
};

G_DEFINE_TYPE(CdosPlaceDeviceInfo, cdos_place_device_info, CDOS_TYPE_PLACE_INFO);

#define CDOS_PLACE_DEVICE_INFO_GET_PRIVATE(obj)    \
        (G_TYPE_INSTANCE_GET_PRIVATE((obj), CDOS_TYPE_PLACE_DEVICE_INFO, CdosPlaceDeviceInfoPrivate))

static void
cdos_place_device_info_remove_finish(GObject *object,
                                     GAsyncResult *res,
                                     gpointer info);

static void 
cdos_place_device_info_finalize(GObject *gobject);

static void
cdos_place_device_info_class_init(CdosPlaceDeviceInfoClass *klass);

static void 
cdos_place_device_info_init(CdosPlaceDeviceInfo *info);

static void 
cdos_place_device_info_init(CdosPlaceDeviceInfo *info)
{
        info->priv = CDOS_PLACE_DEVICE_INFO_GET_PRIVATE(info);
        info->priv->mount = NULL;
}

static void
cdos_place_device_info_class_init(CdosPlaceDeviceInfoClass *klass)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

        g_type_class_add_private(klass, sizeof(CdosPlaceDeviceInfoPrivate));

        gobject_class->finalize = cdos_place_device_info_finalize;
}

static void 
cdos_place_device_info_finalize(GObject *gobject)
{
        CdosPlaceDeviceInfoPrivate *priv = CDOS_PLACE_DEVICE_INFO(gobject)->priv;

        if (priv->mount)
        {
            g_object_unref(priv->mount);
        }
        
        G_OBJECT_CLASS(cdos_place_device_info_parent_class)->finalize(gobject);    
}

/**
 * cdos_place_device_info_new_with_params:
 * @mount: Mount Manager
 * 
 * 新建一个 #CdosPlaceDeviceInfo 类的实例。
 *
 * Returns: 返回一个 #CdosPlaceDeviceInfo 的实例。
 */
CdosPlaceDeviceInfo *
cdos_place_device_info_new_with_params(GMount *mount)
{
        GObject *gobject;
        gobject = g_object_new(CDOS_TYPE_PLACE_DEVICE_INFO, NULL);
        cdos_place_device_info_set_params(CDOS_PLACE_DEVICE_INFO(gobject), mount);

        return gobject;
}

/**
 * cdos_place_device_info_set_params:
 * @info: A #CdosPlaceDeviceInfo
 * @mount: #GMount 
 *
 * 设置info的mount Manager(挂载点管理器)。
 *
 */
CdosPlaceDeviceInfo *
cdos_place_device_info_set_params(CdosPlaceDeviceInfo *info,
                                  GMount *mount)
{
        CdosPlaceDeviceInfoPrivate *priv = info->priv;
        if (priv->mount) 
        {
            g_object_unref(priv->mount);
        }

        priv->mount = mount;
        g_object_ref(priv->mount);

        gchar *name = g_mount_get_name(mount);
        GFile *file  = g_mount_get_root(mount);
        gchar *uri = g_file_get_uri(file);
        gchar *id = g_strconcat("mount:", uri, NULL);      
        cdos_place_info_set_params(info, id, name, NULL, FALSE);

        g_free(name);
        g_object_unref(file);
        g_free(uri);
        g_free(id);

        return info;
}

/**
 * cdos_place_device_info_icon_factory:
 * @info: A #CdosPlaceDeviceInfo
 *
 * 通过获取info的mount成员,获取mount的图片
 *
 * Returns: #ClutterActor　图片对象
 */
ClutterActor *
cdos_place_device_info_icon_factory(CdosPlaceDeviceInfo *info,
                                    gint size)
{
        GIcon *icon = g_mount_get_icon(info->priv->mount);
        StTextureCache *cache = st_texture_cache_get_default();
        return  st_texture_cache_load_gicon(cache, NULL, icon, size);
}

/**
 * cdos_place_device_info_launch:
 * @info: A #CdosPlaceDeviceInfo
 * @params: 工作区的参数 
 *
 * 启动info中的mount
 *
 */
void
cdos_place_device_info_launch(CdosPlaceDeviceInfo *info,
                              CdosPlaceInfoParams *params)
{
        gchar *uri = g_file_get_uri(g_mount_get_root(info->priv->mount));
        GAppLaunchContext *context = cdos_place_info_make_launch_conext(params);
        g_app_info_launch_default_for_uri(uri, context, NULL);
}

/**
 * cdos_place_device_info_is_removable:
 * @info: A #CdosPlaceDeviceInfo
 *
 * 判断info的mount是否可以移除
 *
 * Returns: $gboolean True:可以移除 False:不可以移除
 */
gboolean 
cdos_place_device_info_is_removable(CdosPlaceDeviceInfo *info)
{
        if (NULL == info)
            return FALSE;

        return g_mount_can_unmount(info->priv->mount);
}

/**
 * cdos_place_device_info_remove:
 * @info: A #CdosPlaceDeviceInfo 
 *
 * 卸载info的mount
 *
 */
void
cdos_place_device_info_remove(CdosPlaceDeviceInfo *info)
{
        CdosPlaceDeviceInfoPrivate *priv = info->priv;
        if(!cdos_place_device_info_is_removable(info)) 
                return;

        if (g_mount_can_eject(priv->mount)) 
                g_mount_eject(priv->mount, G_MOUNT_UNMOUNT_NONE, NULL, 
                                cdos_place_device_info_remove_finish, info);
        else 
                g_mount_unmount(priv->mount, G_MOUNT_UNMOUNT_NONE, NULL, 
                                cdos_place_device_info_remove_finish, info);
}

static void
cdos_place_device_info_remove_finish(GObject *object,
                                     GAsyncResult *res,
                                     gpointer info)
{
        CdosPlaceDeviceInfoPrivate *priv=CDOS_PLACE_DEVICE_INFO(info)->priv;
        
        if (g_mount_can_eject(priv->mount)) {
            g_mount_eject_finish(priv->mount, res, NULL);
        } else {
            g_mount_unmount_finish(priv->mount, res, NULL);
        }
}


