#include "applet-place-info.h"

/**
 * SECTION: applet-place-info.c
 * @Title: CdosPlaceInfo
 * @short_description: place 的基本信息. place　大多数的时候是一个书签条目，
 * 但也可能是一个mount或volume,还有可能是一个目录就像/HOME 
 *
 * CdosPlaceInfo　保存place的基本信息
 */
 
struct _CdosPlaceInfoPrivate
{
        gchar       *id;               //place的id
        gchar       *name;             //place的名字
        gchar       *lowerName;        //将name字段中的大写转出小写
        gpointer     iconFactory;       //回调函数用以创建一个icon texture
        gpointer     launch;            //回调函数用来打开一个书签或者目录等　
};

G_DEFINE_TYPE(CdosPlaceInfo, cdos_place_info, G_TYPE_OBJECT);

#define CDOS_PLACE_INFO_GET_PRIVATE(obj)    \
        (G_TYPE_INSTANCE_GET_PRIVATE((obj), CDOS_TYPE_PLACE_INFO, CdosPlaceInfoPrivate))

static void 
cdos_place_info_init(CdosPlaceInfo *info);

static void
cdos_place_info_class_init(CdosPlaceInfoClass *klass);

static void 
cdos_place_info_finalize(GObject *gobject);

static gint 
index_of(const gchar *haystack, const gchar *needle);

static void 
cdos_place_info_init(CdosPlaceInfo *info)
{
        info->priv = CDOS_PLACE_INFO_GET_PRIVATE(info);

        info->priv->id = NULL;
        info->priv->name = NULL;
        info->priv->lowerName = NULL;
        info->priv->iconFactory = NULL;
        info->priv->launch = NULL;
}

static void
cdos_place_info_class_init(CdosPlaceInfoClass *klass)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
        GParamSpec *psec;

        g_type_class_add_private(klass, sizeof(CdosPlaceInfoPrivate));
        gobject_class->finalize = cdos_place_info_finalize;
}

static void 
cdos_place_info_finalize(GObject *gobject)
{
        CdosPlaceInfoPrivate *priv = CDOS_PLACE_INFO(gobject)->priv;

        if (priv->id) {
                g_free(priv->id);
        }

        if (priv->name) {
                g_free(priv->name);
        }

        if (priv->lowerName) {
                g_free(priv->lowerName);
        }

        G_OBJECT_CLASS(cdos_place_info_parent_class)->finalize(gobject);
}

/**
 * cdos_place_info_new_with_params:
 * @id: place的id 
 * @name: place的名字
 * @iconFactory: 回调函数 
 * @launch: 回调函数
 * 
 * 新建一个 #CdosPlaceInfo 类的实例。
 *
 * Returns: 返回一个 #CdosPlaceInfo 的实例。
 */
CdosPlaceInfo *
cdos_place_info_new_with_params(gchar   *id, 
                                gchar   *name, 
                                gpointer iconFactory,
                                gpointer launch)
{
        GObject *gobject;
        gobject = g_object_new(CDOS_TYPE_PLACE_INFO, NULL);
        cdos_place_info_set_params(CDOS_PLACE_INFO(gobject), id, name, iconFactory, launch);
        return CDOS_PLACE_INFO(gobject);
}

/**
 * cdos_place_info_set_params:
 * @info: A#CdosPlaceInfo
 * @id: place的id 
 * @name: place的名字
 * @iconFactory: 回调函数 
 * @launch: 回调函数
 * 
 * 为一个#CdosPlaceInfo对象设置必备属性
 *
 * Returns: 返回一个 #CdosPlaceInfo 的实例。
 */
CdosPlaceInfo *
cdos_place_info_set_params(CdosPlaceInfo *info,
                           gchar *id,
                           gchar *name,
                           gpointer iconFactory,
                           gpointer launch)
{
        CdosPlaceInfoPrivate *priv = info->priv;
        if (priv->id)
        {
            g_free(priv->id);
        }
        priv->id = g_strdup(id);

        if (priv->name)
        {
            g_free(priv->name);
        }
        priv->name = g_strdup(name);

        if (priv->lowerName) 
        {
            g_free(priv->lowerName);
        }
        priv->lowerName = g_utf8_strdown(priv->name, -1);

        priv->iconFactory = iconFactory;   
        priv->launch = launch;

        return info;
}

/**
 * cdos_place_info_match_terms:
 * @info: A #CdosPlaceInfo
 * @terms: 比较的字符串数组 
 *
 * 判断一个字符串数组里所有的字符串是不是都是CdosPlaceInfo对象的lowerName子串
 *
 * Returns: ＃MATCHTYPE　NONE=0,SUBSTRING,MULTIPLE_SUBSTRING,PREFIX,MULTIPLE_PREFIX
 */
MATCHTYPE 
cdos_place_info_match_terms(CdosPlaceInfo *info, GArray *terms)
{
        MATCHTYPE mtype = NONE;
        int i;
        gchar *term;
        gint idx;
        for (i = 0; i < terms->len; i++) {
                term = g_array_index(terms, gchar *, i);
                idx = index_of(info->priv->lowerName, term);

                if (idx == 0) {
                        mtype = PREFIX;
                } else if (idx > 0) {
                        if (mtype == NONE)
                                mtype = MULTIPLE_SUBSTRING;
                } else {
                        mtype = NONE;
                        return mtype;
                }
        }
        return mtype;
}

static gint 
index_of(const gchar *haystack, const gchar *needle)
{
        if (haystack == NULL || needle == NULL)
                return -1;

        gchar *m = g_strrstr(haystack, needle);
        if(m){
                return m-haystack+1;
        }

        return -1;
}

/**
 * cdos_place_info_is_removeable:
 * @info: A #CdosPlaceInfo
 *
 * 获取info是否可以移除
 *
 * Returns: FALSE 
 */
gboolean 
cdos_place_info_is_removeable(CdosPlaceInfo *info)
{
        return FALSE;
}

/**
 * cdos_place_info_make_launch_conext:
 * @params: A #CdosPlaceInfoParams
 *
 * 将launcher 参数转化为GAppLaunchContext
 *
 * Returns: GAppLaunchContext 
 */
GAppLaunchContext *
cdos_place_info_make_launch_conext(CdosPlaceInfoParams *params)
{
        GdkAppLaunchContext *launch = NULL;

/*       launch = (GdkAppLaunchContext *)cdos_global_create_app_launch_context(cdos_global_get());

        if (params->workspace != -1) 
               gdk_app_launch_context_set_desktop(launch, params->workspace);
       if (params->timestamp != 0)
               gdk_app_launch_context_set_timestamp(launch, params->timestamp);*/

        return launch;
}

/**
 * cdos_place_info_get_id:
 * @info: A #CdosPlaceInfo
 *
 * 获取info的id成员
 *
 * Returns: #gchar *　id字符串用完之后不需要释放，如果需要长久保存请g_strdup()拷贝一份. 
 */
gchar *
cdos_place_info_get_id(CdosPlaceInfo *info)
{
        return info->priv->id;
}

/**
 * cdos_place_info_get_name:
 * @info: A #CdosPlaceInfo
 *
 * 获取info的name成员
 *
 * Returns: #gchar *　name字符串用完之后不需要释放，如果需要长久保存请g_strdup()拷贝一份. 
 */
gchar *
cdos_place_info_get_name(CdosPlaceInfo *info)
{
        return info->priv->name;
}

/**
 * cdos_place_info_new:
 * 
 * 新建一个 #CdosPlaceInfo 类的实例。
 * 对象的数据都是空的，需要使用cdos_place_info_set_params方法设置参数
 *
 * Returns: 返回一个 #CdosPlaceInfo 的实例。
 */
CdosPlaceInfo *
cdos_place_info_new(void)
{
    return CDOS_PLACE_INFO(g_object_new(CDOS_TYPE_PLACE_INFO, NULL));
}
