#include <gio/gio.h>
#include "applet-places-manager.h"
#include "applet-place-device-info.h"
#include "cdos-base-util.h"
#include "cdos-sound-manager.h"

G_DEFINE_TYPE(CdosPlacesManager, cdos_places_manager, G_TYPE_OBJECT);

#define CDOS_PLACES_MANAGER_GET_PRIVATE(obj)    \
        (G_TYPE_INSTANCE_GET_PRIVATE((obj), CDOS_TYPE_PLACES_MANAGER, CdosPlacesManagerPrivate))

static gchar *homeUri;
static GIcon *homeIcon;
static gchar *desktopUri;
static GIcon *desktopIcon;

static GIcon *icon;
static gchar *bookmark;

struct _CdosPlacesManagerPrivate
{
        GPtrArray       *defaultPlaces;
        GPtrArray       *mounts;
        GPtrArray       *bookmarks;
        CdosPlaceInfo   *home;
        CdosPlaceInfo   *desktopMenu;
        CdosPlaceInfo   *connect;
        GVolumeMonitor  *volumeMonitor;
        gchar           *bookmarksPath;
        GFile           *bookmarksFile;
        GFileMonitor    *monitor;
        gint             bookmarksTimeoutId;
};

enum
{
         MOUNTS_UPDATED,
         PLACES_UPDATED,
         BOOKMARKS_UPDATED,
  
  LAST_SIGNAL
};

static guint cdos_places_manager_signals[LAST_SIGNAL] =  { 0, };

static void 
cdos_places_manager_init(CdosPlacesManager *manager);

static void
cdos_callback(GFileMonitor *monitor,
              GFile *file,
              GFile *other,
              GFileMonitorEvent event_type,
              CdosPlacesManager *manager);

static gboolean
cdos_timeout_callback(CdosPlacesManager *manager);

static ClutterActor*
cdos_places_manager_gicon(gint size);

static void 
cdos_places_manager_uri(CdosPlaceInfoParams *params);

static void 
cdos_places_manager_class_init(CdosPlacesManagerClass *klass);

static ClutterActor*
cdos_new_home_first(gint size);

static gboolean
cdos_new_home_second(CdosPlaceInfoParams *params);

static ClutterActor*
cdos_new_desktop_first(gint size);

static gboolean
cdos_new_desktop_second(CdosPlaceInfoParams *params);

static GObject*
cdos_new_connect_first(gint size);

static void
cdos_new_connect_second(void);

static void
cdos_places_manager_finalize(GObject *gobject);

static CdosPlacesManager *
cdos_places_manager_new(void);

static void
cdos_places_manager_on_drive_connected(GVolumeMonitor    *volumeMonitor,
                                          GDrive             *drives,
                                          CdosPlacesManager *manager);

static void
cdos_places_manager_on_drive_disconnected(GVolumeMonitor    *volumeMonitor,
                                             GDrive              *drives,
                                             CdosPlacesManager  *manager);

static void
cdos_places_manager_update_devices(GVolumeMonitor       *volumeMonitor, 
                                      GDrive                *drive,
                                      CdosPlacesManager    *manager);

static void
cdos_places_manager_reload_bookmarks(CdosPlacesManager *manager);

static gint
get_str_length(gchar **str);

static void 
cdos_places_manager_add_mount(CdosPlacesManager *manager, GMount *mount);

static gint
cdos_places_manager_lookup_index_by_id(GPtrArray *sourceArray, gchar *id);

static gchar *
substring(gchar *des, int start, int end);

static gint 
index_of(const gchar *haystack, const gchar *needle);

static void
cdos_places_manager_remove_by_id(CdosPlacesManager *manager, 
                                 GPtrArray            *sourceArray,
                                 gchar             *id);

static ClutterActor*
cdos_new_place_info_first(gint size);

static gboolean
cdos_new_place_info_second1(CdosPlaceInfoParams *params);

static gboolean
cdos_new_place_info_second2(CdosPlaceInfoParams *params);


static void 
cdos_places_manager_init(CdosPlacesManager *manager)
{
        CdosPlacesManagerPrivate *priv;
        manager->priv = priv = CDOS_PLACES_MANAGER_GET_PRIVATE(manager);

        priv->defaultPlaces = g_ptr_array_new_full(5, g_object_unref);

        priv->mounts = g_ptr_array_new_full(5, g_object_unref);

        priv->bookmarks = g_ptr_array_new_full(5, g_object_unref);
	
        GFile *homeFile = g_file_new_for_path(g_get_home_dir());
        homeUri = g_file_get_uri(homeFile);
        gchar *homeLabel = cdos_base_util_get_label_for_uri(homeUri);
        homeIcon = cdos_base_util_get_icon_for_uri(homeUri);

        priv->home = cdos_place_info_new_with_params("special:home", 
                                                homeLabel, 
                                                cdos_new_home_first,
                                                cdos_new_home_second);

        const gchar *desktopPath = g_get_user_special_dir(G_USER_DIRECTORY_DESKTOP);
        GFile *desktopFile = g_file_new_for_path(desktopPath);
        desktopUri = g_file_get_uri(desktopFile);
        gchar *desktopLabel = cdos_base_util_get_label_for_uri(desktopUri);
        desktopIcon = cdos_base_util_get_icon_for_uri(desktopUri);
        
        priv->desktopMenu = cdos_place_info_new_with_params("special:desktop",
                                                        desktopLabel,
                                                        cdos_new_desktop_first,
                                                        cdos_new_desktop_second);

        priv->connect = cdos_place_info_new_with_params("special:connect", 
                                                        "Connect to...",
                                                        cdos_new_connect_first,
                                                        cdos_new_desktop_second);

        g_ptr_array_add(priv->defaultPlaces, priv->home);
        g_ptr_array_add(priv->defaultPlaces, priv->desktopMenu);
        g_ptr_array_add(priv->defaultPlaces, priv->connect);

        priv->volumeMonitor = g_volume_monitor_get();

        g_signal_connect(priv->volumeMonitor, 
                        "volume-added", 
                        G_CALLBACK(cdos_places_manager_update_devices),
                        manager);

        g_signal_connect(priv->volumeMonitor,
                        "volume-removed",
                        G_CALLBACK(cdos_places_manager_update_devices),
                        manager);

        g_signal_connect(priv->volumeMonitor,
                        "volume-changed",
                        G_CALLBACK(cdos_places_manager_update_devices),
                        manager);

        g_signal_connect(priv->volumeMonitor,
                        "mount-added",
                        G_CALLBACK(cdos_places_manager_update_devices),
                        manager);

        g_signal_connect(priv->volumeMonitor,
                        "mount-removed",
                        G_CALLBACK(cdos_places_manager_update_devices),
                        manager);

        g_signal_connect(priv->volumeMonitor,
                        "mount-changed",
                        G_CALLBACK(cdos_places_manager_update_devices),
                        manager);

        g_signal_connect(priv->volumeMonitor,
                        "drive-connected",
                        G_CALLBACK(cdos_places_manager_on_drive_connected),
                        manager);

        g_signal_connect(priv->volumeMonitor,
                        "drive-disconnected",
                        G_CALLBACK(cdos_places_manager_on_drive_disconnected),
                        manager);

        g_signal_connect(priv->volumeMonitor, 
                        "drive-changed",
                        G_CALLBACK(cdos_places_manager_update_devices),
                        manager);

        cdos_places_manager_update_devices(priv->volumeMonitor, NULL, manager);

        priv->bookmarksPath = g_build_filename(g_get_user_config_dir(), "gtk-3.0", "bookmarks", NULL); 

        priv->bookmarksFile = g_file_new_for_path(priv->bookmarksPath);

        if (!g_file_query_exists(priv->bookmarksFile, NULL)) 
        {
                priv->bookmarksPath = g_build_filename(g_get_user_config_dir(), "bookmarks", NULL); 
                priv->bookmarksFile = g_file_new_for_path(priv->bookmarksPath);
        }

        priv->monitor = g_file_monitor_file(priv->bookmarksFile, G_FILE_MONITOR_NONE, NULL, NULL);
        priv->bookmarksTimeoutId = 0;

        g_signal_connect(priv->monitor, "changed", G_CALLBACK(cdos_callback), manager);       

        g_free(homeLabel);
        g_free(desktopLabel);
        g_object_unref(homeFile);
        g_object_unref(desktopFile);
        cdos_places_manager_reload_bookmarks(manager);
}

static void
cdos_callback(GFileMonitor *monitor,
              GFile *file,
              GFile *other,
              GFileMonitorEvent event_type,
              CdosPlacesManager *manager)
{
        CdosPlacesManagerPrivate *priv = manager->priv;
        if (priv->bookmarksTimeoutId > 0) 
        {
                return;
        }

        guint times = g_timeout_add(100, cdos_timeout_callback, manager);
        priv->bookmarksTimeoutId = times;
}

static gboolean
cdos_timeout_callback(CdosPlacesManager *manager)
{
        CdosPlacesManagerPrivate *priv = manager->priv;
        priv->bookmarksTimeoutId = 0;
        cdos_places_manager_reload_bookmarks(manager);

        return FALSE;
}

static ClutterActor*
cdos_places_manager_gicon(gint size)
{
        StTextureCache *cache = st_texture_cache_get_default();
        return  st_texture_cache_load_gicon(cache, NULL, homeIcon, size);
}

static void 
cdos_places_manager_uri(CdosPlaceInfoParams *params)
{
        g_app_info_launch_default_for_uri(homeUri, cdos_place_info_make_launch_conext, NULL);
}

static void 
cdos_places_manager_class_init(CdosPlacesManagerClass *klass)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

        GParamSpec *pspec;
        g_type_class_add_private(klass, sizeof(CdosPlacesManagerPrivate));

        gobject_class->finalize = cdos_places_manager_finalize;

        cdos_places_manager_signals[MOUNTS_UPDATED] = g_signal_new (
                        "mounts-updated",
                        G_TYPE_FROM_CLASS(klass),
                        G_SIGNAL_RUN_LAST,
                        G_STRUCT_OFFSET(CdosPlacesManagerClass, mounts_updated),
                        NULL,
                        NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE,
                        0);

        cdos_places_manager_signals[MOUNTS_UPDATED] = g_signal_new (
                        "places-updated",
                        G_TYPE_FROM_CLASS(klass),
                        G_SIGNAL_RUN_LAST,
                        G_STRUCT_OFFSET(CdosPlacesManagerClass, places_updated),
                        NULL,
                        NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE,
                        0);

        cdos_places_manager_signals[BOOKMARKS_UPDATED] = g_signal_new (
                        "bookmarks-updated",
                        G_TYPE_FROM_CLASS(klass),
                        G_SIGNAL_RUN_LAST,
                        G_STRUCT_OFFSET(CdosPlacesManagerClass, bookmarks_updated),
                        NULL,
                        NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE,
                        0);
}

static ClutterActor*
cdos_new_home_first(gint size)
{
        StTextureCache* cache = st_texture_cache_get_default ();
        return st_texture_cache_load_gicon(cache, NULL, homeIcon, size);
}

static gboolean
cdos_new_home_second(CdosPlaceInfoParams *params)
{
        GAppLaunchContext *context = cdos_place_info_make_launch_conext(params);
        return g_app_info_launch_default_for_uri(homeUri, context, NULL);
}

static ClutterActor*
cdos_new_desktop_first(gint size)
{
        StTextureCache* cache = st_texture_cache_get_default ();
        return st_texture_cache_load_gicon(cache, NULL, desktopIcon, size);
}

static gboolean
cdos_new_desktop_second(CdosPlaceInfoParams *params)
{
        GAppLaunchContext *context = cdos_place_info_make_launch_conext(params);
        return g_app_info_launch_default_for_uri(desktopUri, context, NULL);
}

static GObject*
cdos_new_connect_first(gint size)
{
        return g_object_new(ST_TYPE_ICON, 
                        "icon-name","applications-internet",
                        "icon-type", ST_ICON_FULLCOLOR,
                        "icon-size", size,
                        NULL);
}

static void
cdos_new_connect_second(void)
{
        cdos_util_spawn("nemo-connect-server");
}

static void
cdos_places_manager_finalize(GObject *gobject)
{
        CdosPlacesManagerPrivate *priv = CDOS_PLACES_MANAGER(gobject)->priv;

        if (priv->defaultPlaces) 
        {
            g_ptr_array_unref(priv->defaultPlaces);
        }

        if (priv->mounts) 
        {
            g_ptr_array_unref(priv->mounts);
        }

        if (priv->bookmarks) 
        {
            g_ptr_array_unref(priv->bookmarks);
        }

        if (priv->home) 
        {
            g_object_unref(priv->home);
        }

        if (priv->desktopMenu) 
        {
            g_object_unref(priv->desktopMenu);
        }

        if (priv->connect) 
        {
            g_object_unref(priv->connect);
        }

        if (priv->volumeMonitor) 
        {
            g_object_unref(priv->volumeMonitor);
        }

        if (priv->bookmarksPath) 
        {
            g_free(priv->bookmarksPath);
        }

        if (priv->bookmarksFile) 
        {
            g_object_unref(priv->bookmarksFile);
        }

        if (priv->monitor) 
        {
            g_object_unref(priv->monitor);
        }

        if (homeUri) 
        {
            g_free(homeUri);
            homeUri = NULL;
        }

        if (desktopUri)
        {
            g_free(desktopUri);
            desktopUri = NULL;
        }

        if (bookmark)
        {
            g_free(bookmark);
            bookmark = NULL;
        }

        if (homeIcon)
        {
            g_object_unref(homeIcon);
            homeIcon = NULL;
        }

        if (desktopIcon)
        {
            g_object_unref(desktopIcon);
            desktopIcon = NULL;   
        }

        if (icon)
        {
            g_object_unref(icon);
            icon = NULL;
        }

        G_OBJECT_CLASS(cdos_places_manager_parent_class)->finalize(gobject);
}

static void
cdos_places_manager_on_drive_connected(GVolumeMonitor    *volumeMonitor,
                                          GDrive             *drives,
                                          CdosPlacesManager *manager)
{
        cdos_sound_manager_play("plug");
        cdos_places_manager_update_devices(volumeMonitor, drives, manager);
}

static void
cdos_places_manager_on_drive_disconnected(GVolumeMonitor     *volumeMonitor,
                                              GDrive              *drives,
                                              CdosPlacesManager  *manager)
{
        cdos_sound_manager_play("unplug");
        cdos_places_manager_update_devices(volumeMonitor, drives, manager);
}

static void
cdos_places_manager_update_devices(GVolumeMonitor    *volumeMonitor, 
                                   GDrive           *drive,
                                   CdosPlacesManager *manager)
{
        CdosPlacesManagerPrivate *priv = CDOS_PLACES_MANAGER_GET_PRIVATE(manager);
        int i = 0;

        while(priv->mounts->len) 
        {
            g_ptr_array_remove_index(priv->mounts, priv->mounts->len-1);
        } 

        GList  *drives = g_volume_monitor_get_connected_drives(priv->volumeMonitor);
        
        GList *l;
        GList *j;

        for (l = drives; l != NULL; l = l->next)
        {
                GList *volumes = g_drive_get_volumes(l->data);
                for (j = volumes; j != NULL; j = j->next) 
                {
                        char *dev_name = g_volume_get_identifier(j->data, "unix-device");
                        if (!g_strcmp0("/dev/sr0", dev_name))
                            continue;

                        GMount *mount = g_volume_get_mount(j->data);

                        if (NULL != mount) {
                            cdos_places_manager_add_mount(manager, mount);
                            g_object_unref(mount);
                        }
                }
                g_list_free_full(volumes, g_object_unref);
        }

        g_list_free(drives);
        
        GList *volumes = g_volume_monitor_get_volumes(priv->volumeMonitor);
        for (l = volumes; l !=NULL; l = l->next) 
        {
                if (g_volume_get_drive(l->data) != NULL)
                        continue;

                        GMount *mount = g_volume_get_mount(l->data);
                        if (NULL != mount){
                            cdos_places_manager_add_mount(manager, mount);
                            g_object_unref(mount);
                        }
                                
        }
        g_list_free_full(volumes, g_object_unref);

       
        GList *mounts = g_volume_monitor_get_mounts(priv->volumeMonitor);
       
        for (l = mounts; l!=NULL; l = l->next) 
        {

                if(g_mount_is_shadowed(l->data))
                        continue;

                if (g_mount_get_volume(l->data))
                        continue;

                cdos_places_manager_add_mount(manager, l->data);
        }
        g_list_free_full(mounts, g_object_unref);

        g_signal_emit_by_name(manager, "mounts-updated");
        g_signal_emit_by_name(manager, "places-updated");
        
}

static void
cdos_places_manager_reload_bookmarks(CdosPlacesManager *manager)
{
        
        CdosPlacesManagerPrivate *priv = CDOS_PLACES_MANAGER_GET_PRIVATE(manager);
      
        while(priv->bookmarks->len) 
        {
                g_ptr_array_remove_index(priv->bookmarks, priv->bookmarks->len-1);
        }

        if (!g_file_test(priv->bookmarksPath, G_FILE_TEST_EXISTS))
                return;
        
        gchar *bookmarksContent = cdos_base_util_get_file_contents_utf8_sync(priv->bookmarksPath, NULL);
        gchar **bookmarks = g_strsplit(bookmarksContent, "\n", -1);
        gint length = get_str_length(bookmarks);
        GHashTable *bookmarksToLabel = g_hash_table_new_full(NULL, NULL, g_free, g_free);
        GPtrArray *bookmarksOrder = g_ptr_array_new_full(5, g_free);
        int j = 0; 
	    int i = 0;
        gboolean flag = FALSE;

        for (i = 0; i < length - 1; i++)
        {               
                gchar *bookmarkLine = *(bookmarks + i);
               
                gchar **components = g_strsplit(bookmarkLine, " ", -1);
                gchar *bookmark_tmp = *components;

                if (g_hash_table_contains(bookmarksToLabel, bookmark_tmp))
                        continue;

                gchar *label = NULL;
                gchar *res = NULL;
                gint len_com = get_str_length(components);

                if (len_com > 1) {
                        for (j = 1; j < len_com; j++) 
                        {
                                gchar *tmp = g_strconcat(*(components + j), " ", NULL);
                                res = label;
                                label = g_strconcat(label, tmp, NULL); 
                                g_free(res);
                                g_free(tmp);
                        }
                }
                
                g_hash_table_insert(bookmarksToLabel, g_strdup(bookmark_tmp), label);
                g_ptr_array_add(bookmarksOrder, g_strdup(bookmark_tmp));
                g_strfreev(components);
        }
    
        guint len_order = bookmarksOrder->len;
        CdosPlaceInfo *item;
        for (i = 0; i < len_order; i ++) 
        {
                
                gchar *bookmark = g_ptr_array_index(bookmarksOrder, i);
                gchar *label = g_hash_table_lookup(bookmarksToLabel, bookmark);
                GFile *file = g_file_new_for_uri(bookmark);
    
                if (NULL == label)
                {
                        label = cdos_base_util_get_label_for_uri(bookmark);
                        flag = TRUE;
                }
                if (NULL == label)
                        continue; 
                
                gchar *name = g_strconcat("bookmark:", bookmark, NULL);
                if (icon)
                {
                        g_object_unref(icon);
                }

                if (g_file_query_exists(file, NULL)) 
                {
                        icon = cdos_base_util_get_icon_for_uri(bookmark);
                        
                        item = cdos_place_info_new_with_params(name, label,
                                                    cdos_new_place_info_first,
                                                    cdos_new_place_info_second1);

                } else {
                        icon = g_themed_icon_new("gnome-fs-network");
                        item = cdos_place_info_new_with_params(name, label,
                                                    cdos_new_place_info_first,
                                                    cdos_new_place_info_second2);
                }
                
                g_object_unref(file);
                g_ptr_array_add(priv->bookmarks, item);
                g_free(name);
                if (flag)
                {
                    g_free(label);
                } 
        }
        
        g_free(bookmarksContent);
        g_strfreev(bookmarks);
        g_hash_table_destroy(bookmarksToLabel);
        g_ptr_array_unref(bookmarksOrder);
        g_signal_emit_by_name(manager, "bookmarks-updated");
        g_signal_emit_by_name(manager, "places-updated");
        
} 

static ClutterActor*
cdos_new_place_info_first(gint size)
{
        StTextureCache* cache = st_texture_cache_get_default ();
        return st_texture_cache_load_gicon(cache, NULL, icon, size);
}

static gboolean
cdos_new_place_info_second1(CdosPlaceInfoParams *params)
{
        GAppLaunchContext *context = cdos_place_info_make_launch_conext(params);
        return g_app_info_launch_default_for_uri(bookmark, context, NULL);
}

static gboolean
cdos_new_place_info_second2(CdosPlaceInfoParams *params)
{
        GAppInfo *appInfo = g_app_info_get_default_for_uri_scheme ("file");
        GList *list = NULL;
        list = g_list_append(list, bookmark);

        if (appInfo) {
            GAppLaunchContext *context = cdos_place_info_make_launch_conext(params);
            return g_app_info_launch_uris(appInfo, list, context, NULL);
        }
        
        return FALSE;
}

static gint
get_str_length(gchar **str)
{
        gint length = 0;
        while(*str++)
                length++;

        return length;
}

static void 
cdos_places_manager_add_mount(CdosPlacesManager *manager, GMount *mount)
{
        CdosPlaceDeviceInfo *devItem;
        devItem = cdos_place_device_info_new_with_params(mount);
        CdosPlacesManagerPrivate *priv = CDOS_PLACES_MANAGER_GET_PRIVATE(manager);
        g_ptr_array_add(priv->mounts, devItem);
}

GPtrArray *
cdos_places_manager_get_all_places(CdosPlacesManager *manager)
{
        GPtrArray *arr = g_ptr_array_new_full(5, g_object_unref);

        GPtrArray *defaultPlaces = cdos_places_manager_get_default_places(manager);
        GPtrArray *bookmarks = cdos_places_manager_get_bookmarks(manager);
        GPtrArray *mounts = cdos_places_manager_get_mounts(manager);

        int i = 0;
        for (i = 0; i < defaultPlaces->len; i++)
        {
                g_ptr_array_add(arr, g_ptr_array_index(defaultPlaces, i));
        }

        for (i = 0; bookmarks && i < bookmarks->len; i++)
        {
                g_ptr_array_add(arr, g_ptr_array_index(bookmarks, i));
        }

        for (i = 0; i < mounts->len; i ++)
        {
                g_ptr_array_add(arr, g_ptr_array_index(mounts, i));
        }

        g_ptr_array_unref(defaultPlaces);
        g_ptr_array_unref(bookmarks);
        g_ptr_array_unref(mounts);

        return arr;
}

GPtrArray *
cdos_places_manager_get_default_places(CdosPlacesManager *manager)
{
        if (NULL == manager)
                return NULL;

        return g_ptr_array_ref(manager->priv->defaultPlaces);
}

GPtrArray *
cdos_places_manager_get_bookmarks(CdosPlacesManager *manager)
{
        if (NULL == manager)
                return NULL;

        return g_ptr_array_ref(manager->priv->bookmarks);
}

GPtrArray *
cdos_places_manager_get_mounts(CdosPlacesManager *manager)
{
        if (NULL == manager)
                return NULL;

        return g_ptr_array_ref(manager->priv->mounts);
}

static gint
cdos_places_manager_lookup_index_by_id(GPtrArray *sourceArray, gchar *id)
{
    if (!sourceArray)
        return -1;
        int i;
        CdosPlaceInfo *place;

        for (i = 0; i < sourceArray->len; i++)
        {
                place = g_ptr_array_index(sourceArray, i);
                if (!g_strcmp0(cdos_place_info_get_id(place), id))
                        return i;
        }

        return -1;
}

CdosPlaceInfo *
cdos_places_manager_lookup_place_by_id(CdosPlacesManager *manager, gchar *id)
{
        CdosPlacesManagerPrivate *priv = manager->priv;
        gint colonId = index_of(id, ":");
        gchar *type = substring(id, 0, colonId);

        GPtrArray *sourceArray = NULL;
        if (!g_strcmp0(type, "special"))
                sourceArray = priv->defaultPlaces;
        else if(!g_strcmp0(type, "mount"))
                sourceArray = priv->mounts;
        else if (!g_strcmp0(type, "bookmark"))
                sourceArray = priv->bookmarks;

        gint index = cdos_places_manager_lookup_index_by_id(sourceArray, id);

        return g_ptr_array_index(sourceArray, index);
}

static gchar *
substring(gchar *des, int start, int end)
{
        gint size = end - start;
        if (size < 0)
                return NULL;

        gchar *res = g_malloc(size + 1);
        *(res+size+1) = '\0';
        int j = 0;
        int i;
        for (i = start; i < end; i++)
        {
                if(!(*(res+j) = *(des+i)))
                        break;
                j++;
        }

        return res;
}

static gint 
index_of(const gchar *haystack, const gchar *needle)
{
        if (haystack == NULL || needle == NULL)
                return -1;
        gchar *m = strstr(haystack, needle);
        if (m) {
                return m-haystack+1;
        }

        return -1;
}

static void
cdos_places_manager_remove_by_id(CdosPlacesManager *manager, 
                                GPtrArray *sourceArray,
                                gchar *id)
{
        gint index = cdos_places_manager_lookup_index_by_id(sourceArray, id);
        g_ptr_array_remove_index(sourceArray, index);
}

static CdosPlacesManager *
cdos_places_manager_new(void)
{
        return CDOS_PLACES_MANAGER(g_object_new(CDOS_TYPE_PLACES_MANAGER, NULL));
}

CdosPlacesManager *
cdos_places_manager_get_default(void)
{
        static CdosPlacesManager *manager = NULL;

        if (!manager) {
            manager = cdos_places_manager_new();
        }

        return manager;
}   

