/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

#include "config.h"

#include "cdos-app-system.h"
#include <string.h>
#include <gio/gio.h>
#include <gio/gdesktopappinfo.h>
#include <gtk/gtk.h>
#include <clutter/clutter.h>
#include <glib/gi18n.h>
#include <meta/display.h>

#include "cdos-app-private.h"
#include "cdos-window-tracker-private.h"
#include "cdos-app-system-private.h"
#include "cdos-global.h"
#include "cdos-base-util.h"
#include "st.h"

/* Vendor prefixes are something that can be preprended to a .desktop
 * file name.  Undo this.
 */
static const char*const vendor_prefixes[] = { "gnome-",
                                              "fedora-",
                                              "mozilla-",
                                              "cdos-",
                                              "org.gnome.",
                                              "org.cdos.",
                                              "org.fedora.",
                                              "org.mozilla.",
                                              NULL };

enum {
   PROP_0,

};

enum {
  APP_STATE_CHANGED,
  INSTALLED_CHANGED,
  LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

struct _CdosAppSystemPrivate {
  GMenuTree *apps_tree;

  GHashTable *running_apps;
  GHashTable *id_to_app;

  GSList *known_vendor_prefixes;

  GMenuTree *settings_tree;
  GHashTable *setting_id_to_app;
    GHashTable *apps;
};

static void cdos_app_system_finalize (GObject *object);
static void on_apps_tree_changed_cb (GMenuTree *tree, gpointer user_data);
static gchar *strip_extension (const gchar *wm_class);

G_DEFINE_TYPE(CdosAppSystem, cdos_app_system, G_TYPE_OBJECT);

static void cdos_app_system_class_init(CdosAppSystemClass *klass)
{
  GObjectClass *gobject_class = (GObjectClass *)klass;

  gobject_class->finalize = cdos_app_system_finalize;

  signals[APP_STATE_CHANGED] = g_signal_new ("app-state-changed",
                                             CDOS_TYPE_APP_SYSTEM,
                                             G_SIGNAL_RUN_LAST,
                                             0,
                                             NULL, NULL,
                                             g_cclosure_marshal_VOID__OBJECT,
                                             G_TYPE_NONE, 1,
                                             CDOS_TYPE_APP);
  signals[INSTALLED_CHANGED] =
    g_signal_new ("installed-changed",
		  CDOS_TYPE_APP_SYSTEM,
		  G_SIGNAL_RUN_LAST,
		  G_STRUCT_OFFSET (CdosAppSystemClass, installed_changed),
		  NULL, NULL,
		  g_cclosure_marshal_VOID__VOID,
		  G_TYPE_NONE, 0);

  g_type_class_add_private (gobject_class, sizeof (CdosAppSystemPrivate));
}

static void
setup_merge_dir_symlink(void)
{
    const gchar *user_config = g_get_user_config_dir();
    gchar *merge_path = g_build_filename (user_config, "menus", "applications-merged", NULL);
    GFile *merge_file = g_file_new_for_path (merge_path);

    g_file_make_directory_with_parents (merge_file, NULL, NULL);

    gchar *sym_path = g_build_filename (user_config, "menus", "cdos-applications-merged", NULL);
    GFile *sym_file = g_file_new_for_path (sym_path);
    if (!g_file_query_exists (sym_file, NULL)) {
        g_file_make_symbolic_link (sym_file, merge_path, NULL, NULL);
    }

    /*
     * VALGRIND shows that 'name' is already free'd
     * Actually this 'name' is a stack var
     * */

    //g_free (merge_path);
    //g_free (sym_path);
    g_object_unref (merge_file);
    g_object_unref (sym_file);
}

static void
cdos_app_system_init (CdosAppSystem *self)
{
  CdosAppSystemPrivate *priv;

  self->priv = priv = G_TYPE_INSTANCE_GET_PRIVATE (self,
                                                   CDOS_TYPE_APP_SYSTEM,
                                                   CdosAppSystemPrivate);

  priv->running_apps = g_hash_table_new_full (NULL, NULL, (GDestroyNotify) g_object_unref, NULL);
  priv->id_to_app = g_hash_table_new_full (g_str_hash, g_str_equal,
                                           NULL,
                                           (GDestroyNotify)g_object_unref);
/* According to desktop spec, since our menu file is called 'cdos-applications', our
 * merged menu folders need to be called 'cdos-applications-merged'.  We'll setup the folder
 * 'applications-merged' if it doesn't exist yet, and a symlink pointing to it in the
 * ~/.config/menus directory
 */
  setup_merge_dir_symlink();

  /* For now, we want to pick up Evince, Nemo, etc.  We'll
   * handle NODISPLAY semantics at a higher level or investigate them
   * case by case.
   */
  priv->apps_tree = gmenu_tree_new ("cdos-applications.menu", GMENU_TREE_FLAGS_INCLUDE_NODISPLAY);
  g_signal_connect (priv->apps_tree, "changed", G_CALLBACK (on_apps_tree_changed_cb), self);

  on_apps_tree_changed_cb (priv->apps_tree, self);
}

static void
cdos_app_system_finalize (GObject *object)
{
  CdosAppSystem *self = CDOS_APP_SYSTEM (object);
  CdosAppSystemPrivate *priv = self->priv;

  g_object_unref (priv->apps_tree);
  g_object_unref (priv->settings_tree);
  g_hash_table_destroy (priv->running_apps);
  g_hash_table_destroy (priv->id_to_app);
  g_hash_table_destroy (priv->setting_id_to_app);
    g_hash_table_destroy (priv->apps);
  g_slist_foreach (priv->known_vendor_prefixes, (GFunc)g_free, NULL);
  g_slist_free (priv->known_vendor_prefixes);
  priv->known_vendor_prefixes = NULL;

  G_OBJECT_CLASS (cdos_app_system_parent_class)->finalize (object);
}

static char *
get_prefix_for_entry (GMenuTreeEntry *entry)
{
  char *prefix = NULL, *file_prefix = NULL;
  const char *id;
  GFile *file;
  char *name;
  int i = 0;
    /*
     * VALGRIND shows that 'name' is already free'd
     * Actually this 'name' is a stack var
     * all char * 
     * */

  id = gmenu_tree_entry_get_desktop_file_id (entry);
  file = g_file_new_for_path (gmenu_tree_entry_get_desktop_file_path (entry));
  name = g_file_get_basename (file);

  if (!name)
    {
      g_object_unref (file);
      return NULL;
    }
  for (i = 0; vendor_prefixes[i]; i++)
    {
      if (g_str_has_prefix (name, vendor_prefixes[i]))
        {
          file_prefix = g_strdup (vendor_prefixes[i]);
          break;
        }
    }

  while (strcmp (name, id) != 0)
    {
      char *t;
      char *pname;
      GFile *parent = g_file_get_parent (file);

      if (!parent)
        {
          g_warn_if_reached ();
          break;
        }

      pname = g_file_get_basename (parent);
      if (!pname)
        {
          g_object_unref (parent);
          break;
        }
      if (!g_strstr_len (id, -1, pname))
        {
          /* handle <LegacyDir prefix="..."> */
          char *t;
          size_t name_len = strlen (name);
          size_t id_len = strlen (id);
          char *t_id = g_strdup (id);

          t_id[id_len - name_len] = '\0';
          t = g_strdup(t_id);
          //g_free (prefix);
          //g_free (t_id);
          //g_free (name);
          name = g_strdup (id);
          prefix = t;

          g_object_unref (file);
          file = parent;
          //g_free (pname);
          //g_free (file_prefix);
          file_prefix = NULL;
          break;
        }

      t = g_strconcat (pname, "-", name, NULL);
      //g_free (name);
      name = t;

      t = g_strconcat (pname, "-", prefix, NULL);
      //g_free (prefix);
      prefix = t;

      g_object_unref (file);
      file = parent;
      //g_free (pname);
    }

  if (file)
    g_object_unref (file);

  if (strcmp (name, id) == 0)
    {
      //g_free (name);
      if (file_prefix && !prefix)
        return file_prefix;
      if (file_prefix)
        {
          char *t = g_strconcat (prefix, "-", file_prefix, NULL);
          //g_free (prefix);
          //g_free (file_prefix);
          prefix = t;
        }
      return prefix;
    }

  //g_free (name);
  //g_free (prefix);
  //g_free (file_prefix);
  g_return_val_if_reached (NULL);
}

static void
get_flattened_entries_recurse (GMenuTreeDirectory *dir,
                               GHashTable         *entry_set)
{
  GMenuTreeIter *iter = gmenu_tree_directory_iter (dir);
  GMenuTreeItemType next_type;

  while ((next_type = gmenu_tree_iter_next (iter)) != GMENU_TREE_ITEM_INVALID)
    {
      gpointer item = NULL;

      switch (next_type)
        {
        case GMENU_TREE_ITEM_ENTRY:
          {
            GMenuTreeEntry *entry;
            item = entry = gmenu_tree_iter_get_entry (iter);
            /* Key is owned by entry */
                const  gchar *file_id = gmenu_tree_entry_get_desktop_file_id (entry);
                g_hash_table_replace (entry_set,
                                  file_id,
                                  gmenu_tree_item_ref (entry));
          }
          break;
        case GMENU_TREE_ITEM_DIRECTORY:
          {
            item = gmenu_tree_iter_get_directory (iter);
            get_flattened_entries_recurse ((GMenuTreeDirectory*)item, entry_set);
          }
          break;
        default:
          break;
        }
      if (item != NULL)
        gmenu_tree_item_unref (item);
    }

  gmenu_tree_iter_unref (iter);
}

static GHashTable *
get_flattened_entries_from_tree (GMenuTree *tree)
{
  GHashTable *table;
  GMenuTreeDirectory *root;

  table = g_hash_table_new_full (g_str_hash, g_str_equal,
                                 (GDestroyNotify) NULL,
                                 (GDestroyNotify) gmenu_tree_item_unref);

  root = gmenu_tree_get_root_directory (tree);
  
  if (root != NULL)
    get_flattened_entries_recurse (root, table);

  gmenu_tree_item_unref (root);
  
  return table;
}

static void
on_apps_tree_changed_cb (GMenuTree *tree,
                         gpointer   user_data)
{
  CdosAppSystem *self = CDOS_APP_SYSTEM (user_data);
  GError *error = NULL;
  GHashTable *new_apps;
  GHashTableIter iter;
  gpointer key, value;
  GSList *removed_apps = NULL;
  GSList *removed_node;

  g_assert (tree == self->priv->apps_tree);

  g_slist_foreach (self->priv->known_vendor_prefixes, (GFunc)g_free, NULL);
  g_slist_free (self->priv->known_vendor_prefixes);
  self->priv->known_vendor_prefixes = NULL;

  if (!gmenu_tree_load_sync (self->priv->apps_tree, &error))
    {
      g_warning ("Failed to load apps: %s", error->message);
      return;
    }
  if(self->priv->apps){
      g_hash_table_destroy (self->priv->apps);
  }

  new_apps = get_flattened_entries_from_tree (self->priv->apps_tree);
  self->priv->apps = new_apps;
  g_hash_table_iter_init (&iter, new_apps);
  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      const char *id = key;
      GMenuTreeEntry *entry = value;
      GMenuTreeEntry *old_entry;
      char *prefix;
      CdosApp *app;
      
      prefix = get_prefix_for_entry (entry);
      
      if (prefix != NULL
          && !g_slist_find_custom (self->priv->known_vendor_prefixes, prefix,
                                   (GCompareFunc)g_strcmp0))
        self->priv->known_vendor_prefixes = g_slist_append (self->priv->known_vendor_prefixes,
                                                            prefix);
      else
        g_free (prefix);
      
      app = g_hash_table_lookup (self->priv->id_to_app, id);
      if (app != NULL)
        {
          /* We hold a reference to the original entry temporarily,
           * because otherwise the hash table would be referencing
           * potentially free'd memory until we replace it below with
           * the new data.
           */
          old_entry = cdos_app_get_tree_entry (app);
          gmenu_tree_item_ref (old_entry);
          _cdos_app_set_entry (app, entry);
          g_object_ref (app);  /* Extra ref, removed in _replace below */
        }
      else
        {
          old_entry = NULL;
          app = _cdos_app_new (entry);
	}
      /* Note that "id" is owned by app->entry.  Since we're always
       * setting a new entry, even if the app already exists in the
       * hash table we need to replace the key so that the new id
       * string is pointed to.
       */
      g_hash_table_replace (self->priv->id_to_app, (char*)id, app);

      if (old_entry)
        gmenu_tree_item_unref (old_entry);
    }
  /* Now iterate over the apps again; we need to unreference any apps
   * which have been removed.  The JS code may still be holding a
   * reference; that's fine.
   */
  g_hash_table_iter_init (&iter, self->priv->id_to_app);
  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      const char *id = key;
      
      if (!g_hash_table_lookup (new_apps, id))
        removed_apps = g_slist_prepend (removed_apps, (char*)id);
    }
  for (removed_node = removed_apps; removed_node; removed_node = removed_node->next)
    {
      const char *id = removed_node->data;
      g_hash_table_remove (self->priv->id_to_app, id);
    }
  g_slist_free (removed_apps);
      
  //g_hash_table_destroy (new_apps);


  g_signal_emit (self, signals[INSTALLED_CHANGED], 0);
}

/**
 * cdos_app_system_get_tree:
 *
 * Return Value: 应用程序的GMenuTree对象
 */
GMenuTree *
cdos_app_system_get_tree (CdosAppSystem *self)
{
  return self->priv->apps_tree;
}

/**
 * cdos_app_system_get_settings_tree:
 *
 * Return Value: 应用程序的GMenuTree对象
 * 已淘汰 - 保留仅仅是为了兼容性
 * 返回空的GMenuTree对象
 */
GMenuTree *
cdos_app_system_get_settings_tree (CdosAppSystem *self)
{
  return self->priv->settings_tree;
}
/**
 * cdos_app_system_lookup_setting:
 *
 * Returns: gnomecc.menu里的应用程序，如果没有，则返回NULL
 * 已淘汰 - 保留仅仅是为了兼容性
 * 如果不在标准应用程序中，则返回NULL
 */
CdosApp *
cdos_app_system_lookup_setting (CdosAppSystem *self,
                                 const char     *id)
{
  CdosApp *app;
  /* Actually defer to the main app set if there's overlap */
  app = cdos_app_system_lookup_app (self, id);
  if (app != NULL)
    return app;
  return NULL;
}

/**
 * cdos_app_system_get_default:
 *
 * Return Value: 全局CdosAppSystem单例模式
 */
CdosAppSystem *
cdos_app_system_get_default ()
{
  static CdosAppSystem *instance = NULL;

  if (instance == NULL)
    instance = g_object_new (CDOS_TYPE_APP_SYSTEM, NULL);

  return instance;
}

gboolean
case_insensitive_search (const char *key,
                         const char *value,
                         gpointer user_data)
{
  char *given_id = (char *) user_data;

  if (g_ascii_strcasecmp(key, given_id) == 0) {
    return TRUE;
  } else {
    return FALSE;
  }
}

/**
 * cdos_app_system_lookup_app:
 *
 * 通过id找到相关的CdosApp对象。
 *
 * Return value: id对应的CdosApp对象，如果没有，则返回NULL
 */
CdosApp *
cdos_app_system_lookup_app (CdosAppSystem   *self,
                             const char       *id)
{
  CdosApp *result;

  result = g_hash_table_lookup (self->priv->id_to_app, id);
  if (result == NULL) {
    result = g_hash_table_find (self->priv->id_to_app, (GHRFunc) case_insensitive_search, id);
  }
  return result;
}

/**
 * cdos_app_system_lookup_settings_app:
 *
 * Return value: id对应的CdosApp对象，如果没有，则返回NULL
 * 已淘汰 - 保留仅仅是为了兼容性
 * 返回值为NULL
 *
 */
CdosApp *
cdos_app_system_lookup_settings_app (CdosAppSystem   *self,
                             const char       *id)
{
  return NULL;
}

/**
 * cdos_app_system_lookup_app_by_tree_entry:
 * @self: CdosAppSystem对象
 *
 * 查找GMenuTreeEntry对应的CdosApp对象
 *
 * Return value: entry对应的CdosApp对象，如果没有，则返回NULL
 */
CdosApp *
cdos_app_system_lookup_app_by_tree_entry (CdosAppSystem  *self,
                                           GMenuTreeEntry  *entry)
{
  /* If we looked up directly in ->entry_to_app, we'd lose the
   * override of running apps.  Thus, indirect through the id.
   */
  return cdos_app_system_lookup_app (self, gmenu_tree_entry_get_desktop_file_id (entry));
}


/**
 * cdos_app_system_lookup_settings_app_by_tree_entry:
 *
 * Return value: entry对应的CdosApp对象，如果没有，则返回NULL
 * 已废弃 - 保留仅仅是为了兼容性
 * 返回值为NULL
 */
CdosApp *
cdos_app_system_lookup_settings_app_by_tree_entry (CdosAppSystem  *self,
                                           GMenuTreeEntry  *entry)
{
  return NULL;
}

/**
 * cdos_app_system_lookup_app_for_path:
 * @system: CdosAppSystem对象
 * @desktop_path: UTF-8编码格式的完全文件名
 *
 * 查找或者创建有标准路径的完全文件名的CdosApp应用程序。对于数据目录的外部文
 * 件，该函数返回NULL。
 *
 * Return value: 有id的CdosApp对象，如果没有，则返回NULL
 */
CdosApp *
cdos_app_system_lookup_app_for_path (CdosAppSystem   *system,
                                      const char       *desktop_path)
{
  const char *basename;
  const char *app_path;
  CdosApp *app;

  basename = g_strrstr (desktop_path, "/");
  if (basename)
    basename += 1;
  else
    basename = desktop_path;

  app = cdos_app_system_lookup_app (system, basename);
  if (!app)
    return NULL;

  app_path = gmenu_tree_entry_get_desktop_file_path (cdos_app_get_tree_entry (app));
  if (strcmp (desktop_path, app_path) != 0)
    return NULL;

  return app;
}

/**
 * cdos_app_system_lookup_heuristic_basename:
 * @system: CdosAppSystem对象
 * @id: 应用程序标识符
 *
 * 查找检测应用程序标识符串的有效应用程序，如果没有，则返回NULL。
 *
 * Returns: 有name的CdosApp对象
 */
CdosApp *
cdos_app_system_lookup_heuristic_basename (CdosAppSystem *system,
                                            const char     *name)
{
  CdosApp *result;
  GSList *prefix;

  result = cdos_app_system_lookup_app (system, name);
  if (result != NULL)
    return result;

  result = cdos_app_system_lookup_settings_app (system, name);
  if (result != NULL)
    return result;

  for (prefix = system->priv->known_vendor_prefixes; prefix; prefix = g_slist_next (prefix))
    {
      char *tmpid = g_strconcat ((char*)prefix->data, name, NULL);
      result = cdos_app_system_lookup_app (system, tmpid);
      //g_free (tmpid);
      if (result != NULL)
        return result;
    }

  return NULL;
}

/**
 * cdos_app_system_get_all:
 * @self: CdosAppSystem对象
 *
 * Returns: 所有安装的应用程序
 */
GSList *
cdos_app_system_get_all (CdosAppSystem  *self)
{
  GSList *result = NULL;
  GHashTableIter iter;
  gpointer key, value;

  g_hash_table_iter_init (&iter, self->priv->id_to_app);
  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      CdosApp *app = value;
      
      if (!g_desktop_app_info_get_nodisplay (cdos_app_get_app_info (app)))
        result = g_slist_prepend (result, app);
    }
  return result;
}

void
_cdos_app_system_notify_app_state_changed (CdosAppSystem *self,
                                            CdosApp       *app)
{
  CdosAppState state = cdos_app_get_state (app);

  switch (state)
    {
    case CDOS_APP_STATE_RUNNING:
      g_hash_table_insert (self->priv->running_apps, g_object_ref (app), NULL);
      break;
    case CDOS_APP_STATE_STARTING:
      break;
    case CDOS_APP_STATE_STOPPED:
      g_hash_table_remove (self->priv->running_apps, app);
      break;
    }

  if (NULL != cdos_app_get_id (app)) {
    g_signal_emit (self, signals[APP_STATE_CHANGED], 0, app);
  }
}

/**
 * cdos_app_system_get_running:
 * @self: CdosAppSystem对象
 *
 * 返回环境中只有有一个打开窗口的应用程序集合。返回列表以cdos_app_compare()排序。
 *
 * Returns: 激活的应用程序
 */
GSList *
cdos_app_system_get_running (CdosAppSystem *self)
{
  gpointer key, value;
  GSList *ret;
  GHashTableIter iter;

  g_hash_table_iter_init (&iter, self->priv->running_apps);

  ret = NULL;
  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      CdosApp *app = key;

      ret = g_slist_prepend (ret, app);
    }

  ret = g_slist_sort (ret, (GCompareFunc)cdos_app_compare);

  return ret;
}

/**
 * normalize_terms:
 * @terms: 输入查询条件
 *
 * Returns: Unicode标准格式的小写术语
 */
static GSList *
normalize_terms (GSList *terms)
{
  GSList *normalized_terms = NULL;
  GSList *iter;
  for (iter = terms; iter; iter = iter->next)
    {
      const char *term = iter->data;
      normalized_terms = g_slist_prepend (normalized_terms, cdos_base_util_normalize_and_casefold (term));
    }
  return normalized_terms;
}

static GSList *
search_tree (CdosAppSystem *self,
             GSList         *terms,
             GHashTable     *apps)
{
  GSList *prefix_results = NULL;
  GSList *substring_results = NULL;
  GSList *normalized_terms;
  GHashTableIter iter;
  gpointer key, value;

  normalized_terms = normalize_terms (terms);

  g_hash_table_iter_init (&iter, apps);
  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      const char *id = key;
      CdosApp *app = value;
      (void)id;
      _cdos_app_do_match (app, normalized_terms,
                           &prefix_results,
                           &substring_results);
    }
  g_slist_foreach (normalized_terms, (GFunc)g_free, NULL);
  g_slist_free (normalized_terms);

  return g_slist_concat (prefix_results, substring_results);
}

/**
 * cdos_app_system_initial_search:
 * @system: CdosAppSystem对象
 * @terms: 术语列表，逻辑与
 *
 * 根据搜索条件查找应用程序。
 *
 * Returns: 应用程序列表
 */
GSList *
cdos_app_system_initial_search (CdosAppSystem  *self,
                                 GSList          *terms)
{
  return search_tree (self, terms, self->priv->id_to_app);
}

/**
 * cdos_app_system_subsearch:
 * @system: CdosAppSystem对象
 * @previous_results: 先前结果列表
 * @terms: 术语列表，逻辑与
 *
 * 查找以前的结果集合，想了解更多，查看js/ui/search.js。
 *   注意，prefs的值必须和传递到cdos_app_system_initial_search()的值一样；
 * Search through a previous result set; for more information, see
 * js/ui/search.js.  Note the value of @prefs must be
 * the same as passed to cdos_app_system_initial_search().  Note that returned
 * strings are only valid until a return to the main loop.
 *
 * Returns: 应用程序标识符列表
 */
GSList *
cdos_app_system_subsearch (CdosAppSystem   *system,
                            GSList           *previous_results,
                            GSList           *terms)
{
  GSList *iter;
  GSList *prefix_results = NULL;
  GSList *substring_results = NULL;
  GSList *normalized_terms = normalize_terms (terms);

  for (iter = previous_results; iter; iter = iter->next)
    {
      CdosApp *app = iter->data;
      
      _cdos_app_do_match (app, normalized_terms,
                           &prefix_results,
                           &substring_results);
    }
  g_slist_foreach (normalized_terms, (GFunc)g_free, NULL);
  g_slist_free (normalized_terms);

  /* Note that a shorter term might have matched as a prefix, but
     when extended only as a substring, so we have to redo the
     sort rather than reusing the existing ordering */
  return g_slist_concat (prefix_results, substring_results);
}

/**
 * cdos_app_system_search_settings:
 * @system: CdosAppSystem对象
 * @terms: 术语列表，逻辑与
 *
 * 搜索给定查询条件的设置
 *
 * Returns:应用程序设置列表
 */
GSList *
cdos_app_system_search_settings (CdosAppSystem  *self,
                                  GSList          *terms)
{
  return NULL;
}


GMenuTreeEntry *
cdos_app_system_lookup_entry_by_name    (CdosAppSystem *system, gchar *name_str){
    g_return_val_if_fail(CDOS_IS_APP_SYSTEM(system),NULL);
    g_return_val_if_fail(name_str != NULL,NULL);
    GHashTable *apps = system->priv->apps;
    GList *apps_keys = g_hash_table_get_keys(apps);
    GList *by_ids = apps_keys;
    gchar *target_key = NULL;
    GMenuTreeEntry *target_entry = NULL;
    gchar *name = cdos_base_util_str_tolower(name_str);
    name = name_str;
    while (by_ids){
        gchar *id = by_ids->data;
        if(id && g_str_match_string(name,id,FALSE)){
            target_key = id;
            break ;
        }
        by_ids = by_ids->next;
    }
    if(target_key){
        target_entry = g_hash_table_lookup(apps,target_key);
        //g_free(name);
        return target_entry;
    }
    GList *apps_values = g_hash_table_get_values(apps);
    gboolean flag = FALSE;
    while (apps_values){
        GMenuTreeEntry *entry = apps_values->data;
        GDesktopAppInfo *app_info = gmenu_tree_entry_get_app_info(entry);
        gchar *app_name = g_desktop_app_info_get_string(app_info,"Name");
        gchar *wm_class = g_desktop_app_info_get_string(app_info,"StartupWMClass");
        if((app_name && g_str_match_string(name, app_name, FALSE)) || (wm_class && g_str_match_string(name, wm_class, FALSE))) {
            target_entry = entry;
            flag = TRUE;
        }
        //g_free(app_name);
        //g_free(exec);
        if(flag == TRUE){
            break;
        }
        apps_values = apps_values->next;
    }
    /*
     * VALGRIND shows that 'name' is already free'd
     * Actually this 'name' is a stack var
     * */
   // g_free(name);
    return target_entry;
}

static gchar *
strip_extension (const gchar *wm_class)
{
    char *result;
    if (g_str_has_suffix (wm_class, ".py") ||
        g_str_has_suffix (wm_class, ".sh")) {
            result = g_strndup (wm_class, strlen (wm_class) - 3);
    } else {
        result = g_strdup (wm_class);
    }
    return result;
}

CdosApp *
cdos_app_system_lookup_desktop_wmclass(CdosAppSystem *system,
                                       const char *wmclass)
{
  char *canonicalized;
  char *desktop_file;
  char *stripped_name;
  CdosApp *app;

  if (wmclass == NULL)
    return NULL;

  canonicalized = g_ascii_strdown(wmclass, -1);
  stripped_name = strip_extension(canonicalized);
  /* This handles "Fedora Eclipse", probably others.
   * Note g_strdelimit is modify-in-place. */
  g_strdelimit(stripped_name, " ", '-');

  desktop_file = g_strconcat(stripped_name, ".desktop", NULL);
  app = cdos_app_system_lookup_heuristic_basename(system, desktop_file);

  g_free(canonicalized);
  g_free(stripped_name);
  g_free(desktop_file);

  return app;
}

CdosApp *
cdos_app_system_lookup_startup_wmclass (CdosAppSystem *system,
                                        const char *wmclass)
{
  gchar *id = NULL, *lower_wmclass = NULL;
  GDesktopAppInfo *info = NULL;
  CdosApp *result = NULL;
  if (wmclass == NULL)
  {
    return NULL;
  }

  lower_wmclass = g_strdown (wmclass);
  GList *app_list = g_hash_table_get_values (system->priv->id_to_app);

  while (app_list)
  {
    CdosApp *app = app_list->data;
    info = cdos_app_get_app_info (app);
    id = g_desktop_app_info_get_startup_wm_class (info);
    if (id && !g_strcmp0 (id, lower_wmclass))
    {
      result = app;
      g_debug ("\nCDOS-DEBUG: GET IT! %s\n", id);
      break;
    }
    app_list = app_list->next;
  }

  return result;
}
