/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

/**
 * SECTION:cdos-generic-container
 * @short_description: A container class with signals for allocation
 *
 * #CdosGenericContainer is mainly a workaround for the current
 * lack of GObject subclassing + vfunc overrides in gjs.  We
 * implement the container interface, but proxy the virtual functions
 * into signals, which gjs can catch.
 *
 * #CdosGenericContainer is an #StWidget, and automatically takes its
 * borders and padding into account during size request and allocation.
 */

#include "config.h"

#include "cdos-generic-container.h"

#include <clutter/clutter.h>
#include <gtk/gtk.h>
#include <girepository.h>

static void cdos_generic_container_iface_init (ClutterContainerIface *iface);

G_DEFINE_TYPE_WITH_CODE(CdosGenericContainer,
                        cdos_generic_container,
                        ST_TYPE_CONTAINER,
                        G_IMPLEMENT_INTERFACE (CLUTTER_TYPE_CONTAINER,
                                               cdos_generic_container_iface_init));

struct _CdosGenericContainerPrivate {
  GHashTable *skip_paint;
};

/* Signals */
enum
{
  GET_PREFERRED_WIDTH,
  GET_PREFERRED_HEIGHT,
  ALLOCATE,
  LAST_SIGNAL
};

static guint cdos_generic_container_signals [LAST_SIGNAL] = { 0 };

static gpointer
cdos_generic_container_allocation_ref (CdosGenericContainerAllocation *alloc)
{
  alloc->_refcount++;
  return alloc;
}

static void
cdos_generic_container_allocation_unref (CdosGenericContainerAllocation *alloc)
{
  if (--alloc->_refcount == 0)
    g_slice_free (CdosGenericContainerAllocation, alloc);
}

static void
cdos_generic_container_allocate (ClutterActor           *self,
                                  const ClutterActorBox  *box,
                                  ClutterAllocationFlags  flags)
{
  StThemeNode *theme_node;
  ClutterActorBox content_box;

  CLUTTER_ACTOR_CLASS (cdos_generic_container_parent_class)->allocate (self, box, flags);

  theme_node = st_widget_get_theme_node (ST_WIDGET (self));
  st_theme_node_get_content_box (theme_node, box, &content_box);

  /* flags 修改 */
  g_signal_emit (G_OBJECT (self), cdos_generic_container_signals[ALLOCATE], 0,
                 &content_box, (guint)flags);
}

static void
cdos_generic_container_get_preferred_width (ClutterActor *actor,
                                             gfloat        for_height,
                                             gfloat       *min_width_p,
                                             gfloat       *natural_width_p)
{
  CdosGenericContainerAllocation *alloc = g_slice_new0 (CdosGenericContainerAllocation);
  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));

  st_theme_node_adjust_for_height (theme_node, &for_height);

  alloc->_refcount = 1;
  g_signal_emit (G_OBJECT (actor), cdos_generic_container_signals[GET_PREFERRED_WIDTH], 0,
                 for_height, alloc);
  if (min_width_p)
    *min_width_p = alloc->min_size;
  if (natural_width_p)
    *natural_width_p = alloc->natural_size;
  cdos_generic_container_allocation_unref (alloc);

  st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
}

static void
cdos_generic_container_get_preferred_height (ClutterActor *actor,
                                              gfloat        for_width,
                                              gfloat       *min_height_p,
                                              gfloat       *natural_height_p)
{
  CdosGenericContainerAllocation *alloc = g_slice_new0 (CdosGenericContainerAllocation);
  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));

  st_theme_node_adjust_for_width (theme_node, &for_width);

  alloc->_refcount = 1;
  g_signal_emit (G_OBJECT (actor), cdos_generic_container_signals[GET_PREFERRED_HEIGHT], 0,
                 for_width, alloc);
  if (min_height_p)
    *min_height_p = alloc->min_size;
  if (natural_height_p)
    *natural_height_p = alloc->natural_size;
  cdos_generic_container_allocation_unref (alloc);

  st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
}

static void
cdos_generic_container_paint (ClutterActor  *actor)
{
  CdosGenericContainer *self = (CdosGenericContainer*) actor;
  GList *iter, *children;

  CLUTTER_ACTOR_CLASS (cdos_generic_container_parent_class)->paint (actor);

  children = st_container_get_children_list (ST_CONTAINER (actor));
  for (iter = children; iter; iter = iter->next)
    {
      ClutterActor *child = iter->data;

      if (g_hash_table_lookup (self->priv->skip_paint, child))
        continue;

      clutter_actor_paint (child);
    }
}

static void
cdos_generic_container_pick (ClutterActor        *actor,
                              const ClutterColor  *color)
{
  CdosGenericContainer *self = (CdosGenericContainer*) actor;
  GList *iter, *children;

  CLUTTER_ACTOR_CLASS (cdos_generic_container_parent_class)->pick (actor, color);

  children = st_container_get_children_list (ST_CONTAINER (actor));
  for (iter = children; iter; iter = iter->next)
    {
      ClutterActor *child = iter->data;

      if (g_hash_table_lookup (self->priv->skip_paint, child))
        continue;

      clutter_actor_paint (child);
    }
}

static GList *
cdos_generic_container_get_focus_chain (StContainer *container)
{
  CdosGenericContainer *self = CDOS_GENERIC_CONTAINER (container);
  GList *children, *focus_chain;

  focus_chain = NULL;
  for (children = st_container_get_children_list (container); children; children = children->next)
    {
      ClutterActor *child = children->data;

      if (CLUTTER_ACTOR_IS_VISIBLE (child) &&
          !cdos_generic_container_get_skip_paint (self, child))
        focus_chain = g_list_prepend (focus_chain, child);
    }

  return g_list_reverse (focus_chain);
}

/**
 * cdos_generic_container_get_n_skip_paint:
 * @self:  CdosGenericContainer对象
 *
 * Returns: 没被着色的actor数目
 */
guint
cdos_generic_container_get_n_skip_paint (CdosGenericContainer  *self)
{
  return g_hash_table_size (self->priv->skip_paint);
}

/**
 * cdos_generic_container_get_skip_paint:
 * @self: CdosGenericContainer对象
 * @child: ClutterActor对象
 *
 * 当着色时，获取actor是否要被着色的状态
 *
 * Return value: TRUE 或 FALSE
 */
gboolean
cdos_generic_container_get_skip_paint (CdosGenericContainer  *self,
                                        ClutterActor           *child)
{
  return g_hash_table_lookup (self->priv->skip_paint, child) != NULL;
}

/**
 * cdos_generic_container_set_skip_paint:
 * @self: CdosGenericContainer对象
 * @child: ClutterActor对象
 * @skip: 如果不着色，值为TRUE
 *
 * 设置是否应该跳过actor的着色。缺少gjs能力工作区用vfunc替代。
 */
void
cdos_generic_container_set_skip_paint (CdosGenericContainer  *self,
                                        ClutterActor           *child,
                                        gboolean                skip)
{
  gboolean currently_skipping;

  currently_skipping = g_hash_table_lookup (self->priv->skip_paint, child) != NULL;
  if (!!skip == currently_skipping)
    return;

  if (!skip)
    g_hash_table_remove (self->priv->skip_paint, child);
  else
    g_hash_table_insert (self->priv->skip_paint, child, child);

  clutter_actor_queue_redraw (CLUTTER_ACTOR (self));
}

static void
cdos_generic_container_finalize (GObject *object)
{
  CdosGenericContainer *self = (CdosGenericContainer*) object;

  g_hash_table_destroy (self->priv->skip_paint);

  G_OBJECT_CLASS (cdos_generic_container_parent_class)->finalize (object);
}

static void
cdos_generic_container_class_init (CdosGenericContainerClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
  StContainerClass *container_class = ST_CONTAINER_CLASS (klass);

  gobject_class->finalize = cdos_generic_container_finalize;

  actor_class->get_preferred_width = cdos_generic_container_get_preferred_width;
  actor_class->get_preferred_height = cdos_generic_container_get_preferred_height;
  actor_class->allocate = cdos_generic_container_allocate;
  actor_class->paint = cdos_generic_container_paint;
  actor_class->pick = cdos_generic_container_pick;

  container_class->get_focus_chain = cdos_generic_container_get_focus_chain;

  /**
   * CdosGenericContainer::get-preferred-width:
   * @self: CdosGenericContainer对象
   * @for_height: clutter_actor_get_preferred_width()的值
   * @alloc: 填充到CdosGenericContainerAllocation对象的值
   *
   * 当调用when clutter_actor_get_preferred_width()时发送信号。
   * 用最小值和中间宽度值填充alloc的值。
   * Emitted when clutter_actor_get_preferred_width() is called
   * on @self. You should fill in the fields of @alloc with the
   * your minimum and natural widths. #CdosGenericContainer
   * will deal with taking its borders and padding into account
   * for you.
   *
   * alloc结构值被初始化为0，因此除非已经填充指定width值了，否则需要连
   * 接该信号为了填充这个值。
   */
  cdos_generic_container_signals[GET_PREFERRED_WIDTH] =
    g_signal_new ("get-preferred-width",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  0,
                  NULL, NULL,
                  gi_cclosure_marshal_generic,
                  G_TYPE_NONE, 2, G_TYPE_FLOAT, CDOS_TYPE_GENERIC_CONTAINER_ALLOCATION);

  /**
   * CdosGenericContainer::get-preferred-height:
   * @self: CdosGenericContainer对象
   * @for_width: clutter_actor_get_preferred_height()的值
   * @alloc: 填充到CdosGenericContainerAllocation对象的值
   *
   * 当clutter_actor_get_preferred_height()被调用时发送信号。
   * 用最小值和中间高度填充alloc结构值。
   * Emitted when clutter_actor_get_preferred_height() is called
   * on @self. You should fill in the fields of @alloc with the
   * your minimum and natural heights. #CdosGenericContainer
   * will deal with taking its borders and padding into account
   * for you.
   *
   * alloc结构体被初始化为0,因此除非已填充指定的height值，否则需要连接
   * 该信号为了填充这个值。
   */
  cdos_generic_container_signals[GET_PREFERRED_HEIGHT] =
    g_signal_new ("get-preferred-height",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  0,
                  NULL, NULL,
                  gi_cclosure_marshal_generic,
                  G_TYPE_NONE, 2, G_TYPE_FLOAT, CDOS_TYPE_GENERIC_CONTAINER_ALLOCATION);

  /**
   * CdosGenericContainer::allocate:
   * @self: CdosGenericContainer对象
   * @box: self的内部盒子
   * @flags: 分配标志
   *
   * 链接到父分配方法后，如果self对象分配完成，则发送信号。
   *
   * 注意：box是self的内部盒子，不是本身分配的。
   */
  /* [leit 02.16] CLUTTER_TYPE_ALLOCATION_FLAGS -> G_TYPE_UINT */
  cdos_generic_container_signals[ALLOCATE] =
    g_signal_new ("allocate",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  0,
                  NULL, NULL,
                  gi_cclosure_marshal_generic,
                  G_TYPE_NONE, 2, CLUTTER_TYPE_ACTOR_BOX, G_TYPE_UINT /*CLUTTER_TYPE_ALLOCATION_FLAGS*/);

  g_type_class_add_private (gobject_class, sizeof (CdosGenericContainerPrivate));
}

static void
cdos_generic_container_actor_removed (ClutterContainer *container,
                                       ClutterActor     *actor)
{
  CdosGenericContainerPrivate *priv = CDOS_GENERIC_CONTAINER (container)->priv;

  g_hash_table_remove (priv->skip_paint, actor);
}

static void
cdos_generic_container_iface_init (ClutterContainerIface *iface)
{
  iface->actor_removed = cdos_generic_container_actor_removed;
}

static void
cdos_generic_container_init (CdosGenericContainer *area)
{
  area->priv = G_TYPE_INSTANCE_GET_PRIVATE (area, CDOS_TYPE_GENERIC_CONTAINER,
                                            CdosGenericContainerPrivate);
  area->priv->skip_paint = g_hash_table_new (NULL, NULL);
}

GType
cdos_generic_container_allocation_get_type (void)
{
  static GType gtype = G_TYPE_INVALID;
  if (gtype == G_TYPE_INVALID)
    {
      gtype = g_boxed_type_register_static ("CdosGenericContainerAllocation",
         (GBoxedCopyFunc)cdos_generic_container_allocation_ref,
         (GBoxedFreeFunc)cdos_generic_container_allocation_unref);
    }
  return gtype;
}

ClutterActor *
cdos_generic_container_new (void)
{
    return g_object_new (CDOS_TYPE_GENERIC_CONTAINER, NULL);
}
