/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*- */

#include "config.h"

#define GST_USE_UNSTABLE_API
#include <gst/base/gstpushsrc.h>

#include "cdos-recorder-src.h"

struct _CdosRecorderSrc
{
  GstPushSrc parent;

  GMutex *mutex;

  GstClock *clock;
  GstClockTime last_frame_time;

  GstCaps *caps;
  GAsyncQueue *queue;
  gboolean closed;
  guint memory_used;
  guint memory_used_update_idle;
};

struct _CdosRecorderSrcClass
{
  GstPushSrcClass parent_class;
};

enum {
  PROP_0,
  PROP_CAPS,
  PROP_MEMORY_USED
};

/* Special marker value once the source is closed */
#define RECORDER_QUEUE_END ((GstBuffer *)1)

G_DEFINE_TYPE(CdosRecorderSrc, cdos_recorder_src, GST_TYPE_PUSH_SRC);

static void
cdos_recorder_src_init (CdosRecorderSrc      *src)
{
  gst_base_src_set_format (GST_BASE_SRC (src), GST_FORMAT_TIME);
  gst_base_src_set_live (GST_BASE_SRC (src), TRUE);

  src->clock = gst_system_clock_obtain ();
  src->last_frame_time = 0;

  src->queue = g_async_queue_new ();
  src->mutex = g_mutex_new ();
}

static gboolean
cdos_recorder_src_memory_used_update_idle (gpointer data)
{
  CdosRecorderSrc *src = data;

  g_mutex_lock (src->mutex);
  src->memory_used_update_idle = 0;
  g_mutex_unlock (src->mutex);

  g_object_notify (G_OBJECT (src), "memory-used");

  return FALSE;
}

/* The memory_used property is used to monitor buffer usage,
 * so we marshal notification back to the main loop thread.
 */
static void
cdos_recorder_src_update_memory_used (CdosRecorderSrc *src,
				       int               delta)
{
  g_mutex_lock (src->mutex);
  src->memory_used += delta;
  if (src->memory_used_update_idle == 0)
    src->memory_used_update_idle = g_idle_add (cdos_recorder_src_memory_used_update_idle, src);
  g_mutex_unlock (src->mutex);
}

/* The create() virtual function is responsible for returning the next buffer.
 * We just pop buffers off of the queue and block if necessary.
 */
static GstFlowReturn
cdos_recorder_src_create (GstPushSrc  *push_src,
			   GstBuffer  **buffer_out)
{
  CdosRecorderSrc *src = CDOS_RECORDER_SRC (push_src);
  GstBuffer *buffer;

  if (src->closed)
    return GST_FLOW_EOS;

  buffer = g_async_queue_pop (src->queue);
  if (src->last_frame_time == 0)
    src->last_frame_time = gst_clock_get_time (GST_CLOCK (src->clock));

  if (buffer == RECORDER_QUEUE_END)
    {
      /* Returning UNEXPECTED here will cause a EOS message to be sent */
      src->closed = TRUE;
      return GST_FLOW_EOS;
    }

  cdos_recorder_src_update_memory_used (src,
					 - (int)(gst_buffer_get_size(buffer) / 1024));

  *buffer_out = buffer;

  GST_BUFFER_DURATION(*buffer_out) = GST_CLOCK_DIFF (src->last_frame_time, gst_clock_get_time (GST_CLOCK (src->clock)));

  src->last_frame_time = gst_clock_get_time (GST_CLOCK (src->clock));

  return GST_FLOW_OK;
}

static void
cdos_recorder_src_set_caps (CdosRecorderSrc *src,
			     const GstCaps    *caps)
{
  if (caps == src->caps)
    return;

  if (src->caps != NULL)
    {
      gst_caps_unref (src->caps);
      src->caps = NULL;
    }

  if (caps)
    {
      /* The capabilities will be negotated with the downstream element
       * and set on the pad when the first buffer is pushed.
       */
      src->caps = gst_caps_copy (caps);
    }
  else
    src->caps = NULL;
}

static void
cdos_recorder_src_finalize (GObject *object)
{
  CdosRecorderSrc *src = CDOS_RECORDER_SRC (object);

  if (src->memory_used_update_idle) {
    g_source_remove (src->memory_used_update_idle);
    src->memory_used_update_idle = 0;
  }

  cdos_recorder_src_set_caps (src, NULL);
  g_async_queue_unref (src->queue);

  g_mutex_free (src->mutex);

  gst_object_unref (src->clock);

  G_OBJECT_CLASS (cdos_recorder_src_parent_class)->finalize (object);
}

static void
cdos_recorder_src_set_property (GObject      *object,
				 guint         prop_id,
				 const GValue *value,
				 GParamSpec   *pspec)
{
  CdosRecorderSrc *src = CDOS_RECORDER_SRC (object);

  switch (prop_id)
    {
    case PROP_CAPS:
      cdos_recorder_src_set_caps (src, gst_value_get_caps (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
cdos_recorder_src_get_property (GObject         *object,
				 guint            prop_id,
				 GValue          *value,
				 GParamSpec      *pspec)
{
  CdosRecorderSrc *src = CDOS_RECORDER_SRC (object);

  switch (prop_id)
    {
    case PROP_CAPS:
      gst_value_set_caps (value, src->caps);
      break;
    case PROP_MEMORY_USED:
      g_mutex_lock (src->mutex);
      g_value_set_uint (value, src->memory_used);
      g_mutex_unlock (src->mutex);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
cdos_recorder_src_class_init (CdosRecorderSrcClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
  GstPushSrcClass *push_src_class = GST_PUSH_SRC_CLASS (klass);

  static GstStaticPadTemplate src_template =
    GST_STATIC_PAD_TEMPLATE ("src",
			     GST_PAD_SRC,
			     GST_PAD_ALWAYS,
			     GST_STATIC_CAPS_ANY);

  object_class->finalize = cdos_recorder_src_finalize;
  object_class->set_property = cdos_recorder_src_set_property;
  object_class->get_property = cdos_recorder_src_get_property;

  push_src_class->create = cdos_recorder_src_create;

  g_object_class_install_property (object_class,
                                   PROP_CAPS,
                                   g_param_spec_boxed ("caps",
						       "Caps",
						       "Fixed GstCaps for the source",
						       GST_TYPE_CAPS,
						       G_PARAM_READWRITE));
  g_object_class_install_property (object_class,
                                   PROP_MEMORY_USED,
                                   g_param_spec_uint ("memory-used",
						     "Memory Used",
						     "Memory currently used by the queue (in kB)",
						      0, G_MAXUINT, 0,
						      G_PARAM_READABLE));
  gst_element_class_add_pad_template (element_class,
				      gst_static_pad_template_get (&src_template));

  gst_element_class_set_details_simple (element_class,
					"CdosRecorderSrc",
					"Generic/Src",
					"Feed screen capture data to a pipeline",
					"Owen Taylor <otaylor@redhat.com>");
}

/**
 * cdos_recorder_src_add_buffer:
 *
 * 增加一个缓冲到内部队列，为了下次机会时被推出。因为没有流的控制，所以队列缓冲可以用任意数量的内存。
 * 缓冲区的内容必须与GstCaps对象集合一致。
 */
void
cdos_recorder_src_add_buffer (CdosRecorderSrc *src,
			       GstBuffer        *buffer)
{
  g_return_if_fail (CDOS_IS_RECORDER_SRC (src));
  g_return_if_fail (src->caps != NULL);

  gst_base_src_set_caps (GST_BASE_SRC (src), src->caps);
  cdos_recorder_src_update_memory_used (src,
					 (int)(gst_buffer_get_size(buffer) / 1024));

  g_async_queue_push (src->queue, gst_buffer_ref (buffer));
}

/**
 * cdos_recorder_src_close:
 *
 * 指示输入流结束。所有添加到缓存的数据都被推出，结束流的消息会被发送。
 */
void
cdos_recorder_src_close (CdosRecorderSrc *src)
{
  /* We can't send a message to the source immediately or buffers that haven't
   * been pushed yet will be discarded. Instead stick a marker onto our own
   * queue to send an event once everything has been pushed.
   */
  g_async_queue_push (src->queue, RECORDER_QUEUE_END);
}

static gboolean
plugin_init (GstPlugin *plugin)
{
  gst_element_register(plugin, "cdosrecordersrc", GST_RANK_NONE,
		       CDOS_TYPE_RECORDER_SRC);

  return TRUE;
}

/**
 * cdos_recorder_src_register:
 * 注册一个在应用程序中单独使用的插件。能够安全的被多次调用。
 */
void
cdos_recorder_src_register (void)
{
  static gboolean registered = FALSE;
  if (registered)
    return;

  gst_plugin_register_static (GST_VERSION_MAJOR, GST_VERSION_MINOR,
			      "cdosrecorder",
			      "Plugin for CdosRecorder",
			      plugin_init,
			      "0.1",
			      "LGPL",
			      "cdos", "cdos", "http://cdos.linuxmint.com");

  registered = TRUE;
}
