#include "cdos-background.h"
#include "cdos-global.h"
#include "cdos-layout-manager.h"
#include <gdk/gdk.h>
/**
 * SECTION:cdos-background
 * @title: CdosBackground
 * @short_description: 加载桌面的背景图片
 *
 * CdosBackground的功能为当启动桌面时,通过CdosBackground类来加载桌面背景图片
 */

#define PRIMARY_COLOR      "primary-color"
#define SECONDARY_COLOR    "secondary-color"
#define COLOR_SHADING_TYPE "color-shading-type"
#define BACKGROUND_STYLE   "picture-options"
#define PICTURE_OPACITY    "picture-opacity"
#define PICTURE_URI        "picture-uri"

struct _CdosBackgroundManager {
    MetaScreen *screen;
    GList *backgrounds;
    gulong monitor_changed_id;
    ClutterActor *container;
    gboolean vignette;
    gboolean control_position;
    gboolean in_destroy;
};

void
cdos_background_destroy (CdosBackground *self)
{
    g_signal_handler_disconnect (self->screen, self->monitor_changed_id);
    g_signal_handler_disconnect (self->settings, self->settings_changed_id);
    if(self->settings != NULL)
        g_object_unref(self->settings);
    g_object_unref (self->file);
    clutter_actor_destroy (self->actor);
    g_object_unref (self->meta_background);
    g_free(self);
}

static ClutterColor *
cdos_background_parse_clutter_color (const gchar *color_string)
{
    GdkRGBA gdk_rgba;
    ClutterColor *clutter_color = clutter_color_alloc();

    gdk_rgba_parse (&gdk_rgba, color_string);

    clutter_color->red = gdk_rgba.red * 255;
    clutter_color->green = gdk_rgba.green * 255;
    clutter_color->blue = gdk_rgba.blue * 255;
    clutter_color->alpha = gdk_rgba.alpha * 255;

    return clutter_color;
}

static void
on_settings_changed(GSettings *settings,
                    gchar *settings_key,
                    CdosBackground *self)
{
    if (self->file)
    {
        g_object_unref (self->file);
        self->file = NULL;
    }
    GDesktopBackgroundStyle style = g_settings_get_enum (self->settings, BACKGROUND_STYLE);

    CdosGlobal *global = cdos_global_get ();
    MetaScreen *screen = cdos_global_get_screen (global);

    if (self->meta_background){
        g_object_unref(self->meta_background);
    }
    self->meta_background = meta_background_new (screen);

    if (style == G_DESKTOP_BACKGROUND_STYLE_NONE)
    {
        gchar *primary_color = g_settings_get_string (self->settings, PRIMARY_COLOR);
        GDesktopBackgroundShading shading_type = g_settings_get_enum (self->settings, COLOR_SHADING_TYPE);

        ClutterColor *bg_color = cdos_background_parse_clutter_color (primary_color);

        if(shading_type == G_DESKTOP_BACKGROUND_SHADING_SOLID){
            meta_background_set_color (self->meta_background, bg_color);
        } else {
            gchar *secondary_color = g_settings_get_string (self->settings, SECONDARY_COLOR);
            ClutterColor *secondary_bg_color = cdos_background_parse_clutter_color (secondary_color);
            meta_background_set_gradient (self->meta_background, shading_type,bg_color, secondary_bg_color);
            clutter_color_free (secondary_bg_color);
            g_free (secondary_color);
        }
        clutter_color_free(bg_color);
        g_free(primary_color);

    } else {
        gchar *uri = g_settings_get_string (self->settings, PICTURE_URI);
        self->file = g_file_new_for_commandline_arg (uri);
        meta_background_set_file (self->meta_background, self->file, style);
        g_free(uri);
    }

    if (!self->actor)
    {
        ClutterActor *background_actor = meta_background_actor_new (screen, self->monitor_index);
        if (self->vignette)
            meta_background_actor_set_vignette ((MetaBackgroundActor *)background_actor, 1, 1.0, 0.0);
        self->actor = background_actor;
	if (CLUTTER_IS_CONTAINER (self->container))
	    clutter_container_add_actor (self->container, self->actor);
	else
	    clutter_actor_add_child (self->container, self->actor);
    }
    meta_background_actor_set_background ((MetaBackgroundActor *)self->actor, self->meta_background);
}

static void
cdos_background_screen_on_monitors_changed(MetaScreen *meta_screen, CdosBackground *self)
{
    MetaRectangle geometry;
    on_settings_changed(self->settings,NULL,self);
    meta_screen_get_monitor_geometry (meta_screen, self->monitor_index, &geometry);
    clutter_actor_set_size(self->actor, geometry.width, geometry.height);
    clutter_actor_set_position(self->actor, geometry.x, geometry.y);
}

/**
 * cdos_background_new:
 *
 * 创建新的背景
 *
 * Returns:新创建的 #CdosBackground
 *
 * Since:1.0.0
 */
CdosBackground *
cdos_background_new (ClutterActor *container,
                     guint monitor_index,
                     gboolean vignette,
                     gboolean control_position)
{
    CdosBackground *self;
    self = g_new0 (CdosBackground, 1);
    if (!self)
        return NULL;
    self->container = container;
    self->settings = g_settings_new ("org.cdos.desktop.background");
    self->vignette = vignette;
    self->monitor_index = monitor_index;
    self->control_position = control_position;

    CdosGlobal *global = cdos_global_get ();
    MetaScreen *screen = cdos_global_get_screen (global);
    self->screen = screen;
    self->monitor_changed_id = g_signal_connect(screen,"monitors-changed",G_CALLBACK(cdos_background_screen_on_monitors_changed),self);

    self->settings_changed_id = g_signal_connect(self->settings,
            "changed",
            G_CALLBACK (on_settings_changed),
            self);
    on_settings_changed (NULL, NULL, self);
    MetaRectangle geometry;
    meta_screen_get_monitor_geometry (screen, monitor_index, &geometry);
    clutter_actor_set_size(self->actor, geometry.width, geometry.height);
    clutter_actor_set_position(self->actor, geometry.x, geometry.y);
    return self;
}


/*
 * CdosBackgroundManager:
 */

static void 
on_actor_destroy (ClutterActor *actor,
                  CdosBackgroundManager *self)
{
    if (self->in_destroy)
        return;
    for (GList *iter = self->backgrounds; iter; iter = iter->next)
    {
        CdosBackground *background = iter->data;
        if (background->actor == actor)
        {
            self->backgrounds = g_list_remove (self->backgrounds, actor);
            if (self->backgrounds == NULL)
                cdos_background_manager_destroy (self);
            return;
        }
    }
}

static void 
cdos_background_manager_on_monitors_changed (MetaScreen *screen,
                                             CdosBackgroundManager *self)
{
    for (GList *iter = self->backgrounds; iter; iter = iter->next)
    {
        cdos_background_destroy (iter->data);
    }
    g_list_free (self->backgrounds);

    self->backgrounds = NULL;
    int n = meta_screen_get_n_monitors (self->screen);
    for (int i = 0; i < n; i++)
    {
        CdosBackground *background = 
            cdos_background_new (self->container, i, self->vignette, self->control_position);
        self->backgrounds = g_list_prepend (self->backgrounds, background);
        g_signal_connect (background->actor, "destroy", G_CALLBACK (on_actor_destroy), self);
    }
}

CdosBackgroundManager *
cdos_background_manager_new (ClutterActor *container,
                             gboolean vignette,
                             gboolean control_position)
{
    CdosBackgroundManager *self = g_new0 (CdosBackgroundManager, 1);
    self->container = container;
    self->vignette = vignette;
    self->control_position = control_position;

    CdosGlobal *global = cdos_global_get ();
    MetaScreen *screen = cdos_global_get_screen (global);
    self->screen = screen;
    self->monitor_changed_id = g_signal_connect(screen, "monitors-changed", 
            G_CALLBACK(cdos_background_manager_on_monitors_changed), self);
    cdos_background_manager_on_monitors_changed (screen, self);
    return self;
}

void
cdos_background_manager_destroy (CdosBackgroundManager *self)
{
    self->in_destroy = 1;
    g_signal_handler_disconnect (self->screen, self->monitor_changed_id);
    for (GList *iter = self->backgrounds; iter; iter = iter->next)
        cdos_background_destroy (iter->data);
    g_list_free (self->backgrounds);
    g_free (self);
}
