/**
 * SECTION: cdos-box-pointer
 * @short_description: 为弹出框提供了带有明显指向标识（如指向关联元素的小三角）的外加容器。
 * @title: CdosBoxPointer
 * @include: cdos-box-pointer.h
 *
 * #CdosBoxPointer为弹出框(#CdosPopupMenu)提供带有小三角标识的外围容器元素，通常作为PopupMenu的包裹元素。
 *
 * CdosBoxPointer的结构成员包括：actor,container,bin,border。
 *
 * CdosBoxPointer结构关系:
 *
 * |__ actor:StBin:主容器元素，不显示
 *
 * |____ container:CdosGernerContainer:子容器元素，不显示
 *
 * |________ bin:StBin:包裹容器元素
 *
 * |________ border:StDrawingArea:标识元素
 *
 */
#include <math.h>
#include "cdos-box-pointer.h"
#include "cdos-base-util.h"
#include "cdos-layout-manager.h"
#include "cdos-debug.h"

#define CDOS_BOX_POINTER_GET_PRIVATE(obj) \
	(G_TYPE_INSTANCE_GET_PRIVATE ((obj), CDOS_TYPE_BOX_POINTER, CdosBoxPointerPrivate))

typedef struct 
{
	gfloat x;
	gfloat y;
}Pos;


struct _CdosBoxPointerPrivate 
{
	gint                 arrow_side;
	gdouble              arrow_origin;
	CdosGenericContainer *container;
	gfloat               x_offset;
	gfloat               y_offset;
	gfloat               x_position;
	gfloat               y_position;
	guint                source_alignment;
	ClutterActor         *source_actor;
	guint                arrow_alignment;
	guint                opacity;
	Pos                  *pos;
};

enum {
	PROP_0,

	PROP_X_OFFSET,
	PROP_Y_OFFSET,
	PROP_X_POSITION,
	PROP_Y_POSITION,
	PROP_OPACITY
};

G_DEFINE_TYPE(CdosBoxPointer, cdos_box_pointer, G_TYPE_OBJECT);

static void cdos_box_pointer_set_x_offset_real(CdosBoxPointer * self, const gfloat offset);
static void cdos_box_pointer_set_y_offset_real(CdosBoxPointer * self, const gfloat offset);
static void cdos_box_pointer_set_x_position_real(CdosBoxPointer *self, const gfloat position);
static void cdos_box_pointer_set_y_position_real(CdosBoxPointer *self, const gfloat position);
static void cdos_box_pointer_set_opacity_real(CdosBoxPointer *self, const gint opacity);


static void 
cdos_box_pointer_adjust_allocation_for_arrow(CdosBoxPointer * self,
					     gboolean       is_width,
		                             CdosGenericContainerAllocation *alloc) 
{

	CdosBoxPointerPrivate *priv;
	gdouble border_width, rise;
	priv = self->priv;
	StThemeNode *theme_node = st_widget_get_theme_node(ST_WIDGET(self->actor));
	border_width = st_theme_node_get_length(theme_node, "-arrow-border-width");
	alloc->min_size += border_width * 2;
	alloc->natural_size += border_width * 2;
	if ((!is_width 
		&& (priv->arrow_side == ST_SIDE_TOP
					|| priv->arrow_side == ST_SIDE_BOTTOM))
			|| (is_width
					&& (priv->arrow_side == ST_SIDE_LEFT
							|| priv->arrow_side == ST_SIDE_RIGHT))) {
		rise = st_theme_node_get_length(theme_node, "-arrow-rise");
		alloc->min_size += rise;
		alloc->natural_size += rise;
	}
}

/**
 * cdos_box_pointer_shift_actor:
 * @self: 当前实例对象
 *
 * 为主容器元素设置一个锚点
 */
void  cdos_box_pointer_shift_actor(CdosBoxPointer * self) {
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv;
	priv = self->priv;
	clutter_actor_set_anchor_point(CLUTTER_ACTOR(self->actor),
				       -(priv->x_position + priv->x_offset),
			               -(priv->y_position + priv->y_offset));
}

static gboolean
cdos_box_pointer_reposition(CdosBoxPointer * self,
		            ClutterActor * source_actor, 
		            gint alignment) 
{
    CdosBoxPointerPrivate *priv;
	priv = self->priv;

	gfloat     res_x = 0;
	gfloat     res_y = 0;
    if(priv->pos)
	{
        priv->x_position = priv->pos->x;
        priv->y_position = priv->pos->y;
        goto finish;
    }
	StThemeNode  *source_node;
	gfloat source_center_x; 
	gfloat source_center_y;
	ClutterActorBox allocation_box;
	ClutterActorBox source_content_box;
	ClutterActorBox source_allocation;
	gfloat min_width;
	gfloat min_height; 
	gfloat nat_width; 
	gfloat nat_height;

	StThemeNode *theme_node;
	gdouble    border_width;
	gdouble    arrow_base;
	gdouble    border_raidus;
	gdouble    margin;
	gdouble    half_margin;
	gdouble    gap;

	source_node = st_widget_get_theme_node(ST_WIDGET(source_actor));

	clutter_actor_get_allocation_box(source_actor, &allocation_box);
	st_theme_node_get_content_box(source_node, &allocation_box, &source_content_box);

	cdos_base_util_get_transformed_allocation(source_actor, &source_allocation);

	source_center_x = source_allocation.x1 + source_content_box.x1
			+ (source_content_box.x2 - source_content_box.x1) * priv->source_alignment;
	source_center_y = source_allocation.y1 + source_content_box.y1
			+ (source_content_box.y2 - source_content_box.y1) * priv->source_alignment;

	clutter_actor_get_preferred_size(CLUTTER_ACTOR(self->actor), &min_width, &min_height, &nat_width, &nat_height);

    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    MetaRectangle *monitor = cdos_layout_manager_find_monitor_for_actor (layout, source_actor);
	
	theme_node = st_widget_get_theme_node(ST_WIDGET(self->actor));

	border_width = st_theme_node_get_length(theme_node,
			"-arrow-border-width");
	arrow_base = st_theme_node_get_length(theme_node, "-arrow-base");
	border_raidus = st_theme_node_get_length(theme_node,
			"arrow-border-radius");

	margin = (4 * border_raidus + border_width + arrow_base);
	half_margin = margin / 2.0;
	gap = st_theme_node_get_length(theme_node, "-boxpointer-gap");

	switch (priv->arrow_side) {
		case ST_SIDE_TOP:
			res_y = source_allocation.y2 + gap;
			break;
		case ST_SIDE_BOTTOM:
			res_y = source_allocation.y1 - nat_height - gap;
			break;
		case ST_SIDE_LEFT:
			res_x = source_allocation.x2 + gap;
			break;
		case ST_SIDE_RIGHT:
			res_x = source_allocation.x1 - nat_width - gap;
			break;
	}
	// Now align and position the pointing axis, making sure
	// it fits on screen
	switch (priv->arrow_side) {
	case ST_SIDE_TOP:
	case ST_SIDE_BOTTOM:
		res_x = source_center_x
				- (half_margin + (nat_width - margin) * alignment);
		//res_x = MAX(res_x, monitor.x + 10);
		/*
		*调整CdosBoxPointer在x轴方向的偏移量为2 pix。如果偏移量<2，开始菜单在显示时会挤开屏幕上已经全屏显示的应用程序
		*/
        if (monitor != NULL)
        {
		    res_x = MAX(res_x, monitor->x + 2);
		    res_x = MIN(res_x, monitor->x + monitor->width - (10 + nat_width));
        }
		cdos_box_pointer_set_arrow_origin(self, source_center_x - res_x);
		break;
	case ST_SIDE_LEFT:
	case ST_SIDE_RIGHT:
		res_y = source_center_y
				- (half_margin + (nat_height - margin) * alignment);
        if (monitor != NULL)
        {
		    res_y = MAX(res_y, monitor->y + 10);
		    res_y = MIN(res_y, monitor->y + monitor->height - (10 + nat_height));
        }
		cdos_box_pointer_set_arrow_origin(self, source_center_y - res_y);
		break;
	}
finish:
    if (priv->pos)
    {
        res_x = priv->pos->x;
        res_y = priv->pos->y;
    }
	// xiaohou
    ClutterActor *parent = clutter_actor_get_parent(CLUTTER_ACTOR(self->actor));
	gboolean success = FALSE;

	gfloat x, y;
	while (!success) {
		success = clutter_actor_transform_stage_point(parent, res_x, res_y, &x, &y);
		parent = clutter_actor_get_parent(parent);
	}
    priv->x_position = floor(x);
	priv->y_position = floor(y);

	cdos_box_pointer_shift_actor(self);
	return TRUE;
}

static void
get_container_preferred_width(CdosGenericContainer *container,
        		      gfloat for_height,
                              CdosGenericContainerAllocation *alloc,
                              gpointer data)
{
	CdosBoxPointer *self = (CdosBoxPointer *)data;
	gfloat min_internal_size, nat_internal_size;

	clutter_actor_get_preferred_width(CLUTTER_ACTOR(self->bin), for_height, &min_internal_size,
			&nat_internal_size);
	alloc->min_size = min_internal_size;
	alloc->natural_size = nat_internal_size;
	cdos_box_pointer_adjust_allocation_for_arrow(self, TRUE, alloc);
}

static void
get_container_preferred_height(CdosGenericContainer *container,
        		       gfloat for_width,
                               CdosGenericContainerAllocation *alloc,
                               gpointer data)
{
	CdosBoxPointer *self = (CdosBoxPointer *)data;
	gfloat min_internal_size, nat_internal_size;

	clutter_actor_get_preferred_height(CLUTTER_ACTOR(self->bin), for_width, &min_internal_size,
			&nat_internal_size);
	alloc->min_size = min_internal_size;
	alloc->natural_size = nat_internal_size;
	cdos_box_pointer_adjust_allocation_for_arrow(self, FALSE, alloc);
}

static void 
container_allocate(CdosGenericContainer *container,
	           ClutterActorBox * box,
		   ClutterAllocationFlags flags,
		   gpointer data) 
{

	CdosBoxPointer * self = (CdosBoxPointer *)data;
	CdosBoxPointerPrivate *priv = self->priv;
	StThemeNode *theme_node = st_widget_get_theme_node(ST_WIDGET(self->actor));
	gdouble border_width = st_theme_node_get_length(theme_node,
			"-arrow-border-width");
	gdouble rise = st_theme_node_get_length(theme_node, "-arrow-rise");
	gfloat avail_width = box->x2 - box->x1;
	gfloat avail_height = box->y2 - box->y1;

	ClutterActorBox child_box = {0, 0, avail_width, avail_height};
	clutter_actor_allocate(CLUTTER_ACTOR(self->border), &child_box, flags);

	child_box.x1 = border_width;
	child_box.y1 = border_width;
	child_box.x2 = avail_width - border_width;
	child_box.y2 = avail_height - border_width;

	switch (priv->arrow_side) {
	case ST_SIDE_TOP:
		child_box.y1 += rise;
		break;
	case ST_SIDE_BOTTOM:
		child_box.y2 -= rise;
		break;
	case ST_SIDE_LEFT:
		child_box.x1 += rise;
		break;
	case ST_SIDE_RIGHT:
		child_box.x2 -= rise;
		break;
	}

	clutter_actor_allocate(CLUTTER_ACTOR(self->bin), &child_box, flags);

	if (priv->source_actor && clutter_actor_is_mapped (priv->source_actor))
		cdos_box_pointer_reposition(self, priv->source_actor,
				priv->arrow_alignment);

}

static void 
draw_border(StDrawingArea * area, CdosBoxPointer * self) 
{
	CdosBoxPointerPrivate *priv = self->priv;
	StThemeNode *theme_node = st_widget_get_theme_node(ST_WIDGET(self->actor));

	gdouble border_width = st_theme_node_get_length(theme_node,
			"-arrow-border-width");
	gdouble base = st_theme_node_get_length(theme_node, "-arrow-base");
	gdouble rise = st_theme_node_get_length(theme_node, "-arrow-rise");
	gdouble border_raidus = st_theme_node_get_length(theme_node,
			"-arrow-border-radius");

	gdouble half_border = border_width / 2;
	gdouble half_base = floor(base / 2);

	ClutterColor border_color, background_color;
	GdkColor gdk_border_color,gdk_background_color;
	st_theme_node_get_color(theme_node, "-arrow-border-color", &border_color);
	st_theme_node_get_color(theme_node, "-arrow-background-color",
			&background_color);

	gdk_border_color.red =(guint16) border_color.red;
	gdk_border_color.green =(guint16) border_color.green;
	gdk_border_color.blue =(guint16) border_color.blue;

	gdk_background_color.red = (guint16) background_color.red;
	gdk_background_color.green = (guint16) background_color.green;
	gdk_background_color.blue = (guint16) background_color.blue;

	guint width, height;
	st_drawing_area_get_surface_size(area, &width, &height);
	guint box_width, box_height;

	box_width = width;
	box_height = height;
	if (priv->arrow_side == ST_SIDE_TOP || priv->arrow_side == ST_SIDE_BOTTOM)
		box_height -= rise;
	else
		box_width -= rise;

	cairo_t *cr = st_drawing_area_get_context(area);
	//Clutter.cairo_set_source_color(cr, borderColor);

	gdk_cairo_set_source_color(cr, &gdk_border_color);

	if (priv->arrow_side == ST_SIDE_TOP)
		cairo_translate(cr, 0, rise);
	else
		cairo_translate(cr, rise, 0);

	gdouble x1 = half_border, y1 = half_border;
	gdouble x2 = box_width - half_border;
	gdouble y2 = box_height - half_border;
	cairo_move_to(cr, x1 + border_raidus, y1);

	if (priv->arrow_side == ST_SIDE_TOP) {
		if (priv->arrow_origin < (x1 + (border_raidus + half_base))) {
			cairo_line_to(cr, priv->arrow_origin, y1 - rise);
			cairo_line_to(cr,
			MAX(x1 + border_raidus, priv->arrow_origin) + half_base, y1);
		} else if (priv->arrow_origin > (x2 - (border_raidus + half_base))) {
			cairo_line_to(cr,
			MIN(x2 - border_raidus, priv->arrow_origin) - half_base, y1);
			cairo_line_to(cr, priv->arrow_origin, y1 - rise);
		} else {
			cairo_line_to(cr, priv->arrow_origin - half_base, y1);
			cairo_line_to(cr, priv->arrow_origin, y1 - rise);
			cairo_line_to(cr, priv->arrow_origin + half_base, y1);
		}
	}
	cairo_line_to(cr, x2 - border_raidus, y1);
	cairo_arc(cr, x2 - border_raidus, y1 + border_raidus, border_raidus,
			3 * G_PI / 2, G_PI / 2);

	if (priv->arrow_side == ST_SIDE_RIGHT) {
		if (priv->arrow_origin < (y1 + (border_raidus + half_base))) {
			cairo_line_to(cr, x2 + rise, priv->arrow_origin);
			cairo_line_to(cr, x2,
			MAX(y1 + border_raidus, priv->arrow_origin) + half_base);
		} else if (priv->arrow_origin > (y2 - (border_raidus + half_base))) {
			cairo_line_to(cr, x2,
			MIN(y2 - border_raidus, priv->arrow_origin) - half_base);
			cairo_line_to(cr, x2 + rise, priv->arrow_origin);
		} else {
			cairo_line_to(cr, x2, priv->arrow_origin - half_base);
			cairo_line_to(cr, x2 + rise, priv->arrow_origin);
			cairo_line_to(cr, x2, priv->arrow_origin + half_base);
		}
	}
	cairo_line_to(cr, x2, y2 - border_raidus);
	cairo_arc(cr, x2 - border_raidus, y2 - border_raidus, border_raidus, 0,
	G_PI / 2);

	if (priv->arrow_side == ST_SIDE_BOTTOM) {
		if (priv->arrow_origin < (x1 + (border_raidus + half_base))) {
			cairo_line_to(cr,
			MAX(x1 + border_raidus, priv->arrow_origin) + half_base, y2);
			cairo_line_to(cr, priv->arrow_origin, y2 + rise);
		} else if (priv->arrow_origin > (x2 - (border_raidus + half_base))) {
			cairo_line_to(cr, priv->arrow_origin, y2 + rise);
			cairo_line_to(cr,
			MIN(x2 - border_raidus, priv->arrow_origin) - half_base, y2);
		} else {
			cairo_line_to(cr, priv->arrow_origin + half_base, y2);
			cairo_line_to(cr, priv->arrow_origin, y2 + rise);
			cairo_line_to(cr, priv->arrow_origin - half_base, y2);
		}
	}
	cairo_line_to(cr, x1 + border_raidus, y2);
	cairo_arc(cr, x1 + border_raidus, y2 - border_raidus, border_raidus, G_PI / 2,
	G_PI / 2);

	if (priv->arrow_side == ST_SIDE_LEFT) {
		if (priv->arrow_origin < (y1 + (border_raidus + half_base))) {
			cairo_line_to(cr, x1,
			MAX(y1 + border_raidus, priv->arrow_origin) + half_base);
			cairo_line_to(cr, x1 - rise, priv->arrow_origin);
		} else if (priv->arrow_origin > (y2 - (border_raidus + half_base))) {
			cairo_line_to(cr, x1 - rise, priv->arrow_origin);
			cairo_line_to(cr, x1,
			MIN(y2 - border_raidus, priv->arrow_origin) - half_base);
		} else {
			cairo_line_to(cr, x1, priv->arrow_origin + half_base);
			cairo_line_to(cr, x1 - rise, priv->arrow_origin);
			cairo_line_to(cr, x1, priv->arrow_origin - half_base);
		}
	}
	cairo_line_to(cr, x1, y1 + border_raidus);
	cairo_arc(cr, x1 + border_raidus, y1 + border_raidus, border_raidus, G_PI,
			3 * G_PI / 2);

	gdk_cairo_set_source_color(cr, &gdk_background_color);
	cairo_fill_preserve(cr);
	gdk_cairo_set_source_color(cr, &gdk_border_color);
	cairo_set_line_width(cr, border_width);
	cairo_stroke(cr);
}

static void 
cdos_box_pointer_set_property(GObject * gobject, 
			      guint prop_id,
		              const GValue * value, 
		              GParamSpec * pspec) 
{
	CdosBoxPointer *obj = CDOS_BOX_POINTER(gobject);

	switch (prop_id) {
	case PROP_X_OFFSET:
		cdos_box_pointer_set_x_offset_real(obj, g_value_get_float(value));	// 公共接口
		break;
	case PROP_Y_OFFSET:
		cdos_box_pointer_set_y_offset_real(obj, g_value_get_float(value));
		break;
	case PROP_OPACITY:
		cdos_box_pointer_set_opacity_real(obj, g_value_get_int(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
	}
}

static void 
cdos_box_pointer_get_property(GObject * gobject, 
			      guint prop_id,
		              GValue * value, 
		              GParamSpec * pspec) 
{
	CdosBoxPointerPrivate *priv = CDOS_BOX_POINTER(gobject)->priv;

	switch (prop_id) {
		break;
	case PROP_X_OFFSET:
		g_value_set_float(value, priv->x_offset);
		break;
	case PROP_Y_OFFSET:
		g_value_set_float(value, priv->y_offset);
		break;
	case PROP_X_POSITION:
		g_value_set_float(value, priv->x_position);
		break;
	case PROP_Y_POSITION:
		g_value_set_float(value, priv->y_position);
		break;
	case PROP_OPACITY:
		g_value_set_int(value, priv->opacity);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
		break;
	}
}

static void
cdos_box_pointer_finalize(GObject *gobject)
{
	CdosBoxPointer *self = CDOS_BOX_POINTER(gobject);
	CdosBoxPointerPrivate *priv = self->priv;

	g_free(priv->pos);

	G_OBJECT_CLASS(cdos_box_pointer_parent_class)->finalize(gobject);
}

static void 
cdos_box_pointer_dispose(GObject *gobject) 
{
	CdosBoxPointer *self = CDOS_BOX_POINTER(gobject);
	CdosBoxPointerPrivate *priv = self->priv;

	if(priv->container){
		clutter_actor_remove_child(CLUTTER_ACTOR(priv->container),
								   CLUTTER_ACTOR(self->bin));
		clutter_actor_remove_child(CLUTTER_ACTOR(priv->container),
								   CLUTTER_ACTOR(self->border));
	}
	clutter_actor_destroy (self->actor);

	G_OBJECT_CLASS(cdos_box_pointer_parent_class)->dispose(gobject);
}

static void 
cdos_box_pointer_class_init(CdosBoxPointerClass * klass) 
{
	GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
	//ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS(klass);
	/*
	 js connect;
	 */
	GParamSpec *pspec;

	gobject_class->set_property = cdos_box_pointer_set_property;
	gobject_class->get_property = cdos_box_pointer_get_property;
	gobject_class->dispose = cdos_box_pointer_dispose;
	gobject_class->finalize = cdos_box_pointer_finalize;

	g_type_class_add_private(klass, sizeof(CdosBoxPointerPrivate));

	/**
	 * CdosBoxPointer:x_offset:
	 *
	 * 锚点在X方向的偏移量
	 */
	pspec = g_param_spec_float("x_offset", "XOffset", "x_offset of the object", 0,
			100, 0, G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, PROP_X_OFFSET, pspec);// 新建属性并安装

	/**
	 * CdosBoxPointer:y_offset:
	 *
	 * 锚点在Y方向的偏移量
	 */
	pspec = g_param_spec_float("y_offset", "YOffset", "y_offset of the object", 0,
			100, 0, G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, PROP_Y_OFFSET, pspec);

	/**
	 * CdosBoxPointer:opacity:
	 *
	 * 元素显示的透明度。
	 */
	pspec = g_param_spec_uint("opacity", "Opacity", "opacity of the object", 0,
			255, 0, G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class, PROP_OPACITY, pspec);

}

static void 
cdos_box_pointer_init(CdosBoxPointer * self) 
{
	CdosBoxPointerPrivate *priv;
	priv = self->priv = CDOS_BOX_POINTER_GET_PRIVATE(self);

	self->actor = NULL;
	self->bin = NULL;
	priv->container = NULL;
	self->border = NULL;
	priv->source_actor = NULL;

	priv->arrow_side = 0;
	priv->arrow_origin = 0;
	priv->x_offset = 0;
	priv->y_offset = 0;
	priv->x_position = 0;
	priv->y_position = 0;
	priv->pos = NULL;
}
static void cdos_box_pointer_set_x_offset_real(CdosBoxPointer * self, const gfloat offset){
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv = self->priv;

	priv->x_offset = offset;
}
static void cdos_box_pointer_set_y_offset_real(CdosBoxPointer * self, const gfloat offset){
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv = self->priv;

	priv->y_offset = offset;
}
static void cdos_box_pointer_set_x_position_real(CdosBoxPointer *self, const gfloat position){
	CdosBoxPointerPrivate *priv = self->priv;

	priv->x_position = position;
}
static void cdos_box_pointer_set_y_position_real(CdosBoxPointer *self, const gfloat position){
	CdosBoxPointerPrivate *priv = self->priv;

	priv->y_position = position;
}
static void cdos_box_pointer_set_opacity_real(CdosBoxPointer *self, const gint opacity){
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));

	CdosBoxPointerPrivate *priv = self->priv;
	if (opacity && opacity >=0 && opacity <= 255 ){
		priv->opacity = opacity;
	}
}

/**
 *  cdos_box_pointer_set_x_offset:
 *  @self:当前实例对象
 *  @offset: 偏移量
 *
 *  设置锚点相对X方向的偏移量。
 */
void cdos_box_pointer_set_x_offset(CdosBoxPointer * self, const gfloat offset) {
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv = self->priv;
	cdos_box_pointer_set_x_offset_real(self,offset);
	g_object_set(self,"x_offset",offset,NULL);
}


/**
 * cdos_box_pointer_get_x_offset:
 * @self: 当前实例对象
 *
 * 获取锚点在X方向的偏移量。
 *
 * Returns:返回X方向的偏移量。
 */
const gfloat cdos_box_pointer_get_x_offset(CdosBoxPointer * self) {
	// 同上
	g_return_val_if_fail(CDOS_IS_BOX_POINTER(self), 0);

	return self->priv->x_offset;
}

/**
 *  cdos_box_pointer_set_y_offset:
 *  @self:当前实例对象
 *  @offset: 偏移量
 *
 *  设置锚点相对Y方向的偏移量。
 */
void cdos_box_pointer_set_y_offset(CdosBoxPointer * self, const gfloat offset) {

	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv = self->priv;
	cdos_box_pointer_set_y_offset_real(self,offset);
	g_object_set(self,"y_offset",offset,NULL);
}

/**
 * cdos_box_pointer_get_y_offset:
 * @self: 当前实例对象
 *
 * 获取锚点在Y方向的偏移量。
 * Returns:Y方向的偏移量
 */
const gfloat cdos_box_pointer_get_y_offset(CdosBoxPointer * self) {
	g_return_val_if_fail(CDOS_IS_BOX_POINTER(self), 0);
	return self->priv->y_offset;
}
/**
 * cdos_box_pointer_set_x_position:
 * @self: 当前实例对象
 * @position: X方向的位置
 *
 * 设置主容器相对关联元素的X方向的位置。
 */
void cdos_box_pointer_set_x_position(CdosBoxPointer *self, const gfloat position) {
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv = self->priv;
	cdos_box_pointer_set_x_position_real(self,position);
}


/**
 * cdos_box_pointer_get_x_position:
 * @self: 当前实例对象
 *
 * 获取主容器相对关联元素的X方向的位置。
 *
 * Returns: X方向的位置。
 */
const gfloat cdos_box_pointer_get_x_position(CdosBoxPointer *self) {
	g_return_val_if_fail(CDOS_IS_BOX_POINTER(self), 0);
	return self->priv->x_position;
}
/**
 * cdos_box_pointer_set_y_position:
 * @self: 当前实例对象
 * @position: X方向的位置
 *
 * 设置主容器相对关联元素的X方向的位置。
 */
void cdos_box_pointer_set_y_position(CdosBoxPointer *self, const gfloat position) {
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv = self->priv;
	cdos_box_pointer_set_y_position_real(self,position);
}


/**
 * cdos_box_pointer_get_y_position:
 * @self:当前实例对象
 *
 * 获取主容器相对关联元素的Y方向的位置。
 *
 * Returns: Y方向的位置。
 */
const gfloat cdos_box_pointer_get_y_position(CdosBoxPointer *self) {
	g_return_val_if_fail(CDOS_IS_BOX_POINTER(self), 0);
	return self->priv->y_position;
}
/**
 *  cdos_box_pointer_set_opacity:
 *  @self: 当前实例对象
 *  @opacity: 透明度值，取值范围：0～255;
 *
 *  设置当前容器的透明度。
 */
void cdos_box_pointer_set_opacity(CdosBoxPointer *self, const gint opacity) {
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv = self->priv;
	cdos_box_pointer_set_opacity_real(self,opacity);
	if (opacity && opacity >=0 && opacity <= 255 ){
		g_object_set(self,"opacity",opacity,NULL);
	}
}

/**
 *  cdos_box_pointer_get_opacity:
 *  @self:当前实例对象
 *
 *  获取当前容器的透明度；
 *
 *  Returns:透明度值，范围：0～255;
 */
const gint cdos_box_pointer_get_opacity(CdosBoxPointer *self) {
	g_return_val_if_fail(CDOS_IS_BOX_POINTER(self), 0);
	return self->priv->opacity;
}
/**
 * cdos_box_pointer_new:
 * @arrow_side: 锚点所在的方位，取值为StSide类型成员
 *
 * 创建#CdosBoxPointer实例对象
 *
 * Returns:返回一个#CdosBoxPointer实例。
 */
CdosBoxPointer *
cdos_box_pointer_new(gint arrow_side) 
{
	CdosBoxPointer *self;
	self = g_object_new(CDOS_TYPE_BOX_POINTER, NULL);
	CdosBoxPointerPrivate *priv = self->priv;

	cdos_box_pointer_set_arrow_side(self, arrow_side);

	self->actor = g_object_new(ST_TYPE_BIN, "x-fill", TRUE, "y-fill", TRUE, NULL);
	st_bin_set_fill(self->actor, TRUE, TRUE);
	//self->bin = ST_BIN(st_bin_new());
	self->bin = g_object_new(ST_TYPE_BIN, "x-fill", TRUE, "y-fill", TRUE, "x-align", ST_ALIGN_START, NULL);
	st_bin_set_fill(self->bin, TRUE, TRUE);

	st_bin_set_alignment(self->bin, ST_ALIGN_START, ST_ALIGN_START);
	priv->container = g_object_new (CDOS_TYPE_GENERIC_CONTAINER, NULL);
	self->border = g_object_new(ST_TYPE_DRAWING_AREA, NULL);
    //xunli
    clutter_actor_set_reactive (CLUTTER_ACTOR (self->bin), TRUE);
	st_bin_set_child(self->actor, CLUTTER_ACTOR(priv->container));

	g_signal_connect(priv->container, "get-preferred-width", G_CALLBACK(get_container_preferred_width), self);
	g_signal_connect(priv->container, "get-preferred-height", G_CALLBACK(get_container_preferred_height),self);
	g_signal_connect(priv->container, "allocate",G_CALLBACK(container_allocate), self);

	g_signal_connect(self->border, "repaint", G_CALLBACK(draw_border), self);

	clutter_actor_add_child(CLUTTER_ACTOR(priv->container),
			CLUTTER_ACTOR(self->bin));
	clutter_actor_add_child(CLUTTER_ACTOR(priv->container),
			CLUTTER_ACTOR(self->border));
	clutter_actor_raise(CLUTTER_ACTOR(self->bin), CLUTTER_ACTOR(self->border));

	return self;
}

/**
 * cdos_box_pointer_show:
 * @self: 当前实例对象
 * @animate: 是否使用过渡动画，TRUE-使用，FALSE-不使用
 *
 * 显示当前实例主容器元素。
 */
void cdos_box_pointer_show(CdosBoxPointer * self, gboolean animate) {
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv;
	priv = self->priv;

	StThemeNode *theme_node = st_widget_get_theme_node(ST_WIDGET(self->actor));
	gdouble rise = st_theme_node_get_length(theme_node, "-arrow-rise");

	if (animate) {
		priv->opacity = 0;
		clutter_actor_show(CLUTTER_ACTOR(self->actor));

		switch (priv->arrow_side) {
		case ST_SIDE_TOP:
			priv->x_offset = -rise;
			break;
		case ST_SIDE_BOTTOM:
			priv->x_offset = rise;
			break;
		case ST_SIDE_LEFT:
			priv->x_offset = -rise;
			break;
		case ST_SIDE_RIGHT:
			priv->x_offset = rise;
			break;
		}

	} else {
		priv->opacity = 255;
		clutter_actor_show(CLUTTER_ACTOR(self->actor));
	}
}

/**
 * cdos_box_pointer_hide:
 * @self: 当前实例对象
 * @animate: 是否使用过渡动画，TRUE-使用，FALSE-不使用
 *
 * 隐藏当前实例主容器元素。
 */
void cdos_box_pointer_hide(CdosBoxPointer * self, gboolean animate) {
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	clutter_actor_hide(CLUTTER_ACTOR(self->actor));
}

/**
 * cdos_box_pointer_set_arrow_side:
 * @self: 当前实例对象；
 * @side: 方位值，取值为StSide枚举成员
 *
 * 设置当前实例锚点所在方位；
 */
void cdos_box_pointer_set_arrow_side(CdosBoxPointer * self, gint side) {
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv;
	priv = self->priv;
	priv->arrow_side = side;
}

/**
 *  cdos_box_pointer_set_position:
 *  @self: 当前实例对象
 *  @source_actor: 给定元素
 *  @alignment: 对齐方式
 *
 *  根据给定元素和对齐方式，重置当前实例元素相对给定元素的位置。
 */
gboolean cdos_box_pointer_set_position(CdosBoxPointer * self,
		ClutterActor * source_actor, gint alignment)
{
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	// We need to show it now to force an allocation,
	// so that we can query the correct size.
	CdosBoxPointerPrivate *priv;
	priv = self->priv;
	clutter_actor_show(CLUTTER_ACTOR(self->actor));
	priv->source_actor = source_actor;
	gboolean ret = cdos_box_pointer_reposition(self, source_actor, alignment);
	return ret;
}

/**
 *  cdos_box_pointer_set_position_x_y:
 *  @self: 当前实例对象
 *  @source_actor: 给定元素
 *  @alignment: 对齐方式
 *  @x: x方向的位置偏移量
 *  @y: y方向的位置偏移量
 *
 *  根据给定元素，对齐方式和相对偏移量，重置当前实例元素和给定元素之间的位置。
 */
void cdos_box_pointer_set_position_x_y(CdosBoxPointer * self,
		ClutterActor * source_actor, gint alignment, gfloat x, gfloat y)
{
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	// We need to show it now to force an allocation,
	// so that we can query the correct size.
	CdosBoxPointerPrivate *priv;
	priv = self->priv;
	clutter_actor_show(CLUTTER_ACTOR(self->actor));
	priv->source_actor = source_actor;
	priv->pos = g_new0(Pos,1);
	priv->pos->x = x;
	priv->pos->y = y;
	cdos_box_pointer_reposition(self, source_actor, alignment);
}

/**
 * cdos_box_pointer_set_source_alignment:
 * @self:当前实例对象
 * @alignmentz:对其方式
 *
 * 根据给定的对其方式，重置当前实例元素和关联元素之间的位置。
 */
void cdos_box_pointer_set_source_alignment(CdosBoxPointer * self, gint alignment) {
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv;
	priv = self->priv;
	priv->source_alignment = alignment;
	// !!!!!
	if (!priv->source_actor)
		return;
	// We need to show it now to force an allocation,
	// so that we can query the correct size.
	clutter_actor_show(CLUTTER_ACTOR(self->actor));
	cdos_box_pointer_reposition(self, priv->source_actor, priv->source_alignment);
}

/**
 * cdos_box_pointer_set_arrow_origin:
 * @self: 当前实例对象
 * @origin: 原点位置
 *
 * 统一设置原点所在的偏移距离
 */
void cdos_box_pointer_set_arrow_origin(CdosBoxPointer * self, gfloat origin) {
	g_return_if_fail(CDOS_IS_BOX_POINTER(self));
	CdosBoxPointerPrivate *priv;
	priv = self->priv;
	if (priv->arrow_origin != origin) {
		priv->arrow_origin = origin;
		st_drawing_area_queue_repaint(ST_DRAWING_AREA(self->border));
	}
}

/**
 * cdos_box_pointer_get_arrow_origin:
 * @self:当前实例对象
 *
 * 获取原点所在的偏移距离。
 * Returns: 返回偏离量
 */
gdouble cdos_box_pointer_get_arrow_origin (CdosBoxPointer *self){
	g_return_val_if_fail(CDOS_IS_BOX_POINTER(self),0);
    CdosBoxPointerPrivate *priv = self->priv;

    return priv->arrow_origin;
}
/**
 * cdos_box_pointer_get_container:
 * @self:当前实例对象
 *
 * 获取当前实例元素的子容器，该容器主要用于包裹元素和显示标识元素的父容器；
 *
 * Returns:返回一个#CdosGenericContainer类型的实例对象。
 */
CdosGenericContainer *
cdos_box_pointer_get_container(CdosBoxPointer *self)
{
	g_return_val_if_fail(CDOS_IS_BOX_POINTER(self),NULL);
    CdosBoxPointerPrivate *priv = self->priv;

    return priv->container;
}
