#include <math.h>

#include "st.h"
#include "st-marshal.h"

#include "cdos-global.h"
#include "cdos-debug.h"
#include "cdos-base-util.h"

#include "cdos.h"
#include "cdos-start.h"

/**
 * SECTION: cdos-dnd
 * @Title: CdosDraggable
 * @short_description: 对 ClutterActor 的拖动事件处理
 *
 * CdosDraggable 控制一个 #ClutterActor 拖动时以及当其他 #ClutterActor 拖动到该 actor 上的事件处理。
 */

#define SCALE_ANIMATION_TIME 250
#define SNAP_BACK_ANIMATION_TIME 250
#define REVERT_ANIMATION_TIME 750

#define DRAG_EVENT_TIMEOUT_MS 5000
#define DND_ANIMATION_TIME 200

static ClutterActor *event_handler_actor = NULL;
static CdosDraggable *current_draggable = NULL;
static gulong current_event_handler = 0L;
static GList *drag_monitors = NULL;
static GHashTable *dnd_func_table = NULL;

typedef struct _DNDFuncTable {
    gpointer             delegate;
    GetDragActor         get_drag_actor;
    GetDragActorSource   get_drag_actor_source;
    HandleDragOver       handle_drag_over;
    AcceptDrop           accept_drop;
}DNDFuncTable;

/**
 * cdos_draggable_set_delegate:
 * @actor: a #ClutterActor
 * @delegate: 该 actor 的 delegate
 *
 * 设置一个 DND 对象的 actor 的 delegate。
 *
 * Since: 1.0.0
 */
void
cdos_draggable_set_delegate (void *actor, void *delegate)
{
    g_return_if_fail ((actor != NULL) && (delegate != NULL));
    g_return_if_fail (dnd_func_table != NULL);
    DNDFuncTable *table = NULL;
    if (g_hash_table_contains (dnd_func_table, actor) == 0)
    {
        table = g_new0 (DNDFuncTable, 1);
        g_hash_table_replace (dnd_func_table, actor, table);
    }else 
        table = g_hash_table_lookup (dnd_func_table, actor);
    table->delegate = delegate;
}

/**
 * cdos_draggable_set_func:
 * @actor: a #ClutterActor
 * @func: 函数指针
 * @type: 函数类型
 *
 * 设置 DND 对象的 actor 的四个函数
 *
 * Since: 1.0.0
 */
void 
cdos_draggable_set_func (void *actor, void *func, DNDFuncType type)
{
    g_return_if_fail ((actor != NULL) && (func != NULL));

    g_return_if_fail (dnd_func_table != NULL);
    if (g_hash_table_contains (dnd_func_table, actor) == 0)
        g_hash_table_replace (dnd_func_table, actor, NULL);
    
    DNDFuncTable *table = g_hash_table_lookup (dnd_func_table, actor);
    if (table == NULL)
    {
        table = g_new0 (DNDFuncTable, 1);
        g_hash_table_replace (dnd_func_table, actor, table);
    }

    switch (type)
    {
        case DND_GET_DRAG_ACTOR:
            table->get_drag_actor = func;
            break;
        case DND_GET_DRAG_ACTOR_SRC:
            table->get_drag_actor_source = func;
            break;
        case DND_HANDLE_DRAG_OVER:
            table->handle_drag_over = func;
            break;
        case DND_ACCEPT_DROP:
            table->accept_drop = func;
            break;
    }
}

/**
 * cdos_draggable_get_delegate:
 * @actor: a #ClutterActor
 *
 * 从 DND 对象的 actor 获取其 delegate
 *
 * Returns: 该 actor 的 delegate
 *
 * Since: 1.0.0
 */
void *
cdos_draggable_get_delegate (ClutterActor *actor)
{
    DNDFuncTable *table = g_hash_table_lookup (dnd_func_table, actor);
    if (G_LIKELY(table == NULL))
        return NULL;
    else
        return table->delegate;
}

/**
 * cdos_draggable_get_delegate_actor:
 * @delegate: delegate
 *
 * 获取 delegate 关联的 actor
 *
 * Returns: a #ClutterActor
 *
 * Since: 1.0.0
 */
ClutterActor *
cdos_draggable_get_delegate_actor (void *delegate)
{
    GHashTableIter iter;
    gpointer key, value;

    g_hash_table_iter_init (&iter, dnd_func_table);
    while (g_hash_table_iter_next (&iter, &key, &value))
    {
        DNDFuncTable *table = value;
        if (table->delegate == delegate)
            return (ClutterActor *)key;
    }
    return NULL;
}

/**
 * cdos_draggable_get_func:
 * @actor: a #ClutterActor
 * @type: 函数类型
 *
 * 获取与 @actor 关联的 @type 类型的函数
 *
 * Returns: 返回 #DNDFuncType 类型函数指针
 */
void *
cdos_draggable_get_func (ClutterActor *actor, DNDFuncType type)
{
    if (actor == NULL)
        return NULL;
    DNDFuncTable *table = g_hash_table_lookup (dnd_func_table, actor);
    if (G_LIKELY(table == NULL))
        return NULL;
    else {
        switch (type)
        {
            case DND_GET_DRAG_ACTOR:
                return table->get_drag_actor;
            case DND_GET_DRAG_ACTOR_SRC:
                return table->get_drag_actor_source;
            case DND_HANDLE_DRAG_OVER:
                return table->handle_drag_over;
            case DND_ACCEPT_DROP:
                return table->accept_drop;
        }
        return NULL;
    }
}

G_DEFINE_TYPE (CdosDraggable, cdos_draggable, G_TYPE_OBJECT);

enum {
    DRAG_BEGIN,
    DRAG_END,
    DRAG_CANCELLED,
    LAST_SIGNAL
};

static guint cdos_draggable_signals [LAST_SIGNAL] = {0};

static gboolean drag_actor_dropped (CdosDraggable *drag, ClutterEvent *event);

static gboolean cdos_draggable_on_event (ClutterActor *actor, ClutterEvent *event, CdosDraggable *drag);

static void
cdos_draggable_finalize (GObject *object)
{
    /* FIXME */
    G_OBJECT_CLASS (cdos_draggable_parent_class)->finalize (object);
}

static void
cdos_draggable_init (CdosDraggable *drag)
{

}

uintptr_t
cdos_draggable_get_actor_delegate (ClutterActor *actor)
{
    uintptr_t delegate = 0;
    if(g_object_class_find_property (G_OBJECT_GET_CLASS (actor), "delegate"))
        g_object_get (actor, "delegate", &delegate, NULL);
    return delegate;
}

/*
gboolean
is_draggable ()
{
    return current_draggable != NULL;
}
*/

static ClutterActor *
get_event_handler_actor ()
{
    /* js 里不需要这段代码。
     * 在 c 里，对每个新的 Draggable 都要重新 关联信号
     */
    if (current_event_handler != 0)
    {
        g_signal_handler_disconnect (event_handler_actor, current_event_handler);
        current_event_handler = g_signal_connect (event_handler_actor, "event",
                G_CALLBACK (cdos_draggable_on_event), current_draggable);
    }

    if (!event_handler_actor)
    {
        //event_handler_actor = clutter_rectangle_new ();
        event_handler_actor = clutter_actor_new ();        
        clutter_actor_set_width (event_handler_actor, 0.0);
        clutter_actor_set_height (event_handler_actor, 0.0);

        st_container_add_actor((ClutterContainer *)ui_group, event_handler_actor);
        current_event_handler = g_signal_connect (event_handler_actor, "event",
                G_CALLBACK (cdos_draggable_on_event), current_draggable);

    }
    return event_handler_actor;
}

void
cdos_add_drag_monitor (void *monitor)
{
    drag_monitors = g_list_append (drag_monitors, monitor);
}

void
cdos_remove_drag_monitor (void *monitor)
{
    drag_monitors = g_list_remove (drag_monitors, monitor);
}

/* CdosDraggable */

static void set_timer(CdosDraggable *drag, gboolean renew);
static void cancel_drag (CdosDraggable *drag, guint32 timestamp);
static void drag_complete (CdosDraggable *drag);
static gboolean update_drag_position (CdosDraggable *drag, ClutterEvent *event);
static void ungrab_actor (CdosDraggable *drag);
static void grab_actor (CdosDraggable *drag);
static gboolean maybe_start_drag (CdosDraggable *drag, ClutterEvent *event);
//static gboolean check_threshold (CdosDraggable *drag, gint x, gint y);
static void restore_drag_actor (CdosDraggable *drag, guint32 time);

static gboolean drop_on_target (CdosDraggable *drag,
                                ClutterActor *target,
                                ClutterEvent *event,
                                float drop_x,
                                float drop_y);


static void cdos_draggable_start_drag (CdosDraggable *draggable,
                                       gfloat      stage_x,
                                       float      stage_y,
                                       guint32    time);


static void
on_button_hover_changed (StButton   *button,
                         GParamSpec *pspec,
                         gpointer    draggable)
{
    if (st_widget_get_hover ((StWidget *)button) || !st_button_get_pressed (button))
        return;
    st_button_fake_release (button);
    CdosDraggable *drag = draggable;
    cdos_draggable_start_drag (drag, drag->start.x, drag->start.y,
            cdos_global_get_current_time (global));
}

static gboolean
cdos_draggable_on_event (ClutterActor *actor,
                         ClutterEvent *event,
                         CdosDraggable    *drag)
{
    //set_timer (drag, true);
    ClutterEventType type = clutter_event_type (event);
    if (type == CLUTTER_BUTTON_RELEASE)
    {
        drag->button_down = false;
        if (drag->drag_in_progress)
        {
            // TODO 需要desklet
            // dnd.js: 232
            // else 
            //cancel_drag (drag, clutter_event_get_time (event));
            //return true;
            return drag_actor_dropped (drag, event);
        }else if (drag->drag_actor != NULL && !drag->animation_in_progress)
        {
            drag_complete (drag);
            return true;
        }else {
            ungrab_actor (drag);
            return false;
        }
    }
    else if (type == CLUTTER_MOTION)
    {
        if (drag->drag_in_progress)
            return update_drag_position (drag, event);
        else if (drag->drag_actor == NULL)
                return maybe_start_drag (drag, event);
    }
    else if (type == CLUTTER_KEY_PRESS && drag->drag_in_progress)
    {
        if (clutter_event_get_key_symbol (event) == CLUTTER_KEY_Escape)
            cancel_drag (drag, clutter_event_get_time (event));
        return true;
    }
    return false;
}

static void 
fake_release (CdosDraggable *drag)
{
    drag->button_down = false;
    ungrab_actor (drag);
}

static void
grab_actor (CdosDraggable *drag)
{
    clutter_grab_pointer (drag->actor);
    drag->on_event_id = g_signal_connect (drag->actor, "event",
            G_CALLBACK (cdos_draggable_on_event), drag);
}

static void
ungrab_actor (CdosDraggable *drag)
{
    clutter_ungrab_pointer ();
    if (!drag->on_event_id)
        return;
    g_signal_handler_disconnect (drag->actor, drag->on_event_id);
    drag->on_event_id = 0;
}

static void
grab_events (CdosDraggable *drag)
{
    if (!drag->events_grabbed)
    {
        ClutterActor *eha = get_event_handler_actor ();
        clutter_grab_pointer (eha);
        drag->previous_key_focus_actor = 
            clutter_stage_get_key_focus (cdos_global_get_stage (global));
        clutter_actor_grab_key_focus (eha);
        drag->events_grabbed = true;
    }
}

static void
ungrab_events (CdosDraggable *drag)
{
    if (drag->events_grabbed)
    {
        if (drag->previous_key_focus_actor)
            clutter_actor_grab_key_focus (drag->previous_key_focus_actor);
        clutter_ungrab_pointer ();
        drag->events_grabbed = false;
    }
}

static gboolean
on_button_press (ClutterActor *actor,
                 ClutterEvent *event,
                 gpointer draggable)
{
    ClutterEvent *e = event;
    CdosDraggable *drag = CDOS_DRAGGABLE(draggable);

    if (drag->inhibit)
        return false;
    if (clutter_event_get_button (e) != 1)
        return false;
    drag->button_down = 1;
    /*
    if (ST_IS_BUTTON (drag->actor))
        g_signal_connect ((StButton *)drag->actor, "notify::hover",
                G_CALLBACK (on_button_hover_changed), drag);
    else
    */
        grab_actor (drag);
    clutter_event_get_coords (e, &drag->start.x, &drag->start.y);
    return false;
}

static void
disconnect_signals (CdosDraggable *drag)
{

}

static void
on_actor_destroyed (gpointer actor,
                    gpointer draggable)
{
    CdosDraggable *drag = CDOS_DRAGGABLE (draggable);
    drag->actor_destroyed = true;
    if (drag->actor == drag->first_leave_actor)
        drag->first_leave_actor = drag->orig_parent;
    if (drag->drag_in_progress)
        cancel_drag (drag, cdos_global_get_current_time (cdos_global_get ()));
    disconnect_signals (drag);
}


static void
on_drag_threshold_changed (gpointer settings,
                           char    *key,
                           gpointer draggable)
{
    GSettings *s = settings ? settings : 
                    cdos_global_get_settings (global);

    CdosDraggable *drag = draggable;
    gint value = g_settings_get_int (s, (char *)key);
    drag->drag_threshold = value;
}

static CdosDraggable *
cdos_draggable_new (ClutterActor *actor,
                    ClutterActor *target,
                    gboolean      manual_mode,
                    gboolean      restore_on_success,
                    gint          drag_actor_max_size,
                    float         drag_actor_opacity)
{
    CdosDraggable *drag = g_object_new (CDOS_TYPE_DRAGGABLE, NULL);

    drag->restore_on_success  = restore_on_success;
    drag->drag_actor_max_size = drag_actor_max_size;
    drag->drag_actor_opacity  = drag_actor_opacity;

    drag->actor = CLUTTER_ACTOR (actor);
    if (target)
        drag->target = CLUTTER_ACTOR (target);

    if (!manual_mode)
        drag->button_press_event_id = g_signal_connect (actor, "button-press-event",
                G_CALLBACK (on_button_press), drag);
    drag->destroy_event_id = g_signal_connect (actor, "destroy", G_CALLBACK (on_actor_destroyed), drag);

    drag->drag_threshold = 8;

    g_signal_connect (cdos_global_get_settings (global),
            "changed::dnd-drag-threshold", G_CALLBACK (on_drag_threshold_changed), drag);
    on_drag_threshold_changed (NULL, "dnd-drag-threshold", drag);

    if (dnd_func_table == NULL)
        dnd_func_table = g_hash_table_new (NULL, NULL);

    return drag;
}

typedef struct _ScaleData {
    CdosDraggable *drag;
    vec2f      offset;
    gdouble    orig_scale;
}ScaleData;

static void
on_update (ClutterTimeline *timeline,
           gint             msecs,
           ScaleData       *data)
{
    CdosDraggable *drag = data->drag;
    ClutterActor *actor = drag->drag_actor;
    vec2d scale;
    clutter_actor_get_scale (actor, &scale.x, &scale.y);
    gdouble current_scale = scale.x / data->orig_scale;
    drag->offset.x = current_scale * data->offset.x;
    drag->offset.y = current_scale * data->offset.y;
    clutter_actor_set_position (actor,
            drag->x + drag->offset.x, drag->y + drag->offset.y);
}

static void
on_completed (ClutterTimeline *timeline,
              ScaleData       *data)
{
    CdosDraggable *drag = data->drag;
    clutter_actor_remove_transition (drag->drag_actor, "drag-start");
    g_free (data);
}

static void
cdos_draggable_start_drag (CdosDraggable *drag,
                           gfloat      stage_x,
                           gfloat      stage_y,
                           guint32    time)
{
    current_draggable = drag;
    drag->drag_in_progress = true;

    if (ST_IS_BUTTON (drag->actor))
    {
        st_button_fake_release ((StButton *)drag->drag_actor);
        st_widget_set_hover ((StWidget *)drag->drag_actor, 0);
    }
    g_signal_emit (drag, cdos_draggable_signals[DRAG_BEGIN], 0, time);

    if (drag->on_event_id)
        ungrab_actor (drag);
    grab_events (drag);
    meta_screen_set_cursor (cdos_global_get_screen (global), META_CURSOR_DND_IN_DRAG);

    drag->x = drag->start.x = stage_x;
    drag->y = drag->start.y = stage_y;

    void *delegate = cdos_draggable_get_delegate (drag->actor);
    GetDragActor get_drag_actor = cdos_draggable_get_func (drag->actor, DND_GET_DRAG_ACTOR);
    if (delegate && get_drag_actor)
    {
        drag->drag_actor = get_drag_actor (delegate);
        
        GetDragActorSource get_drag_actor_source = NULL;
        get_drag_actor_source = cdos_draggable_get_func (drag->actor, DND_GET_DRAG_ACTOR_SRC);

        if (get_drag_actor_source)
        {
            vec2f source;
            drag->drag_actor_source = get_drag_actor_source (delegate);
            clutter_actor_get_transformed_position (drag->drag_actor_source, &source.x, &source.y);
            float x, y;
            float w, h;
            clutter_actor_get_size (drag->drag_actor, &w, &h);
            if ((stage_x > source.x && stage_x <= source.x + w) &&
                    (stage_y > source.y && stage_y <= source.y + h))
            {
                x = source.x;
                y = source.y;
            }
            else{
                x = stage_x - w / 2.0;
                y = stage_y - h / 2.0;
            }
            clutter_actor_set_position (drag->drag_actor, x, y);
        }
        else
            drag->drag_actor_source = drag->actor;
        
        drag->orig_parent = NULL;//clutter_actor_get_parent (drag->actor);
        float x, y;
        clutter_actor_get_transformed_position (drag->actor, &x, &y);
        drag->offset.x = x - drag->start.x;
        drag->offset.y = y - drag->start.y;
    }
    else {
        /* dnd.js 这里错误, 需要用 clutter_clone_new () */
        drag->drag_actor = clutter_clone_new (drag->actor);
        drag->drag_actor_source = NULL;
        drag->orig_parent = clutter_actor_get_parent (drag->actor);
        /* 记住原来的后继节点 */
        if (ST_IS_CONTAINER (drag->orig_parent))
        {
            GList *children = st_container_get_children_list ((StContainer *)drag->orig_parent);
            drag->orig_pos = g_list_index(children,drag->actor);
        }
        clutter_actor_get_position (drag->actor, &drag->orig.x, &drag->orig.y);
        gdouble tmp;
        clutter_actor_get_scale (drag->actor, &drag->orig_scale, &tmp);

        float x, y;
        clutter_actor_get_transformed_position (drag->actor, &x, &y);
        drag->offset.x = x - drag->start.x;
        drag->offset.y = y - drag->start.y;
    }
    {
        /* 使 drag_actor 与 actor 大小相同 */
        float scale_width, scale_height;
        clutter_actor_get_transformed_size (drag->actor, &scale_width, &scale_height);
        float w, h;
        clutter_actor_get_size (drag->actor, &w, &h);
        clutter_actor_set_size (drag->drag_actor, w, h);
        clutter_actor_set_scale (drag->drag_actor, scale_width / w, scale_height / h);
    }
    /* 参考 clutter-clone.c.
     * ClutterClone 的 paint 函数调用其 source 的 paint:
     *   clone->priv->clone_source->priv->parent 不能为 NULL,
     *   所以不能使 drag->actor 的 parent=NULL
     */
    st_container_add_actor ((ClutterContainer *)ui_group, drag->drag_actor);
    st_container_raise ((ClutterContainer *)ui_group, drag->drag_actor, NULL);

    cdos_base_util_set_hidden_from_pick (drag->drag_actor, true);

    drag->orig_opacity = clutter_actor_get_opacity (drag->drag_actor);
    if (drag->drag_actor_opacity > 0.f)
        clutter_actor_set_opacity (drag->drag_actor, drag->drag_actor_opacity);

    drag->snap_back_x = drag->start.x + drag->offset.x;
    drag->snap_back_y = drag->start.y + drag->offset.y;
    gdouble scale_y;

    clutter_actor_get_scale (drag->drag_actor, &drag->snap_back_scale, &scale_y);

    if (drag->drag_actor_max_size > 0)
    {
        sizef scale_size;
        clutter_actor_get_transformed_size (drag->drag_actor, &scale_size.width, &scale_size.height);
        float current_size = max (scale_size.width, scale_size.height);
        if (current_size > drag->drag_actor_max_size)
        {
            float scale = drag->drag_actor_max_size / current_size;
            gdouble scale_x, scale_y;
            clutter_actor_get_scale (drag->drag_actor, &scale_x, &scale_y);
            gdouble orig_scale = scale_x;
            ScaleData *data = g_new0 (ScaleData, 1);
            data->offset = drag->offset;
            data->drag = drag;
            data->orig_scale = orig_scale;

            /*
            ClutterAnimation *anim = clutter_actor_animate (drag->drag_actor,
                    CLUTTER_EASE_OUT_QUAD,
                    SCALE_ANIMATION_TIME,
                    "scale-x", scale * orig_scale,
                    "scale-y", scale * orig_scale,
                    NULL);
            ClutterTimeline *timeline = clutter_animation_get_timeline (anim);
            g_signal_connect (timeline, "new-frame", G_CALLBACK (on_update), data);
            g_signal_connect (timeline, "completed", G_CALLBACK (on_completed), data);
            */
            clutter_actor_save_easing_state (drag->drag_actor);

            ClutterTransition *trans = clutter_transition_group_new ();

            ClutterTransition *trans_x = clutter_property_transition_new ("scale-x");
            ClutterTransition *trans_y = clutter_property_transition_new ("scale-y");

            clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_x);
            clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_y);

            clutter_actor_set_easing_mode (drag->drag_actor, CLUTTER_EASE_OUT_QUAD);
            clutter_timeline_set_duration (CLUTTER_TIMELINE (trans), SCALE_ANIMATION_TIME);
            clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (trans), 0);

            clutter_transition_set_from (trans_x, G_TYPE_DOUBLE, orig_scale);
            clutter_transition_set_to (trans_x, G_TYPE_DOUBLE, scale * orig_scale);

            clutter_transition_set_from (trans_y, G_TYPE_DOUBLE, orig_scale);
            clutter_transition_set_to (trans_y, G_TYPE_DOUBLE, scale * orig_scale);

            clutter_actor_add_transition (drag->drag_actor, "drag-start", trans);
            clutter_actor_restore_easing_state (drag->drag_actor);
            g_signal_connect (trans, "new-frame", G_CALLBACK (on_update), data);
            g_signal_connect (trans, "completed", G_CALLBACK (on_completed), data);

        }
    }
}

static gboolean
maybe_start_drag (CdosDraggable *drag, ClutterEvent *event)
{
    float stage_x, stage_y;
    clutter_event_get_coords (event, &stage_x, &stage_y);
    /* 参照 Cinnamon-2.8.6 */
    if ((fabsf(stage_x - drag->start.x) > drag->drag_threshold) ||
            fabsf (stage_y - drag->start.y) > drag->drag_threshold)
    {
        cdos_draggable_start_drag (drag, stage_x, stage_y, clutter_event_get_time (event));
        update_drag_position (drag, event);
    }
    return true;
}

static gboolean
set_cursor (DragMotionResult result)
{
    if (result != DRAG_MOTION_CONTINUE)
    {
        MetaCursor cursor_map[3] = {
            META_CURSOR_DND_UNSUPPORTED_TARGET,
            META_CURSOR_DND_COPY,
            META_CURSOR_DND_MOVE};
        meta_screen_set_cursor (cdos_global_get_screen (global), cursor_map[result]);
        return true;
    }
    return false;
}
typedef struct _DragEvent {
    float stage_x;
    float stage_y;
    ClutterActor *drag_actor;
    ClutterActor *source;
    ClutterActor *target_actor;
}DragEvent;

static gboolean
update_drag_position (CdosDraggable *drag, ClutterEvent *event)
{
    float stage_x, stage_y;
    clutter_event_get_coords (event, &stage_x, &stage_y);
    guint32 time = clutter_event_get_time (event);
    drag->x = stage_x;
    drag->y = stage_y;

    if (drag->drag_actor)
    {
        clutter_actor_set_position (drag->drag_actor,
                stage_x + drag->offset.x, stage_y + drag->offset.y);
        ClutterActor *stage = clutter_actor_get_stage (drag->drag_actor);
        ClutterActor *target = clutter_stage_get_actor_at_pos ((ClutterStage *)stage,
                CLUTTER_PICK_ALL, stage_x, stage_y);

        /* motionFunc dragMotion 不知道什么用, 没有写 */
        //void *actor_delegate = (void *)cdos_draggable_get_actor_delegate (drag->actor);
        void *actor_delegate = cdos_draggable_get_delegate (drag->actor);
        if (drag->target)
        {
            //void *target_delegate = (void *)cdos_draggable_get_actor_delegate (drag->actor);
            void *target_delegate = cdos_draggable_get_delegate (drag->target);
            /*
            HandleDragOver handle_drag_over = NULL;
            if (G_IS_OBJECT (target_delegate))
            {
                if (g_object_class_find_property (G_OBJECT_GET_CLASS (target_delegate),
                        "handle-drag-over") != NULL)
                    g_object_get (target_delegate, "handle-drag-over", &handle_drag_over, NULL);
            }
            */
            HandleDragOver handle_drag_over = cdos_draggable_get_func (target_delegate, DND_HANDLE_DRAG_OVER);

            if (target_delegate && handle_drag_over)
            {
                gfloat targ_x, targ_y;
                gboolean r = clutter_actor_transform_stage_point (drag->target,
                        stage_x, stage_y, &targ_x, &targ_y);
                DragMotionResult result = handle_drag_over (target_delegate,
                        actor_delegate, drag->drag_actor, targ_x, targ_y, time);
                if (set_cursor (result))
                    return true;
            }
        }
        while (target)
        {
            /*
            void *target_delegate = (void *)cdos_draggable_get_actor_delegate (target);
            HandleDragOver handle_drag_over = NULL;
            if (G_IS_OBJECT (target_delegate))
            {
                if (g_object_class_find_property (G_OBJECT_GET_CLASS (target_delegate),
                        "handle-drag-over") != NULL)
                    g_object_get (target_delegate, "handle-drag-over", &handle_drag_over, NULL);
            }
            */
            if (drag->actor != target)
            {
                void *target_delegate = cdos_draggable_get_delegate (target);
                HandleDragOver handle_drag_over = NULL;
                if (target_delegate)
                    handle_drag_over = cdos_draggable_get_func (target,
                            DND_HANDLE_DRAG_OVER);
                if (target_delegate && handle_drag_over)
                {
                    gfloat targ_x, targ_y;
                    gboolean r = clutter_actor_transform_stage_point (target,
                            stage_x, stage_y, &targ_x, &targ_y);
                    DragMotionResult result = 
                        handle_drag_over (target_delegate, actor_delegate, 
                                drag->drag_actor, targ_x, targ_y, time);
                    if (set_cursor (result))
                        return true;
                }
            }
            target = clutter_actor_get_parent (target);
        }
        meta_screen_set_cursor (cdos_global_get_screen (global), META_CURSOR_DND_IN_DRAG);
    }
    return true;
}

static gboolean
drag_actor_dropped (CdosDraggable *drag, ClutterEvent *event)
{
    float drop_x, drop_y;
    clutter_event_get_coords (event, &drop_x, &drop_y);
    ClutterActor *stage = clutter_actor_get_stage (drag->drag_actor);
    ClutterActor *target = clutter_stage_get_actor_at_pos ((ClutterStage *)stage,
            CLUTTER_PICK_ALL, drop_x, drop_y);
    /* dragDrop 同 dragMotion, TODO */

    if (drag->target && drop_on_target (drag, drag->target, event, drop_x, drop_y))
        return true;

    while (target)
    {
        if (drop_on_target (drag, target, event, drop_x, drop_y))
            return true;
        target = clutter_actor_get_parent (target);
    }

    cancel_drag (drag, clutter_event_get_time (event));
    return true;
}

static gboolean
drop_on_target (CdosDraggable *drag,
                ClutterActor *target,
                ClutterEvent *event,
                float drop_x,
                float drop_y)
{
    if (target == drag->actor)
        return false;
    /*
    void *target_delegate = cdos_draggable_get_actor_delegate (target);

    AcceptDrop accept_drop = NULL;
    if (G_IS_OBJECT (target_delegate))
    {
        if (g_object_class_find_property (G_OBJECT_GET_CLASS (target_delegate),
                        "accept-drop") != NULL)
            g_object_get ((void *)target_delegate, "accept-drop", &accept_drop, NULL);
    }
    */
    void *target_delegate = cdos_draggable_get_delegate (target);
    AcceptDrop accept_drop = cdos_draggable_get_func (target, DND_ACCEPT_DROP);

    if(target_delegate && accept_drop)
    {
        gfloat targ_x, targ_y;
        gboolean r = clutter_actor_transform_stage_point (target,
                    drop_x, drop_y, &targ_x, &targ_y);
        guint32 time = clutter_event_get_time (event);

        if (accept_drop (target_delegate, 
                cdos_draggable_get_delegate (drag->actor),
                drag->drag_actor, targ_x, targ_y, time))
        {
            if (drag->actor_destroyed)
                return true;
            if (CLUTTER_IS_ACTOR (drag->drag_actor) &&
                    clutter_actor_get_parent (drag->drag_actor) == ui_group)
            {
                if (drag->restore_on_success)
                {
                    restore_drag_actor (drag, time);
                    return true;
                }
                else
                {
                    clutter_actor_destroy (drag->drag_actor);
                    drag->drag_actor = NULL;
                }
            }
            drag->drag_in_progress = false;
            meta_screen_set_cursor (cdos_global_get_screen (global), META_CURSOR_DEFAULT);
            g_signal_emit (drag, cdos_draggable_signals[DRAG_END], 0, time, true);
            drag_complete (drag);
            return true;
        }
    }
    return false;
}

static void
get_restore_location (CdosDraggable *drag, float *x, float *y, gdouble *scale)
{
    /* CLUTTER_ACTOR_IS_VISIBLE() 在后面的版本不再使用.
     * 当前版本没有 clutter_actor_is_visible() 
     */
    if (drag->drag_actor_source && 
            (drag->drag_actor_source->flags & CLUTTER_ACTOR_VISIBLE))
    {
        clutter_actor_get_transformed_position (drag->drag_actor_source, x, y);
        float w, h;
        clutter_actor_get_transformed_size (drag->drag_actor_source, &w, &h);
        *scale = clutter_actor_get_width (drag->drag_actor) / w;
    }
    else if (drag->orig_parent)
    {
        float parent_x, parent_y;
        clutter_actor_get_transformed_position (drag->orig_parent, &parent_x, &parent_y);
        float parent_width, parent_height;
        clutter_actor_get_size (drag->orig_parent, &parent_width, &parent_height);
        float parent_scale_width, parent_scale_height;
        clutter_actor_get_transformed_size (drag->orig_parent,
                &parent_scale_width, &parent_scale_height);
        float parent_scale = 1.0;
        if (parent_width != 0)
            parent_scale = parent_scale_width / parent_width;

        *x = parent_x + parent_scale * drag->orig.x;
        *y = parent_y + parent_scale * drag->orig.y;
        *scale = drag->orig_scale * parent_scale;
    }
    else{
        *x = drag->snap_back_x;
        *y = drag->snap_back_y;
        *scale = drag->snap_back_scale;
    }
}

static void
on_animation_complete (ClutterTimeline *tml, CdosDraggable *drag)
{
    /* Desklet TODO 
     * need Desklet."

    uintptr_t delegate;
    g_object_get (drag->drag_actor, "delegate", &delegate, NULL);
    if (CDOS_TYPE_IS_DESKLET (delegate))
       desklet_cancel_drag (delegate, drag->drag_actor);
    */
    st_container_remove_actor((ClutterContainer *)ui_group,
                    drag->drag_actor);

    if (drag->orig_parent)
    {
        if (ST_IS_CONTAINER (drag->orig_parent))
        {
            //st_container_move_child ((StContainer *)drag->orig_parent,
            //        drag->drag_actor, drag->orig_pos);
        }
    } 
    /*
    clutter_actor_set_position (drag->actor, drag->orig.x, drag->orig.y);
    clutter_actor_set_fixed_position_set(drag->actor, 0);
    clutter_actor_set_scale (drag->actor, drag->orig_scale, drag->orig_scale);
    */

    //else
        clutter_actor_destroy (drag->drag_actor);
    meta_screen_set_cursor (cdos_global_get_screen (global), META_CURSOR_DEFAULT);
    guint32 time = cdos_global_get_current_time (global);
    g_signal_emit (drag, cdos_draggable_signals[DRAG_END], 0, time, 0);

    drag->drag_in_progress = 0;
    drag_complete (drag);
}

static void
on_animation_cancel (ClutterTimeline *tml, CdosDraggable *drag)
{
    clutter_actor_remove_transition (drag->drag_actor, "drag-cancelled");
    on_animation_complete (tml, drag);
}
static void
cancel_drag (CdosDraggable *drag, guint32 event_time_opt)
{
    guint32 event_time = event_time_opt ? event_time_opt : cdos_global_get_current_time (global);
    g_signal_emit (drag, cdos_draggable_signals[DRAG_CANCELLED], 0, event_time);

    drag->drag_in_progress = false;
    float snap_back_x, snap_back_y;
    gdouble snap_back_scale;
    get_restore_location (drag, &snap_back_x, &snap_back_y, &snap_back_scale);

    if (drag->actor_destroyed)
    {
        /* Desklet TODO */
        /*
        void *delegate = cdos_draggable_get_target_delegate (drag);
        if (CDOS_TYPE_IS_DESKLET (delegate))
            desklet_cancel_drag (delegate, drag->drag_actor);
        */
        meta_screen_set_cursor (cdos_global_get_screen (global), META_CURSOR_DEFAULT);
        if (!drag->button_down)
            drag_complete (drag);
        g_signal_emit (drag, cdos_draggable_signals[DRAG_END], 0, event_time, 0);

        if ((drag->orig_parent == NULL) && drag->drag_actor != NULL)
            clutter_actor_destroy (drag->drag_actor);
        return;
    }
    
    drag->animation_in_progress = true;

    /* No target, so snap back */
    /*
    ClutterAnimation *anim = clutter_actor_animate (drag->drag_actor,
                                CLUTTER_EASE_OUT_QUAD,
                                SNAP_BACK_ANIMATION_TIME,
                                "x", snap_back_x,
                                "y", snap_back_y,
                                "scale-x", snap_back_scale,
                                "scale-y", snap_back_scale,
                                "opacity", drag->orig_opacity,
                                "signal::completed", on_animation_complete, drag,
                                NULL);
    */
    clutter_actor_save_easing_state (drag->drag_actor);

    ClutterTransition *trans = clutter_transition_group_new ();

    ClutterTransition *trans_x  = clutter_property_transition_new ("x");
    ClutterTransition *trans_y  = clutter_property_transition_new ("y");
    ClutterTransition *trans_a  = clutter_property_transition_new ("opacity");    
    ClutterTransition *trans_sx = clutter_property_transition_new ("scale-x");
    ClutterTransition *trans_sy = clutter_property_transition_new ("scale-y");

    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_x);
    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_y);
    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_a);    
    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_sx);
    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_sy);

    clutter_actor_set_easing_mode (drag->drag_actor, CLUTTER_EASE_OUT_QUAD);
    clutter_timeline_set_duration (CLUTTER_TIMELINE (trans), SNAP_BACK_ANIMATION_TIME);
    clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (trans), 0);

    clutter_transition_set_to (trans_x, G_TYPE_FLOAT, snap_back_x);
    clutter_transition_set_to (trans_y, G_TYPE_FLOAT, snap_back_y);
    clutter_transition_set_to (trans_a, G_TYPE_UCHAR, drag->orig_opacity);
    
    clutter_transition_set_to (trans_sx, G_TYPE_DOUBLE, snap_back_scale);
    clutter_transition_set_to (trans_sy, G_TYPE_DOUBLE, snap_back_scale);

    clutter_actor_add_transition (drag->drag_actor, "drag-cancelled", trans);
    clutter_actor_restore_easing_state (drag->drag_actor);
    g_signal_connect (trans, "completed", G_CALLBACK (on_animation_cancel), drag);

}

static void
on_animation_restore (ClutterTimeline *tml, CdosDraggable *drag)
{
    clutter_actor_remove_transition (drag->drag_actor, "drag-restore");
    on_animation_complete (tml, drag);
}

static void
restore_drag_actor (CdosDraggable *drag, guint32 event_time)
{
    drag->drag_in_progress = 0;
    vec3fd restore;
    get_restore_location (drag, &restore.x, &restore.y, &restore.scale);

    // fade the actor back in at its original location
    clutter_actor_set_position (drag->drag_actor, restore.x, restore.y);
    clutter_actor_set_scale (drag->drag_actor, restore.scale, restore.scale);
    clutter_actor_set_opacity (drag->drag_actor, 0);

    drag->animation_in_progress = 1;
    /*
    ClutterAnimation *anim = clutter_actor_animate (drag->drag_actor,
                            CLUTTER_EASE_OUT_QUAD, REVERT_ANIMATION_TIME,
                            "opacity", drag->orig_opacity,
                            "signal::completed", on_animation_complete, drag,
                            NULL);
    */
    clutter_actor_save_easing_state (drag->drag_actor);
    ClutterTransition *transition = clutter_property_transition_new ("opacity");
    clutter_actor_set_easing_mode (drag->drag_actor, CLUTTER_EASE_OUT_QUAD);
    clutter_timeline_set_duration (CLUTTER_TIMELINE (transition), REVERT_ANIMATION_TIME);
    clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (transition), 0);
    clutter_transition_set_from (transition, G_TYPE_UCHAR, 0x0);
    clutter_transition_set_to (transition, G_TYPE_UCHAR, drag->orig_opacity);
    clutter_actor_add_transition (drag->drag_actor, "drag-restore", transition);
    g_signal_connect (transition, "completed", G_CALLBACK (on_animation_restore), drag);
    clutter_actor_restore_easing_state (drag->drag_actor);
}

static void
sync_hover (CdosDraggable *drag, ClutterActor *actor)
{
    while (actor)
    {
        ClutterActor *parent = clutter_actor_get_parent (actor);
        if (ST_IS_WIDGET (actor))
            st_widget_sync_hover ((StWidget *)actor);
        actor = parent;
    }
}

static void
drag_complete (CdosDraggable *drag)
{
    if (drag->actor_destroyed != 0 && drag->drag_actor != NULL)
        cdos_base_util_set_hidden_from_pick (drag->drag_actor, 0);

    ungrab_events (drag);

    if (drag->first_leave_actor != NULL)
    {
        sync_hover (drag, drag->first_leave_actor);
        drag->first_leave_actor = NULL;
    }
    if (drag->last_enter_actor != NULL)
    {
        sync_hover (drag, drag->last_enter_actor);
        drag->last_enter_actor = NULL;
    }
    drag->drag_actor = NULL;
    current_draggable = NULL;
}

static void
cdos_draggable_class_init (CdosDraggableClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->finalize = cdos_draggable_finalize;

    /**
     * CdosDraggable::drag-begin:
     * @draggable: 正在拖动的DND对象
     * @time: 事件发生的时间
     *
     * 在开始拖动时发出 ::drag-begin 信号。
     *
     * Since: 1.0.0
     */
    cdos_draggable_signals[DRAG_BEGIN] = 
      g_signal_new ("drag-begin",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__UINT,
                    G_TYPE_NONE, 1,
                    G_TYPE_UINT);
    /**
     * CdosDraggable::drag-cancelled:
     * @draggable: 正在拖动的DND对象
     * @time: 事件发生的时间
     *
     * 在拖动取消时发出 ::drag-cancelled 信号。
     *
     * Since: 1.0.0
     */
    cdos_draggable_signals[DRAG_CANCELLED] = 
      g_signal_new ("drag-cancelled",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__UINT,
                    G_TYPE_NONE, 1,
                    G_TYPE_UINT);
    /**
     * CdosDraggable::drag-end:
     * @draggable: 正在拖动的DND对象
     * @time: 事件发生的时间
     * @snap: 拖动取消时为 %FALSE, 正常结束为 %TRUE。
     *
     * 在拖动结束时发出 ::drag-end 信号。
     *
     * Since: 1.0.0
     */
    cdos_draggable_signals[DRAG_END] = 
      g_signal_new ("drag-end",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    _st_marshal_VOID__UINT_UINT,
                    G_TYPE_NONE, 2,
                    G_TYPE_UINT, G_TYPE_UINT);
}

/**
 * cdos_make_draggable:
 * @actor: Source actor
 * @target: (optional) 接受拖放事件的目标 actor
 * @n_args: 可变参数个数
 *
 * Create an object which controls drag and drop for the given actor.
 *
 * 可变参数列表：
 *
 *   #gboolean      manual_mode,
 *
 *   #gboolean      restore_on_success,
 *
 *   #gint          drag_actor_max_size,
 *
 *   #float         drag_actor_opacity
 *
 * 如果 manualMode 为 %true, 不会在点击时自动开始拖动。
 *
 * 如果 drag_actor_max_size 不为 0, drag actor 将会根据其值进行缩放，且不会超过原 actor 的大小
 *
 * 如果设置了 drag_actor_opacity，drag actor 的透明度将会设置为该值。
 *
 * Returns: #CdosDraggable 对象。
 */
CdosDraggable *
cdos_make_draggable (ClutterActor *actor,
                     ClutterActor *target,
                     gint n_args,
                     ...)
                        
{
    va_list ap;
    va_start(ap, n_args);
    gint manual = 0;
    gint restore = 0;
    gint max_size = -1;
    gfloat opacity = -1.0;

    gint i;
    gint m = min(4, n_args);
    for (i = 0; i < m; i++)
    {
        if (i == 0)
        {
            manual = va_arg(ap, gint);
            continue;
        }
        if (i == 1)
        {
            restore = va_arg(ap, gint);
            continue;
        }
        if (i == 2)
        {
            max_size = va_arg(ap, gint);
            continue;
        }
        if (i == 3)
            opacity = (gfloat)va_arg(ap, gdouble);
        if (i >= 4)
            break;
    }
    va_end(ap);
    CdosDraggable *drag = cdos_draggable_new (actor,
                      target,
                      manual,
                      restore,
                      max_size,
                      opacity);
    return drag;
}


/* CdosGenericDragItemContainer */

/*
G_DEFINE_TYPE (CdosGenericDragItemContainer, cdos_generic_drag_item_container, CDOS_TYPE_GENERIC_CONTAINER);

static void
generic_drag_item_container_get_preferred_width (gpointer  actor,
                                                 gfloat    for_width,
                                                 gpointer  allocation)
{
    CdosGenericContainerAllocation *alloc = allocation;
    alloc->min_size = 0;
    alloc->natural_size = 0;
    
    CdosGenericDragItemContainer *self = actor;
    if (self->child == NULL)
        return;

    gfloat min_width, nat_width;
    clutter_actor_get_preferred_width (self->child, for_width, &min_width, &nat_width);
    gdouble scale_y;
    clutter_actor_get_scale (self->child, NULL, &scale_y);
    alloc->min_size += min_width * scale_y;
    alloc->natural_size = nat_width * scale_y;
}

static void
generic_drag_item_container_get_preferred_height (gpointer actor,
                                                  gfloat   for_height,
                                                  gpointer allocation)
{
    CdosGenericContainerAllocation *alloc = allocation;
    alloc->min_size = 0;
    alloc->natural_size = 0;
    
    CdosGenericDragItemContainer *self = actor;
    if (self->child == NULL)
        return;

    gfloat min_height, nat_height;
    clutter_actor_get_preferred_height (self->child, for_height, &min_height, &nat_height);
    gdouble scale_y;
    clutter_actor_get_scale (self->child, NULL, &scale_y);
    alloc->min_size += min_height * scale_y;
    alloc->natural_size = nat_height * scale_y;
}

static void
generic_drag_item_container_allocate (gpointer  actor,
                                      ClutterActorBox  *box,
                                      guint     flags)
{
    CdosGenericDragItemContainer *self = actor;

    if (self->child == NULL)
        return;

    gfloat avail_width = box->x2 - box->x1;
    gfloat avail_height = box->y2 - box->y1;

    gfloat min_child_width, min_child_height, nat_child_width, nat_child_height;
    clutter_actor_get_preferred_size (self->child,
            &min_child_width, &min_child_height,
            &nat_child_width, &nat_child_height);

    gdouble child_scale_x, child_scale_y;
    clutter_actor_get_scale (self->child, &child_scale_x, &child_scale_y);

    gfloat child_width = min(nat_child_width * child_scale_x, avail_width);
    gfloat child_height = min(nat_child_height * child_scale_y, avail_height);

    gfloat x1 = (avail_width - child_width) / 2;
    gfloat y1 = (avail_height - child_height) / 2;
    gfloat x2 = x1 + child_width;
    gfloat y2 = y1 + child_height;
    ClutterActorBox *child_box = clutter_actor_box_new (x1, y1, x2, y2);

    clutter_actor_allocate (self->child, child_box, (ClutterAllocationFlags)flags);
}

static void
cdos_generic_drag_item_container_init (CdosGenericDragItemContainer *self)
{
    self->child = NULL;
    self->child_scale = 1;
    self->child_opacity = 255;
    self->animating_out = 0;
}

static CdosGenericDragItemContainer *
cdos_generic_drag_item_container_new ()
{
    CdosGenericDragItemContainer *container = g_object_new (CDOS_TYPE_GENERIC_DRAG_ITEM_CONTAINER, NULL);
    st_widget_set_style_class_name ((StWidget *)container, "drag-item-container");

    g_signal_connect ((CdosGenericContainer *)container, "get-preferred-width",
            G_CALLBACK (generic_drag_item_container_get_preferred_width), NULL);
    g_signal_connect ((CdosGenericContainer *)container, "get-preferred-height",
            G_CALLBACK (generic_drag_item_container_get_preferred_height), NULL);
    g_signal_connect ((CdosGenericContainer *)container, "allocate",
            G_CALLBACK (generic_drag_item_container_allocate), NULL);


    return container;
}

static void
cdos_generic_drag_item_container_dispose (GObject *object)
{
    G_OBJECT_CLASS(cdos_generic_drag_item_container_parent_class)->dispose (object);
}

static void
cdos_generic_drag_item_container_finalize (GObject *object)
{
    G_OBJECT_CLASS(cdos_generic_drag_item_container_parent_class)->finalize (object);
}

static void
cdos_generic_drag_item_container_class_init(CdosGenericDragItemContainerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->dispose = cdos_generic_drag_item_container_dispose;
    gobject_class->finalize = cdos_generic_drag_item_container_finalize;

}

void
cdos_generic_drag_item_container_set_child (CdosGenericDragItemContainer *container,
                                            ClutterActor *actor)
{
    if (container->child == actor)
        return;
    clutter_actor_destroy_all_children ((ClutterActor *)container);
    container->child = actor;
    clutter_actor_add_child ((ClutterActor *)container, actor);
}

static void
on_animate_redraw (ClutterTimeline *tml, CdosGenericDragItemContainer *container)
{
    //ClutterActor *child = container->child;

    clutter_actor_queue_relayout ((ClutterActor *)container);
    clutter_actor_queue_redraw ((ClutterActor *)container);
}

static void
on_animate_in (ClutterTimeline *tml, ClutterActor *actor)
{
    clutter_actor_remove_transition (actor, "animate-in");
}

void
cdos_generic_drag_item_container_animate_in (CdosGenericDragItemContainer *container)
{
    if (container->child == NULL)
        return;
    //container->child_scale = 0;
    cdos_generic_drag_item_container_set_child_opacity (container, 0);
    cdos_generic_drag_item_container_set_child_scale (container, 0.0f);

    clutter_actor_set_pivot_point (container->child, 0.5, 0.5);
    clutter_actor_set_scale (container->child, 0.f, 0.f);
    clutter_actor_show (container->child);
    clutter_actor_save_easing_state (container->child);

    ClutterTransition *trans = clutter_transition_group_new ();

    ClutterTransition *trans_a = clutter_property_transition_new ("opacity");    
    ClutterTransition *trans_x = clutter_property_transition_new ("scale-x");
    ClutterTransition *trans_y = clutter_property_transition_new ("scale-y");

    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_a);    
    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_x);
    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_y);

    clutter_actor_set_easing_mode (container->child, CLUTTER_EASE_OUT_QUAD);
    clutter_timeline_set_duration (CLUTTER_TIMELINE (trans), DND_ANIMATION_TIME);
    clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (trans), 0);

    clutter_transition_set_from (trans_x, G_TYPE_DOUBLE, 0.f);
    clutter_transition_set_to (trans_x, G_TYPE_DOUBLE, 1.0f);

    clutter_transition_set_from (trans_y, G_TYPE_DOUBLE, 0.0f);
    clutter_transition_set_to (trans_y, G_TYPE_DOUBLE, 1.0f);

    clutter_transition_set_from (trans_a, G_TYPE_UCHAR, 0);
    clutter_transition_set_to (trans_a, G_TYPE_UCHAR, 0xff);

    clutter_actor_add_transition (container->child, "animate-in", trans);
    clutter_actor_restore_easing_state (container->child);

    g_signal_connect (trans, "new-frame", G_CALLBACK (on_animate_redraw), container);    
    g_signal_connect (trans, "completed", G_CALLBACK (on_animate_in), container->child);

}

static void
on_animate_out (ClutterTimeline *tml, ClutterActor *actor)
{
    CdosGenericDragItemContainer *container = (void *)actor;
    
    clutter_actor_remove_transition (container->child, "animate-out");
    clutter_actor_destroy (container->child);
    clutter_actor_destroy (actor);
}
void
cdos_generic_drag_item_container_animate_out_and_destroy (CdosGenericDragItemContainer *container)
{
    if (container->child == NULL)
    {
        clutter_actor_destroy ((ClutterActor *)container);
        return;
    }
    container->animating_out = 1;
    //container->child_scale = 1.0;
    cdos_generic_drag_item_container_set_child_scale (container, 1.0f);    
  
    clutter_actor_set_pivot_point (container->child, 0.5, 0.5);    
    clutter_actor_set_scale (container->child, 1.0f, 1.0f);
    clutter_actor_show (container->child);
    clutter_actor_save_easing_state (container->child);

    ClutterTransition *trans = clutter_transition_group_new ();

    ClutterTransition *trans_a = clutter_property_transition_new ("opacity");    
    ClutterTransition *trans_x = clutter_property_transition_new ("scale-x");
    ClutterTransition *trans_y = clutter_property_transition_new ("scale-y");

    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_a);    
    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_x);
    clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_y);

    clutter_actor_set_easing_mode (container->child, CLUTTER_EASE_OUT_QUAD);
    clutter_timeline_set_duration (CLUTTER_TIMELINE (trans), DND_ANIMATION_TIME);
    clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (trans), 0);

    clutter_transition_set_from (trans_x, G_TYPE_DOUBLE, 1.f);
    clutter_transition_set_to (trans_x, G_TYPE_DOUBLE, 0.0f);

    clutter_transition_set_from (trans_y, G_TYPE_DOUBLE, 1.0f);
    clutter_transition_set_to (trans_y, G_TYPE_DOUBLE, 0.0f);

    clutter_transition_set_from (trans_a, G_TYPE_UCHAR, 0xff);
    clutter_transition_set_to (trans_a, G_TYPE_UCHAR, 0x0);

    clutter_actor_add_transition (container->child, "animate-out", trans);
    clutter_actor_restore_easing_state (container->child);
    g_signal_connect (trans, "new-frame", G_CALLBACK (on_animate_redraw), container);
    g_signal_connect (trans, "completed", G_CALLBACK (on_animate_out), container);
}

void
cdos_generic_drag_item_container_set_child_scale (CdosGenericDragItemContainer *container, gdouble scale)
{
    container->child_scale = scale;
    if (container->child == NULL)
        return;
    clutter_actor_set_pivot_point (container->child, 0.5, 0.5);
    clutter_actor_queue_relayout ((ClutterActor *)container);
}

gdouble
cdos_generic_drag_item_container_get_child_scale (CdosGenericDragItemContainer *container)
{
    return container->child_scale;
}

void
cdos_generic_drag_item_container_set_child_opacity (CdosGenericDragItemContainer *container, guint8 opacity)
{
    container->child_opacity = opacity;
    if (container->child == NULL)
        return;
    clutter_actor_set_opacity (container->child, opacity);
    clutter_actor_queue_redraw ((ClutterActor *)container);
}

guint8
cdos_generic_drag_item_container_get_child_opacity (CdosGenericDragItemContainer *container)
{
    return container->child_opacity;
}

*/

/* CdosGenericDragPlaceholderItem */

/*
G_DEFINE_TYPE (CdosGenericDragPlaceholderItem, cdos_generic_drag_placeholder_item, \
        CDOS_TYPE_GENERIC_DRAG_ITEM_CONTAINER);

static void
cdos_generic_drag_placeholder_item_init (CdosGenericDragPlaceholderItem *item)
{

}

static void
cdos_generic_drag_placeholder_item_class_init (CdosGenericDragPlaceholderItemClass *klass)
{

}

CdosGenericDragPlaceholderItem *
cdos_generic_drag_placeholder_item_new ()
{
    CdosGenericDragItemContainer *container = 
        cdos_generic_drag_item_container_new ();
    StWidget *child = st_bin_new ();
    st_widget_set_style_class_name (child, "drag-placeholder");
    cdos_generic_drag_item_container_set_child (container, (ClutterActor *)child);

    return (CdosGenericDragPlaceholderItem *)container;
}

*/

