#ifndef __CDOS_DND_H__
#define __CDOS_DND_H__

#include <clutter/clutter.h>
#include "cdos-generic-container.h"
#include "cdos-types.h"


/**
 * DragMotionResult:
 * @DRAG_MOTION_NO_DROP: 不接受 Drop 事件
 * @DRAG_MOTION_COPY_DROP: 光标样式为 copy-drop
 * @DRAG_MOTION_MOVE_DROP: 光标样式为 move-drop
 * @DRAG_DROP_CONTINUE: 光标样式为 continue
 *
 * 函数 #HandleDragOver 的返回值。
 *
 * Since: 1.0.0
 */
typedef enum _DragMotionResult {
    DRAG_MOTION_NO_DROP,
    DRAG_MOTION_COPY_DROP,
    DRAG_MOTION_MOVE_DROP,
    DRAG_MOTION_CONTINUE
}DragMotionResult;

/**
 * DragDropResult:
 * @DRAG_DROP_FAILURE:
 * @DRAG_DROP_SUCCESS:
 * @DRAG_DROP_CONTINUE:
 *
 * 没有用到。
 *
 * Since: 1.0.0
 */
typedef enum _DragDropResult {
    DRAG_DROP_FAILURE,
    DRAG_DROP_SUCCESS,
    DRAG_DROP_CONTINUE
}DragDropResult;


//gboolean is_draggable ();

GType cdos_draggable_get_type ();

#define CDOS_TYPE_DRAGGABLE      (cdos_draggable_get_type ())
#define CDOS_DRAGGABLE(obj)          (G_TYPE_CHECK_INSTANCE_CAST ((obj), \
                         CDOS_TYPE_DRAGGABLE, CdosDraggable))
#define CDOS_IS_DRAGGABLE(obj)       (G_TYPE_CHECK_INSTANCE_TYPE ((obj), \
            CDOS_TYPE_DRAGGABLE))
#define CDOS_DRAGGABLE_CLASS(klass)     (G_TYPE_CHECK_CLASS_CAST ((klass), \
            CDOS_TYPE_DRAGGABLE, CdosDraggableClass))
#define CDOS_IS_DRAGGABLE_CLASS(klass)  (G_TYPE_CHECK_CLASS_TYPE ((klass), \
            CDOS_TYPE_DRAGGABLE))
#define CDOS_DRAGGABLE_GET_CLASS(obj)   (G_TYPE_INSTANCE_GET_CLASS ((obj), \
            CDOS_TYPE_DRAGGABLE, CdosDraggable))

typedef struct _CdosDraggable CdosDraggable;
typedef struct _CdosDraggableClass CdosDraggableClass;

struct _CdosDraggable {
    GObject       parent;
    gboolean inhibit;
    ClutterActor *actor;
    ClutterActor *target;
    /* <private> */
    ClutterActor *first_leave_actor;
    ClutterActor *last_enter_actor;
    ClutterActor *orig_parent;
    ClutterActor *previous_key_focus_actor;
    ClutterActor *drag_actor;
    ClutterActor *drag_actor_source;

    gint     orig_pos;

    gboolean drag_in_progress;
    gboolean actor_destroyed;
    gulong on_event_id;
    gulong button_press_event_id;
    gulong destroy_event_id;

    gboolean restore_on_success;

    gboolean button_down;
    gboolean animation_in_progress;
    gboolean events_grabbed;

    /* drag */
    vec2f    start;
    float    x;
    float    y;
    vec2f    offset;
    gint     drag_actor_max_size;
    guint8   drag_actor_opacity;
    gulong   drag_check_id;
    guint    drag_event_timeout_id;
    gint     drag_threshold;

    /* orig */
    vec2f    orig;
    gdouble  orig_scale;
    guint8   orig_opacity;
    /* snap */
    float    snap_back_x;
    float    snap_back_y;
    gdouble  snap_back_scale;
};

struct _CdosDraggableClass {
    GObjectClass    parent_class;
};
/**
 * DNDFuncType:
 * @DND_GET_DRAG_ACTOR: #GetDragActor
 * @DND_GET_DRAG_ACTOR_SRC: #GetDragActorSource
 * @DND_HANDLE_DRAG_OVER: #HandleDragOver
 * @DND_ACCEPT_DROP: #AcceptDrop
 *
 * 用于 cdos_draggable_set_func(), 设置DND对象的四个函数。
 *
 * Since: 1.0.0
 */
typedef enum {
    DND_GET_DRAG_ACTOR,
    DND_GET_DRAG_ACTOR_SRC,
    DND_HANDLE_DRAG_OVER,
    DND_ACCEPT_DROP
}DNDFuncType;

/**
 * HandleDragOver:
 * @self: 正在拖动的对象
 * @source: 目标对象
 * @actor: 没有用到，参照 js 代码保留了该参数
 * @x: 光标位置坐标 x
 * @y: 光标位置 y
 * @time: ClutterEvent 时间
 *
 * 拖动 A 到 B，先调用 A 的 get_drag_actor()、get_drag_actor_source(),
 * 拖动到 B 时，调用 B 的 handle_drag_over(), accept_drop(),
 * (参照 cdos-2.8.6 及 gnome-shell 的 dnd.js, 删除了
 * clear_drag_placeholder() 和 hide_drag_placeholder())
 * 下面函数中的 self 为 A，source 为 B.
 *
 * Returns: 根据返回结果改变光标样式。
 */
typedef DragMotionResult (*HandleDragOver) (void         *self,
                                            void         *source,/* delegate */
                                            ClutterActor *actor,
                                            float         x,
                                            float         y,
                                            guint32       time);
/**
 * AcceptDrop:
 * @self: 正在拖动的对象
 * @source: 目标对象
 * @actor: 没有用到，参照 js 代码保留了该参数
 * @x: 光标位置坐标 x
 * @y: 光标位置 y
 * @time: ClutterEvent 时间
 *
 * 当拖动 A 到 B 时, 执行 B 的该函数
 *
 * Returns: 根据返回结果判断是否接受 Drop 事件。
 */
typedef gboolean (*AcceptDrop) (void         *self,
                                void         *source, /* delegate */
                                ClutterActor *actor,
                                float         x,
                                float         y,
                                guint32       time);


/**
 * GetDragActor:
 * @delegate: ClutterActor 的 delegate。
 *
 * 从 delegate 获取一个 actor，用于拖动时在光标下显示。
 *
 * Returns: 一个 ClutterActor。
 */
typedef ClutterActor *(*GetDragActor) (void *delegate);

/**
 * GetDragActorSource:
 * @delegate: ClutterActor 的 delegate。
 *
 * 从 delegate 获取一个 actor。
 *
 * Returns: 一个 ClutterActor, 一般为该 delegate 的 actor。
 */
typedef ClutterActor *(*GetDragActorSource) (void *delegate);

void cdos_draggable_set_delegate (void *actor, void *delegate);
void cdos_draggable_set_func (void *actor, void *func, DNDFuncType type);

void *cdos_draggable_get_delegate (ClutterActor *actor);
ClutterActor *cdos_draggable_get_delegate_actor (void *delegate);
void *cdos_draggable_get_func (ClutterActor *actor, DNDFuncType type);


/*
void cdos_draggable_start_drag (CdosDraggable *draggable,
                               float      stage_x,
                               float      stage_y,
                               guint32    time);

*/

CdosDraggable *
cdos_make_draggable (ClutterActor *actor,
                     ClutterActor *target,
                     gint n_args,
                         ...);
/*
                         gboolean      manual_mode,
                         gboolean      restore_on_success,
                         gint          drag_actor_max_size,
                         float         drag_actor_opacity);
*/

/* GenericDragPlaceholderItem */

/*
typedef struct _CdosGenericDragItemContainer CdosGenericDragItemContainer;
typedef struct _CdosGenericDragItemContainerClass CdosGenericDragItemContainerClass;

#define CDOS_TYPE_GENERIC_DRAG_ITEM_CONTAINER              (cdos_generic_drag_item_container_get_type ())
#define CDOS_GENERIC_DRAG_ITEM_CONTAINER(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), CDOS_TYPE_GENERIC_DRAG_ITEM_CONTAINER, CdosGenericDragItemContainer))
#define CDOS_GENERIC_DRAG_ITEM_CONTAINER_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), CDOS_TYPE_GENERIC_DRAG_ITEM_CONTAINER, CdosGenericDragItemContainerClass))
#define CDOS_IS_GENERIC_DRAG_ITEM_CONTAINER(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), CDOS_TYPE_GENERIC_DRAG_ITEM_CONTAINER))
#define CDOS_IS_GENERIC_DRAG_ITEM_CONTAINER_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), CDOS_TYPE_GENERIC_DRAG_ITEM_CONTAINER))
#define CDOS_GENERIC_DRAG_ITEM_CONTAINER_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), CDOS_TYPE_GENERIC_DRAG_ITEM_CONTAINER, CdosGenericDragItemContainerClass))

GType cdos_generic_drag_item_container_get_type (void) G_GNUC_CONST;

struct _CdosGenericDragItemContainer {
    CdosGenericContainer parent;

    float child_scale;
    guint8 child_opacity;
    gboolean animating_out;

    ClutterActor *child;
};

void cdos_generic_drag_item_container_set_child (CdosGenericDragItemContainer *container, ClutterActor *actor);

void cdos_generic_drag_item_container_animate_in (CdosGenericDragItemContainer *container);

void cdos_generic_drag_item_container_animate_out_and_destroy (CdosGenericDragItemContainer *container);

void cdos_generic_drag_item_container_set_child_scale (CdosGenericDragItemContainer *container, gdouble scale);

gdouble cdos_generic_drag_item_container_get_child_scale (CdosGenericDragItemContainer *container);

void cdos_generic_drag_item_container_set_child_opacity (CdosGenericDragItemContainer *container, guint8 opacity);

guint8 cdos_generic_drag_item_container_get_child_opacity (CdosGenericDragItemContainer *container);

struct _CdosGenericDragItemContainerClass {
    CdosGenericContainerClass parent_class;
};
*/
/* CdosGenericDragPlaceholderItem */
/*
typedef struct _CdosGenericDragPlaceholderItem CdosGenericDragPlaceholderItem;
typedef struct _CdosGenericDragPlaceholderItemClass CdosGenericDragPlaceholderItemClass;

#define CDOS_TYPE_GENERIC_DRAG_PLACEHOLDER_ITEM              (cdos_generic_drag_placeholder_item_get_type ())
#define CDOS_GENERIC_DRAG_PLACEHOLDER_ITEM(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), CDOS_TYPE_GENERIC_DRAG_PLACEHOLDER_ITEM, CdosGenericDragPlaceholderItem))
#define CDOS_GENERIC_DRAG_PLACEHOLDER_ITEM_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), CDOS_TYPE_GENERIC_DRAG_PLACEHOLDER_ITEM, CdosGenericDragPlaceholderItemClass))
#define CDOS_IS_GENERIC_DRAG_PLACEHOLDER_ITEM(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), CDOS_TYPE_GENERIC_DRAG_PLACEHOLDER_ITEM))
#define CDOS_IS_GENERIC_DRAG_PLACEHOLDER_ITEM_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), CDOS_TYPE_GENERIC_DRAG_PLACEHOLDER_ITEM))
#define CDOS_GENERIC_DRAG_PLACEHOLDER_ITEM_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), CDOS_TYPE_GENERIC_DRAG_PLACEHOLDER_ITEM, CdosGenericDragPlaceholderItemClass))

struct _CdosGenericDragPlaceholderItem {
    CdosGenericDragItemContainer parent;
};

struct _CdosGenericDragPlaceholderItemClass {
    CdosGenericDragItemContainerClass parent_class;
};

GType cdos_generic_drag_placeholder_item_get_type (void) G_GNUC_CONST;
CdosGenericDragPlaceholderItem *cdos_generic_drag_placeholder_item_new ();
*/
#endif // __UI_DND_H__
