#include "cdos-doc-info.h"
#include <gtk/gtk.h>

#include <meta/display.h>
#include "cdos-debug.h"
#include "cdos-global.h"

struct _CdosDocInfoPrivate
{
        GtkRecentInfo                  *recentInfo;
        GnomeDesktopThumbnailFactory   *factory;
        time_t                          timestamp;
        gchar                          *name;
        gchar                          *lowerName;
        const gchar                          *uri;
        gchar                          *mimeType;
        gint                            mtime;
};

G_DEFINE_TYPE(CdosDocInfo, cdos_doc_info, G_TYPE_OBJECT);

#define CDOS_DOC_INFO_GET_PRIVATE(obj)    \
        (G_TYPE_INSTANCE_GET_PRIVATE((obj), CDOS_TYPE_DOC_INFO, CdosDocInfoPrivate))

static void 
cdos_doc_info_init(CdosDocInfo *info);

static void 
cdos_doc_info_class_init(CdosDocInfoClass *klass);

static void 
cdos_doc_info_dispose(GObject *gobject);

static gint 
cdos_doc_info_fetch_mtime(CdosDocInfo *info);

static gboolean
cdos_doc_info_real_launch(CdosDocInfo *info);

static void 
cdos_doc_info_onmount_callback(CdosDocInfo *info,
                               GFile *file,
                               const char *result,
                               gpointer data);

static gint 
index_of(const gchar *haystack, const gchar *needle);

static void 
cdos_doc_info_init(CdosDocInfo *info)
{
        CdosDocInfoPrivate *priv;
        info->priv = priv = CDOS_DOC_INFO_GET_PRIVATE(info);

        priv->recentInfo = NULL;
        priv->factory = NULL;
        priv->timestamp = 0;
        priv->name = NULL;
        priv->lowerName = NULL;
        priv->uri = NULL;
        priv->mimeType = NULL;
        priv->mtime = 0;
}

static void 
cdos_doc_info_class_init(CdosDocInfoClass *klass)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

        gobject_class->dispose = cdos_doc_info_dispose;
        g_type_class_add_private(klass, sizeof(CdosDocInfoPrivate));
}

static void
cdos_doc_info_dispose(GObject *gobject)
{
        CdosDocInfoPrivate *priv = CDOS_DOC_INFO(gobject)->priv;

        if (priv->recentInfo)
        {
            gtk_recent_info_unref(priv->recentInfo);
        }

        if (priv->factory)
        {
            g_object_unref(priv->factory);
        }

        if (priv->name)
        {
            g_free(priv->name);
        }

        if (priv->mimeType)
        {
            g_free(priv->mimeType);
        }
        if (priv->lowerName)
        {
            g_free(priv->lowerName);
        }

        G_OBJECT_CLASS(cdos_doc_info_parent_class)->dispose(gobject);
}

/**
 * cdos_doc_info_new
 * @recentInfo: #GtkRecentInfo对象，与最近使用的文件有关
 * @factory: #GnomeDesktopThumbnailFactory对象，生成和查找文件和目录的缩略图
 *
 * 创建#CdosDocInfo对象
 *
 * Returns: #CdosDocInfo对象
 */

CdosDocInfo *
cdos_doc_info_new(GtkRecentInfo *recentInfo,
                  GnomeDesktopThumbnailFactory   *factory)
{
        CdosDocInfo *gobject;
        g_return_val_if_fail(recentInfo != NULL,NULL);
        //g_return_val_if_fail(GNOME_DESKTOP_IS_THUMBNAIL_FACTORY(factory),NULL);

        gobject = g_object_new(CDOS_TYPE_DOC_INFO, NULL);
        cdos_doc_info_set_params(CDOS_DOC_INFO(gobject), recentInfo, factory);

        return gobject;        
}

/**
 * cdos_doc_info_set_params
 * @info: #CdosDocInfo对象
 * @recentInfo: #GtkRecentInfo对象，与最近使用的文件有关
 * @factory: #GnomeDesktopThumbnailFactory对象，生成和查找文件和目录的缩略图
 *
 * 设置#CdosDocInfo对象的私有属性值
 *
 * Returns: #CdosDocInfo对象
 */

CdosDocInfo *
cdos_doc_info_set_params(CdosDocInfo *info,
                         GtkRecentInfo *recentInfo,
                         GnomeDesktopThumbnailFactory   *factory)
{
        g_return_val_if_fail(CDOS_IS_DOC_INFO(info),NULL);
        CdosDocInfoPrivate *priv = info->priv;
        priv->recentInfo = recentInfo;
        gtk_recent_info_ref(priv->recentInfo);

        priv->factory = factory;
        g_object_ref(priv->factory);

        priv->timestamp = gtk_recent_info_get_modified(recentInfo);
        priv->name = g_strdup(gtk_recent_info_get_display_name(recentInfo));

        priv->lowerName = g_utf8_strdown(priv->name, -1);
        priv->uri = gtk_recent_info_get_uri(recentInfo);
        priv->mimeType = g_strdup(gtk_recent_info_get_mime_type(recentInfo));
        //priv->mtime = cdos_doc_info_fetch_mtime(info);

        return info;
}

static gint 
cdos_doc_info_fetch_mtime(CdosDocInfo *info)
{
        GFileInfo *file_info;
        GFile     *file;
        gint ret = -1;

        CdosDocInfoPrivate *priv = info->priv;
        if (g_str_has_prefix(priv->uri, "file://")) {
                file = g_file_new_for_uri(priv->uri);
                file_info = g_file_query_info(file, G_FILE_ATTRIBUTE_TIME_MODIFIED, 
                                              G_FILE_CREATE_NONE, NULL, NULL);
                if (file_info) {
                        ret = g_file_info_get_attribute_uint64(file_info, G_FILE_ATTRIBUTE_TIME_MODIFIED);
                }

                g_object_unref(file_info);
                g_object_unref(file);
        }

        return ret;
}

/**
 * cdos_doc_info_get_uri
 * @info: #CdosDocInfo对象
 *
 * 获取#CdosDocInfo对象的uri属性值，即资源所对应的uri
 *
 * Returns: #CdosDocInfo对象
 */

const gchar *
cdos_doc_info_get_uri(CdosDocInfo *info)
{
        g_return_val_if_fail(CDOS_IS_DOC_INFO(info), NULL);
        return info->priv->uri;
}

/**
 * cdos_doc_info_create_icon
 * @info: #CdosDocInfo对象
 * @size: 主题中对应的图标大小
 *
 * 根据指定的size创建文档图标
 *
 * Returns: 返回#ClutterActor对象类型的图标
 */

ClutterActor*
cdos_doc_info_create_icon(CdosDocInfo *info, gint size)
{
        g_return_val_if_fail(CDOS_IS_DOC_INFO(info), NULL);
        CdosDocInfoPrivate *priv = info->priv;
        gchar *existing = NULL;
        ClutterActor *actor;
   
        StTextureCache* texture_cache = st_texture_cache_get_default ();      
        GIcon *gicon = gtk_recent_info_get_gicon(priv->recentInfo);

        actor = st_texture_cache_load_gicon(texture_cache, NULL, gicon, size);
        g_object_unref(gicon);
        return actor;
}

static gboolean
cdos_doc_info_real_launch(CdosDocInfo *info)
{
        GdkAppLaunchContext * context;
        //context = cdos_global_create_app_launch_context (cdos_global_get());

        CdosGlobal *cdos_global = cdos_global_get ();
        GdkScreen  *gdk_screen = cdos_global_get_gdk_screen(cdos_global);
        GdkDisplay *display = gdk_screen_get_display(gdk_screen);

        context = gdk_display_get_app_launch_context (display);
        gboolean res = FALSE;
        const gchar *uri = cdos_doc_info_get_uri(info);

        if (uri)
            res =  g_app_info_launch_default_for_uri(uri, (GAppLaunchContext *)context, NULL);

        g_object_unref (context);
        return res;
}

static void 
cdos_doc_info_onmount_callback(CdosDocInfo *info,
                               GFile *file,
                               const char *result,
                               gpointer data)
{
        gchar *error;
        if (!cdos_doc_info_real_launch(info)) {
                //GQuark *quark = g_quark_from_string("g_vfs_error_quark");
                error = g_file_get_uri(file);
                g_print("docinfo : Failed to mount: %s", error);
                g_free(error);
        }
}

/**
 * cdos_doc_info_launch
 * @info: #CdosDocInfo对象
 *
 * 启动应用程序或是挂载含有文件位置的卷
 *
 */

void
cdos_doc_info_launch(CdosDocInfo *info)
{
        g_return_if_fail(CDOS_IS_DOC_INFO(info));
        CdosDocInfoPrivate *priv = info->priv;
        if (priv->mtime == -1) {
                GFile *file = g_file_new_for_uri(priv->uri);
                g_file_mount_enclosing_volume(file, G_MOUNT_MOUNT_NONE, NULL, NULL, (GAsyncReadyCallback)cdos_doc_info_onmount_callback, NULL);
                g_object_unref(file);
        } else {
                cdos_doc_info_real_launch(info);
        }
}

/**
 * cdos_doc_info_match_terms
 * @info: #CdosDocInfo对象
 * @terms: 准备用于匹配的#GArray数组
 *
 * 匹配#CdosDocInfo的属性lowerName名称是否在给定的terms中
 *
 * Returns: 返回PREFIX 或 MULTIPLE_SUBSTRING 或 NONE
 */

gint 
cdos_doc_info_match_terms(CdosDocInfo *info, GList *terms)
{
        g_return_val_if_fail(CDOS_IS_DOC_INFO(info),0);
        MATCHTYPE mtype = NONE;
        gint idx;
        if(terms){
                GList *temp = terms;
                while(temp){
                        gchar *term = terms->data;
                        idx = index_of(info->priv->lowerName,term);
                        if (idx == 0) {
                                mtype = PREFIX;
                                break ;
                        } else if (idx > 0) {
                                mtype = MULTIPLE_SUBSTRING;
                                break ;
                        }
                        temp = temp->next;
                }

        }
        return mtype;
}

/**
 * cdos_doc_info_get_name
 * @info: #CdosDocInfo对象
 *
 * 获取#CdosDocInfo的属性name，文档名称
 *
 * Returns: name对应的字符串
 */

gchar *
cdos_doc_info_get_name(CdosDocInfo *info)
{
        g_return_val_if_fail(CDOS_IS_DOC_INFO(info),NULL);
        CdosDocInfoPrivate *priv = info->priv;
        return priv->name;
}

static gint 
index_of(const gchar *haystack, const gchar *needle)
{
        if (haystack == NULL || needle == NULL)
                return -1;
        gchar *m = g_strrstr(haystack, needle);
        if (m) {
                return m-haystack+1;
        }

        return -1;
}



