#include "cdos-doc-manager.h"
#include "cdos-doc-system.h"

/**
 * SECTION: cdos-doc-manager
 * @Title: CdosDocManager
 * @short_description: DocManager围绕着DocSystem，主要用于公开DocInfo对象
 * @include: cdos-doc-manager.h
 *
 * CdosDocManager 用来管理文档信息对象
 */

struct _CdosDocManagerPrivate
{
        CdosDocSystem                        *docSystem;
        GnomeDesktopThumbnailFactory         *thumbnail_factory;
        GList                                *infosByTimestamp;
        GHashTable                           *infosByUri;
};

enum
{
        CHANGED,

  LAST_SIGNAL
};

static guint cdos_doc_manager_signals[LAST_SIGNAL] =  { 0, };

G_DEFINE_TYPE(CdosDocManager, cdos_doc_manager, G_TYPE_OBJECT);

#define CDOS_DOC_MANAGER_GET_PRIVATE(obj)    \
        (G_TYPE_INSTANCE_GET_PRIVATE((obj), CDOS_TYPE_DOC_MANAGER, CdosDocManagerPrivate))

static void 
cdos_doc_manager_finalize(GObject *gobject);
static void
cdos_doc_manager_dispose(GObject *gobject);

static void
cdos_doc_manager_reload(CdosDocSystem *doc, CdosDocManager *doc_manager);

static void 
cdos_doc_manager_init(CdosDocManager *manager)
{         
        CdosDocManagerPrivate *priv;
        priv = manager->priv = CDOS_DOC_MANAGER_GET_PRIVATE(manager);

        priv->docSystem = cdos_doc_system_get_default();

        priv->thumbnail_factory = gnome_desktop_thumbnail_factory_new(GNOME_DESKTOP_THUMBNAIL_SIZE_NORMAL);
        priv->infosByTimestamp = NULL;
        priv->infosByUri = g_hash_table_new_full(g_str_hash, NULL, g_free, g_object_unref);
}

static void
cdos_doc_manager_class_init(CdosDocManagerClass *klass)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

        gobject_class->finalize = cdos_doc_manager_finalize;
        gobject_class->dispose = cdos_doc_manager_dispose;
        g_type_class_add_private(klass, sizeof(CdosDocManagerPrivate));
 
        cdos_doc_manager_signals[CHANGED] = g_signal_new (
                        "changed",
                        G_TYPE_FROM_CLASS(klass),
                        G_SIGNAL_RUN_LAST,
                        G_STRUCT_OFFSET(CdosDocManagerClass, changed),
                        NULL,
                        NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE,
                        0);
}

static void
cdos_doc_manager_dispose(GObject *gobject){
        CdosDocManagerPrivate *priv = CDOS_DOC_MANAGER(gobject)->priv;
        if (priv->docSystem)
        {
                g_clear_object(&priv->docSystem);
        }
        if (priv->thumbnail_factory)
        {
                g_clear_object(&priv->thumbnail_factory);
        }
        if (priv->infosByTimestamp)
        {
                g_list_free_full(priv->infosByTimestamp, (GDestroyNotify)g_object_unref);
        }

        if (priv->infosByUri)
        {
                g_hash_table_destroy(priv->infosByUri);
        }

        G_OBJECT_CLASS(cdos_doc_manager_parent_class)->dispose(gobject);
}

static void 
cdos_doc_manager_finalize(GObject *gobject)
{
        //CdosDocManagerPrivate *priv = CDOS_DOC_MANAGER(gobject)->priv;

        G_OBJECT_CLASS(cdos_doc_manager_parent_class)->finalize(gobject);
}

static CdosDocManager *
cdos_doc_manager_new(void)
{
        CdosDocManager *manager;
        
        manager = CDOS_DOC_MANAGER(g_object_new(CDOS_TYPE_DOC_MANAGER, NULL));
        g_signal_connect(manager->priv->docSystem, "changed", G_CALLBACK(cdos_doc_manager_reload), manager);
        cdos_doc_manager_reload(NULL, manager);

        return manager;
}

static void
cdos_doc_manager_reload(CdosDocSystem *doc, CdosDocManager *doc_manager)
{
        GSList          *l;
        GtkRecentInfo  *recentInfo;
        CdosDocInfo    *docInfo;
        g_return_if_fail(CDOS_IS_DOC_MANAGER(doc_manager));
        CdosDocManager *manager = doc_manager;

        CdosDocManagerPrivate *priv = CDOS_DOC_MANAGER_GET_PRIVATE(manager);

        GSList *docs = cdos_doc_system_get_all(priv->docSystem);

        if (priv->infosByTimestamp) 
        {
            g_list_free(priv->infosByTimestamp);
            priv->infosByTimestamp = NULL;
        }

        if (priv->infosByUri)
        {
            g_hash_table_remove_all(priv->infosByUri);
        }

        for (l = docs; l != NULL; l = l->next)
        {
                recentInfo = l->data;
                docInfo = cdos_doc_info_new(recentInfo, priv->thumbnail_factory);           
                priv->infosByTimestamp = g_list_append(priv->infosByTimestamp, docInfo);
                gchar *url = g_strdup(cdos_doc_info_get_uri(docInfo));
                g_hash_table_insert(priv->infosByUri, url, docInfo);
        }
        
        g_signal_emit_by_name(manager, "changed");
}

/**
 * cdos_doc_manager_get_time_stamp_ordered_infos
 * @manager: #CdosDocManager对象
 *
 * 获取以时间戳为顺序的文档信息
 *
 * Returns: GPtrArray信息数组
 */

GList *
cdos_doc_manager_get_time_stamp_ordered_infos(CdosDocManager *manager)
{
        g_return_val_if_fail(CDOS_IS_DOC_MANAGER(manager), NULL);
        return manager->priv->infosByTimestamp;
}

/**
 * cdos_doc_manager_get_infos_by_uri
 * @manager: #CdosDocManager对象
 *
 * 获取#CdosDocManager的私有属性infosByUri
 *
 * Returns: 以哈希表的形式返回infosByUri
 */

GHashTable *
cdos_doc_manager_get_infos_by_uri(CdosDocManager *manager)
{
        g_return_val_if_fail(CDOS_IS_DOC_MANAGER(manager), NULL);
        return manager->priv->infosByUri;
}

/**
 * cdos_doc_manager_lookup_by_uri
 * @manager: #CdosDocManager对象
 * @uri: 需要查找的uri
 *
 * 根据uri查找对应的#CdosDocInfo
 *
 * Returns: 查找到的文档信息对象
 */

CdosDocInfo *
cdos_doc_manager_lookup_by_uri(CdosDocManager *manager,
                               const gchar * uri)
{
        g_return_val_if_fail(CDOS_IS_DOC_MANAGER(manager), NULL);
        g_return_val_if_fail(uri != NULL, NULL);

        return g_hash_table_lookup(manager->priv->infosByUri, uri);        
}

/**
 * cdos_doc_manager_get_default
 *
 * 获取#CdosDocManager对象
 *
 * Returns: #CdosDocManager对象
 */

CdosDocManager *
cdos_doc_manager_get_default(void)
{
        static CdosDocManager *manager = NULL;

        if (manager == NULL) {
            manager = cdos_doc_manager_new();
        }

        return manager;   
}
