
#include "cdos-expo-thumbnails-box.h"
#include "cdos-expo-workspace-thumbnail.h"
#include "cdos-monkey-plugin.h"
#include "cdos-window-manager.h"
#include "cdos-layout-manager.h"
#include "cdos-window-overlay.h"
#include "cdos-base-util.h"
#include "cdos-workspace.h"

#include "cdos-global.h"

#include <meta/window.h>
#include <meta/workspace.h>

#include <math.h>


/**
 * SECTION:cdos-expo-thumbnails-box
 * @Title: CdosExpoThumbnailsBox
 * @Short_description: 存放工作区缩略图的容器
 * 
 * CdosExpoThumbnailsBox 用于工作区的展示方式，目前工作区默认按横排显示，
 * 通过修改 org.cdos.workspace-expo-view-as-grid 的值，可以网格显示工作区缩略图
 * 成员 thumnbnails 保存系统中所有的 CdosExpoWorkspaceThumbnail 对象，默认有四个工作区，也就是四个工作区缩略图
 */
typedef struct {

	gpointer param1;
	gpointer param2;

} Params_expo;


enum {
	ALLOCATED,
	SET_OVERVIEW_MODE,
    STICKTY_DETECTED,
    LAST_SIGNAL
};

static gint cdos_expo_thumbnails_box_signals [LAST_SIGNAL] = { 0 };

enum {
    PROP_0,
    PROP_SCALE
};

// The maximum size of a thumbnail is 1/8 the width and height of the screen
#define MAX_THUMBNAIL_SCALE 0.9

#define POINTER_LEAVE_MILLISECONDS_GRACE 500
#define POINTER_ENTER_MILLISECONDS_GRACE 150

#define RESCALE_ANIMATION_TIME 200
// 全局变量 persistent throughout session
static guint forceOverviewMode = FALSE; //0


enum {	//DND.DragMotionResult

    DRAG_MOTION_RESULT_NO_DROP,
    DRAG_MOTION_RESULT_COPY_DROP,
    DRAG_MOTION_RESULT_MOVE_DROP,
    DRAG_MOTION_RESULT_CONTINUE
};


G_DEFINE_TYPE  ( CdosExpoThumbnailsBox, cdos_expo_thumbnails_box, CDOS_TYPE_GENERIC_CONTAINER);


static gboolean updateStates (CdosExpoThumbnailsBox *self);

static void getNumberOfColumnsAndRows (CdosExpoThumbnailsBox *self, guint nWorkspaces, guint *nRows, guint *nColumns) {

	GSettings *global_settings = cdos_global_get_settings (cdos_global_get ());
	gboolean asGrid;
	gfloat divisor;
	gfloat screenRatio;
	gfloat boxRatio;

	asGrid = g_settings_get_boolean (global_settings, "workspace-expo-view-as-grid");

	*nColumns = asGrid ? ceil (sqrt (nWorkspaces)) : nWorkspaces;
	*nRows = ceil ((float)nWorkspaces / *nColumns);

    // in case of a very wide screen, we can try and optimize the screen 
    // utilization by switching the columns and rows, but only if there's a
    // big difference. If the user doesn't want a grid we are even more conservative.
    gint sw, sh;
    cdos_global_get_screen_size (global, &sw, &sh);
    divisor = 1.25;
    screenRatio = (gfloat)sw / (gfloat)sh;

	if (self->box)
		boxRatio = (self->box->x2 - self->box->x1) / (self->box->y2 - self->box->y1);
	else
		boxRatio = 1.6;

    if (nWorkspaces <= floor (screenRatio)) {
		*nRows = 1;
		*nColumns = nWorkspaces;
    } 
	else if (!asGrid || (screenRatio / divisor) <= boxRatio) {
        return;
    } 
	else 
	{	//将行与列的结果相反传回
		guint tmp;
		tmp = *nColumns;
		*nColumns = *nRows;
		*nRows = tmp;
    }
}

static void
getPreferredHeight (CdosExpoThumbnailsBox *self, gfloat *forWidth, CdosGenericContainerAllocation *alloc)
{
	gdouble spacing;
	MetaScreen *global_screen = cdos_global_get_screen (cdos_global_get());
	guint nWorkspaces;
	gfloat totalSpacing;
	gfloat avail;
	gfloat scale;
	CdosLayoutManager *layout_manager = cdos_monkey_plugin_get_layout_manager (monkey_plugin_get ());
	MetaRectangle *metaRect;
	guint nRows, nColumns;
	gfloat height;

        // See comment about this.background in _init()
        StThemeNode *themeNode = st_widget_get_theme_node (ST_WIDGET (self->background));

        st_theme_node_adjust_for_width (themeNode, forWidth);

        // Note that for getPreferredWidth/Height we cheat a bit and skip propagating
        // the size request to our children because we know how big they are and know
        // that the actors aren't depending on the virtual functions being called.

        if (g_list_length (self->thumbnails) == 0)
            	return;

	spacing = st_theme_node_get_length (st_widget_get_theme_node (ST_WIDGET (self)), "spacing" );

	nWorkspaces = meta_screen_get_n_workspaces (global_screen);
	totalSpacing = (nWorkspaces - 1) * spacing;

	metaRect = cdos_layout_manager_get_primary_monitor (layout_manager);

	avail = metaRect->width - totalSpacing;

	getNumberOfColumnsAndRows (self, nWorkspaces, &nRows, &nColumns);

        scale = (avail / nColumns) / self->porthole->width;

        height = round (self->porthole->height * scale);

        //[alloc.min_size, alloc.natural_size] =
        //    themeNode.adjust_preferred_height(400, Main.layoutManager.primaryMonitor.height);
	//st_theme_node_adjust_preferred_height (themeNode, alloc->min_size, alloc->natural_size);
}

static void
getPreferredWidth (CdosExpoThumbnailsBox *self, gfloat *forHeight, CdosGenericContainerAllocation *alloc)
{
        gdouble spacing;
        MetaScreen *global_screen = cdos_global_get_screen (cdos_global_get());
        guint nWorkspaces;
        gfloat totalSpacing;
        gfloat avail;
        gfloat scale;
        CdosLayoutManager *layout_manager = cdos_monkey_plugin_get_layout_manager (monkey_plugin_get ());
        MetaRectangle *metaRect;
        guint nRows, nColumns;
	gfloat maxWidth;
	gfloat width;

        // See comment about this.background in _init()
        StThemeNode *themeNode = st_widget_get_theme_node (ST_WIDGET (self->background));

        if (g_list_length (self->thumbnails) == 0)
                return;

        // We don't animate our preferred width, which is always reported according
        // to the actual number of current workspaces, we just animate within that
	spacing = st_theme_node_get_length (st_widget_get_theme_node (ST_WIDGET (self)), "spacing" );

        nWorkspaces = meta_screen_get_n_workspaces (global_screen);
        totalSpacing = (nWorkspaces - 1) * spacing;

        metaRect = cdos_layout_manager_get_primary_monitor (layout_manager);

        avail = metaRect->width - totalSpacing;

        getNumberOfColumnsAndRows (self, nWorkspaces, &nRows, &nColumns);

        scale = (avail / nColumns) / self->porthole->width;

        width = round (self->porthole->height * scale);

        maxWidth = (width) * nWorkspaces;

        //[alloc.min_size, alloc.natural_size] =
        //    themeNode.adjust_preferred_width(totalSpacing, Main.layoutManager.primaryMonitor.width);
	//st_theme_node_adjust_preferred_height (themeNode, alloc->min_size, alloc->natural_size);
}


static void setThumbnailState (CdosExpoThumbnailsBox *self, CdosExpoWorkspaceThumbnail *thumbnail, guint state) {

        self->stateCounts [thumbnail->state]--;
        thumbnail->state = state;
        self->stateCounts [thumbnail->state]++;
}

static void iterateStateThumbnails (CdosExpoThumbnailsBox *self, gint state, 
					void (*callback)(CdosExpoThumbnailsBox *, CdosExpoWorkspaceThumbnail *)) {

        if (self->stateCounts[state] == 0)
            	return;

	if ( !self->thumbnails ){
		return;
	}
        GList *th = self->thumbnails;
        //for (i = 0; i < g_list_length (self->thumbnails); i++) {
        while(th != NULL){
            	
            //CdosExpoWorkspaceThumbnail *thumbnail_index = g_list_nth_data (self->thumbnails, i);
            CdosExpoWorkspaceThumbnail *thumbnail_index = th->data;

            if (thumbnail_index->state == state)
                callback (self, thumbnail_index);

            th = th->next;
        }
}

static void queueUpdateStates (CdosExpoThumbnailsBox *self) {
    if (self->stateUpdateQueued)
        return;
    meta_later_add (META_LATER_BEFORE_REDRAW, (GSourceFunc) updateStates, self, NULL);

    self->stateUpdateQueued = TRUE;
}

static void
tweenScaleCb (ClutterTimeline *tml, CdosExpoThumbnailsBox *self)
{
    clutter_actor_remove_transition ((ClutterActor *)self, "tween-scale");
    queueUpdateStates (self);
}

static void tweenScale (CdosExpoThumbnailsBox *self)
{
    cdos_transition_to ((ClutterActor *)self,
                        CLUTTER_EASE_OUT_QUAD,
                        RESCALE_ANIMATION_TIME,
                        0,
                        (TransCallback)tweenScaleCb, self,
                        "tween-scale",
                        "scale", TRANS_PROP_DOUBLE, self->targetScale,
                        NULL);
}

static void slide_out_cb (CdosExpoThumbnailsBox *self, CdosExpoWorkspaceThumbnail *thumbnail) {

    clutter_actor_hide (CLUTTER_ACTOR (thumbnail->title));

    setThumbnailState (self, thumbnail, THUMBNAIL_STATE_ANIMATING_OUT);

}

static void start_scaling_cb (CdosExpoThumbnailsBox *self, CdosExpoWorkspaceThumbnail *thumbnail) {

	//设定忽略actor的painting,是因為gjs缺少override _paint vfunc
	cdos_generic_container_set_skip_paint ((CdosGenericContainer *)self, (ClutterActor *)thumbnail->actor, TRUE);
		
        setThumbnailState (self, thumbnail, THUMBNAIL_STATE_COLLAPSING);

		//进行动画
                /*Tweener.addTween(thumbnail,
                                 { time: RESCALE_ANIMATION_TIME,
                                   transition: 'easeOutQuad',
                                   onComplete: function() {
                                       this.stateCounts[thumbnail.state]--;
                                       thumbnail.state = ThumbnailState.DESTROYED;

                                       let index = this.thumbnails.indexOf(thumbnail);
                                       this.thumbnails.splice(index, 1);
                                       thumbnail.destroy();

                                       if (index < this.kbThumbnailIndex ||
                                           (index === this.kbThumbnailIndex &&
                                               index === this.thumbnails.length))
                                       {
                                           --this.kbThumbnailIndex;
                                       }

                                       this.queueUpdateStates();
                                   },
                                   onCompleteScope: this
                                 });
		*/
}


static void slide_new_cb (CdosExpoThumbnailsBox *self, CdosExpoWorkspaceThumbnail *thumbnail) {

	setThumbnailState (self, thumbnail, THUMBNAIL_STATE_ANIMATING_IN);

	//进行动画
        /*Tweener.addTween(thumbnail,
                                 { slidePosition: 0,
                                   time: SLIDE_ANIMATION_TIME,
                                   transition: 'easeOutQuad',
                                   onComplete: function() {
                                       this.setThumbnailState(thumbnail, ThumbnailState.NORMAL);
                                   },
                                   onCompleteScope: this
                                 });
	*/  
}


static void refresh_cb (CdosExpoThumbnailsBox *self, CdosExpoWorkspaceThumbnail *thumbnail) {

	//调用CdosExpoWorkspaceThumbnail中的refresh方法
	cdos_expo_workspace_thumbnail_refresh (thumbnail);
}

static gboolean isShowingModalDialog (CdosExpoThumbnailsBox *self) {

	// the normal value is 1 while Expo is active
        //return Main.modalCount > 1; 参考cdos-start中的 modal_count
	return TRUE;
}

static gboolean updateStates (CdosExpoThumbnailsBox *self) {

    self->stateUpdateQueued = FALSE;

    // Then slide out any thumbnails that have been destroyed
    iterateStateThumbnails (self, THUMBNAIL_STATE_REMOVING, slide_out_cb);

    // As long as things are sliding out, don't proceed
    if (self->stateCounts [THUMBNAIL_STATE_ANIMATING_OUT] > 0)
        return FALSE;

    // Once that's complete, we can start scaling to the new size and collapse any removed thumbnails
    iterateStateThumbnails (self, THUMBNAIL_STATE_ANIMATED_OUT, start_scaling_cb);

    if (self->pendingScaleUpdate) {
        tweenScale (self);
        self->pendingScaleUpdate = FALSE;
    }

    // Wait until that's done
    if (self->scale != self->targetScale || self->stateCounts [THUMBNAIL_STATE_COLLAPSING] > 0)
        return FALSE;

    // And then slide in any new thumbnails
    iterateStateThumbnails (self, THUMBNAIL_STATE_NEW, slide_new_cb);

    iterateStateThumbnails (self, THUMBNAIL_STATE_NORMAL, refresh_cb);

    CdosExpoWorkspaceThumbnail *thumbnail = g_list_nth_data (self->thumbnails, self->kbThumbnailIndex);        

    //调用thumbnail中的showKeyboardSelectedState方法;
    cdos_expo_workspace_thumbnail_show_keyboard_selected_state (thumbnail, TRUE);

    if (!isShowingModalDialog (self)) {
        // we may inadvertently have lost keyboard focus during the reshuffling
        ClutterStage *global_stage = cdos_global_get_stage (cdos_global_get ());
        clutter_stage_set_key_focus (global_stage, CLUTTER_ACTOR (self));
    }
    return FALSE;

   cdos_debug ();
	meta_later_add (META_LATER_BEFORE_REDRAW, (GSourceFunc) updateStates, self, NULL);

   self->stateUpdateQueued = TRUE;
}

static void
allocate (CdosExpoThumbnailsBox *self, ClutterActorBox *box, ClutterAllocationFlags flags)
{
    ClutterActor *actor = (ClutterActor *)self;

    self->box = box;

    gboolean rtl = (st_widget_get_default_direction () == ST_TEXT_DIRECTION_RTL);

    if (g_list_length (self->thumbnails) == 0) // not visible
        return;

    gint portholeWidth = self->porthole->width;
    gint portholeHeight = self->porthole->height;

    gdouble spacing = st_theme_node_get_length (st_widget_get_theme_node (ST_WIDGET (self)), "spacing" );

    // We must find out every setting that may affect the height of 
    // the workspace title:

    //let firstThumbnailTitleThemeNode = this.thumbnails[0].title.get_theme_node();
    CdosExpoWorkspaceThumbnail *thumbmail_0 = g_list_nth_data (self->thumbnails, 0);
    StThemeNode *firstThumbnailTitleThemeNode = st_widget_get_theme_node (ST_WIDGET (thumbmail_0->title));

    gdouble thTitleHeight = st_theme_node_get_length (firstThumbnailTitleThemeNode, "height");
    gdouble thTitleTopPadding = st_theme_node_get_padding (firstThumbnailTitleThemeNode, ST_SIDE_TOP);
    gdouble thTitleBottomPadding = st_theme_node_get_padding (firstThumbnailTitleThemeNode, ST_SIDE_BOTTOM);
    gdouble thTitleMargin = thTitleBottomPadding;
    gdouble thTitleBorderHeight = st_theme_node_get_border_width (firstThumbnailTitleThemeNode, ST_SIDE_BOTTOM) * 2;
    gdouble extraHeight = thTitleHeight + thTitleTopPadding + thTitleBottomPadding + thTitleMargin + thTitleBorderHeight;

    // Compute the scale we'll need once everything is updated
    guint nWorkspaces = g_list_length (self->thumbnails);

    guint nColumns, nRows;
    getNumberOfColumnsAndRows (self, nWorkspaces, &nRows, &nColumns);

    gfloat totalSpacingX = (nColumns - 1) * spacing;
    gfloat availX = (box->x2 - box->x1) - totalSpacingX - (spacing * 2) ;
    gfloat availY = (box->y2 - box->y1) - 2 * spacing - nRows * extraHeight - (nRows - 1) * thTitleMargin;
    gfloat screen = (box->x2 - box->x1);

    gfloat newScaleX = (gfloat)(availX / nColumns) / portholeWidth;
    gfloat newScaleY = (gfloat)(availY / nRows) / portholeHeight;
    gfloat newScale = fmin (newScaleX, newScaleY); //MAX_THUMBNAIL_SCALE 没用到?
    //主要為了算出newScale

    if (newScale != self->targetScale) {

        if (self->targetScale > 0) {
            // We don't do the tween immediately because we need to observe the ordering
            // in queueUpdateStates - if workspaces have been removed we need to slide them
            // out as the first thing.
            self->targetScale = newScale;
            self->pendingScaleUpdate = TRUE;
        } 
        else 
        {
            self->targetScale = newScale;
            cdos_expo_thumbnails_box_set_scale(self, newScale);

        }
        queueUpdateStates (self);
    }

    gfloat thumbnailHeight = round (portholeHeight * self->scale);
    gfloat thumbnailWidth = round (portholeWidth * self->scale);

    // The background is horizontally restricted to correspond to the current thumbnail size
    // but otherwise covers the entire allocation
    ClutterActorBox *childBox = clutter_actor_box_new ( 	box->x1, 
            box->y1,
            box->x2,
            box->y2 + thTitleHeight); //前面算出来的
    //设置background
    clutter_actor_allocate (CLUTTER_ACTOR (self->background), childBox, flags);


    gfloat x = 0.f;
    gfloat y = spacing + floor ((availY - nRows * thumbnailHeight) / 2);
    guint th_length = g_list_length (self->thumbnails);
    int i;
    //对每一个CdosExpoWorkspaceThumbnail都做处理
    for (i = 0; i < th_length; i++) {
        guint column = i % nColumns;
        guint row = floor (i / nColumns);
        guint cItemsInRow = min (th_length - (row * nColumns), nColumns);

        //x = column > 0 ? x : calcPaddingX(cItemsInRow)
        if (column <= 0) //calcPaddingX(cItemsInRow)
        {
            gfloat neededX = (thumbnailWidth * cItemsInRow) + totalSpacingX + (spacing * 2);
            gfloat extraSpaceX = (box->x2 - box->x1) - neededX;
            x = spacing + extraSpaceX/2.0;
        }

        guint rowMultiplier = row + 1;

        CdosExpoWorkspaceThumbnail *thumbnail = g_list_nth_data (self->thumbnails, i);

        // We might end up with thumbnailHeight being something like 99.33
        // pixels. To make this work and not end up with a gap at the bottom,
        // we need some thumbnails to be 99 pixels and some 100 pixels height;
        // we compute an actual scale separately for each thumbnail.
        gfloat x1 = round (x + (thumbnailWidth * thumbnail->slidePosition / 2));
        gfloat x2 = round (x + thumbnailWidth);
        gfloat y1 = y;
        gfloat y2 = y1 + thumbnailHeight;

        // Allocating a scaled actor is funny - x1/y1 correspond to the origin
        // of the actor, but x2/y2 are increased by the *unscaled* size.
        childBox->x1 = x1;
        childBox->x2 = x1 + portholeWidth;
        childBox->y1 = y1;
        childBox->y2 = y1 + portholeHeight;

        gfloat scale = self->scale * (1 - thumbnail->slidePosition);
        clutter_actor_set_scale (CLUTTER_ACTOR (thumbnail->actor), scale, scale);

        //设置thumbnail
        clutter_actor_allocate (CLUTTER_ACTOR (thumbnail->actor), childBox, flags);

        StThemeNode *framethemeNode = st_widget_get_theme_node (ST_WIDGET (thumbnail->frame));
        gfloat borderWidth = st_theme_node_get_border_width (framethemeNode, ST_SIDE_BOTTOM);

        childBox->x1 = x1 - borderWidth;
        childBox->x2 = x2 + borderWidth;
        childBox->y1 = y1 - borderWidth;
        childBox->y2 = y2 + borderWidth;

        gfloat scale_f = 1 - thumbnail->slidePosition;
        clutter_actor_set_scale (CLUTTER_ACTOR (thumbnail->frame), scale_f, scale_f);

        //设置frame大小
        clutter_actor_allocate (CLUTTER_ACTOR (thumbnail->frame), childBox, flags);

        gfloat thumbnailx = round (x + (thumbnailWidth * thumbnail->slidePosition / 2));
        childBox->x1 = fmax (thumbnailx, thumbnailx + round (thumbnailWidth/2) 
                - round (clutter_actor_get_width (CLUTTER_ACTOR (thumbnail->title)) / 2));
        childBox->x2 = fmin (thumbnailx + thumbnailWidth, childBox->x1 + clutter_actor_get_width (CLUTTER_ACTOR (thumbnail->title)));
        childBox->y1 = y + thumbnailHeight + thTitleMargin;
        childBox->y2 = childBox->y1 + clutter_actor_get_height (CLUTTER_ACTOR (thumbnail->title));

        //设置title大小
        clutter_actor_allocate (CLUTTER_ACTOR (thumbnail->title), childBox, flags);

        x += thumbnailWidth + spacing;
        y += (i + 1) % nColumns > 0 ? 0 : thumbnailHeight + extraHeight + thTitleMargin;
    }

    x = 0;
    y = 0;

    //StThemeNode *buttonTN = st_widget_get_theme_node (ST_WIDGET (self->button));

    gfloat buttonWidth = 0.f;//st_theme_node_get_length (buttonTN, "width");
    gfloat buttonHeight = 0.f;//st_theme_node_get_length (buttonTN, "height");
    gfloat buttonOverlap = 0.f;//st_theme_node_get_length (buttonTN, "-cinnamon-close-overlap");

    if (self->lastHovered && self->lastHovered->actor != NULL && !self->lastHovered->doomed) {
        //x = this.lastHovered.actor.allocation.x1 + 
        //	((this.lastHovered.actor.allocation.x2 - this.lastHovered.actor.allocation.x1) * this.lastHovered.actor.get_scale()[0])
        // 	- buttonOverlap;
        ClutterActorBox tmp_box;
        clutter_actor_get_allocation_box (CLUTTER_ACTOR (self->lastHovered->actor), &tmp_box);
        gdouble scale_x, scale_y;
        clutter_actor_get_scale (CLUTTER_ACTOR (self->lastHovered->actor), &scale_x, &scale_y);
        x = tmp_box.x1 + ((tmp_box.x2 - tmp_box.x1) * scale_x) - buttonOverlap;

        //y = this.lastHovered.actor.allocation.y1 - (buttonHeight - buttonOverlap);
        y = tmp_box.y1 - (buttonHeight - buttonOverlap);
    } 
    else 
    {
        //clutter_actor_hide (CLUTTER_ACTOR (self->button));
    }
    /*
    childBox->x1 = x;
    childBox->x2 = childBox->x1 + buttonWidth;
    childBox->y1 = y;
    childBox->y2 = childBox->y1 + buttonHeight;

    //设置button大小
    clutter_actor_allocate (CLUTTER_ACTOR (self->button), childBox, flags);
    */
    //完成，发出allocted信号
    g_signal_emit_by_name (self, "allocated");
    clutter_actor_box_free(childBox);
}


//尚未完成 source类型未知
static guint background_handleDragOver (CdosExpoThumbnailsBox *self, void *source, ClutterActor *actor, gint x, gint y, guint time) {
    /*
       MetaWindow *metaWindow = source->metaWindow;

       if (metaWindow != NULL && !meta_window_is_on_all_workspaces (metaWindow)) 
       return DRAG_MOTION_RESULT_MOVE_DROP;
       else
       return DRAG_MOTION_RESULT_CONTINUE;
*/
    return 0;
}

//尚未完成 source类型未知
static gboolean background_acceptDrop (CdosExpoThumbnailsBox *self, void *source, ClutterActor *actor, gint x, gint y, guint time) {
/*
	MetaWindow *metaWindow = source->metaWindow;
        
	//这句不懂
	if (background_handleDragOver (self, source, actor, x, y, time) == DRAG_MOTION_RESULT_MOVE_DROP) {
                CdosDraggable *draggable = source._draggable;

                ClutterActor *parent = clutter_actor_get_parent (actor);
		clutter_actor_remove_child (parent, actor);

                draggable._dragOrigParent.add_actor(actor);
		clutter_actor_add_child ();

                clutter_actor_set_opacity (actor, draggable._dragOrigOpacity);

                meta_window_stick (metaWindow);

                return TRUE;
	}
*/	return FALSE;
}

static gboolean onEnterEvent (GObject *gobject, gpointer user_data) {

	CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (user_data);

	clutter_actor_show (CLUTTER_ACTOR (self->button));
    return FALSE;
} 

static gboolean onLeaveEvent (GObject *gobject, gpointer user_data) {
 
	CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (user_data);

	clutter_actor_hide (CLUTTER_ACTOR (self->button));
    return FALSE;
}

static void onClicked (GObject *gobject, gpointer user_data) {

	CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (user_data);

	//调用cdosExpoWorkspaceThumbnial中的remove
	cdos_expo_workspace_thumbnail_remove (self->lastHovered);

	clutter_actor_hide (CLUTTER_ACTOR (self->button));
}

static void onScrollEvent (GObject *gobject, ClutterEvent *event, gpointer user_data) {

        switch (clutter_event_get_scroll_direction (event)) {

            case CLUTTER_SCROLL_UP:
            	cdos_wm_action_move_workspace_up ();
            	break;

            case CLUTTER_SCROLL_DOWN:
            	cdos_wm_action_move_workspace_up ();
            	break;
            default:
                break;
        }
}

static gboolean set_overview_mode_cb (gpointer user_data) {

	CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (user_data);
	//CdosExpoWorkspaceThumbnail *thumbnail;
	
	g_signal_emit_by_name (self,"set-overview-mode",(forceOverviewMode == 1)); //gboolean

	//thumbnail = g_list_nth_data (self->thumbnails, self->kbThumbnailIndex);

	//调用CdosExpoWorkspaceThumbnail的函数
	//cdos_expo_workspace_thumbnail_show_keyboard_selected_state (thumbnail, TRUE);

	return FALSE;
} 

static void onAllocated (GObject *gobject, gpointer user_data)
{
    CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (user_data);
    if (self->allocId > 0){
        g_signal_handler_disconnect (self, self->allocId);
        self->allocId = 0L;
    }
    g_timeout_add (0, (GSourceFunc) set_overview_mode_cb, self);
    //set_overview_mode_cb (self);
}

static void toggleGlobalOverviewMode (CdosExpoThumbnailsBox *self) {

	forceOverviewMode = (forceOverviewMode + 1) % 2;

	g_signal_emit_by_name (self, "set-overview-mode", (forceOverviewMode == 1));
}

static gboolean onButtonReleaseEvent (GObject *gobject, ClutterEvent *event, gpointer user_data) {

	CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (user_data);

	if (cdos_base_util_get_event_state (event) & CLUTTER_BUTTON2_MASK) {
                toggleGlobalOverviewMode (self);
        }
    return FALSE;
}


// 固定方法
static void cdos_expo_thumbnails_box_dispose (GObject *object) {
    CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (object);

    if(self->thumbnails){
        GList *iter = self->thumbnails;
        while(iter != NULL){
            CdosExpoWorkspaceThumbnail *thumbnail = iter->data;
            g_clear_object(&thumbnail);
            iter = iter->next;
        }
    }
    g_clear_object(&self->background);
    //g_clear_object(&self->button);

    G_OBJECT_CLASS (cdos_expo_thumbnails_box_parent_class)->dispose (object);
}

static void cdos_expo_thumbnails_box_finalize (GObject *object) {
    CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (object);

    g_free(self->porthole);
    g_list_free (self->thumbnails);

    G_OBJECT_CLASS (cdos_expo_thumbnails_box_parent_class)->finalize (object);
}

static void
cdos_expo_thumbnails_box_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec  *pspec)
{
    CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (object);
    switch (prop_id)
    {
        case PROP_SCALE:
            cdos_expo_thumbnails_box_set_scale (self, g_value_get_double (value));
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }
}

static void
cdos_expo_thumbnails_box_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec  *pspec)
{
    CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (object);
    switch (prop_id)
    {
        case PROP_SCALE:
            g_value_set_double (value, self->scale);
            break;
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }
}

static void
cdos_expo_thumbnails_box_allocate (ClutterActor           *self,
                                  const ClutterActorBox  *box,
                                  ClutterAllocationFlags  flags)
{
  StThemeNode *theme_node;
  ClutterActorBox content_box;

  CLUTTER_ACTOR_CLASS (cdos_expo_thumbnails_box_parent_class)->allocate (self, box, flags);

  theme_node = st_widget_get_theme_node (ST_WIDGET (self));
  st_theme_node_get_content_box (theme_node, box, &content_box);

  /*
  g_signal_emit (G_OBJECT (self), cdos_generic_container_signals[ALLOCATE], 0,
                 &content_box, (guint)flags);
  */
  allocate ((CdosExpoThumbnailsBox *)self, &content_box, flags);
}

static void
cdos_expo_thumbnails_box_get_preferred_width (ClutterActor *actor,
                                             gfloat        for_height,
                                             gfloat       *min_width_p,
                                             gfloat       *natural_width_p)
{
  CdosGenericContainerAllocation *alloc = g_slice_new0 (CdosGenericContainerAllocation);
  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));

  st_theme_node_adjust_for_height (theme_node, &for_height);

  /*
  g_signal_emit (G_OBJECT (actor), cdos_generic_container_signals[GET_PREFERRED_WIDTH], 0,
                 for_height, alloc);
  */
  getPreferredWidth ((CdosExpoThumbnailsBox *)actor, &for_height, alloc);
  if (min_width_p)
    *min_width_p = alloc->min_size;
  if (natural_width_p)
    *natural_width_p = alloc->natural_size;
  g_slice_free (CdosGenericContainerAllocation, alloc);

  st_theme_node_adjust_preferred_width (theme_node, min_width_p, natural_width_p);
}

static void
cdos_expo_thumbnails_box_get_preferred_height (ClutterActor *actor,
                                              gfloat        for_width,
                                              gfloat       *min_height_p,
                                              gfloat       *natural_height_p)
{
  CdosGenericContainerAllocation *alloc = g_slice_new0 (CdosGenericContainerAllocation);
  StThemeNode *theme_node = st_widget_get_theme_node (ST_WIDGET (actor));

  st_theme_node_adjust_for_width (theme_node, &for_width);

  /*
  g_signal_emit (G_OBJECT (actor), cdos_generic_container_signals[GET_PREFERRED_HEIGHT], 0,
                 for_width, alloc);
  */
  getPreferredHeight ((CdosExpoThumbnailsBox *)actor, &for_width, alloc);

  if (min_height_p)
    *min_height_p = alloc->min_size;
  if (natural_height_p)
    *natural_height_p = alloc->natural_size;
  g_slice_free (CdosGenericContainerAllocation, alloc);

  st_theme_node_adjust_preferred_height (theme_node, min_height_p, natural_height_p);
}

static void
cdos_expo_thumbnails_box_class_init (CdosExpoThumbnailsBoxClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = cdos_expo_thumbnails_box_finalize;
    gobject_class->dispose = cdos_expo_thumbnails_box_dispose;

    gobject_class->set_property = cdos_expo_thumbnails_box_set_property;
    gobject_class->get_property = cdos_expo_thumbnails_box_get_property;

    ClutterActorClass *parent_class = CLUTTER_ACTOR_CLASS (klass);

    parent_class->get_preferred_width = cdos_expo_thumbnails_box_get_preferred_width;
    parent_class->get_preferred_height = cdos_expo_thumbnails_box_get_preferred_height;
    parent_class->allocate = cdos_expo_thumbnails_box_allocate;

    cdos_expo_thumbnails_box_signals [ALLOCATED] =
        g_signal_new ("allocated",
                G_TYPE_FROM_CLASS (klass),
                G_SIGNAL_RUN_LAST,
                0,
                NULL, NULL,
                g_cclosure_marshal_VOID__VOID,
                G_TYPE_NONE, 0 );

    cdos_expo_thumbnails_box_signals [SET_OVERVIEW_MODE] =
        g_signal_new ("set-overview-mode",
                G_TYPE_FROM_CLASS (klass),
                G_SIGNAL_RUN_LAST,
                0,
                NULL, NULL,
                g_cclosure_marshal_VOID__BOOLEAN,
                G_TYPE_NONE, 1 ,
                G_TYPE_BOOLEAN );

     cdos_expo_thumbnails_box_signals [STICKTY_DETECTED] =
         g_signal_new ("sticky-detected",
                 G_TYPE_FROM_CLASS (klass),
                 G_SIGNAL_RUN_LAST,
                 0,
                 NULL, NULL,
                 g_cclosure_marshal_VOID__OBJECT,
                 G_TYPE_NONE, 1 ,
                 META_TYPE_WINDOW );

    g_object_class_install_property (gobject_class,
                                     PROP_SCALE,
                                     g_param_spec_enum ("scale",
                                                        "CdosExpoThumbnailsBox scale",
                                                        "The CdosExpoThumbnailsBox scale",
                                                        G_TYPE_DOUBLE,
                                                        1.0f,
                                                        G_PARAM_READWRITE));
}

static void
cdos_expo_thumbnails_box_init (CdosExpoThumbnailsBox *self)
{

}

static void
cdos_expo_thumbnails_box_post_init (CdosExpoThumbnailsBox *self)
{
	MetaScreen *global_screen = cdos_global_get_screen (cdos_global_get());
	int i;

	st_widget_add_style_class_name (ST_WIDGET (self), "workspace-thumbnails");	
	clutter_actor_set_reactive ((ClutterActor *)self, TRUE);
	clutter_actor_set_request_mode ((ClutterActor *)self, CLUTTER_REQUEST_WIDTH_FOR_HEIGHT);

    // When we animate the scale, we don't animate the requested size of the thumbnails, rather
    // we ask for our final size and then animate within that size. This slightly simplifies the
    // interaction with the main workspace windows (instead of constantly reallocating them
    // to a new size, they get a new size once, then use the standard window animation code
    // allocate the windows to their new positions), however it causes problems for drawing
    // the background and border wrapped around the thumbnail as we animate - we can't just pack
    // the container into a box and set style properties on the box since that box would wrap
    // around the final size not the animating size. So instead we fake the background with
    // an actor underneath the content and adjust the allocation of our children to leave space
    // for the border and padding of the background actor.

    //在container中加入StBin (background)
    self->background = ST_BIN (st_bin_new ());
    clutter_actor_set_reactive (CLUTTER_ACTOR (self->background), TRUE);
    st_container_add_actor ((ClutterContainer *)self, CLUTTER_ACTOR (self->background));

    self->background_delegate = CLUTTER_ACTOR (self->background);

    //在container中加入StButton (button)
    /*self->button = ST_BUTTON (st_button_new ());
    st_widget_set_style_class_name (ST_WIDGET (self->button), "workspace-close-button");
    st_container_add_actor ((ClutterContainer *)self, CLUTTER_ACTOR (self->button));

    g_signal_connect (self->button, "enter-event", G_CALLBACK (onEnterEvent), self);
    g_signal_connect (self->button, "leave-event", G_CALLBACK (onLeaveEvent), self);
    g_signal_connect (self->button, "clicked", G_CALLBACK (onClicked), self);


    clutter_actor_hide (CLUTTER_ACTOR (self->button));
*/
//	g_signal_connect (self->actor, "scroll-event", G_CALLBACK (onScrollEvent), self);

    self->targetScale = 0;
    self->scale = 0;
    self->pendingScaleUpdate = FALSE;
    self->stateUpdateQueued = FALSE;

    //初始化stateCount内的8种状态的值為0
    for (i = 0; i <= THUMBNAIL_STATE_DESTROYED; i++)
        self->stateCounts [i] = 0;

    self->thumbnails = NULL;

    // The "porthole" is the portion of the screen that we show in the workspaces
    self->porthole = g_new0 (ExpoThumbnailPorthole, 1);
    self->porthole->x = 0;
    self->porthole->y = 0;
    self->porthole->width = gdk_screen_width ();
    self->porthole->height = gdk_screen_height ();

    self->kbThumbnailIndex = meta_screen_get_active_workspace_index (global_screen);

    // apparently we get no direct call to show the initial
    // view, so we must force an explicit overviewMode On/Off display
    // after it has been allocated
    self->allocId = g_signal_connect (self, "allocated", G_CALLBACK (onAllocated), self); 

    g_signal_connect (self, "button-release-event", G_CALLBACK (onButtonReleaseEvent), self);

}

static void activateSelectedWorkspace (CdosExpoThumbnailsBox *self) {

	CdosExpoWorkspaceThumbnail *thumbnail = g_list_nth_data (self->thumbnails, self->kbThumbnailIndex);
	
	cdos_expo_workspace_thumbnail_activate_workspace (thumbnail);
}

static void editWorkspaceTitle (CdosExpoThumbnailsBox *self) {

	CdosExpoWorkspaceThumbnail *thumbnail = g_list_nth_data (self->thumbnails, self->kbThumbnailIndex);
	
    cdos_debug ();
	clutter_actor_grab_key_focus (CLUTTER_ACTOR (thumbnail->title));
}

static gboolean selectNextWorkspace (CdosExpoThumbnailsBox *self, guint symbol) {

    gint prevIndex = self->kbThumbnailIndex;
    guint length = g_list_length (self->thumbnails);
    gint lastIndex = length - 1;

    guint nColumns, nRows;
    getNumberOfColumnsAndRows (self, length, &nRows, &nColumns);

    //在workspace.h中 
    //let nextIndex = GridNavigator.nextIndex(this.thumbnails.length, nColumns, prevIndex, symbol);
    gint nextIndex = grid_navigator_next_index (length, nColumns, prevIndex, symbol);
    //	gint nextIndex = 0;

    cdos_debug ("next index  %d, prev index %d\n", nextIndex, prevIndex);
    if (nextIndex >= 0) {
        self->kbThumbnailIndex = nextIndex;
    }
    else 
    {
        gint index = symbol - 48 - 1; // convert '1' to index 0, etc

        if (index >= 0 && index < 10) {
            // OK
        }
        else 
        {
            index = symbol - CLUTTER_KP_1; // convert Num-pad '1' to index 0, etc

            if (index < 0 || index > 9) {
                return FALSE; // not handled
            }
        }

        if (index > lastIndex) {
            return TRUE; // handled, but out of range
        }

        self->kbThumbnailIndex = index;
        cdos_debug ("%d", index);
        activateSelectedWorkspace (self);

        //Main.wm.showWorkspaceOSD(); 可能有问题
        MetaScreen *global_screen = cdos_global_get_screen (cdos_global_get());
        show_workspace_OSD (global_screen);

        return TRUE; // handled
    }

    if (prevIndex != self->kbThumbnailIndex) {

        CdosExpoWorkspaceThumbnail *thumbnail_prev = g_list_nth_data (self->thumbnails ,prevIndex);
        if (thumbnail_prev)
            cdos_expo_workspace_thumbnail_show_keyboard_selected_state (thumbnail_prev, FALSE);

        CdosExpoWorkspaceThumbnail *thumbnail_kb = g_list_nth_data (self->thumbnails ,self->kbThumbnailIndex);
        if (thumbnail_kb)
            cdos_expo_workspace_thumbnail_show_keyboard_selected_state (thumbnail_kb, TRUE);            	
    }
    return TRUE; // handled
}

static void
activeWorkspaceChanged (CdosBaseWM *cdos_base_wm,
                        gint from,
                        gint to,
                        gint direction,
                        CdosExpoThumbnailsBox *self)
{
	MetaScreen *global_screen = cdos_global_get_screen (cdos_global_get());

	CdosExpoWorkspaceThumbnail *thumbnail = g_list_nth_data (self->thumbnails, self->kbThumbnailIndex);
	cdos_expo_workspace_thumbnail_show_keyboard_selected_state (thumbnail, FALSE);	

	self->kbThumbnailIndex = meta_screen_get_active_workspace_index (global_screen);
	thumbnail = g_list_nth_data (self->thumbnails, self->kbThumbnailIndex);
	cdos_expo_workspace_thumbnail_show_keyboard_selected_state (thumbnail, TRUE);

    MetaWorkspace *activeWorkspace = meta_screen_get_active_workspace (global_screen);

	
	CdosExpoWorkspaceThumbnail *thumbnail_index = NULL;
    GList *iter;
    for (iter = self->thumbnails; iter; iter = iter->next)
    {
        thumbnail = iter->data;

        if (thumbnail->metaWorkspace == activeWorkspace) {
            thumbnail_index = thumbnail;
            break;
        }
    }

    if (self->lastActiveWorkspace) {
        cdos_expo_workspace_thumbnail_set_active (self->lastActiveWorkspace, FALSE);
    }

    cdos_expo_workspace_thumbnail_set_active (thumbnail_index, TRUE);

    self->lastActiveWorkspace = thumbnail_index;
} 

//判断是否為内部事件   
static gboolean
isInternalEvent (CdosExpoWorkspaceThumbnail *thumbnail, ClutterActor *actor, ClutterEvent *event) {

	ClutterActor *relatedActor = clutter_event_get_related (event);

    return actor == relatedActor || clutter_actor_contains (CLUTTER_ACTOR (thumbnail->actor), relatedActor);
}

static void setOverviewTimeout (CdosExpoThumbnailsBox *self, CdosExpoWorkspaceThumbnail *thumbnail, gint timeout, GSourceFunc func) {

	if (thumbnail->overviewTimeoutId) {
        	g_source_remove (thumbnail->overviewTimeoutId);
		thumbnail->overviewTimeoutId = 0;
	}

       	if (timeout && func) {
                //传递参数到timeout func
                Params_expo *params = g_new0 (Params_expo, 1);
                params->param1 = self;
                params->param2 = thumbnail;
		
        	thumbnail->overviewTimeoutId = g_timeout_add (timeout, func, params);
	}
}

static gboolean timeout_overview_on_func_1 (Params_expo *params) {
    g_free (params);
    return FALSE;
}


static void onThumbnailDestroy (ClutterActor *actor, Params_expo *params) {

		CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (params->param1);
		CdosExpoWorkspaceThumbnail *thumbnail = CDOS_EXPO_WORKSPACE_THUMBNAIL (params->param2); 

        setOverviewTimeout (self, thumbnail, 0, (GSourceFunc) timeout_overview_on_func_1);

		st_container_remove_actor ((ClutterContainer *)self, CLUTTER_ACTOR (thumbnail->frame));
		st_container_remove_actor ((ClutterContainer *)self, actor);
		clutter_actor_destroy (CLUTTER_ACTOR (thumbnail->title));

		g_free (params);
}

static void onThumbnailDragOver (CdosExpoWorkspaceThumbnail *thumbnail, CdosExpoThumbnailsBox *self) {

        cdos_expo_workspace_thumbnail_shade (thumbnail, FALSE);

        if (self->lastHovered && self->lastHovered != thumbnail) {
                    cdos_expo_workspace_thumbnail_shade (self->lastHovered, FALSE);
        }

	self->lastHovered = thumbnail;
}

static gboolean timeout_overview_on_func (Params_expo *params) {

        CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (params->param1);
        CdosExpoWorkspaceThumbnail *thumbnail = CDOS_EXPO_WORKSPACE_THUMBNAIL (params->param2);

        if (thumbnail->hovering)
            cdos_expo_workspace_thumbnail_overview_mode_on (thumbnail);

        g_free (params);
        return FALSE;
}

static gboolean timeout_overview_off_func (Params_expo *params) {

    CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (params->param1);
    CdosExpoWorkspaceThumbnail *thumbnail = CDOS_EXPO_WORKSPACE_THUMBNAIL (params->param2);

    if (!thumbnail->hovering)
        cdos_expo_workspace_thumbnail_overview_mode_off (thumbnail, FALSE, FALSE);

    /* FIXME */
    g_free (params);
    return FALSE;
}


static gboolean showButton (CdosExpoThumbnailsBox *self) {

    MetaScreen *global_screen = cdos_global_get_screen (cdos_global_get());
        return TRUE;
    if ( meta_screen_get_n_workspaces (global_screen) <= 1)
        return FALSE;

    clutter_actor_queue_relayout (CLUTTER_ACTOR (self));

    st_container_raise ((ClutterContainer *)self, CLUTTER_ACTOR (self->button), NULL);

    clutter_actor_show (CLUTTER_ACTOR (self->button));

    return TRUE;
}

static void onThumbnailMotionEvent (ClutterActor *actor, ClutterEvent *event, Params_expo *params) {

    CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (params->param1);
    CdosExpoWorkspaceThumbnail *thumbnail = CDOS_EXPO_WORKSPACE_THUMBNAIL (params->param2);

    if (!cdos_pointer_tracker_has_moved (thumbnail->pointerTracker))
        return;

    if (!thumbnail->hovering) {
        thumbnail->hovering = TRUE;
        self->lastHovered = thumbnail;
        showButton (self);
        cdos_expo_workspace_thumbnail_highlight (thumbnail);
        setOverviewTimeout (self, thumbnail, POINTER_ENTER_MILLISECONDS_GRACE, (GSourceFunc) timeout_overview_on_func);
    }
}

static void onThumbnailLeaveEvent (ClutterActor *actor, ClutterEvent *event, Params_expo *params) {

	CdosExpoThumbnailsBox *self = CDOS_EXPO_THUMBNAILS_BOX (params->param1);
        CdosExpoWorkspaceThumbnail *thumbnail = CDOS_EXPO_WORKSPACE_THUMBNAIL (params->param2);


	if (!cdos_pointer_tracker_has_moved (thumbnail->pointerTracker))
		 return;

	if (!isShowingModalDialog(self)) 
		return;

    if (thumbnail->hovering && !isInternalEvent (thumbnail, actor, event)) {

        thumbnail->hovering = FALSE;

        //clutter_actor_hide (CLUTTER_ACTOR (self->button));

        cdos_expo_workspace_thumbnail_shade (thumbnail, FALSE);

        setOverviewTimeout (self, thumbnail, POINTER_LEAVE_MILLISECONDS_GRACE, (GSourceFunc) timeout_overview_off_func);
	}
}

static void
addThumbnails (CdosExpoThumbnailsBox *self, gint start, gint count)
{
	MetaScreen *global_screen = cdos_global_get_screen (cdos_global_get());

	gint k;
	for (k = start; k < start + count; k++) {

        cdos_debug ("workspace add thumbnails %d", count);
        MetaWorkspace *metaWorkspace = meta_screen_get_workspace_by_index (global_screen, k);
		CdosExpoWorkspaceThumbnail *thumbnail = cdos_expo_workspace_thumbnail_new (metaWorkspace, self);

        self->thumbnails = g_list_append (self->thumbnails, thumbnail);


        if (metaWorkspace == meta_screen_get_active_workspace (global_screen)) {
            self->lastActiveWorkspace = thumbnail;
            cdos_expo_workspace_thumbnail_set_active (thumbnail, TRUE);
        }


        //destroy 传递参数用
        Params_expo *params = g_new0 (Params_expo, 1);
        params->param1 = self;
        params->param2 = thumbnail; 

        g_signal_connect (thumbnail->actor, "destroy", G_CALLBACK (onThumbnailDestroy), params);

        st_container_add_actor ((ClutterContainer *)self, (ClutterActor *)thumbnail->frame);
        st_container_add_actor ((ClutterContainer *)self, (ClutterActor *)thumbnail->actor);
		st_container_add_actor ((ClutterContainer *)self, (ClutterActor *)thumbnail->title);

        //		g_signal_connect (thumbnail, "drag-over", G_CALLBACK (onThumbnailDragOver), self);

        // We want to ignore spurious events caused by animations
        // (when the contents are moving and not the pointer).
        thumbnail->pointerTracker = cdos_pointer_tracker_new ();

        g_signal_connect (thumbnail->actor, "motion-event", G_CALLBACK (onThumbnailMotionEvent), params);

		g_signal_connect (thumbnail->actor, "leave-event", G_CALLBACK (onThumbnailLeaveEvent), params);

        if (start > 0) { // not the initial fill
            thumbnail->state = THUMBNAIL_STATE_NEW;
            thumbnail->slidePosition = 1; // start slid out
        } 
        else 
        {
            thumbnail->state = THUMBNAIL_STATE_NORMAL;
        }

        self->stateCounts[thumbnail->state]++;
    }

    if (start > 0) {
        updateStates (self);
    }
    else 
    {
        queueUpdateStates (self);
    }
}

static void onWorkspaceAdded (CdosBaseWM *cdos_base_wm, gint screen, gint index, CdosExpoThumbnailsBox *self) {

    addThumbnails (self, index, 1);

}

static void onWorkspaceRemoved (CdosBaseWM *cdos_base_wm, CdosExpoThumbnailsBox *self) {

	MetaScreen *global_screen = cdos_global_get_screen (cdos_global_get());

	//clutter_actor_hide (CLUTTER_ACTOR (self->button));

        // just handling the single workspace removed is not enough
        gint removedCount = 0;

	gint i = 0;
	//for (i = 0; i < g_list_length (self->thumbnails); i++) {
    GList *list = self->thumbnails;
    while(list != NULL){
		MetaWorkspace *metaWorkspace = meta_screen_get_workspace_by_index (global_screen, i - removedCount);
		CdosExpoWorkspaceThumbnail *thumbnail = list->data;
		
		if (thumbnail->metaWorkspace != metaWorkspace) {
			++removedCount;
			
			if (thumbnail->state <= THUMBNAIL_STATE_NORMAL)
				setThumbnailState (self, thumbnail, THUMBNAIL_STATE_REMOVING);
		}
        i++;
        list = list->next;
	}

        updateStates (self);
}

//公有函数

CdosExpoThumbnailsBox* cdos_expo_thumbnails_box_new ( ) {

    CdosExpoThumbnailsBox *instance = g_object_new (CDOS_TYPE_EXPO_THUMBNAILS_BOX, NULL);

    cdos_expo_thumbnails_box_post_init (instance);

    return instance;
}

/**
 * cdos_expo_thumbnails_box_set_scale:
 * @self: #CdosExpoThumbnailsBox 对象
 * @scale: 浮点型，工作区的缩放比例，值域：(0,1];
 *
 * 设置缩略图的缩放比例，显示工作区缩略图时，先获取所有的工作区列表，根据工作区的个数，屏幕的宽高，
 * 计算出每个缩略图的缩放比例并设置
 *
 * Returns: none
 */
void cdos_expo_thumbnails_box_set_scale (CdosExpoThumbnailsBox *self, gfloat scale) {
        g_return_if_fail(CDOS_IS_EXPO_THUMBNAILS_BOX(self));
        if(scale > 0 && scale <= 1.0){
            self->scale = scale;
            clutter_actor_queue_relayout ((ClutterActor *)self);
        }
}

/**
 * cdos_expo_thumbnails_box_get_scale:
 * @self: #CdosExpoThumbnailsBox 对象
 *
 * 获取工作区缩略图的缩放比例（小于 1 的浮点数）
 *
 * Returns: 缩放比例
 */
gdouble cdos_expo_thumbnails_box_get_scale (CdosExpoThumbnailsBox *self) {
    g_return_val_if_fail(CDOS_IS_EXPO_THUMBNAILS_BOX(self),(double)1.0);
    return self->scale;
}

/**
 * cdos_expo_thumbnails_box_remove_selected_workspace:
 * @self: #CdosExpoThumbnailsBox 对象
 *
 * 将选中的工作区删除（未实现）
 *
 * Returns: none
 */
void cdos_expo_thumbnails_box_remove_selected_workspace (CdosExpoThumbnailsBox *self) {
        g_return_if_fail(CDOS_IS_EXPO_THUMBNAILS_BOX(self));
        CdosExpoWorkspaceThumbnail *thumbnail = g_list_nth_data (self->thumbnails, self->kbThumbnailIndex);

        cdos_expo_workspace_thumbnail_remove (thumbnail);
}

/**
 * cdos_expo_thumbnails_box_get_last_active_workspace:
 * @self: #CdosExpoThumbnailsBox 对象
 *
 * 获取上一个被激活的工作区
 *
 * Return value: #CdosExpoWorkspaceThumbnail 对象
 */
CdosExpoWorkspaceThumbnail* cdos_expo_thumbnails_box_get_last_active_workspace (CdosExpoThumbnailsBox *self) {
        g_return_val_if_fail(CDOS_IS_EXPO_THUMBNAILS_BOX(self),NULL);
        return self->lastActiveWorkspace;
}

/**
 * cdos_expo_thumbnails_box_handle_key_press_event:
 * @self: #CdosExpoThumbnailsBox 对象
 * @actor: 可选
 * @event: 键盘事件
 *
 * 键盘事件的回调函数， 在 #CdosExpo 中调用，它的 group 容器容纳了一系列的组件，包括 coverpane（遮罩功能的一层容器），
 * #CdosExpo 的私有成员 expo，就是我们这个 CdosExpoThumbnailsBox对象，这个对象的 actor，添加到 group 中作为子控件
 * 当用户输入 Escape、回车、空格键时，退出缩略图。上下左右键切换到其他工作区的缩略图。
 *
 * Return value: 布尔值
 */
gboolean cdos_expo_thumbnails_box_handle_key_press_event (CdosExpoThumbnailsBox *self, ClutterActor *actor, ClutterEvent *event) {

        g_return_val_if_fail(CDOS_IS_EXPO_THUMBNAILS_BOX(self),FALSE);
        ClutterModifierType modifiers = cdos_base_util_get_event_state (event);

        ClutterModifierType ctrlAltMask = CLUTTER_CONTROL_MASK | CLUTTER_MOD1_MASK;

        guint symbol = clutter_event_get_key_symbol (event);

        if (symbol == CLUTTER_Return || symbol == CLUTTER_KEY_space || symbol == CLUTTER_KP_Enter) {
                cdos_debug ();
                activateSelectedWorkspace (self);
                return TRUE;
        }

        if (symbol == CLUTTER_F2) {

                editWorkspaceTitle (self);
                return TRUE;
        }

        if ((symbol == CLUTTER_o || symbol == CLUTTER_O) && modifiers & CLUTTER_CONTROL_MASK) {

                toggleGlobalOverviewMode (self);
                return TRUE;
        }

        if (modifiers & ctrlAltMask) {
                return FALSE;
        }

        return selectNextWorkspace (self, symbol);
}

/**
 * cdos_expo_thumbnails_box_show:
 * @self: #CdosExpoThumbnailsBox 对象
 *
 * 显示工作区缩略图
 *
 */
void cdos_expo_thumbnails_box_show (CdosExpoThumbnailsBox *self) {

    g_return_if_fail(CDOS_IS_EXPO_THUMBNAILS_BOX(self));
	CdosBaseWM *cdos_base_wm = cdos_global_get_window_manager (cdos_global_get ());
	MetaScreen *global_screen = cdos_global_get_screen (cdos_global_get());
        if(self->switchWorkspaceNotifyId == 0){
            self->switchWorkspaceNotifyId = g_signal_connect (cdos_base_wm, "switch-workspace", G_CALLBACK (activeWorkspaceChanged), self);
        }
        if(self->workspaceAddedId == 0){
            self->workspaceAddedId = g_signal_connect (global_screen, "workspace-added", G_CALLBACK (onWorkspaceAdded), self);

        }
        if(self->workspaceRemovedId == 0){
            self->workspaceRemovedId = g_signal_connect (global_screen, "workspace-removed", G_CALLBACK (onWorkspaceRemoved), self);
        }
        //self->stateCounts = NULL;

	gint i;
	for (i = 0;i < 8 ; i++) {
		self->stateCounts [i] = 0;
	}

        addThumbnails (self, 0, meta_screen_get_n_workspaces (global_screen));
        
	/*ClutterActor *parent = clutter_actor_get_parent (CLUTTER_ACTOR (self->button));
	clutter_actor_set_child_above_sibling (parent, CLUTTER_ACTOR (self->button), NULL);
*/
	ClutterStage *global_stage = cdos_global_get_stage (cdos_global_get ());
    clutter_stage_set_key_focus (global_stage, (ClutterActor *)self);
}

/**
 * cdos_expo_thumbnails_box_hide:
 * @self: #CdosExpoThumbnailsBox 对象
 *
 * 隐藏工作区缩略图
 *
 */
void
cdos_expo_thumbnails_box_hide (CdosExpoThumbnailsBox *self)
{
    g_return_if_fail(CDOS_IS_EXPO_THUMBNAILS_BOX(self));
	CdosBaseWM *cdos_base_wm = cdos_global_get_window_manager (cdos_global_get ());

    if(self->switchWorkspaceNotifyId){
        g_signal_handler_disconnect (cdos_base_wm, self->switchWorkspaceNotifyId);
        self->switchWorkspaceNotifyId = 0;
    }

    MetaScreen *screen = cdos_global_get_screen (cdos_global_get ());
    if(self->workspaceAddedId){
        g_signal_handler_disconnect (screen, self->workspaceAddedId);
        self->workspaceAddedId = 0;
    }

    if(self->workspaceRemovedId){
        g_signal_handler_disconnect (screen, self->workspaceRemovedId);
        self->workspaceRemovedId = 0;
    }
    GList *iter;
    for (iter = self->thumbnails; iter; iter = iter->next)
    {
		CdosExpoWorkspaceThumbnail *thumbnail_index = iter->data;
        cdos_expo_workspace_thumbnail_destroy (thumbnail_index);
    }
    g_list_free(self->thumbnails);
    self->thumbnails = NULL;
}

/**
 * cdos_expo_thumbnails_box_get_force_overview_mode:
 * @self: #CdosExpoThumbnailsBox 对象
 *
 * 是否强制进入缩略图模式
 *
 */
guint cdos_expo_thumbnails_box_get_force_overview_mode (void) {

	return forceOverviewMode;
}


