
#include "cdos-expo-window-clone.h"
#include "cdos-window-utils.h"

#include "cdos-global.h"
#include "cdos-window-tracker.h"
#include "cdos-debug.h"
#include "cdos-base-util.h"
#include "cdos-start.h"

#include <math.h>

/**
 * SECTION:cdos-expo-window-clone
 * @Title: CdosExpoWindowClone
 * @Short_description: 工作区中打开窗口的克隆窗口
 * 
 * CdosExpoWindowClone 就是每个工作区内，窗口的克隆体。我们通过组合键进入工作区缩略图模式，
 * 除了每个工作区的缩略图，还需要显示这个工作区内所有打开的窗口的缩略图（同 CdosOverview 的工作，不过是缩小版的），
 * 这个对象就是每个窗口的克隆体，由于是克隆体，关闭这个窗口时，需要将这个克隆对象销毁，
 * 更改窗口大小时，需要刷新 clone 的大小坐标鼠标等。
 */
#define ICON_OPACITY	230 /* round(255 * 0.9) */
#define DRAGGING_WINDOW_OPACITY	204 /* round(255 * 0.8) */
#define WINDOW_DND_SIZE	256
#define ICON_SIZE 128

#define DEMANDS_ATTENTION_CLASS_NAME "window-list-item-demands-attention"


/*signal*/

enum {
	WORKSPACE_CHANGED,
	HOVERING,
	DEMANDING_ATTENTION,
	DRAG_BEGIN,
	DRAG_CANCELLED,
	DRAG_END,
	SELECTED,
	MIDDLE_BUTTON_RELEASE,
	LAST_SIGNAL
};

static guint cdos_expo_window_clone_signals [LAST_SIGNAL] = { 0 };


G_DEFINE_TYPE  ( CdosExpoWindowClone, cdos_expo_window_clone, CLUTTER_TYPE_ACTOR);


static void cdos_expo_window_clone_demand_attention (CdosExpoWindowClone *self);

static void onPositionChanged (MetaWindowActor *rw, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);
	gfloat width, height;

	self->origX = clutter_actor_get_x (CLUTTER_ACTOR (self->realWindow));
	self->origY = clutter_actor_get_y (CLUTTER_ACTOR (self->realWindow));
	
	clutter_actor_set_position ((ClutterActor *)self, self->origX, self->origY);

	width = clutter_actor_get_width (CLUTTER_ACTOR (self->realWindow));
	height = clutter_actor_get_height (CLUTTER_ACTOR (self->realWindow));
	
	clutter_actor_set_size ((ClutterActor *)self, width, height);
}

static void onSizeChanged (MetaWindowActor *rw, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);
	gfloat width, height;
	
	width = clutter_actor_get_width (CLUTTER_ACTOR (self->realWindow));
        height = clutter_actor_get_height (CLUTTER_ACTOR (self->realWindow));

	clutter_actor_set_size ((ClutterActor *)self, width, height);
}


static void onRealWindowDestroyed (GObject *gobject, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);

	self->orphaned = TRUE;
}

//function(w, oldws)  oldws类型不确定
static void onWorkspaceChanged (GObject *gobject, guint oldws, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);

	//this.emit('workspace-changed', oldws);
	g_signal_emit_by_name ( self, "workspace-changed", oldws);
}


static void disconnectWindowSignalsFunc (GObject *gobject, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);
    if(self->workspaceChangedId > 0){
        g_signal_handler_disconnect ( self->metaWindow, self->workspaceChangedId );
        self->workspaceChangedId = 0L;
    }


        if (self->orphaned) 
		return;
    if(self->sizeChangedId > 0){
        g_signal_handler_disconnect (self->realWindow, self->sizeChangedId);
        self->sizeChangedId = 0L;
    }
    if(self->positionChangedId > 0){
        g_signal_handler_disconnect (self->realWindow, self->positionChangedId);
        self->positionChangedId = 0L;
    }
    if(self->realWindowDestroyedId > 0){
        g_signal_handler_disconnect (self->realWindow, self->realWindowDestroyedId);
        self->realWindowDestroyedId = 0L;
    }
}


static gboolean onButtonPressEvent (GObject *gobject, ClutterEvent *event, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);
	ClutterActor *actor = CLUTTER_ACTOR (gobject);
	
	self->lastButtonPressActor = actor;
	self->lastButtonPressTime = clutter_event_get_time (event);
    return TRUE;
}


static gboolean onButtonReleaseEvent (GObject *gobject, ClutterEvent *event, gpointer user_data) { 

        CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);
	ClutterActor *actor = CLUTTER_ACTOR (gobject);

	if (self->lastButtonPressActor == actor && (clutter_event_get_time (event) - self->lastButtonPressTime) < 500) {
                
		cdos_expo_window_clone_on_button_release (self, event);
	}

	return TRUE;
}


static gboolean onMotionEvent (ClutterActor *actor, ClutterEvent *event) {

    CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (actor);

	if (cdos_pointer_tracker_has_moved (self->pointerTracker)) {
		//发送hovering表示现在是hover状态
		g_signal_emit_by_name ( self, "hovering", TRUE);
    }
	return FALSE;    
}

static gboolean onLeaveEvent (ClutterActor *actor, ClutterEvent *event, gpointer user_data) {
    CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (actor);
    if (cdos_pointer_tracker_has_moved (self->pointerTracker)) {
        //现在不是hovering状态
        g_signal_emit_by_name (self, "hovering", FALSE);
    }

    return FALSE;
}


static void onDestroy (GObject *gobject, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);

        self->disconnectWindowSignals (gobject, self);

        //this.actor._delegate = null;

        if (self->inDrag) {

            	self->inDrag = FALSE;
		g_signal_emit_by_name (self, "drag-end");
        }

        //this.disconnectAll(); 尚未实现
}

//void cdos_expo_window_clone_on_drag_begin (CdosExpoWindowClone *self,gboolean draggable,gint64 time)
//void cdos_expo_window_clone_on_drag_cancelled (CdosExpoWindowClone *self,gboolean draggable,gint64 time)
//void cdos_expo_window_clone_on_drag_end (CdosExpoWindowClone *self,gboolean draggable,gint64 time, gpointer *snapback)

static void onDragBegin (GObject *gobject, guint32 time, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);

        self->inDrag = TRUE;
        self->dragCancelled = FALSE;
        
   	g_signal_emit_by_name (self, "drag-begin");
}


static void onDragEnd (GObject *gobject, guint32 time, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);

        self->inDrag = FALSE;

	g_signal_emit_by_name (self, "drag-end");
}


static void onDragCancelled (GObject *gobject, guint32 time, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);

	self->dragCancelled = TRUE;
	g_signal_emit_by_name (self, "drag-cancelled");
}


static void disconnectAttentionSignalsFunc (GObject *gobject, gpointer user_data) {

        CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);
        CdosGlobal *global = cdos_global_get ();
        MetaDisplay *display = cdos_global_get_display (global);
        if(self->attentionId > 0){
            g_signal_handler_disconnect (display, self->attentionId);
            self->attentionId = 0L;
        }
        if(self->urgentId > 0){
            g_signal_handler_disconnect (display, self->urgentId);
            self->urgentId = 0L;
        }


}

static void onWindowDemandsAttention (GObject *gobject, MetaWindow *metaWindow, gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);

    if (metaWindow != self->metaWindow) 
		return;

	cdos_expo_window_clone_demand_attention (self);
} 


//固定的类方法
static void cdos_expo_window_clone_dispose (GObject *object){
    CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (object);
    cdos_expo_window_clone_kill_urgency_timeout (self);
    self->disconnectAttentionSignals (NULL, self);
    self->disconnectWindowSignals (NULL, self);
    if(self->icon){
        clutter_actor_destroy (self->icon);
    }
    if(self->clone){
        GList *clones = cdos_create_window_clone ( self->metaWindow, 0, 0, TRUE, FALSE );
        GList *iter;
        for (iter = clones; iter; iter = iter->next)
        {
            ClutterActor *clone = iter->data;
            ClutterActor *parent = clutter_actor_get_parent(clone);
            if(self->clone == parent){
                g_object_ref(clone);
                clutter_actor_remove_child (self->clone, clone);
            }
        }
        g_clear_object(&self->clone);
    }

    g_clear_object(&self->draggable);
    g_clear_object(&self->pointerTracker);

    G_OBJECT_CLASS (cdos_expo_window_clone_parent_class)->dispose (object);
}

static void cdos_expo_window_clone_finalize (GObject *object){
    CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (object);

    self->disconnectAttentionSignals = NULL;
    self->disconnectWindowSignals = NULL;


    G_OBJECT_CLASS (cdos_expo_window_clone_parent_class)->finalize (object);
}

static void
cdos_expo_window_clone_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec  *pspec)
{
    switch (prop_id)
    {
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }
}

static void
cdos_expo_window_clone_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec  *pspec)
{
    switch (prop_id)
    {
        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
            break;
    }
}


static void
cdos_expo_window_clone_class_init (CdosExpoWindowCloneClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = cdos_expo_window_clone_finalize;
    gobject_class->dispose = cdos_expo_window_clone_dispose;
    gobject_class->set_property = cdos_expo_window_clone_set_property;
    gobject_class->get_property = cdos_expo_window_clone_get_property;

        cdos_expo_window_clone_signals [WORKSPACE_CHANGED] =
                g_signal_new ("workspace-changed",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__INT,
                        G_TYPE_NONE, 1 ,
                        G_TYPE_INT );

        cdos_expo_window_clone_signals [HOVERING] =
                g_signal_new ("hovering",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__BOOLEAN,
                        G_TYPE_NONE, 1,
			G_TYPE_BOOLEAN );

        cdos_expo_window_clone_signals [DEMANDING_ATTENTION] =
                g_signal_new ("demanding-attention",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE, 0 );

        cdos_expo_window_clone_signals [DRAG_BEGIN] =
                g_signal_new ("drag-begin",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE, 0 );

        cdos_expo_window_clone_signals [DRAG_CANCELLED] =
                g_signal_new ("drag-cancelled",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE, 0 );

        cdos_expo_window_clone_signals [DRAG_END] =
                g_signal_new ("drag-end",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE, 0 );

        cdos_expo_window_clone_signals [SELECTED] =
                g_signal_new ("selected",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__UINT,
                        G_TYPE_NONE, 1, 
			G_TYPE_UINT);

        cdos_expo_window_clone_signals [MIDDLE_BUTTON_RELEASE] =
                g_signal_new ("middle-button-release",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__UINT,
                        G_TYPE_NONE, 1,
			G_TYPE_UINT);


}


/* 实例初始化 */
static void
cdos_expo_window_clone_init (CdosExpoWindowClone *self)
{
}


static void
cdos_expo_window_clone_post_init (CdosExpoWindowClone *self, MetaWindowActor *realWindow) {

    MetaDisplay *display = cdos_global_get_display (global);

    //this.actor._delegate = this; 尚未处理
    self->realWindow = realWindow;
    self->metaWindow = meta_window_actor_get_meta_window (self->realWindow);


    cdos_expo_window_clone_refresh_clone (self, FALSE); //withTransient

    self->positionChangedId = g_signal_connect (self->realWindow, "position-changed", G_CALLBACK (onPositionChanged), self);
    self->sizeChangedId = g_signal_connect (self->realWindow, "position-changed", G_CALLBACK (onSizeChanged), self);

    self->orphaned = FALSE;

    self->realWindowDestroyedId = g_signal_connect (self->realWindow, "destroy", G_CALLBACK (onRealWindowDestroyed), self);
    self->workspaceChangedId = g_signal_connect (self->metaWindow, "workspace-changed", G_CALLBACK (onWorkspaceChanged), self);

    self->disconnectWindowSignals = disconnectWindowSignalsFunc;

    onPositionChanged (self->realWindow, self);
    onSizeChanged (self->realWindow, self);

    self->lastButtonPressActor = NULL;
    self->lastButtonPressTime = 0;

    g_signal_connect (self, "button-press-event", G_CALLBACK (onButtonPressEvent), self);
    g_signal_connect (self, "button-release-event", G_CALLBACK (onButtonReleaseEvent), self);

    self->pointerTracker = cdos_pointer_tracker_new ();

    g_signal_connect (self, "motion-event", G_CALLBACK (onMotionEvent), NULL);
    g_signal_connect (self, "leave-event", G_CALLBACK (onLeaveEvent), self);
    g_signal_connect (self, "destroy", G_CALLBACK (onDestroy), self);

    //让物件可以被dnd
    self->draggable = cdos_make_draggable ((ClutterActor *)self, 
            NULL, //target
            0,
            FALSE,
            WINDOW_DND_SIZE,
            DRAGGING_WINDOW_OPACITY);

    g_signal_connect (self->draggable, "drag-begin", G_CALLBACK (onDragBegin), self);
    g_signal_connect (self->draggable, "drag-end", G_CALLBACK (onDragEnd), self);
    g_signal_connect (self->draggable, "drag-cancelled", G_CALLBACK (onDragCancelled), self);

    self->inDrag = FALSE;
    self->dragCancelled = FALSE;

    self->icon = (ClutterActor *)st_group_new (); 
    //self->icon = clutter_actor_new (); 
    clutter_actor_add_child ((ClutterActor *)self, self->icon);
    clutter_actor_hide (self->icon);

    ClutterActor *iconActor = NULL;

    //app = this.metaWindow._expoApp; // will be non-null if the window comes from another ws

    if (!self->app) {
        CdosWindowTracker* tracker = cdos_window_tracker_get_default ();
        self->app = cdos_window_tracker_get_window_app (tracker, self->metaWindow);
        // Cache the app, as the tracker has difficulty in finding the app for windows
        // that come from recently removed workspaces.

        //this.metaWindow._expoApp = app;
    }

    if (self->app) {
        iconActor = cdos_app_create_icon_texture (self->app, ICON_SIZE);
    }

    if (!iconActor) {

        iconActor = (ClutterActor *)st_icon_new ();
        st_icon_set_icon_name ((StIcon *)iconActor, "applications-other");
        st_icon_set_icon_type ((StIcon *)iconActor, ST_ICON_FULLCOLOR);
        st_icon_set_icon_size ((StIcon *)iconActor, ICON_SIZE);
    }

    st_container_add_actor ((ClutterContainer *)self->icon, iconActor);
    clutter_actor_set_opacity (iconActor, ICON_OPACITY);

    self->attentionId = g_signal_connect (display, "window-demands-attention", G_CALLBACK(onWindowDemandsAttention), self);
    self->urgentId = g_signal_connect (display, "window-marked-urgent", G_CALLBACK(onWindowDemandsAttention), self);

    self->disconnectAttentionSignals = disconnectAttentionSignalsFunc;

    self->urgencyTimeout = 0;
}

/**
 * cdos_expo_window_clone_new:
 * @windowActor: #MetaWindowActor 对象
 *
 * 根据传递过来的 #MetaWindowActor，也就是实际的窗口对象，创建克隆体
 *
 * Return value: #CdosExpoWindowClone 对象
 */
CdosExpoWindowClone* cdos_expo_window_clone_new (MetaWindowActor *windowActor)
{
    CdosExpoWindowClone *instance = g_object_new (CDOS_TYPE_EXPO_WINDOW_CLONE, "reactive", TRUE, NULL);
	cdos_expo_window_clone_post_init (instance, windowActor);

    return instance;
}

/**
 * cdos_expo_window_clone_refresh_clone:
 * @self: #CdosExpoWindowClone 对象
 * @with_transient: 是否包含临时窗口
 *
 * 不同于 #CdosWindowClone （每个 clone 对应一个实际的窗口）， #CdosExpoWindowClone 存放的是当前工作区窗口以及这个窗口打开的临时窗口的克隆列表，
 * 它的 clone 是个 group 容器。 本函数是在每次进入退出缩略图时，刷新工作区中窗口调用的。如果已经有了 clone 链表，需要先销毁。
 * #CdosExpoWorkspaceThumbnail 对象初始化时，需要先获取工作区的所有窗口，如下：
 *        1、先获取所有 window actors 
 *        2、过滤这个链表，先过滤非本工作区的，再过滤非普通窗口的，获取符合工作区显示条件的那些窗口
 * 	  3、遍历这个链表，获取每个窗口（以及窗口的临时窗口）的克隆体
 * 重新设置每个克隆体的大小和位置。
 */
void cdos_expo_window_clone_refresh_clone (CdosExpoWindowClone *self, gboolean with_transients) {

    g_return_if_fail(CDOS_IS_EXPO_WINDOW_CLONE(self));
    gfloat pwidth, pheight;
    GList *clones;
    //gint i, num_clones;


    //如果self->clone已存在，先销毁
    if ( self->clone ) {
        clutter_actor_destroy (self->clone);
    }

	//创建新的clone
    //self->clone = st_group_new (); 
    self->clone = clutter_actor_new (); 
	clutter_actor_set_reactive (self->clone, FALSE);

	clutter_actor_add_child ((ClutterActor *)self, self->clone);

	pwidth = clutter_actor_get_width ( CLUTTER_ACTOR(self->realWindow) );
    pheight = clutter_actor_get_height ( CLUTTER_ACTOR(self->realWindow) );

	//调用window_utils中的createWindowClone创建
	clones = cdos_create_window_clone ( self->metaWindow, 0, 0, with_transients, FALSE );

    GList *iter;
    for (iter = clones; iter; iter = iter->next)
    {
        gfloat width, height;
        ClutterActor *clone = iter->data;

        clutter_actor_add_child (self->clone, clone);
        clutter_actor_get_size (clone, &width, &height);
        clutter_actor_set_position (clone, round ((pwidth - width) / 2.0), round ((pheight - height) / 2.0));
    }
}

/**
 * cdos_expo_window_clone_kill_urgency_timeout:
 * @self: #CdosExpoWindowClone 对象
 *
 * 停止回调函数 cdos_expo_window_clone_show_urgency_state
 */
void cdos_expo_window_clone_kill_urgency_timeout (CdosExpoWindowClone *self) {

        g_return_if_fail(CDOS_IS_EXPO_WINDOW_CLONE(self));
        if (self->urgencyTimeout) {
        	g_source_remove (self->urgencyTimeout);
        }
        self->urgencyTimeout = 0;
}


static gboolean onTimeoutFunc (gpointer user_data) {

	CdosExpoWindowClone *self = CDOS_EXPO_WINDOW_CLONE (user_data);

	cdos_expo_window_clone_show_urgency_state (self, !self->force, self->temp_param_reps - (self->force ? 0 : 1));
    return FALSE;
}

/**
 * cdos_expo_window_clone_show_urgency_state:
 * @self: #CdosExpoWindowClone 对象
 * 
 * 未知
 */
void cdos_expo_window_clone_show_urgency_state (CdosExpoWindowClone *self, gboolean param_showUrgent, guint param_reps) {

    g_return_if_fail(CDOS_IS_EXPO_WINDOW_CLONE(self));
	MetaWindow *mw;
	gboolean isUrgent;
	gboolean isNotUrgent;
	gboolean hasStyle;

    if (param_reps == 0) {
        // probably the easiest way to just show the current state and stop repeating        	
        //cdos_expo_window_clone_show_urgency_state (self, TRUE, 0);
        return;
    }

    mw = self->metaWindow;
    // Until urgency-query support is generally available in muffin,
    // this is more than a little complicated to get right.
    isUrgent = meta_window_is_above (mw) ;

    if (isUrgent && !self->demanding_attention) {

        cdos_expo_window_clone_demand_attention (self);
        return;
    }

    isNotUrgent = meta_window_is_above(mw);

    StWidget *actor = (StWidget *)self->icon;

    hasStyle = st_widget_has_style_class_name (actor, DEMANDS_ATTENTION_CLASS_NAME);

    if (!hasStyle && isNotUrgent) {
        self->demanding_attention = FALSE;
        return; // window is no longer urgent, so stop alerting
    }

    self->force = param_showUrgent;

    if (!hasStyle && (self->force || isUrgent)) {
        st_widget_add_style_class_name(actor, DEMANDS_ATTENTION_CLASS_NAME);
    }

    if (hasStyle && (isNotUrgent || !param_showUrgent)) {
        st_widget_remove_style_class_name (actor, DEMANDS_ATTENTION_CLASS_NAME);
    }

    if (param_reps > 0)
    {
        cdos_expo_window_clone_kill_urgency_timeout (self);

        self->temp_param_reps = param_reps; //当成参数传递

        self->urgencyTimeout = g_timeout_add (750, (GSourceFunc) onTimeoutFunc, self);
    }
}

static void cdos_expo_window_clone_demand_attention (CdosExpoWindowClone *self) {

	self->demanding_attention = TRUE;

    cdos_expo_window_clone_show_urgency_state (self, TRUE, 50);

	g_signal_emit_by_name (self, "demanding-attention");
}

/**
 * cdos_expo_window_clone_set_stack_above:
 * @self: #CdosExpoWindowClone 对象
 * @actor: 需要放在窗口栈顶的 #ClutterActor 对象
 *
 * 将 actor 置顶（同一 parent 的所有 actor 控件）
 */
void cdos_expo_window_clone_set_stack_above (CdosExpoWindowClone *self, ClutterActor *actor)
{
    g_return_if_fail(CDOS_IS_EXPO_WINDOW_CLONE(self));
    if (clutter_actor_get_parent (actor) != clutter_actor_get_parent ((ClutterActor *)self))
        return;

    self->stackAbove = actor;

    if (self->stackAbove == NULL)
        //this.actor.lower_bottom();
        clutter_container_lower_child ((ClutterContainer *)clutter_actor_get_parent ((ClutterActor *)self), (ClutterActor *)self, NULL);
    else
        //this.actor.raise(this.stackAbove);
        clutter_container_raise_child ((ClutterContainer *)clutter_actor_get_parent ((ClutterActor *)self), (ClutterActor *)self, self->stackAbove);
}

/**
 * cdos_expo_window_clone_destroy:
 * @self: #CdosExpoWindowClone 对象
 * 
 * 销毁 #CdosExpoWindowClone 对象，比如当关闭窗口后，他的 clone 需要销毁掉，在工作区缩略图中不可再显示
 */
void cdos_expo_window_clone_destroy (CdosExpoWindowClone *self)
{
    g_return_if_fail(CDOS_IS_EXPO_WINDOW_CLONE(self));
    clutter_actor_destroy ((ClutterActor *)self);

}

/**
 * cdos_expo_window_clone_on_button_release:
 * @self: #CdosExpoWindowClone 对象
 * @event: 鼠标点击事件
 *
 * 缩略图中的窗口对鼠标点击事件做出响应，比如单击滚轮关闭当前窗口等（然而未做）
 */
gboolean cdos_expo_window_clone_on_button_release (CdosExpoWindowClone *self, ClutterEvent *event)
{
    g_return_val_if_fail(CDOS_IS_EXPO_WINDOW_CLONE(self),FALSE);
    if ((cdos_base_util_get_event_state (event) & CLUTTER_BUTTON1_MASK) ||

            (cdos_base_util_get_event_state (event) & CLUTTER_BUTTON3_MASK))
    {
        g_signal_emit_by_name (self, "selected", clutter_event_get_time (event));
    }

    if ((cdos_base_util_get_event_state (event) & CLUTTER_BUTTON2_MASK))
    {
        g_signal_emit_by_name (self, "middle-button-release", clutter_event_get_time (event));
    }

    return TRUE;
}

