#include "cdos-extension.h"
#include "cdos-key-value.h"
#include "cdos-test.h"
#include "applets/applets.h"
#include "config.h"

#include <gio/gio.h>
#include <girepository.h>
#include <time.h>

#define SYSTEM_PATH_LEN  50

G_DEFINE_TYPE (CdosExtension, cdos_extension, G_TYPE_OBJECT);
G_LOCK_DEFINE_STATIC (last_thread);

enum {
    STATE_INITIALIZING= 0,
    STATE_LOADED,
    STATE_ERROR,
    STATE_OUT_OF_DATE,
    STATES_NUM
};

enum
{
  STATUS_MENU_LOADED,
  LAST_SIGNAL
};

static guint cdos_extension_signals [LAST_SIGNAL] = { 0 };
gboolean init_flag = FALSE;

/* Maps uuid -> imports object (extension directory tree) */
static GHashTable *import_objects = NULL;

/* Maps uuid -> Extension object */
static GHashTable *objects = NULL;

/* Maps uuid -> metadata object */
static GHashTable *extension_meta = NULL; //const meta = {};

typedef struct _StrExtension {
  gchar *key;
  CdosExtension *value;
}StrExtension;

static CdosExtensionType *type_extension = NULL;
static CdosExtensionType *type_applet = NULL;
static CdosExtensionType *type_desklet = NULL;

static void cdos_extension_set_module (CdosExtension *extension, gint module);

/**
 * const Type
 * name: Upper case first character name for printing messages
 * floder: The floder name within the system and user cdos floders
 * required_functions: Functions that must exist in the start file
 * required_properties: Properties that must be set in the metadata.json file
 * nice_to_have_properties: Properties that are encouraged to be set in the metadata.json file
 * roles: Roles an extension can assume. Values will be set internally, set to null.
 *        Key => name of the role, Value => reference to the extension object
 * callbacks: callbacks used to do some manual actions on load / unload
 * 
 * Extension types with some attributes helping to load these extension types.
 * Properties are nested, with lowerCameCase properties (e.g required_functions) as sub-properties of CAPITAL one (EXTENSION). Thus they are refered to as, e.g., Type.EXTENSION.required_functions
 */
/**
 * cdos_extension_get_extension_type:
 * @type: guint type
 *
 * get a #CdosExtensionType instance by type
 *
 * Returns: a instance of #CdosExtensionType
 */
CdosExtensionType *
cdos_extension_get_extension_type (guint type)
{
  if (type < TYPE_NUM) {
    switch (type) {
      case TYPE_EXTENSION:
        if (type_extension == NULL)
          type_extension = cdos_extension_type_new (TYPE_EXTENSION);
        return type_extension;
      case TYPE_APPLET:
        if (type_applet == NULL)
          type_applet = cdos_extension_type_new (TYPE_APPLET);
        return type_applet;
      case TYPE_DESKLET:
        if (type_desklet == NULL)
          type_desklet = cdos_extension_type_new (TYPE_DESKLET);
        return type_desklet;
      default:
        break;
    }
  }

  return NULL;
}

/**
 * cdos_extension_get_object_by_uuid:
 * @uuid: uuid of extension or applet
 *
 * get object by uuid like that objects[uuid] from js
 *
 * Returns: a instance of #CdosExtension
 */
CdosExtension *
cdos_extension_get_object_by_uuid (gchar *uuid)
{
  if (objects == NULL)
  {
    objects = g_hash_table_new(g_str_hash, g_str_equal);
    return NULL;
  }
  if (uuid != NULL) {
    return g_hash_table_lookup (objects, uuid);
  }
  return NULL;
}

static GFile *
cdos_extension_find_extension_directory_in (gchar *uuid, GFile *dir)
{
  if (uuid == NULL || dir == NULL)
    return NULL;

  // if (G_IS_FILE (dir))
  //   g_print ("dir is a file..., path: %s, basename: %s\n", g_file_get_path (dir), g_file_get_basename (dir));

  GFileEnumerator *file_enum = g_file_enumerate_children (dir, "standard::*", 
				G_FILE_QUERY_INFO_NONE,
				NULL, 
				NULL);

  if (file_enum == NULL)
    return NULL;

  GFile *directory = NULL;
  GFileInfo *info;
  while ((info = g_file_enumerator_next_file (file_enum, NULL, NULL)) != NULL) {
  // info = g_file_enumerator_next_file (file_enum, NULL, NULL);
  // while (info != NULL) {
    GFileType file_type = g_file_info_get_file_type (info);
    if (file_type != G_FILE_TYPE_DIRECTORY) {
      continue;
    }
    // g_print ("%s, uuid: %s\n", __func__, uuid);
    const char *name = g_file_info_get_name (info);
    if (uuid && name && g_strcmp0 (uuid, name) == 0) {
      GFile *child = g_file_get_child (dir, name);
      directory = child;
      break;
    } 
    // info = g_file_enumerator_next_file (file_enum, NULL, NULL);
    g_object_unref (info);
  }
  g_file_enumerator_close (file_enum, NULL, NULL);
  g_object_unref (file_enum);

  return directory;
}

/**
 * cdos_extension_find_extension_directory:
 * @uuid: uuid
 * @type: A #CdosExtensionType 
 *
 * find extension directory by uuid and type
 *
 * Returns: directory,使用完成之后，调用g_object_unref()予以释放。
 */
GFile *
cdos_extension_find_extension_directory (gchar *uuid, CdosExtensionType *type)
{
  GFile *directory = cdos_extension_find_extension_directory_in (uuid, type->user_dir);
  if (directory == NULL) {
    // g_print ("%s, directory is NULL...\n", __func__);
    const gchar * const * system_data_dirs = g_get_system_data_dirs ();
    guint i;
    guint len = g_strv_length ((gchar **)system_data_dirs);
    for (i = 0; i < len; i++) {
      gchar *dir_path = g_strconcat (system_data_dirs[i], "/cdos-de/", type->folder, NULL);
      GFile *dir = g_file_new_for_path (dir_path);
      g_free (dir_path);
      if (g_file_query_exists (dir, NULL)) {
        directory = cdos_extension_find_extension_directory_in (uuid, dir);
        if (directory != NULL) {
          g_object_unref (dir);
          break;
        }
      }else{
        g_object_unref (dir);
      }
    }
  }

  return directory;
}

static void
cdos_extension_load_style_sheet (CdosExtension *extension, GFile *file)
{
  if (g_file_query_exists (file, NULL)) {
    StThemeContext *theme_context = st_theme_context_get_for_stage (
				cdos_global_get_stage (cdos_global_get ()));
    if (theme_context) {
      StTheme *theme = st_theme_context_get_theme (theme_context);
      extension->theme = theme;
    } else {
      // throw this.logError('Error trying to get theme', e);
      g_print ("Error trying to get theme\n");
    }

    gchar *file_path = g_file_get_path (file);
    if (st_theme_load_stylesheet (extension->theme, file_path, NULL)) {
      extension->style_sheet = g_file_get_path (file);
    } else {
      // throw this.logError('Stylesheet parse error', e);
      g_print ("Stylesheet parse error.\n");
    }
    g_free (file_path);
  }
}

static void
cdos_extension_unload_style_sheet (CdosExtension *extension)
{
  // TODO: 
}

/*
static gchar *
load_local_stylesheet (GFile *stylefile, GFile *metadataFile){
	if(!ensure_file_exists(metadataFile)){
                return NULL;
        }

	CdosGlobal *global;
    ClutterStage *stage;
	StThemeContext * themeContext;
	StTheme *theme;
	gchar *stylesheet;

    global = cdos_global_get();
    stage = cdos_global_get_stage(global);
	themeContext = st_theme_context_get_for_stage (stage);
	theme = st_theme_context_get_theme (themeContext);

	stylesheet = g_file_get_path(stylefile);
	if(st_theme_load_stylesheet(theme, stylesheet, NULL)){
		return stylesheet;
	}else {
		return NULL;
	}
}
*/

/**
 * cdos_extension_lock_role:
 * @extension: CdosExtension * extension
 * @role_provider: CdosApplet * role_provider
 *
 * lock role
 *
 * Returns: TRUE/FALSE;
 */
gboolean
cdos_extension_lock_role (CdosExtension *extension, CdosApplet *role_provider)
{
  gchar *role = extension->meta->role;
  CdosExtensionType *type = extension->type;
 
  if (role && type->roles != NULL ){
    CdosExtension *tmp = cdos_kv_se_find_by_key (type->roles, role);
    if (tmp != extension) {
      if (tmp != NULL) {
        g_print ("Role %s already taken by (TODO: the lowerType): %s\n", role, tmp->uuid);
        return FALSE;
      }
      if (role_provider != NULL) {
        if (cdos_kv_se_remove_by_key (type->roles, role) != -1) {
          CdosSExtension *se = g_new0 (CdosSExtension, 1);
          se->key = role;
          se->value = extension;
          type->roles = g_slist_append (type->roles, se);

          extension->role_provider = role_provider;
        } else {
              g_print ("error: the member of roles change failed.\n");
        }
        g_print ("Role locked: %s\n", role);
      }
    }
  }

  return TRUE;
}

static void
cdos_extension_unlock_role (CdosExtension *extension) 
{
  // TODO
}

static gboolean
cdos_extension_ensure_file_exists (GFile *file)
{
  if (file) {
    return g_file_query_exists (file, NULL);
  }
  return FALSE;
}

static MetaData *
cdos_extension_create_meta_dummy(gchar *uuid, gchar *path, guint state_id){
  MetaData *meta_dummy = NULL;
  meta_dummy = g_malloc0 (sizeof(MetaData));
  gchar *description = "Metadata load failed";

  if (meta_dummy){
    meta_dummy->name = g_strdup (uuid);
    meta_dummy->uuid = g_strdup (uuid);
    meta_dummy->description = g_strdup (description);
    meta_dummy->state = state_id;
    meta_dummy->path = g_strdup (path);
    meta_dummy->error = NULL;
  }
  return meta_dummy;
}

static void
cdos_extension_check_properties (char **properties, gboolean fatal)
{
  //TODO
}

static void
cdos_extension_vali_data_meta_data (CdosExtension *extension)
{
  g_return_if_fail (CDOS_IS_EXTENSION (extension));
  if (extension) {
    CdosExtensionType *type = extension->type;
    /* Some properties are required to run */
    cdos_extension_check_properties (type->required_properties, TRUE);
  
    /* Others are nice to have */
    cdos_extension_check_properties (type->nice_to_have_properties, FALSE);

    MetaData *meta = extension->meta;
    if (g_strcmp0 (meta->uuid, extension->uuid) != 0) {
      // throw this.logError('uuid "' + this.meta.uuid + '" from metadata.json does not match directory name.');
      g_print ("uuid  '%s ' from metadata.json does not match directory name.\n", meta->uuid);
      g_print ("path of meta: %s\n", meta->path);
      return ;
    }
    //TODO
  }
}

static void 
cdos_extension_load_meta_data (CdosExtension *extension, GFile *meta_data_file) 
{
  if(!cdos_extension_ensure_file_exists(meta_data_file)){
    return ;
  }

  // MetaData *meta_data = g_new0 (MetaData, 1);
  guint old_state = extension->meta->state;
  gchar *old_path = extension->meta->path;  

  //TODO: let metadataContents = Cinnamon.get_file_contents_utf8_sync(metadataFile.get_path());
  //TODO: this.meta = JSON.parse(metadataContents);
  // g_print ("%s, meta_data_file path: %s\n", __func__, g_file_get_path (meta_data_file));
  // gchar *meta_data_contents = cdos_base_util_get_file_contents_utf8_sync (g_file_get_path (meta_data_file));
  // g_print ("meta_data_contents: %s\n", meta_data_contents);

  /* Store some additional crap here */
  extension->meta->state = old_state;
  extension->meta->path = old_path;

  if (!g_hash_table_insert (extension_meta, extension->uuid, extension->meta))
    ;

  // TODO： 有必要判断每个STEP的执行结果，模仿js的try & catch模式，来避免一些结构与算法缺陷

}

static void
cdos_extension_finalize (GObject *object)
{
  CdosExtension *extension = CDOS_EXTENSION(object);
  g_free(extension->uuid);
  g_free(extension->lower_type);
  g_free(extension->style_sheet);

  if(extension->meta){
    g_free(extension->meta->name);
    g_free(extension->meta->uuid);
    g_free(extension->meta->description);
    g_free(extension->meta->path);
    g_free(extension->meta);
  }
  G_OBJECT_CLASS (cdos_extension_parent_class)->finalize (object);
}
static void
cdos_extension_dispose (GObject *object)
{
  CdosExtension *extension = CDOS_EXTENSION(object);

  g_object_unref(extension->type);
  if(extension->loaded_definitions != NULL){
    g_hash_table_remove_all(extension->loaded_definitions);
    g_hash_table_unref(extension->loaded_definitions);
  }
  g_clear_object(&extension->theme);

  G_OBJECT_CLASS (cdos_extension_parent_class)->dispose (object);
}
static void
cdos_extension_init (CdosExtension *extension)
{  
  extension->start_time = time (NULL);
}

static void
cdos_extension_class_init (CdosExtensionClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->finalize = cdos_extension_finalize;
  object_class->dispose = cdos_extension_dispose;

  cdos_extension_signals[STATUS_MENU_LOADED] =
            g_signal_new ("status-menu-loaded",
                  G_TYPE_FROM_CLASS (klass),
                  G_SIGNAL_RUN_LAST,
                  G_STRUCT_OFFSET (CdosExtensionClass, status_menu_loaded),
                  NULL, NULL,
		  g_cclosure_marshal_VOID__OBJECT,
                  G_TYPE_NONE, 1,
                  G_TYPE_OBJECT
                  );

}

static void 
cdos_extension_set_uuid (CdosExtension *extension, gchar *uuid)
{
  g_return_if_fail (CDOS_IS_EXTENSION (extension));
  if (uuid) {
    extension->uuid = uuid;
  }
}

static void
cdos_extension_set_lower_type (CdosExtension *extension, gchar *lower_type)
{
  // g_print ("%s\n", __func__);
  g_return_if_fail (CDOS_IS_EXTENSION (extension));
  if (lower_type) {
    extension->lower_type = lower_type;
  }
}

static void
cdos_extension_set_type (CdosExtension *extension, CdosExtensionType *type)
{
  g_return_if_fail (CDOS_IS_EXTENSION (extension));
  if (type) {
    extension->type = type;
  }
}

static void
cdos_extension_set_dir (CdosExtension *extension, GFile *dir)
{
  g_return_if_fail (CDOS_IS_EXTENSION (extension));
  // g_print ("%s, basename: %s\n", __func__, g_file_get_basename (dir));
  if (dir) {
    extension->dir = dir;
  }
}

// 这里根据applet的uuid去设置module, 决定创建具体的applet类型（cdos_extension_module_main创建具体的applet）
// 后面可能将直接取消module, 用uuid代替其起到的作用
/* Cinnamon applets have a uuid property; this key applet which should be loaded.*/
static void
cdos_extension_set_module_by_uuid (CdosExtension *extension, gchar *uuid)
{
  if (extension && uuid) {
    /* default settings key list */
    if (g_strcmp0 (uuid, "cdos-menu@cdos.org") == 0 || 
          g_strcmp0 (uuid, "menu@cdos.org") == 0) {
	// 添加监听 systray 添加 icon 的信号，在这个时候开始 load menu，节省一秒
	// 需要 loaded 是否成功的标志，防止重复添加
	  cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_MENU);
    } else if (g_strcmp0 (uuid, "scale@cdos.org") == 0 ||
		g_strcmp0 (uuid, "cdos-scale@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SCALE);
    } else if (g_strcmp0 (uuid, "window-list@cdos.org") == 0 || 
		g_strcmp0 (uuid, "cdos-window-list@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_WINDOW_LIST);
    } else if (g_strcmp0 (uuid, "notifications@cdos.org") == 0 ||
               g_strcmp0 (uuid, "cdos-notifications@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_NOTIFICATIONS);
    } else if (g_strcmp0 (uuid, "workspace@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_WORKSPACE);
    } else if (g_strcmp0 (uuid, "removable-drives@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_REMOVABLE_DRIVES);
    } else if (g_strcmp0 (uuid, "keyboard@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_KEYBOARD);
    } else if (g_strcmp0 (uuid, "bluetooth@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_BLUETOOTH);
    } else if (g_strcmp0 (uuid, "network@cdos.org") == 0 ||
		g_strcmp0 (uuid, "cdos-network@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_NETWORK);
    } else if (g_strcmp0 (uuid, "sound@cdos.org") == 0 ||
		g_strcmp0 (uuid, "cdos-sound@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SOUND);
    } else if (g_strcmp0 (uuid, "power@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_POWER);
    } else if (g_strcmp0 (uuid, "systray@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SYSTRAY);
    } else if (g_strcmp0 (uuid, "calendar@cdos.org") == 0 ||
		g_strcmp0 (uuid, "cdos-calendar@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_CALENDAR);
    } else if (g_strcmp0 (uuid, "show-desktop@cdos.org") == 0 ||
		g_strcmp0 (uuid, "cdos-show-desktop@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SHOW_DESKTOP);
    } else if (g_strcmp0 (uuid, "grouped-applets@cdos.org") == 0) {
      cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_GROUPED_APPLETS);
    } else {
      /* the others */
      if (g_strcmp0 (uuid, "panel-launchers@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_PANEL_LAUNCHERS);
      } else if (g_strcmp0 (uuid, "settings@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SETTINGS);
      } else if (g_strcmp0 (uuid, "settings-example@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SETTINGS_EXAMPLE);
      } else if (g_strcmp0 (uuid, "separator@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SEPARATOR);
      } else if (g_strcmp0 (uuid, "spacer@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_SPACER);
      } else if (g_strcmp0 (uuid, "recent@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_RECENT);
      } else if (g_strcmp0 (uuid, "a11y@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_A11Y);
      } else if (g_strcmp0 (uuid, "brightness@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_BRIGHTNESS);
      } else if (g_strcmp0 (uuid, "windows-quick-list@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_WINDOWS_QUICK_LIST);
      } else if (g_strcmp0 (uuid, "trash@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_TRASH);
      } else if (g_strcmp0 (uuid, "xrandr@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_XRANDR);
      } else if (g_strcmp0 (uuid, "workspace-switcher@cdos.org") == 0) {
        cdos_extension_set_module (extension, CDOS_EXTENSION_MODULE_WORKSPACE_SWITCHER);
      }
   }
  }
}

static void
cdos_extension_theme_set_signal_callback (CdosThemeManager *theme_manager, gpointer user_data)
{
  CdosExtension *extension = CDOS_EXTENSION (user_data);
  cdos_extension_load_style_sheet (extension, g_file_get_child (extension->dir, "stylesheet.css"));
}

static void
cdos_extension_init_private ()
{
  if (import_objects == NULL) {
    import_objects = g_hash_table_new (g_str_hash, g_str_equal);
  }

  if (objects == NULL) {
    objects = g_hash_table_new (g_str_hash, g_str_equal);
  }

  if (extension_meta == NULL) {
    extension_meta = g_hash_table_new (g_str_hash, g_str_equal);
  }

  init_flag = TRUE;
}

/**
 * cdos_extension_new:
 * @dir: GFile * dir
 * @type: CdosExtensionType * type
 *
 * 新建一个 #CdosExtension 类的实例。
 *
 * Returns: 返回一个 #CdosExtension 的实例。
 */
CdosExtension *
cdos_extension_new (GFile *dir, CdosExtensionType *type)
{

  if (dir && type) {
    int i;
    CdosExtension *extension;

    if (!init_flag) {
      cdos_extension_init_private ();
    }

    extension = g_object_new (CDOS_TYPE_EXTENSION, NULL);
    g_object_ref(type);
    cdos_extension_set_dir (extension, dir);
    cdos_extension_set_type (extension, type);

    cdos_extension_set_uuid (extension, g_file_get_basename (dir));

    /* set extension module by uuid*/
    gchar *file_basename = g_file_get_basename (dir);
    cdos_extension_set_module_by_uuid (extension, file_basename);
    g_free (file_basename);

    gchar *str = g_strdup (type->name);
    cdos_utils_to_lower_case (str);
    cdos_extension_set_lower_type (extension, g_strdup (str));

    //TODO: this.theme = null;
    //TODO: this.stylesheet = null;

    char *dir_basename = g_file_get_basename (dir);
    char *dir_path = g_file_get_path (dir);
    extension->meta = cdos_extension_create_meta_dummy (
				dir_basename, 
				dir_path,
 				STATE_INITIALIZING);

    GFile *meta_data = g_file_get_child (dir, "metadata.json");
    cdos_extension_load_meta_data (extension, meta_data);
    g_object_unref(meta_data);

    cdos_extension_vali_data_meta_data (extension);

    GFile *style_sheet = g_file_get_child (dir, "stylesheet.css");
    cdos_extension_load_style_sheet (extension, style_sheet);
    g_object_unref (style_sheet);    

    if (extension->style_sheet) {
      g_signal_connect (monkey_plugin_get_theme_manager (monkey_plugin_get ()), 
				"theme-set", 
				G_CALLBACK (cdos_extension_theme_set_signal_callback),
				extension);
    }

    /*
    if (!cdos_global_add_extension_importer (cdos_global_get (), 
					"imports.ui.extension.importObjects", 
					extension->uuid, 
					extension->meta->path, 
					NULL)) {
      // throw this.logError('Error importing extension ' + this.uuid + ' from path ' + this.meta.path, e);
      g_print ("Error importing extension %s from path %s.\n", extension->uuid, extension->meta->path);
    } */

    //TODO: this.module = importObjects[this.uuid][this.lowerType]; // get [extension/applet/desklet].js 

    type = extension->type;
    gchar *func = NULL;
    gchar len = g_strv_length (type->required_functions);
    for (i = 0; i < len; i ++ ) {
      func = type->required_functions[i];
      //TODO if (!this.module[func]) {
      //          throw this.logError('Function "' + func + '" is missing');
      //      }
    }

    if (objects == NULL)
      objects = g_hash_table_new (g_str_hash, g_str_equal);

    extension->loaded_menu = FALSE;
    g_hash_table_insert (objects, extension->uuid, extension);

    g_free (str);
    g_free (dir_basename);
    g_free (dir_path);

    return CDOS_EXTENSION (extension);
  }
  return NULL;
}

/*
GType cdos_extension_get_type (void)
{
  static GType extension_type = 0;
  if (!extension_type) {
    static const GTypeInfo extension_info = {
		sizeof (CdosExtensionClass),
		NULL, NULL,
		(GClassInitFunc) cdos_extension_class_init,
		NULL, NULL,
		sizeof (CdosExtension),
		0,
		(GInstanceInitFunc) cdos_extension_init};

    extension_type = g_type_register_static (G_TYPE_OBJECT, "CdosExtension", &extension_info, 0);
  }
  return extension_type;
} */

/**
 * cdos_extension_load_extension:
 * @uuid: gchar * uuid
 * @type: CdosExtenisonType *type
 *
 * load extension by uuid & type
 *
 * Returns: 返回一个 #CdosExtension 的实例。
 */
CdosExtension *
cdos_extension_load_extension (gchar* uuid, CdosExtensionType *type)
{

  G_LOCK (last_thread);
  if (!uuid || !type) {
    g_print ("!uuid || !type is TRUE\n");
    G_UNLOCK (last_thread);
    return NULL;
  }

  CdosExtension *extension = cdos_extension_get_object_by_uuid (uuid);
  if (!extension) {
    // gboolean forget_meta = TRUE;

    GFile *dir = cdos_extension_find_extension_directory (uuid, type);
    if (dir == NULL) {
      g_print ("%s %s not found.\n", type->name, uuid); //global.logError(type.name + ' ' + uuid + ' not found.');

      G_UNLOCK (last_thread);
      return NULL;
    }

    extension = cdos_extension_new (dir, type);

/*    if ( g_strcmp0 (uuid, "network@cdos.org") == 0 
			 || g_strcmp0 (uuid, "sound@cdos.org") == 0){
	return extension;
    }*/

    // forget_meta = FALSE;

    /*暫時只支持 applet 類型的加載
    if (!cdos_extension_type_callbacks_finish_extension_load (type, extension))
      return NULL; */

    if (!cdos_applet_manager_finish_extension_load (extension)){
	
  	G_UNLOCK (last_thread);
        return NULL;
    }

    //cdos_extension_finalize ((GObject *)extension);

    /**
     * catch(e) {
     *  forgetExtension(uuid, forgetMeta);
     *  if(e._alreadyLogged)
     *     e = undefined;
     *  global.logError('Could not load ' + type.name.toLowerCase() + ' ' + uuid, e);
     *  return null;
     * }
     */
  }

  G_UNLOCK (last_thread);
  return extension;
}

static void
cdos_extension_forget_extension (gchar *uuid, gboolean forget_meta)
{
  //TODO:
}

/**
 * cdos_extension_unload_extension:
 * @uuid: gchar * uuid
 *
 * unload extension
 *
 */
void
cdos_extension_unload_extension (gchar *uuid)
{
  if(uuid == NULL){
     return ;
  }
  CdosExtension *extension = cdos_extension_get_object_by_uuid (uuid);
  if (extension) {
    cdos_extension_unlock_role (extension);

    cdos_extension_type_callbacks_prepare_extension_unload (extension->type, extension);

    cdos_extension_unload_style_sheet (extension);
    g_signal_emit_by_name (extension->type, "extension-unloaded", extension->uuid);
    cdos_extension_forget_extension (extension->uuid, TRUE);
  }
}

/**
 * cdos_extension_module_main:
 * @extension: A #CdosExtension
 * @orientation: StSide orientation
 * @panel_height: gfloat panel_height
 * @applet_id: gchar * applet_id
 *
 * create a #CdosApplet instance
 *
 * Returns: a #CdosApplet instance.
 */
CdosApplet *
cdos_extension_module_main (CdosExtension *extension,
                                StSide orientation,
                                gfloat panel_height,
                                gchar *applet_id)
{
  g_return_val_if_fail (CDOS_IS_EXTENSION (extension), NULL);

  gint module = extension->module;
  MetaData *meta = extension->meta;
  CdosApplet *applet = NULL;

#define TIME_START do{\
    clock_t start,end;\
    start= clock()

#define TIME_END end=clock();\
    printf("used time: %gms\n", (double)(end-start) / CLOCKS_PER_SEC * 1000);\
}while(0)

  switch (module) {
    case CDOS_EXTENSION_MODULE_MENU:
      TIME_START;
      applet = (CdosApplet *) applets_menu_main(meta, orientation, panel_height);
      TIME_END;
      break;
    case CDOS_EXTENSION_MODULE_SHOW_DESKTOP:
      TIME_START;
      applet = (CdosApplet *) cdos_showdesktop_main (meta, orientation, panel_height);
      TIME_END;
      break;
    case CDOS_EXTENSION_MODULE_PANEL_LAUNCHERS:	
      break;
    case CDOS_EXTENSION_MODULE_WINDOW_LIST:
      TIME_START;
      applet = (CdosApplet *) applets_window_list_main(meta, orientation, panel_height, applet_id);
      TIME_END;
      break;
    case CDOS_EXTENSION_MODULE_NOTIFICATIONS:
      TIME_START;
      applet = (CdosApplet *) applet_notifications_main (meta, orientation, panel_height, applet_id);
      TIME_END;
      break;
    case CDOS_EXTENSION_MODULE_REMOVABLE_DRIVES:
      TIME_START;
      applet = (CdosApplet *) applet_removable_drives_main(meta, orientation, panel_height);
      TIME_END;
      break;
    case CDOS_EXTENSION_MODULE_KEYBOARD:
      break;
    case CDOS_EXTENSION_MODULE_BLUETOOTH:
      break;
    case CDOS_EXTENSION_MODULE_NETWORK:
      TIME_START;
      applet = (CdosApplet *) applet_network_main (meta, orientation, (guint)panel_height);
      TIME_END;
      break;
    case CDOS_EXTENSION_MODULE_SOUND:
      applet = (CdosApplet *) applet_sound_main (meta, orientation, panel_height, applet_id);
      break;
    case CDOS_EXTENSION_MODULE_POWER:
      TIME_START;
      applet = (CdosApplet *) applets_power_main (meta, orientation, panel_height, applet_id);
      TIME_END;
      break;
    case CDOS_EXTENSION_MODULE_SYSTRAY:
      TIME_START;
      applet = (CdosApplet *) applets_systray_main (meta, orientation, panel_height);
      TIME_END;
      break;
    case CDOS_EXTENSION_MODULE_CALENDAR:
      TIME_START;
      applet = (CdosApplet *) applets_calendar_main(meta, orientation, panel_height);
      TIME_END;
      break;
    case CDOS_EXTENSION_MODULE_WINDOWS_QUICK_LIST:
      break;
    case CDOS_EXTENSION_MODULE_WORKSPACE_SWITCHER:
      break;
    case CDOS_EXTENSION_MODULE_SCALE:
      TIME_START;
      applet = (CdosApplet *) cdos_scale_new (meta,orientation,panel_height,applet_id);
      TIME_END;
      break;
    case CDOS_EXTENSION_MODULE_GROUPED_APPLETS:
      TIME_START;
      applet = (CdosApplet *) applet_group_main(meta,orientation,panel_height);
      TIME_END;
      break;
    default:
      break;
  }
  return applet;
}

/**
 * cdos_extension_get_dir:
 * @extension: A #CdosExtension
 *
 * get dir of extension
 *
 * Returns: dir
 */
GFile *
cdos_extension_get_dir (CdosExtension *extension)
{
  g_return_val_if_fail (CDOS_IS_EXTENSION (extension), NULL);
  return extension->dir;
}

/**
 * cdos_extension_get_uuid:
 * @extension: A #CdosExtension
 *
 * get uuid of extension
 *
 * Returns: uuid
 */
gchar *
cdos_extension_get_uuid (CdosExtension *extension)
{
  g_return_val_if_fail (CDOS_IS_EXTENSION (extension), NULL);
  return extension->uuid;
}

/**
 * cdos_extension_get_meta:
 * @extension: CdosExtension * extension
 *
 * get meta
 *
 * Returns: struct of MetaData
 */
MetaData *
cdos_extension_get_meta (CdosExtension *extension)
{
  g_return_val_if_fail (CDOS_IS_EXTENSION (extension), NULL);
  return extension->meta;
}

/**
 * cdos_extension_get_role_provider:
 * @extension: A #CdosExtension
 *
 * get role provider of extension
 *
 * Returns: a #CdosApplet instance
 */
CdosApplet *
cdos_extension_get_role_provider (CdosExtension *extension)
{
  g_return_val_if_fail (CDOS_IS_EXTENSION (extension), NULL);
  return extension->role_provider;
}

/**
 * cdos_extension_get_loaded_definitions:
 * @extension: A #CdosExtension
 *
 * get loaded definitions of extension
 *
 * Returns: loaded_definitions
 */
GHashTable *
cdos_extension_get_loaded_definitions (CdosExtension *extension)
{
  g_return_val_if_fail (CDOS_IS_EXTENSION (extension), NULL);
  return extension->loaded_definitions;
}

/**
 * cdos_extension_set_loaded_definitions:
 * @extension: A #CdosExtension
 *
 * set loaded definitions of extension
 *
 */
void
cdos_extension_set_loaded_definitions (CdosExtension *extension, GHashTable *loaded_definitions)
{
  g_return_if_fail (CDOS_IS_EXTENSION (extension));
  if(extension->loaded_definitions != NULL){
      g_hash_table_remove_all(extension->loaded_definitions);
      g_hash_table_unref(extension->loaded_definitions);
  }
  extension->loaded_definitions = loaded_definitions;
}

/**
 * cdos_extension_get_extension_meta:
 *
 * get extension meta
 *
 * Returns: extension meta, a object of GHashTable
 */
GHashTable *
cdos_extension_get_extension_meta (void)
{
  return extension_meta;
}

/**
 * cdos_extension_get_import_objects:
 *
 * get import objects
 *
 * Returns: import objects, a object of GHashTable
 */
GHashTable *
cdos_extension_get_import_objects ()
{
  if (import_objects == NULL)
    import_objects = g_hash_table_new (g_str_hash, g_str_equal);
  return import_objects;
}

/**
 * cdos_extension_set_module:
 * @extension: A #CdosExtension
 * @module: module of extension
 *
 * set module of extension
 *
 */
static void
cdos_extension_set_module (CdosExtension *extension, gint module)
{
  g_return_if_fail (CDOS_IS_EXTENSION (extension));

  extension->module = module;
}

/**
 * cdos_extension_get_max_instances:
 * @uuid: uuid of applet
 *
 * get max instances
 *
 * Returns: number of max instances。
 */
gint
cdos_extension_get_max_instances (const gchar *uuid)
{
  if (uuid) {
    MetaData *meta = g_hash_table_lookup (extension_meta, uuid);
    if (meta != NULL) {
      if (meta->max_instances != 0) {
	return meta->max_instances;
      }
    }
  }
  return 1; 
}
