/**
 * SECTION: cdos-focus-grabber
 * @short_description: 显示CdosMessageTray的时候进行焦点的处理
 * @include: cdos-focus-grabber.h
 * 
 * CdosFocusGrabber 类用来处理 #CdosMessageTray 的焦点切换
 */

#include "cdos-focus-grabber.h"
#include "cdos-source.h"
#include "cdos-global.h"
#include "cdos.h"

/* FocusGrabber start: */
G_DEFINE_TYPE (CdosFocusGrabber, cdos_focus_grabber, G_TYPE_OBJECT);


/* signal */
enum {
    FOCUS_GRABBED,       //focus-grabbed
    BUTTON_PRESSED,      //button-pressed
    ESCAPE_PRESSED,         //escape-pressed
    FOCUS_UNGRABBED,        //focus-ungrabbed
    LAST_SIGNAL
};


static guint cdos_focus_grabber_signals[LAST_SIGNAL] = {0};

/* Public Value */
//CdosGlobal *global;
MetaDisplay *display;
ClutterStage *global_stage;

/* declare */
static gboolean onCapturedEvent_cb (GObject *obj, ClutterEvent *event, gpointer user_data);

// Because we grab focus differently in the overview
// and in the main view, we need to change how it is
// done when we move between the two.

static void toggleFocusGrabMode (GObject *gobject, gpointer user_data)
{
    CdosFocusGrabber *self = CDOS_FOCUS_GRABBER (user_data);

    if (self->hasFocus) {
        self->togglingFocusGrabMode = TRUE;
        cdos_focus_grabber_ungrab_focus (self);
        cdos_focus_grabber_grab_focus (self, self->actor);
        self->togglingFocusGrabMode = FALSE;
    }
}

/* ========== GObject ========== */
static void
cdos_focus_grabber_finalize (GObject *gobject)
{
    CdosFocusGrabber *self = CDOS_FOCUS_GRABBER (gobject);

    g_free (self->actor);
    g_free (self->prevFocusedWindow);
    g_free (self->prevKeyFocusActor);

    /* 调用父类的 finalize() 方法 */
    G_OBJECT_CLASS(cdos_focus_grabber_parent_class)->finalize (gobject);
}

static void
cdos_focus_grabber_class_init (CdosFocusGrabberClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

    /**
     * CdosFocusGrabber::focus-grabbed:
     * @object (GObject): 绑定信号的对象
     * @user_data (gpointer): 真对回调函数所传入的参数
     *
     * 当在获取焦点的时候，可以捕获该信号
     */
    cdos_focus_grabber_signals[FOCUS_GRABBED] =
            g_signal_new ("focus-grabbed",
                          G_TYPE_FROM_CLASS (klass),
                          G_SIGNAL_RUN_LAST,
                          0,
                          NULL, NULL,
                          g_cclosure_marshal_VOID__VOID,
                          G_TYPE_NONE, 0);

    /**
     * CdosFocusGrabber::button-pressed:
     * @object (GObject): 绑定信号的对象
     * @source (ClutterActor):
     * @user_data (gpointer): 真对回调函数所传入的参数
     *
     * 当点击按钮时，可以捕获该信号
     */    
    cdos_focus_grabber_signals[BUTTON_PRESSED] =
            g_signal_new ("button-pressed",
                          G_TYPE_FROM_CLASS (klass),
                          G_SIGNAL_RUN_LAST,
                          0,
                          NULL, NULL,
                          g_cclosure_marshal_VOID__OBJECT,
                          G_TYPE_NONE, 1,
                          G_TYPE_OBJECT);

    /**
     * CdosFocusGrabber::escape-pressed:
     * @object (GObject): 绑定信号的对象
     * @user_data (gpointer): 真对回调函数所传入的参数
     *
     * 按Esc键，可以捕获该信号
     */    
    cdos_focus_grabber_signals[ESCAPE_PRESSED] =
            g_signal_new ("escape-pressed",
                          G_TYPE_FROM_CLASS (klass),
                          G_SIGNAL_RUN_LAST,
                          0,
                          NULL, NULL,
                          g_cclosure_marshal_VOID__VOID,
                          G_TYPE_NONE, 0);

    /**
     * CdosFocusGrabber::focus-grabbed:
     * @object (GObject): 绑定信号的对象
     * @user_data (gpointer): 真对回调函数所传入的参数
     *
     * 当在移除焦点的时候，可以捕获该信号
     */
    cdos_focus_grabber_signals[FOCUS_UNGRABBED] =
            g_signal_new ("focus-ungrabbed",
                          G_TYPE_FROM_CLASS (klass),
                          G_SIGNAL_RUN_LAST,
                          0,
                          NULL, NULL,
                          g_cclosure_marshal_VOID__VOID,
                          G_TYPE_NONE, 0);

    gobject_class->finalize = cdos_focus_grabber_finalize;

}

static void
cdos_focus_grabber_init (CdosFocusGrabber *self)
{
    global = cdos_global_get ();
    self->actor = NULL;

    self->hasFocus = FALSE;
    self->prevFocusedWindow = NULL;
    self->prevKeyFocusActor = NULL;

    self->focusActorChangedId = 0;
    self->stageInputModeChangedId = 0;
    self->capturedEventId = 0;
    self->togglingFocusGrabMode = FALSE;

    CdosOverview *overview = cdos_overview_get ();
    //CdosExpo *expo = cdos_monkey_plugin_get_expo ();

    //"hidden" or "showing"
    g_signal_connect (overview, "showing", G_CALLBACK (toggleFocusGrabMode), self);
    g_signal_connect (overview, "hidden", G_CALLBACK (toggleFocusGrabMode), self);
    //g_signal_connect (expo, "expo-showing", G_CALLBACK (toggleFocusGrabMode), self);
    //g_signal_connect (expo, "expo-hidden", G_CALLBACK (toggleFocusGrabMode), self);
}

 /**
  * cdos_focus_grabber_new:
  *
  * 新建一个 #CdosFocusGrabber 类的实例。
  *
  * Returns (CdosUrlHighlighter): 返回一个#CdosFocusGrabber 类的实例。
  */
CdosFocusGrabber* cdos_focus_grabber_new (void)
{
    CdosFocusGrabber *self = g_object_new (CDOS_TYPE_FOCUS_GRABBER, NULL);
    return self;
}

/* ======== GObject END ========== */
static gboolean onCapturedEvent_cb (GObject *obj, ClutterEvent *event, gpointer user_data)
{
    CdosFocusGrabber *self = CDOS_FOCUS_GRABBER (user_data);
    ClutterActor *source = clutter_event_get_source (event);
    guint symbol;
    CdosLayoutManager *layout_manager = NULL;
    g_print ("\n Func:%s event->type: %d \n", __FUNCTION__, event->type);

    switch (event->type) {
       case CLUTTER_BUTTON_PRESS: //Clutter.EventType.BUTTON_PRESS:
            layout_manager = cdos_monkey_plugin_get_layout_manager (monkey_plugin_get ());
            gboolean tmp = clutter_actor_contains (CLUTTER_ACTOR (cdos_layout_manager_get_keyboardbox (layout_manager)), source);
            if (!clutter_actor_contains (self->actor, source) && !tmp) {
                g_signal_emit_by_name (self, "button-pressed", source);
            }

            break;

       case CLUTTER_KEY_PRESS: //Clutter.EventType.KEY_PRESS:

            symbol = clutter_event_get_key_symbol (event);

            if (symbol == CLUTTER_KEY_Escape) {

                g_print ("\n ----- LINE: %d FILE:%s Escape Key Press >>>>>\n", __LINE__, __FILE__);
                g_signal_emit_by_name (self, "escape-pressed");

                return TRUE;
            }

            break;

       default:
            g_print ("\n LINE:%d FUNCTION:%s event->type error!\n", __LINE__, __FUNCTION__);
    }

    return FALSE;
}


static void stageInputModeChanged_cb (GObject *gobject, GParamSpec *pspec, gpointer user_data)
{
    CdosFocusGrabber *self = CDOS_FOCUS_GRABBER (user_data);
    cdos_focus_grabber_ungrab_focus (self);
}

static void focusActorChanged_cb(ClutterStage *stage, GParamSpec *pspec, gpointer user_data)
{
    CdosFocusGrabber *self = CDOS_FOCUS_GRABBER (user_data);
    ClutterActor *focusedActor = clutter_stage_get_key_focus (stage);

    if (!focusedActor || !clutter_actor_contains (self->actor, focusedActor)) {
        self->prevKeyFocusActor = NULL;
        cdos_focus_grabber_ungrab_focus (self);
    }
}

 /**
  * cdos_focus_grabber_grab_focus:
  * @self: A #CdosFocusGrabber
  * @clutter: 待获取焦点的actor
  *
  * 对 #CdosMessageTray 获取焦点
  * 
  */
void cdos_focus_grabber_grab_focus (CdosFocusGrabber *self, ClutterActor *actor)
{
    g_return_if_fail(CDOS_IS_FOCUS_GRABBER(self));
    if (self->hasFocus) {
        return;
    }

    global = cdos_global_get ();
    display = cdos_global_get_display (global);
    global_stage = cdos_global_get_stage (global);

    //CdosStageInputMode input_mode;

    /*保存之前的窗口及key焦点*/
    self->actor = actor;
    self->prevFocusedWindow = meta_display_get_focus_window (display);
    self->prevKeyFocusActor = clutter_stage_get_key_focus (global_stage);

    /*设定取得焦点*/
    //g_object_get (global, "stage-input-mode", &input_mode, NULL);

    /*if (input_mode == CDOS_STAGE_INPUT_MODE_NONREACTIVE || input_mode == CDOS_STAGE_INPUT_MODE_NORMAL) {
        cdos_global_set_stage_input_mode (global, CDOS_STAGE_INPUT_MODE_FOCUSED);
    }*/

    /*监听并信号绑定回调函数*/
    self->capturedEventId = g_signal_connect (global_stage, "captured-event", G_CALLBACK (onCapturedEvent_cb), self);
    self->stageInputModeChangedId = g_signal_connect (global, "notify::stage-input-mode", G_CALLBACK (stageInputModeChanged_cb), self);
    self->focusActorChangedId = g_signal_connect (global_stage, "notify::key-focus", G_CALLBACK (focusActorChanged_cb), self);

    self->hasFocus = TRUE;

    st_widget_navigate_focus (ST_WIDGET(self->actor), NULL, GTK_DIR_TAB_FORWARD, FALSE);
    g_signal_emit_by_name (self, "focus-grabbed");
}

 /**
  * cdos_focus_grabber_grab_focus:
  * @self: A #CdosFocusGrabber
  *
  * 对 #CdosMessageTray 移除相关信号和焦点
  * 
  */
void cdos_focus_grabber_ungrab_focus (CdosFocusGrabber *self)
{
    g_return_if_fail(CDOS_IS_FOCUS_GRABBER(self));
    if (!self->hasFocus)
        return;

    global = cdos_global_get ();
    display = cdos_global_get_display (global);
    global_stage = cdos_global_get_stage (global);

    /*断开信号*/
    if (self->focusActorChangedId > 0) {

        g_signal_handler_disconnect (global_stage, self->focusActorChangedId);

        self->focusActorChangedId = 0;
    }

    if (self->stageInputModeChangedId) {

        g_signal_handler_disconnect (global, self->focusActorChangedId);

        self->stageInputModeChangedId = 0;
    }

    if (self->capturedEventId > 0) {

        g_signal_handler_disconnect (global_stage, self->capturedEventId);

        self->capturedEventId = 0;
    }

    /*设定不取得焦点*/
    self->hasFocus = FALSE;

    g_signal_emit_by_name (self, "focus-ungrabbed");

    /*设定输入焦点為之前保存的焦点*/
    if (self->prevFocusedWindow && !meta_display_get_focus_window (display)) {

        meta_display_set_input_focus_window (display,
                                             self->prevFocusedWindow,
                                             FALSE,
                                             cdos_global_get_current_time (global));

        self->prevFocusedWindow = NULL;
    }

    if (self->prevKeyFocusActor) {

        clutter_stage_set_key_focus (global_stage, self->prevKeyFocusActor);

        self->prevKeyFocusActor = NULL;

    } else {
        // We don't want to keep any actor inside the previously focused actor focused.
        ClutterActor *focusedActor = clutter_stage_get_key_focus (global_stage);

        if (focusedActor && clutter_actor_contains (self->actor, focusedActor)) {
            clutter_stage_set_key_focus (global_stage, NULL);
        }
    }

    if (!self->togglingFocusGrabMode)
        self->actor = NULL;

}

