//
// Created by zhangtao on 16-7-5.
// source:keyboard.js:Keyboard
/**
 * SECTION: cdos-keyboard
 * @short_description: 提供虚拟键盘输入相关功能。
 * @title: CdosKeyboard
 * @include: cdos-keyboard.h
 *
 * #CdosKeyboard主要提供了桌面虚拟键盘输入相关的功能:控制虚拟键盘布局，加载虚拟按键，控制虚拟键盘的是显示和隐藏，响应鼠标和键盘的聚焦事件等。
 *
 * #CdosKeyboard是基于caribou库，在此基础上进行了一定的扩展和修改。
 */


#include <math.h>
#include "cdos-keyboard.h"
#include "cdos-monkey-plugin.h"
#include "cdos-key.h"
#include "cdos-keyboard-source.h"
#include "cdos-global.h"

typedef struct _CdosKeyboardSignalIds CdosKeyboardSignalIds;
/*-----------------信号枚举-------------------*/
enum {
    LAST_SIGNAL
};
/*-----------------实例私有结构体-------------------*/
struct _CdosKeyboardPrivate{
    guint32 timestamp;
    gboolean enable_keyboard;
    guint num_of_horiz_keys;
    guint num_of_vert_keys;
    StWidget *actor;
    GSettings *keyboard_settings;
    GSettings *a11y_applications_settings;
    CdosKeyboardSignalIds *signal_ids;
    CaribouKeyboardModel *keyboard;
    GHashTable *groups;
    StWidget *current_page;
    CdosKeyboardSource *source;

};


/*-----------------信号id-------------------*/
struct _CdosKeyboardSignalIds{
    gulong main_layout_manager_monitors_changed_id;
    gulong keyboard_settings_changed_id;
    gulong a11y_applications_settings_changed_id;
    gulong keyboard_notify_active_group_id;
    gulong global_stage_notify_key_focus_id;
    gulong group_notify_active_level_id;
    gulong tray_button_press_id;
    gulong main_overview_showing_id;
    gulong main_overview_hiding_id;
    gulong key_released_id;
};
/*-----------------属性枚举-------------------*/
enum {
    PROP_0
};

/*-------global variables-------------*/
static gint CdosKeyboard_private_offset = 0;
static gint cdos_keyboard_signals[LAST_SIGNAL] = { 0 };
static gpointer cdos_keyboard_parent_class = NULL;

/*-------class functions---------*/
static void cdos_keyboard_init(CdosKeyboard *self);
static void cdos_keyboard_class_init(CdosKeyboardClass *klass);
static void cdos_keyboard_dispose(GObject *object);
static void cdos_keyboard_finalize(GObject *object);

/*-------------private functions of instance-------------------*/
static void cdos_keyboard_settings_changed(CdosKeyboard *self,GSettings *settings,gchar *key);

static void cdos_keyboard_destroy(CdosKeyboard *self);

static void cdos_keyboard_setup(CdosKeyboard *self,gboolean show);

static void cdos_keyboard_add_keys(CdosKeyboard *self);

static StWidget* cdos_keyboard_get_tray_icon(CdosKeyboard *self);

static void cdos_keyboard_add_rows(CdosKeyboard *self,CaribouRowModel *row,StWidget *layout);

static void cdos_keyboard_load_rows(CdosKeyboard *self,CaribouLevelModel *level,StWidget *layout);

static void cdos_keyboard_redraw(CdosKeyboard *self);

static void cdos_keyboard_set_active_layer(CdosKeyboard *self);

static void cdos_keyboard_create_source(CdosKeyboard *self);

static void cdos_keyboard_destroy_source(CdosKeyboard *self);

static void cdos_keyboard_move_temporarily(CdosKeyboard *self);

static void cdos_keyboard_set_location(CdosKeyboard *self,gfloat x,gfloat y);

static void cdos_keyboard_settings_on_changed(GSettings *settings,gchar *key,CdosKeyboard *self);

static void cdos_keyboard_on_notify_active_group(GObject *keyboard,GParamSpec *pspec,CdosKeyboard *self);

static void cdos_keyboard_global_stage_on_notify_key_focus(GObject *stage,GParamSpec *pspec,CdosKeyboard *self);

static void cdos_keyboard_group_on_notify_active_level(GObject *group,GParamSpec *pspec,CdosKeyboard *self);

static gboolean cdos_keyboard_tray_button_on_press(ClutterActor *actor,ClutterEvent *event,CdosKeyboard *self);

static void cdos_keyboard_main_overview_on_showing(CdosOverview *overview,StWidget *tray_button);

static void cdos_keyboard_main_overview_on_hiding(CdosOverview *overview,StWidget *tray_button);

static void cdos_keyboard_key_on_key_released(CaribouIKeyboardObject* _sender, CaribouKeyModel* key, CdosKeyboard *self);

static gboolean cdos_keyboard_groups_GHRFunc(gchar *key,GHashTable *value,gchar *user_data);

static gboolean cdos_keyboard_layers_GHRFunc(gchar *key,StWidget *value,gchar *user_data);

/*-------functions-------------*/

/**
 * 注册类；
 *
 *
 * */
GType
cdos_keyboard_get_type(void)
{
    static GType cdos_keyboard_type = 0;
    if(G_UNLIKELY(cdos_keyboard_type==0))
    {
        static const GTypeInfo cdos_keyboard_info = {
                sizeof(CdosKeyboardClass),
                NULL,NULL,
                (GClassInitFunc)cdos_keyboard_class_init,
                NULL,NULL,
                sizeof(CdosKeyboard),
                0,
                (GInstanceInitFunc)cdos_keyboard_init
        };
        //类注册
        cdos_keyboard_type = g_type_register_static(G_TYPE_OBJECT,"CdosKeyboard",&cdos_keyboard_info,0);
        //g_type_add_class_private(cdos_keyboard_type, sizeof(CdosKeyboardClassPrivate));
        CdosKeyboard_private_offset = \
                g_type_add_instance_private(cdos_keyboard_type, sizeof(CdosKeyboardPrivate));

    }
    return cdos_keyboard_type;
}
/**
 *  获取实例的私有属性结构
 * */
static inline gpointer
cdos_keyboard_get_instance_private(CdosKeyboard *self){
    return (G_STRUCT_MEMBER_P (self, CdosKeyboard_private_offset));
}


/**
 * 实例初始化；
 *
 * */
static void
cdos_keyboard_init(CdosKeyboard *self)
{
    CdosKeyboardPrivate *priv;
    self->priv=cdos_keyboard_get_instance_private(self);
    priv=self->priv;
    /*--------------实例初始化（构造实例）-------------*/
    priv->signal_ids = g_new0(CdosKeyboardSignalIds,1);
    priv->timestamp = cdos_global_get_current_time(global);
    priv->keyboard_settings = g_settings_new(KEYBOARD_SCHEMA);
    priv->a11y_applications_settings = g_settings_new(A11Y_APPLICATIONS_SCHEMA);

    priv->signal_ids->main_layout_manager_monitors_changed_id = 0;
    /* about:在CdosLayoutManager中暂时无信号“monitors-changed”;
    CdosLayoutManager *main_layout_manager = cdos_monkey_plugin_get_layout_manager(monkey_plugin_get());
    priv->signal_ids->main_layout_manager_monitors_changed_id = \
            g_signal_connect(main_layout_manager,"monitors-changed",G_CALLBACK(cdos_keyboard_main_layout_manager_on_monitors_changed),self);
    */
    priv->signal_ids->keyboard_settings_changed_id = \
            g_signal_connect(priv->keyboard_settings,"changed",G_CALLBACK(cdos_keyboard_settings_on_changed),self);
    priv->signal_ids->a11y_applications_settings_changed_id = \
            g_signal_connect(priv->a11y_applications_settings,"changed",G_CALLBACK(cdos_keyboard_settings_on_changed),self);

}
/**
 *
 * 类结构初始化；
 *
 * */
static void
cdos_keyboard_class_init(CdosKeyboardClass *klass)
{
    //转换为gobject；
    GObjectClass *gobject_class=G_OBJECT_CLASS(klass);
    g_type_class_adjust_private_offset(klass,&CdosKeyboard_private_offset);
    cdos_keyboard_parent_class = g_type_class_peek_parent(klass);

    /*-------------重写父类的虚方法---------------*/
    //gobject_class->constructed=cdos_keyboard_constructed;
    gobject_class->dispose=cdos_keyboard_dispose;
    gobject_class->finalize=cdos_keyboard_finalize;
}

/*------------类方法定义-----------*/
static void cdos_keyboard_dispose(GObject *object){
    CdosKeyboard *self = CDOS_KEYBOARD(object);
    CdosKeyboardPrivate *priv = self->priv;



    G_OBJECT_CLASS(cdos_keyboard_parent_class)->dispose(object);
};

static void cdos_keyboard_finalize(GObject *object){
    CdosKeyboard *self = CDOS_KEYBOARD(object);
    CdosKeyboardPrivate *priv = self->priv;


    if(g_object_is_floating(object)){
        g_warning ("A floating object was finalized. This means that someone called g_object_unref() on an object that had only a floating reference; the initial floating reference is not owned by anyone and must be removed with g_object_ref_sink().");
    }
    G_OBJECT_CLASS(cdos_keyboard_parent_class)->finalize(object);
};

/*-------------private functions--------------*/
static void
cdos_keyboard_settings_changed(CdosKeyboard *self,GSettings *settings,gchar *key){
    CdosKeyboardPrivate *priv = self->priv;
    priv->enable_keyboard = g_settings_get_boolean(priv->a11y_applications_settings,SHOW_KEYBOARD);
    if(CDOS_KEYBOARD_DEBUG){
        priv->enable_keyboard = FALSE;
    }
    if(!priv->enable_keyboard && priv->keyboard == NULL){
        return ;
    }
    gchar *keyboard_type = NULL;
    gchar *keyboard_type_s = g_settings_get_string(priv->keyboard_settings,KEYBOARD_TYPE);
    if(priv->keyboard){
        keyboard_type = caribou_keyboard_model_get_keyboard_type (priv->keyboard);
    }
    if(priv->enable_keyboard && priv->keyboard != NULL && keyboard_type == keyboard_type_s){
        return ;
    }
    if(priv->keyboard != NULL){
        cdos_keyboard_destroy(self);
    }
    if(priv->enable_keyboard){
        gboolean show_keyboard = TRUE;
        if(show_keyboard){
            MetaDisplay *global_display = cdos_global_get_display(global);
            priv->timestamp = meta_display_get_current_time_roundtrip(global_display) + 1;
        }
        cdos_keyboard_setup(self,show_keyboard);
    }else{
        //notice:CdosLayoutManager中暂无hide_keyboard这个接口函数;
        MonkeyPlugin  *main = monkey_plugin_get();
        CdosMessageTray *main_message_tray = cdos_monkey_plugin_get_message_tray(main);
        if(main_message_tray != NULL){
            cdos_message_tray_hide(main_message_tray);
        }
        CdosLayoutManager *main_layout_manager = cdos_monkey_plugin_get_layout_manager(main);
        StBoxLayout *keyboard_box =cdos_layout_manager_get_keyboardbox(main_layout_manager);
        clutter_actor_hide(CLUTTER_ACTOR(keyboard_box));
    }
}

static void
cdos_keyboard_destroy(CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    if(priv->signal_ids->keyboard_notify_active_group_id){
        g_signal_handler_disconnect(priv->keyboard,priv->signal_ids->keyboard_notify_active_group_id);
        priv->signal_ids->keyboard_notify_active_group_id = 0;
    }
    ClutterStage *global_stage = cdos_global_get_stage(global);
    if(priv->signal_ids->global_stage_notify_key_focus_id){
        g_signal_handler_disconnect(global_stage,priv->signal_ids->global_stage_notify_key_focus_id);
        priv->signal_ids->global_stage_notify_key_focus_id = 0;
    }
    g_object_unref(priv->keyboard);
    priv->keyboard = NULL;
    clutter_actor_destroy(CLUTTER_ACTOR(priv->actor));
    priv->actor = NULL;
    cdos_keyboard_destroy_source(self);
}

static void
cdos_keyboard_setup(CdosKeyboard *self,gboolean show){
    CdosKeyboardPrivate *priv = self->priv;
    priv->actor = st_box_layout_new();
    clutter_actor_set_name(CLUTTER_ACTOR(priv->actor),"keyboard");
    clutter_actor_set_reactive(CLUTTER_ACTOR(priv->actor),TRUE);
    st_box_layout_set_vertical(ST_BOX_LAYOUT(priv->actor),TRUE);

    MonkeyPlugin  *main = monkey_plugin_get();
    CdosLayoutManager *main_layout_manager = cdos_monkey_plugin_get_layout_manager(main);
    StBoxLayout *keyboard_box = cdos_layout_manager_get_keyboardbox(main_layout_manager);
    clutter_container_add_actor(CLUTTER_CONTAINER(keyboard_box),CLUTTER_ACTOR(priv->actor));

    //notice:无法调用CdosLayoutManager的track_chrome方法；
    priv->keyboard = caribou_keyboard_model_new();
    //gchar *keyboard_type = g_settings_get_string(priv->keyboard_settings,KEYBOARD_TYPE);
    //g_object_set(priv->keyboard,"keyboard-type",keyboard_type,NULL);

    priv->groups = g_hash_table_new(NULL,NULL);
    priv->current_page = NULL;

    priv->num_of_horiz_keys = 0;
    priv->num_of_vert_keys = 0;
    cdos_keyboard_add_keys(self);

    priv->signal_ids->keyboard_notify_active_group_id = \
            g_signal_connect(G_OBJECT(priv->keyboard),"notify::active-group",G_CALLBACK(cdos_keyboard_on_notify_active_group),self);

    ClutterStage *global_stage = cdos_global_get_stage(global);
    priv->signal_ids->global_stage_notify_key_focus_id = \
            g_signal_connect(G_OBJECT(global_stage),"notify::key-focus",G_CALLBACK(cdos_keyboard_global_stage_on_notify_key_focus),self);

    if(show == TRUE){
        cdos_keyboard_show(self);
    }else{
        cdos_keyboard_create_source(self);
    }
}

static void
cdos_keyboard_add_keys(CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    gint groups_length = 0;
    gchar **groups = caribou_keyboard_model_get_groups (priv->keyboard,&groups_length);
    guint i = 0;
    for(i; i < groups_length; i++){
        gchar *gname = groups[i];
        CaribouGroupModel *group = caribou_keyboard_model_get_group (priv->keyboard,gname);
        g_signal_connect(G_OBJECT(group),"notify::active-level",G_CALLBACK(cdos_keyboard_group_on_notify_active_level),self);
        gint levels_length = 0, j = 0;
        GHashTable *layers = g_hash_table_new(NULL,NULL);
        gchar **levels = caribou_group_model_get_levels (group,&levels_length);
        for(j;j < levels_length; j++){
            gchar *lname = levels[j];
            CaribouLevelModel *level = caribou_group_model_get_level (group,lname);
            StWidget *layout = st_box_layout_new();
            st_widget_set_style_class_name(layout,"keyboard-layout");
            st_box_layout_set_vertical(ST_BOX_LAYOUT(layout),TRUE);
            cdos_keyboard_load_rows(self,level,layout);
            g_hash_table_insert(layers,lname,layout);
            clutter_container_add_actor(CLUTTER_CONTAINER(priv->actor),CLUTTER_ACTOR(layout));
            clutter_container_child_set ((ClutterContainer*) priv->actor, CLUTTER_ACTOR(layout),
                                         "x-fill",FALSE,
                                         NULL);
            clutter_actor_hide(CLUTTER_ACTOR(layout));
        }
        g_hash_table_insert(priv->groups,gname,layers);
    }
    cdos_keyboard_set_active_layer(self);
}

static StWidget*
cdos_keyboard_get_tray_icon(CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    StWidget *tray_button = st_button_new_with_label(_("tray"));
    st_widget_set_style_class_name(tray_button,"keyboard-key");

    priv->signal_ids->tray_button_press_id = \
            g_signal_connect(CLUTTER_ACTOR(tray_button),"button-press-event",G_CALLBACK(cdos_keyboard_tray_button_on_press),self);

    CdosOverview *main_overview = cdos_overview_get ();

    priv->signal_ids->main_overview_showing_id = \
            g_signal_connect(main_overview,"showing",G_CALLBACK(cdos_keyboard_main_overview_on_showing),tray_button);

    priv->signal_ids->main_overview_hiding_id = \
            g_signal_connect(main_overview,"hiding",G_CALLBACK(cdos_keyboard_main_overview_on_hiding),tray_button);

    return tray_button;
}

static void
cdos_keyboard_add_rows(CdosKeyboard *self,CaribouRowModel *row,StWidget *layout){
    CdosKeyboardPrivate *priv = self->priv;
    StWidget *keyboard_row = st_box_layout_new();
    gint keys_length = 0,i = 0;
    CaribouColumnModel **keys = caribou_row_model_get_columns (row,keys_length);
    for(i;i<keys_length;i++){
        CaribouColumnModel *key = keys[i];
        gint children_length = 0,j = 0;
        CaribouIKeyboardObject **children =  caribou_ikeyboard_object_get_children ((CaribouIKeyboardObject*)key,children_length);
        StWidget *right_box = st_box_layout_new();
        StWidget *left_box = st_box_layout_new();
        st_widget_set_style_class_name(right_box,"keyboard-row");
        st_widget_set_style_class_name(left_box,"keyboard-row");

        for(j;j < children_length;j++){
            if(priv->num_of_horiz_keys == 0){
                priv->num_of_horiz_keys = children_length;
            }
            CaribouKeyModel *key = (CaribouKeyModel *)children[i];
            CdosKey *button = cdos_key_new(key);
            gchar *key_align = caribou_key_model_get_align(key);
            gchar *key_name = aribou_key_model_get_name(key);

            StWidget *button_actor = cdos_key_get_actor(button);
            if(g_str_equal(key_align,"right") == TRUE){
                clutter_container_add_actor(CLUTTER_CONTAINER(right_box),CLUTTER_ACTOR(button_actor));
            }else{
                clutter_container_add_actor(CLUTTER_CONTAINER(left_box),CLUTTER_ACTOR(button_actor));
            }
            if(g_str_equal(key_name,"Caribou_Prefs") == TRUE){
                g_signal_connect(key,"key-released",G_CALLBACK(cdos_keyboard_key_on_key_released),self);
                StWidget *tray_icon = cdos_keyboard_get_tray_icon(self);
                clutter_container_add_actor(CLUTTER_CONTAINER(right_box),CLUTTER_ACTOR(tray_icon));
            }
        }
        clutter_container_add_actor(CLUTTER_CONTAINER(keyboard_row),CLUTTER_ACTOR(left_box));
        clutter_container_add_actor(CLUTTER_CONTAINER(keyboard_row),CLUTTER_ACTOR(right_box));
        clutter_container_child_set ((ClutterContainer*) keyboard_row, CLUTTER_ACTOR(left_box),
                                     "x-fill",FALSE,
                                     "x-align",CLUTTER_BIN_ALIGNMENT_START,
                                     "expand",TRUE,
                                     NULL);
        clutter_container_child_set ((ClutterContainer*) keyboard_row, CLUTTER_ACTOR(right_box),
                                     "x-fill",FALSE,
                                     "x-align",CLUTTER_BIN_ALIGNMENT_END,
                                     "expand",TRUE,
                                     NULL);
    }
    clutter_container_add_actor(CLUTTER_CONTAINER(layout),CLUTTER_ACTOR(keyboard_row));
}

static void
cdos_keyboard_load_rows(CdosKeyboard *self,CaribouLevelModel *level,StWidget *layout){
    CdosKeyboardPrivate *priv = self->priv;
    gint rows_length = 0,i = 0;
    CaribouRowModel **rows = caribou_level_model_get_rows (level,&rows_length);
    for(i;i < rows_length;i++){
        CaribouRowModel *row = rows[i];
        if(priv->num_of_vert_keys == 0){
            priv->num_of_vert_keys = rows_length;
        }
        cdos_keyboard_add_rows(self,row,layout);
    }
}

static void
cdos_keyboard_redraw(CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    if(!priv->enable_keyboard){
        return ;
    }
    MonkeyPlugin  *main = monkey_plugin_get();
    CdosLayoutManager *main_layout_manager = cdos_monkey_plugin_get_layout_manager(main);
    MetaRectangle *monitor = cdos_layout_manager_get_primary_monitor (main_layout_manager);
    gfloat max_height = (monitor->height)/3;
    clutter_actor_set_width(CLUTTER_ACTOR(priv->actor),monitor->width);

    StWidget *layout = priv->current_page;
    StThemeNode *layout_node = st_widget_get_theme_node(layout);
    gdouble vertical_spacing = st_theme_node_get_length(layout_node,"spacing");
    gdouble padding = st_theme_node_get_length(layout_node,"padding");

    GList *layout_children = clutter_container_get_children (CLUTTER_CONTAINER(layout));
    ClutterActor *first_child = g_list_first(layout_children)->data;
    GList *first_child_children = clutter_container_get_children (CLUTTER_CONTAINER(first_child));
    ClutterActor *box = g_list_first(first_child_children)->data;

    StThemeNode *box_node = st_widget_get_theme_node(box);
    gdouble horizontal_spacing = st_theme_node_get_length(box_node,"spacing");

    gdouble all_horizontal_spacing = (priv->num_of_horiz_keys - 1) * horizontal_spacing;
    gfloat actor_width = clutter_actor_get_width(CLUTTER_ACTOR(priv->actor));
    gdouble key_width = floor(((double)actor_width - all_horizontal_spacing - 2 * padding)/priv->num_of_horiz_keys);

    gdouble all_vertical_spacing = (priv->num_of_vert_keys - 1) * vertical_spacing;
    gdouble key_height = floor(((double)max_height - all_vertical_spacing - 2 * padding)/priv->num_of_vert_keys);

    gdouble key_size = MIN(key_width,key_height);
    gfloat actor_height = key_size * priv->num_of_vert_keys + all_vertical_spacing + 2 * padding;
    clutter_actor_set_height(CLUTTER_ACTOR(priv->actor),actor_height);

    GList *rows = clutter_container_get_children (CLUTTER_CONTAINER(priv->current_page));
    while (rows != NULL){
        ClutterActor *keyboard_row = rows->data;
        GList *boxes = clutter_container_get_children (CLUTTER_CONTAINER(keyboard_row));
        while (boxes != NULL){
            ClutterActor *box = boxes->data;
            GList *keys = clutter_container_get_children (CLUTTER_CONTAINER(box));
            while(keys != NULL){
                ClutterActor *child = keys->data;
                //gfloat child_width = clutter_actor_get_width(child);
                gdouble *key_width = cdos_key_get_button_key_width(ST_WIDGET(child));
                if((int)key_width == 0){
                    *key_width = 1.0;
                }
                clutter_actor_set_width(child,*key_width * (float)key_size);
                clutter_actor_set_height(child,(float)key_size);
                StWidget *child_s = ST_WIDGET(child);
                if(cdos_key_get_actor_extended_keys(child_s)){
                    GList *extended_keys = clutter_container_get_children (CLUTTER_CONTAINER(child));
                    while(extended_keys != NULL){
                        ClutterActor *extended_key = extended_keys->data;
                        clutter_actor_set_width(extended_key,(float)key_size);
                        clutter_actor_set_height(extended_key,(float)key_size);

                        extended_keys = extended_keys->next;
                    }
                }
                keys = keys->next;
            }
            boxes = boxes->next;
        }
        rows = rows->next;
    }
}

static void
cdos_keyboard_set_active_layer(CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    const gchar *active_group_name = caribou_keyboard_model_get_active_group (priv->keyboard);
    CaribouGroupModel *active_group = caribou_keyboard_model_get_group (priv->keyboard,active_group_name);
    const gchar *active_level = caribou_group_model_get_active_level (active_group);

    GHashTable *layers = g_hash_table_find(priv->groups,cdos_keyboard_groups_GHRFunc,active_group_name);
    if(priv->current_page != NULL){
        clutter_actor_hide(CLUTTER_ACTOR(priv->current_page));
    }

    priv->current_page = g_hash_table_find(layers,cdos_keyboard_layers_GHRFunc,active_level);
    clutter_actor_show(CLUTTER_ACTOR(priv->current_page));
}

static void
cdos_keyboard_create_source(CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    if(priv->source == NULL){
        priv->source = cdos_keyboard_source_new(self);
        cdos_source_set_transient(CDOS_SOURCE(priv->source),TRUE);
        MonkeyPlugin  *main = monkey_plugin_get();
        CdosMessageTray *main_message_tray = cdos_monkey_plugin_get_message_tray(main);
        if(main_message_tray != NULL){
            cdos_message_tray_add(main_message_tray,CDOS_SOURCE(priv->source));
        }
    }
}

static void
cdos_keyboard_destroy_source(CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    if(priv->source != NULL){
        cdos_source_destroy(CDOS_SOURCE(priv->source),DISMISSED);
        g_object_unref(priv->source);
        priv->source = NULL;
    }
}

static void
cdos_keyboard_move_temporarily(CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    MetaScreen *global_screen = cdos_global_get_screen (global);
    MetaDisplay *meta_display = meta_screen_get_display (global_screen);
    MetaWindow *current_window = meta_display_get_focus_window (meta_display);

    MetaRectangle rect;
    meta_window_get_frame_rect(current_window,&rect);
    gint new_x = rect.x;
    gfloat actor_height = clutter_actor_get_height(CLUTTER_ACTOR(priv->actor));
    gint new_y =(int)(3 *  actor_height/2);
    meta_window_move_frame(current_window,TRUE,new_x,new_y);
}

static void
cdos_keyboard_set_location(CdosKeyboard *self,gfloat x,gfloat y){
    CdosKeyboardPrivate *priv = self->priv;
    gfloat actor_height = clutter_actor_get_height(CLUTTER_ACTOR(priv->actor));
    if(y >= 2 * actor_height){
        cdos_keyboard_move_temporarily(self);
    }
}

static void
cdos_keyboard_settings_on_changed(GSettings *settings,gchar *key,CdosKeyboard *self){
    cdos_keyboard_settings_changed(self,settings,key);
}

static void
cdos_keyboard_on_notify_active_group(GObject *keyboard,GParamSpec *pspec,CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    cdos_keyboard_set_active_layer(self);
    cdos_keyboard_redraw(self);
}

static void
cdos_keyboard_global_stage_on_notify_key_focus(GObject *stage,GParamSpec *pspec,CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    ClutterStage *global_stage = CLUTTER_STAGE(stage);
    ClutterActor *focus = NULL;
    g_object_get(global_stage,"key-focus",&focus,NULL);

    if(focus && cdos_key_get_actor_extended_keys(ST_WIDGET(focus)) != NULL ){
        return ;
    }
    guint32 time = cdos_global_get_current_time(global);
    if(CLUTTER_IS_TEXT(CLUTTER_TEXT(focus))){
        cdos_keyboard_Show(self,time);
    }else{
        cdos_keyboard_Hide(self,time);
    }
}
static void
cdos_keyboard_group_on_notify_active_level(GObject *group,GParamSpec *pspec,CdosKeyboard *self){
    cdos_keyboard_set_active_layer(self);
    cdos_keyboard_redraw(self);
}

static gboolean
cdos_keyboard_tray_button_on_press(ClutterActor *actor,ClutterEvent *event,CdosKeyboard *self){

    MonkeyPlugin  *main = monkey_plugin_get();
    CdosMessageTray *main_message_tray = cdos_monkey_plugin_get_message_tray(main);
    if(main_message_tray){
        cdos_message_tray_toggle(main_message_tray);
    }
}

static void
cdos_keyboard_main_overview_on_showing(CdosOverview *overview,StWidget *tray_button){
    clutter_actor_set_reactive(CLUTTER_ACTOR(tray_button),FALSE);
    st_widget_add_style_pseudo_class(tray_button,"grayed");
}

static void
cdos_keyboard_main_overview_on_hiding(CdosOverview *overview,StWidget *tray_button){
    clutter_actor_set_reactive(CLUTTER_ACTOR(tray_button),TRUE);
    st_widget_remove_style_pseudo_class(tray_button,"grayed");
}

static void
cdos_keyboard_key_on_key_released(CaribouIKeyboardObject* _sender, CaribouKeyModel* key, CdosKeyboard *self){
    cdos_keyboard_hide(self);
}

static gboolean cdos_keyboard_groups_GHRFunc(gchar *key,GHashTable *value,gchar *user_data){
    if(key == user_data){
        return  TRUE;
    }
    return FALSE;
}

static gboolean cdos_keyboard_layers_GHRFunc(gchar *key,StWidget *value,gchar *user_data){
    if(key == user_data){
        return  TRUE;
    }
    return FALSE;
}
/*-------------public functions--------------*/

/**
 * cdos_keyboard_new:
 *
 * 创建#CdosKeyboard实例化对象。
 *
 * Returns:实例化对象
 */
CdosKeyboard* cdos_keyboard_new(void)
{
    CdosKeyboard *self;
    self = g_object_new(CDOS_TYPE_KEYBOARD, NULL);
    cdos_keyboard_settings_changed(self,NULL,NULL);

    return self;
}

/**
 * cdos_keyboard_start:
 * @self: 当前实例对象
 *
 * 当前实例对象构造完成之后，在特定场景下，做了部分的补充，该方法通常在实例化之后被调用。
 */
void
cdos_keyboard_start(CdosKeyboard *self){
    cdos_keyboard_redraw(self);
}

/**
 * cdos_keyboard_show:
 * @self:当前实例对象
 *
 * 显示当前桌面的虚拟键盘。
 */
void
cdos_keyboard_show(CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    MonkeyPlugin  *main = monkey_plugin_get();
    CdosLayoutManager *main_layout_mananger = cdos_monkey_plugin_get_layout_manager(main);
    StBoxLayout *keyboard_box = cdos_layout_manager_get_keyboardbox(main_layout_mananger);
    CdosMessageTray *main_message_tray = cdos_monkey_plugin_get_message_tray(main);
    if(main_message_tray != NULL){
        cdos_message_tray_hide(main_message_tray);
    }
    clutter_actor_raise_top(CLUTTER_ACTOR(keyboard_box));
    clutter_actor_show(CLUTTER_ACTOR(keyboard_box));

}

/**
 * cdos_keyboard_hide:
 * @self:当前实例对象
 *
 * 隐藏当前桌面的虚拟键盘。
 */
void
cdos_keyboard_hide(CdosKeyboard *self){
    CdosKeyboardPrivate *priv = self->priv;
    MonkeyPlugin  *main = monkey_plugin_get();
    CdosLayoutManager *main_layout_mananger = cdos_monkey_plugin_get_layout_manager(main);
    StBoxLayout *keyboard_box = cdos_layout_manager_get_keyboardbox(main_layout_mananger);
    CdosMessageTray *main_message_tray = cdos_monkey_plugin_get_message_tray(main);
    if(main_message_tray != NULL){
        cdos_message_tray_hide(main_message_tray);
    }
    clutter_actor_hide(CLUTTER_ACTOR(keyboard_box));
}

/**
 * cdos_keyboard_set_cursor_location:
 * @self:当前实例对象
 * @x:
 * @y:
 * @w:
 * @h:
 *
 * DBUS专用接口函数。
 */
void
cdos_keyboard_set_cursor_location(CdosKeyboard *self,gfloat x,gfloat y,gfloat w,gfloat h){
    CdosKeyboardPrivate *priv = self->priv;
    if(!priv->enable_keyboard){
        return ;
    }
}

/**
 * cdos_keyboard_set_entry_location:
 * @self:当前实例对象
 * @x:
 * @y:
 * @w:
 * @h:
 *
 *
 * DBUS专用接口函数。
 */
void
cdos_keyboard_set_entry_location(CdosKeyboard *self,gfloat x,gfloat y,gfloat w,gfloat h){
    CdosKeyboardPrivate *priv = self->priv;
    if(!priv->enable_keyboard){
        return ;
    }
}

/**
 * cdos_keyboard_Show:
 * @self:实例对象
 * @timestamp:时间戳
 *
 * 通过DBUS，显示当前桌面的虚拟键盘。
 *
 * DBUS专用接口函数。
 *
 */
void
cdos_keyboard_Show(CdosKeyboard *self,guint32 timestamp){
    CdosKeyboardPrivate *priv = self->priv;
    if(!priv->enable_keyboard){
        return ;
    }
    if(timestamp - priv->timestamp < 0){
        return ;
    }
    priv->timestamp = timestamp;
    cdos_keyboard_show(self);
}

/**
 * cdos_keyboard_Hide:
 * @self:实例对象
 * @timestamp:时间戳
 *
 * 通过DBUS，隐藏当前桌面的虚拟键盘。
 *
 * DBUS专用接口函数。
 *
 */
void
cdos_keyboard_Hide(CdosKeyboard *self,guint32 timestamp){
    CdosKeyboardPrivate *priv = self->priv;
    if(!priv->enable_keyboard){
        return ;
    }
    if(timestamp - priv->timestamp < 0){
        return ;
    }
    priv->timestamp = timestamp;
    cdos_keyboard_hide(self);
}
