#include "cdos.h"
#include "cdos-start.h"
#include "config.h"

/**
 * SECTION:cdos-layout-manager
 * @Title: CdosLayoutManager
 * @Short_description: 桌面布局管理器
 * 
 * CdosLayoutManager 负责桌面上所有组件的显示与输入区域更新，比如面板、弹出菜单等，
 * 桌面上有控件显示，如果这个控件需要与用户交互，必须要获取这个控件的起始位置和宽高，
 * 将它添加到 CdosLayoutManager 的 chrome 中，并实时更新输入区域。
 */
G_DEFINE_TYPE (CdosLayoutManager, cdos_layout_manager, G_TYPE_OBJECT);

#define DESKTOPS_ON_PRIMARY "true::false"
#define DESKTOPS_ON_ALL "true::true"
#define DESKTOPS_ON_NON_PRIMARY  "false::true"
#define DESKTOPS_ON_NONE  "false::false"
#define DESKTOPS_DEFAULT	DESKTOPS_ON_PRIMARY

#define META_LAYER_FULLSCREEN	       5
/* 下面这个宏可以用来访问新建对象的私有结构*/
#define CDOS_LAYOUT_MANAGER_GET_PRIVATE(obj) \
        (G_TYPE_INSTANCE_GET_PRIVATE ((obj), CDOS_TYPE_LAYOUT_MANAGER, CdosLayoutManagerPrivate))


typedef struct _actorStruct
{
	gulong visibleId;
	gulong allocationId;
	gboolean isToplevel;
	ClutterActor *actor;
    gint flags;
    /*
    gint track_fullscreen;
    gint affect_struts;
    gint affect_input_region;
    */
}ActorStruct;

struct _CdosLayoutManagerPrivate
{
	guint primaryindex;
    guint bottomindex;
    guint _updateRegionIdle;
	StBoxLayout *panelbox;
	StBoxLayout *keyboardbox;
	MetaRectangle *primarymonitor;
	GList *_trackedActors;	
	GPtrArray *monitors;
    gboolean *fullscreenMonitors;
};

enum {
  MONITORS_CHANGED,
  LAST_SIGNAL
};

static guint cdos_layout_manager_signals[LAST_SIGNAL] = { 0 };

static CdosLayoutManager *layoutmanager = NULL;

static void cdos_layout_manager_queue_update_regions (ClutterActor *actor, gpointer unused, gpointer data);

static void cdos_layout_manager_update_fullscreen (CdosLayoutManager *self);

static void cdos_layout_manager_set_panelbox (CdosLayoutManager    *self, StBoxLayout *panelbox);

static void cdos_layout_manager_set_keyboardbox (CdosLayoutManager    *self, StBoxLayout *keyboardbox);

static void
cdos_layout_manager_dispose (GObject *gobject) {

    CdosLayoutManager *self = CDOS_LAYOUT_MANAGER (gobject);
    CdosLayoutManagerPrivate *priv = self->priv;

    cdos_layout_manager_remove_chrome (self, (ClutterActor *)priv->panelbox);

    g_object_unref (priv->panelbox);
    g_object_unref (priv->keyboardbox);

    GList *actors = NULL;
    actors = priv->_trackedActors;
    while(actors != NULL){
        GList *next = actors->next;
        ActorStruct *tmp;
        tmp = actors->data;
        //g_object_unref(tmp->actor);
        g_free (tmp); // every element
        actors = next;
    }
    g_list_free(priv->_trackedActors);
    g_ptr_array_free (priv->monitors, 1);

    G_OBJECT_CLASS (cdos_layout_manager_parent_class)->dispose (gobject);
}

static void
cdos_layout_manager_finalize (GObject *gobject)
{
    CdosLayoutManager *self = CDOS_LAYOUT_MANAGER (gobject);
  CdosLayoutManagerPrivate *priv = self->priv;

  g_free (priv->primarymonitor);
    g_free (priv->fullscreenMonitors);
  /* 调用父类的 finalize() 方法 */
  G_OBJECT_CLASS (cdos_layout_manager_parent_class)->finalize (gobject);
}

static void
cdos_layout_manager_class_init (CdosLayoutManagerClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);// 自己的父类型是什么？
  
  gobject_class->finalize = cdos_layout_manager_finalize;
  gobject_class->dispose = cdos_layout_manager_dispose;

  g_type_class_add_private (klass, sizeof (CdosLayoutManagerPrivate));

   /**
   * CdosLayoutManager::monitors-changed:
   * @actor: 接受这个信号的对象
   *
   * CdosLayoutManager 监听桌面 MetaScreen 的 monitors-changed 信号并转发
   *
   * Since: 1.0.0
   */
  cdos_layout_manager_signals[MONITORS_CHANGED] = g_signal_new ("monitors-changed",
                                     CDOS_TYPE_LAYOUT_MANAGER,
                                     G_SIGNAL_RUN_LAST,
                                     0,
                                     NULL, NULL,
                                     g_cclosure_marshal_VOID__VOID,
                                     G_TYPE_NONE, 0);
  
} 

static gboolean 
is_above_or_below_primary (CdosLayoutManager *self, MetaRectangle *monitor)
{
    CdosLayoutManagerPrivate *priv = self->priv;
    MetaRectangle *primary = priv->monitors->pdata[priv->primaryindex];
    int m_left = monitor->x;
    int m_right = monitor->x + monitor->width;
    int p_left  = primary->x;
    int p_right = primary->x + primary->width;

    if ((m_left >= p_left && m_left < p_right) ||
        (m_right > p_left && m_right <= p_right) ||
        (p_left >= m_left && p_left < m_right) ||
        (p_right > m_left && p_right <= m_right))
        return TRUE;
    return FALSE;
}

static gint
find_monitor_for_rect (CdosLayoutManager *self, guint x, guint y,
                        guint w, guint h){
    guint centerX = x + w/2;
    guint centerY = y + h/2;

    GPtrArray *monitors = self->priv->monitors;

    gint i = 0;
    while (monitors != NULL && i < monitors->len){
        //GList *next = monitors->next;
        MetaRectangle *tmprect = monitors->pdata[i];

        if (centerX >= tmprect->x && centerX < tmprect->x + tmprect->width &&
            centerY >= tmprect->y && centerY < tmprect->y + tmprect->height)
            return i;

        //monitors = next;
        i++;
    }
    i = 0;
    while (monitors != NULL && i < monitors->len){
        //GList *next = monitors->next;
        MetaRectangle *tmprect = monitors->pdata[i];

        if (x + w > tmprect->x && x < tmprect->x + tmprect->width &&
            y + h > tmprect->y && y < tmprect->y + tmprect->height)
            return i;
        i++;
        //monitors = next;
    }

    return -1;
}

static gint
find_monitor_for_window (CdosLayoutManager *self, MetaWindowActor *window ){
    gfloat x, y, w, h;
    clutter_actor_get_size ((ClutterActor *)window, &w, &h);
    clutter_actor_get_position ((ClutterActor *)window, &x, &y);
    return find_monitor_for_rect (self, x, y, w, h);
}

static gint
find_monitor_for_actor (CdosLayoutManager *self, ClutterActor *actor)
{
    gfloat x, y, w, h;
    clutter_actor_get_transformed_size (actor, &w, &h);
    clutter_actor_get_transformed_position (actor, &x, &y);
    return find_monitor_for_rect (self, x, y, w, h);
}

static void
cdos_layout_manager_updateMonitors(CdosLayoutManager *self){
	MetaScreen *screen;
    MetaRectangle *rect;
	CdosGlobal *global;
	int nMonitors = 0;
	int i;
	CdosLayoutManagerPrivate *priv;

    priv = self->priv;

	global = cdos_global_get();
	screen = cdos_global_get_screen(global);
	nMonitors = meta_screen_get_n_monitors(screen);

    if (priv->monitors)
        g_ptr_array_free (priv->monitors, 1);
    priv->monitors = g_ptr_array_new_full (nMonitors, g_free);
    priv->monitors->len = nMonitors;
	for(i = 0; i< nMonitors; i++){
        rect = g_malloc0(sizeof(MetaRectangle));
		meta_screen_get_monitor_geometry(screen, i, rect);
		//g_ptr_array_add(priv->monitors, rect);
        priv->monitors->pdata[i] = rect;
	}

	if (nMonitors == 1) {
        priv->primaryindex = 0;
        priv->bottomindex = 0;
    } else {
        priv->primaryindex = meta_screen_get_primary_monitor (screen);
        priv->bottomindex = priv->primaryindex;
	    for(i = 0; i< nMonitors; i++){
		    MetaRectangle *monitor = priv->monitors->pdata[i];
            if (is_above_or_below_primary (self, monitor))
            {
                if (monitor->y > ((MetaRectangle *)priv->monitors->pdata[priv->bottomindex])->y)
                    priv->bottomindex = i;
            }
	    }
	}
    priv->primarymonitor = priv->monitors->pdata[priv->primaryindex];
}

/**
 * cdos_layout_manager_update_boxes:
 * @self: #CdosLayoutManager 对象
 *
 * 这个函数用于实时更新显示，例如分辨率发生变化的时候，更改 panel 的显示位置等，
 * 更新桌面上所有控件的接受输入的状态
 */
void 
cdos_layout_manager_update_boxes (CdosLayoutManager *self)
{
    g_return_if_fail(CDOS_IS_LAYOUT_MANAGER(self));
    CdosLayoutManagerPrivate *priv = self->priv;

    ClutterActor *panel = cdos_panel;
    gfloat p1_height = clutter_actor_get_height (panel);
    MetaRectangle *bottom = priv->monitors->pdata[priv->bottomindex];

    /*
    cdos_debug ("%d %d %d %d %f\n",
            priv->primarymonitor->width, priv->primarymonitor->height,
            bottom->width, bottom->height, p1_height);
    */

    ClutterActor *panelbox = (ClutterActor *)priv->panelbox;
    if (g_strcmp0 (desktop_layout, LAYOUT_TRADITIONAL) == 0)
    {
        clutter_actor_set_size (panelbox, (gfloat)bottom->width, p1_height);
        clutter_actor_set_position (panelbox, 
                (gfloat)bottom->x, (gfloat)bottom->y + (gfloat)bottom->height - p1_height);
    }
    else if (g_strcmp0 (desktop_layout, LAYOUT_FLIPPED) == 0)
    {
        clutter_actor_set_size (panelbox, (gfloat)priv->primarymonitor->width, p1_height);
        clutter_actor_set_position (panelbox, 
                (gfloat)priv->primarymonitor->x, (gfloat)priv->primarymonitor->y + (gfloat)priv->primarymonitor->height - p1_height);
    }
    else if (g_strcmp0 (desktop_layout, LAYOUT_CLASSIC) == 0)
    {
        clutter_actor_set_size (panelbox, (gfloat)priv->primarymonitor->width, p1_height);
        clutter_actor_set_position (panelbox, 
                (gfloat)priv->primarymonitor->x, (gfloat)priv->primarymonitor->y + (gfloat)priv->primarymonitor->height - p1_height);
    }

    cdos_layout_manager_queue_update_regions (NULL, NULL, self);
}

static void 
cdos_layout_manager_update_visibility (CdosLayoutManager *self)
{
    CdosLayoutManagerPrivate *priv = self->priv;
    GList *iter = NULL;
    for (iter = priv->_trackedActors; iter; iter = iter->next)
    {
        ActorStruct *actorData = iter->data;
        //if (!actorData->isToplevel)
        //    continue;
        if (!actorData->flags & TRACK_FULLSCREEN)
            continue;
        gboolean visible = TRUE;
        /* FIXME screenSaverActive, inOverview */

        CdosGlobal *global = cdos_global_get ();
        ClutterActor *window_group = cdos_global_get_window_group (global);
        if (clutter_actor_is_visible (window_group))
        {
            gint m = find_monitor_for_actor (self, actorData->actor);
            if (m >= 0 && priv->fullscreenMonitors && priv->fullscreenMonitors[m])
                visible = FALSE;


        }
        cdos_generic_container_set_skip_paint ((CdosGenericContainer *)ui_group, actorData->actor, !visible);
    }
}

void
cdos_layout_manager_on_window_restacked (MetaScreen *screen, gpointer data){

    CdosLayoutManager *self = CDOS_LAYOUT_MANAGER (data);
    g_return_if_fail(CDOS_IS_LAYOUT_MANAGER(self));
    CdosLayoutManagerPrivate *priv = self->priv;
    if (!priv->fullscreenMonitors)
        return;

    gboolean *wasInFullscreen = g_new0 (gboolean, priv->monitors->len);
    gint i;
    for (i = 0; i < priv->monitors->len; i++)
        wasInFullscreen[i] = priv->fullscreenMonitors[i];

    // xunli cdos_layout_manager_update_fullscreen (self);

    gboolean changed = FALSE;
    for (i = 0; i < priv->monitors->len; i++)
    {
        if (wasInFullscreen[i] != priv->fullscreenMonitors[i])
        {
            changed = TRUE;
            break;
        }
    }
    if (changed)
    {
        cdos_layout_manager_update_visibility (self);
        cdos_layout_manager_queue_update_regions (NULL, NULL, self);
    }
    g_free (wasInFullscreen);
}

static void 
cdos_layout_manager_relayout (CdosLayoutManager *self)
{
    cdos_layout_manager_update_fullscreen (self);
    cdos_layout_manager_update_visibility (self);
    cdos_layout_manager_queue_update_regions (NULL, NULL, self);
}
/* 
 * leit
 * 处理 screen 的 monitors-changed 信号
 */
static void 
on_monitors_changed (MetaScreen *screen, 
                     CdosLayoutManager *self)
{
    cdos_layout_manager_updateMonitors (self);
    cdos_layout_manager_update_boxes (self);

    gint current_monitor = meta_screen_get_current_monitor (screen);
    gint primary_monitor = meta_screen_get_primary_monitor (screen);
    gint nMonitors = meta_screen_get_n_monitors(screen);

    g_print ("layout monitor: %d, current monitor: %d, primary monitor: %d\n", 
		    nMonitors, current_monitor, primary_monitor);

    GSettings *nemo_settings = g_settings_new ("org.nemo.desktop");
   
    if (nMonitors > 1){
    	switch (primary_monitor){
		case 0:
			g_settings_set_string (nemo_settings, "desktop-layout", DESKTOPS_DEFAULT);
			break;
		case 1:
			g_settings_set_string (nemo_settings, "desktop-layout", DESKTOPS_ON_NON_PRIMARY);
			break;
		default:
			g_settings_set_string (nemo_settings, "desktop-layout", DESKTOPS_DEFAULT);
			break;

	}
    }else {
		g_settings_set_string (nemo_settings, "desktop-layout", DESKTOPS_DEFAULT);
    }
    //	
    //cdos_layout_manager_update_hotcorners (self);
    //cdos_layout_manager_relayout (self);
    g_signal_emit (self, cdos_layout_manager_signals[MONITORS_CHANGED], 0);
}

static void 
update_panel_barriers (ClutterActor *panelbox, 
                       CdosLayoutManager *self)
{

}

static void
cdos_layout_manager_init(CdosLayoutManager *self){
	CdosLayoutManagerPrivate *priv;
    CdosGlobal *global = cdos_global_get ();
    MetaScreen *screen = cdos_global_get_screen (global);

	priv = self->priv = CDOS_LAYOUT_MANAGER_GET_PRIVATE (self);
	StBoxLayout *panelbox = (StBoxLayout *)st_box_layout_new();
	StBoxLayout *keyboardbox = (StBoxLayout *)st_box_layout_new();

    clutter_actor_set_name ((ClutterActor *)panelbox, "panel");
	cdos_layout_manager_set_panelbox(self, panelbox);
	cdos_layout_manager_set_keyboardbox(self, keyboardbox);
	priv->primaryindex = 0;	
    priv->_updateRegionIdle = 0;
	priv->_trackedActors = NULL;
	//MetaRectangle *rect;
	//rect = g_malloc0(sizeof(MetaRectangle));

    g_signal_connect (screen, "restacked", G_CALLBACK (cdos_layout_manager_on_window_restacked), self);
    g_signal_connect (screen, "monitors-changed", G_CALLBACK(on_monitors_changed), self);
	cdos_layout_manager_updateMonitors(self);
    cdos_layout_manager_add_chrome (self, (ClutterActor *)panelbox, 0x7); 
    cdos_layout_manager_relayout (self);
    g_signal_connect (panelbox, "allocation-changed", G_CALLBACK (update_panel_barriers), self);
}

/**
 * cdos_layout_manager_set_panelbox:
 * @self: #CdosLayoutManager 对象
 * @panelbox: #StBoxLayout 对象，用于放置面板
 *
 * 设置 CdosLayoutManager 对象的面板属性
 *
 * Returns: none.
 */
static void
cdos_layout_manager_set_panelbox (CdosLayoutManager    *self,
                    StBoxLayout *panelbox)
{
    CdosLayoutManagerPrivate *priv;
    priv = self->priv;
    if(panelbox){
        priv->panelbox = panelbox;
    }
}

/**
 * cdos_layout_manager_get_panelbox:
 * @self: #CdosLayoutManager 对象
 *
 * 获取 CdosLayoutManager 对象的 panelBox 属性
 *
 * Returns: (transfer full): 返回 #StBoxLayout 对象.
 */
StBoxLayout *
cdos_layout_manager_get_panelbox (CdosLayoutManager *self)
{
  // 同上
  g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);

  return self->priv->panelbox;
}

/**
 * cdos_layout_manager_set_keyboardbox:
 * @self: #CdosLayoutManager 对象
 * @keyboardbox: 一个 #StBoxLayout 对象
 *
 * 设置 CdosLayoutManager 对象的 keyboardBox 属性
 *
 * Returns: none.
 */
static void
cdos_layout_manager_set_keyboardbox (CdosLayoutManager    *self,
                    StBoxLayout *keyboardbox)
{
  CdosLayoutManagerPrivate *priv;

  g_return_if_fail (CDOS_IS_LAYOUT_MANAGER (self));

  priv = self->priv;
    if(keyboardbox){
        priv->keyboardbox = keyboardbox;
    }

}

/**
 * cdos_layout_manager_get_keyboardbox:
 * @self: #CdosLayoutManager 对象
 *
 * 获取 CdosLayoutManager的 keyboardBox 属性
 *
 * Returns: (transfer full): 返回 #StBoxLayout.
 */
StBoxLayout *
cdos_layout_manager_get_keyboardbox (CdosLayoutManager *self)
{
  // 同上
  g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);

  return self->priv->keyboardbox;
}

/**
 * cdos_layout_manager_get_primary_monitor:
 * @self: #CdosLayoutManager 对象
 *
 * 获取 CdosLayoutManager 的屏幕属性
 *
 * Returns: (transfer full): 返回 #MetaRectangle 结构数据
 */
const MetaRectangle *
cdos_layout_manager_get_primary_monitor (CdosLayoutManager *self)
{
  // 同上
  CdosLayoutManagerPrivate *priv;
  g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);
  
  priv = self->priv;
  return priv->primarymonitor;
} 

const MetaRectangle *
cdos_layout_manager_find_monitor_for_rect (CdosLayoutManager *self,
                                           int x, int y, int width, int height)
{
    CdosLayoutManagerPrivate *priv;
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);
    priv = self->priv;
    GPtrArray *monitors = priv->monitors;
    for (int i = 0; i < monitors->len; i++)
    {
        MetaRectangle *rect = monitors->pdata[i];
        if (rect->x <= x && rect->y <= y &&
            rect->x + rect->width >x &&
            rect->y + rect->height > y)
            return rect;
    }
    return NULL;
}

const MetaRectangle *
cdos_layout_manager_find_monitor_for_actor (CdosLayoutManager *self,
                                            ClutterActor *actor)
{
    CdosLayoutManagerPrivate *priv;
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);
    priv = self->priv;
    GPtrArray *monitors = priv->monitors;
    gint i = find_monitor_for_actor (self, actor);
    if (i >= 0)
        return monitors->pdata[i];
    return NULL;
}

static guint 
cdos_layout_manager_get_monitor_index (CdosLayoutManager *self){
	CdosLayoutManagerPrivate *priv;

    CdosGlobal *global;

	MetaRectangle rect;
	GPtrArray *monitors;

    priv = self->priv;	
    monitors = priv->monitors;
    //GPtrArray *recorder = monitors;

    global = cdos_global_get();
	MetaDisplay *display = cdos_global_get_display (global);
	if (display){
		MetaWindow *focus_window = meta_display_get_focus_window (display);

		if (focus_window){
            meta_window_get_frame_rect (focus_window, &rect);
            int idx = 0;
			while(idx < monitors->len){
                //GList *next = monitors->next;
		        MetaRectangle *tmprect;	
		        tmprect = monitors->pdata[idx];

                if (tmprect->x <= rect.x && tmprect->y <= rect.y &&
                                  tmprect->x + tmprect->width > rect.x &&
                                  tmprect->y + tmprect->height > rect.y){
                    return idx;//g_list_index (recorder, tmprect);
                }
                idx++;
  			}
		}
	}
    return priv->primaryindex;
}

MetaRectangle *
cdos_layout_manager_focus_monitor (CdosLayoutManager *self){
	CdosLayoutManagerPrivate *priv;
	g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);

	priv = self->priv;
    
    guint index = cdos_layout_manager_get_monitor_index (self);
	
    return priv->monitors->pdata[index];//g_list_nth_data (priv->monitors, index);
}


/*
 *  fullscreen 
 */
static void
cdos_layout_manager_update_fullscreen (CdosLayoutManager *self)
{
    CdosLayoutManagerPrivate *priv = self->priv;
    MetaScreen *screen = cdos_global_get_screen (global);
    MetaWorkspace *workspace = meta_screen_get_active_workspace (screen);

    GList *windows =  NULL;
    if (workspace){
    	windows = get_window_actors_for_workspace (workspace);
    }
    /* @windows is sorted bottom to top */
    if (windows == NULL){
    	return;
    }

    windows = g_list_reverse (windows);

    GPtrArray *monitors = self->priv->monitors;

    g_free (priv->fullscreenMonitors);

    priv->fullscreenMonitors = g_new0 (gboolean, monitors->len);
    gint i = 0;

    GList *iter;
    for (iter = windows; iter != NULL; iter = iter->next){
        MetaWindowActor *window = iter->data;
        MetaWindow *mwin = meta_window_actor_get_meta_window (window);
        if (!meta_window_showing_on_its_workspace (mwin)){
            continue;
        }
        MetaStackLayer layer = META_LAYER_FULLSCREEN;
        if ((meta_window_get_layer (mwin) == layer) || meta_window_is_fullscreen (mwin)){
            gint monitor = find_monitor_for_window (self, window);
            if (monitor != -1) {
                priv->fullscreenMonitors[monitor] = TRUE;
            }
        }
        
        if (meta_window_is_override_redirect (mwin)) {
            gfloat x, y, w, h;
            clutter_actor_get_position ((ClutterActor *)window, &x, &y);
            clutter_actor_get_size ((ClutterActor *)window, &w, &h);

            gint sw, sh;
            cdos_global_get_screen_size (global, &sw, &sh);
            gboolean isScreenSized = 
                (x == 0.f && y == 0.f &&
                 (gint)w == sw && (gint)h == sh);
            if (isScreenSized) {
                gint i;
                for (i = 0; i < monitors->len; i++)
                    priv->fullscreenMonitors[i] = FALSE;
            }else {
                gint m = find_monitor_for_window (self, window);
                if (m >= 0) {
                    MetaRectangle *monitor = monitors->pdata[i];
                    if ((gint)x <= monitor->x && 
                        (gint)(x + w) >= monitor->x + monitor->width &&
                        (gint)y <= monitor->y &&
                        (gint)(y + h) >= monitor->y + monitor->height)
                        priv->fullscreenMonitors[i] = TRUE;
                }
            }
        } else 
            break;
    }
    g_list_free (windows);
}

/**
 * cdos_layout_manager_get_monitors:
 * @self: a #CdosLayoutManager
 *
 * Get all monitors
 *
 * Return value: (transfer container) (element-type MetaRectangle): the monitor's list.
 */
GPtrArray *
cdos_layout_manager_get_monitors (CdosLayoutManager *self)
{
  // 同上
  g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), NULL);

  CdosLayoutManagerPrivate *priv = self->priv;
  return priv->monitors;
  /*
  gint i;
  GList *ret = NULL;
  for (i = 0; i < priv->monitors->len; i++)
      ret = g_list_prepend (ret, priv->monitors->pdata[i]);
  return ret;
  */
} 

/**
 * cdos_layout_manager_get_index:
 * @self: a #CdosLayoutManager
 *
 * Get index property
 *
 * Returns: index
 */
gint 
cdos_layout_manager_get_index ()
{
    CdosGlobal *global = cdos_global_get ();
    MetaScreen *screen = cdos_global_get_screen (global);
    gint index = meta_screen_get_current_monitor (screen);
    return index;
}

static void 
cdos_layout_manager_queue_update_regions (ClutterActor *actor, gpointer unused, gpointer data){
    CdosLayoutManager *self = CDOS_LAYOUT_MANAGER (data);
    CdosLayoutManagerPrivate *priv = CDOS_LAYOUT_MANAGER_GET_PRIVATE (self);
    
    if (priv->_updateRegionIdle == 0){
        //priv->_updateRegionIdle = meta_later_add (META_LATER_BEFORE_REDRAW, (GSourceFunc)cdos_layout_manager_update_regions, self, NULL);
        priv->_updateRegionIdle = g_idle_add ((GSourceFunc)cdos_layout_manager_update_regions, self);
    }
    
}

/**
 * cdos_layout_manager_add_chrome:
 * @self:  #CdosLayoutManager 对象
 * @actor: #ClutterActor 对象
 *
 * 将 #ClutterActor 对象添加到显示容器 uiGroup 之中，并修改它的的 input region，
 * 只有给这个对象设置了合适的输入区域后，这个 actor 才会对鼠标等输入事件做响应
 *
 * Returns: none
 */
void 
cdos_layout_manager_add_chrome(CdosLayoutManager *self, 
                               ClutterActor *actor,
                               gint flags)
{
    g_return_if_fail (CDOS_IS_LAYOUT_MANAGER (self));
    CdosLayoutManagerPrivate *priv= CDOS_LAYOUT_MANAGER_GET_PRIVATE(self);
    ActorStruct *actorData;
    gulong handleone, handletwo;
    ClutterActor *uiGroup;

    //FIXME ADD_TO_UI_GROUP 
    GList *iter;
    for (iter = priv->_trackedActors; iter; iter = iter->next)
    {
        ActorStruct *strut = iter->data;
        if (strut->actor == actor)
        {
            g_warning ("trying to re-track existing ClutterActor %p\n", actor);
            return;
        }
    }
    uiGroup = monkey_plugin_get_ui_group (monkey_plugin_get ());
    ClutterActor *parent = clutter_actor_get_parent (actor);
    if (!parent)
        clutter_container_add_actor ((ClutterContainer *) uiGroup, (ClutterActor *)actor);

    actorData = g_malloc0(sizeof(ActorStruct));
    handleone = g_signal_connect(actor, "notify::visible", G_CALLBACK(cdos_layout_manager_queue_update_regions), self);
    handletwo = g_signal_connect(actor, "notify::allocation", G_CALLBACK(cdos_layout_manager_queue_update_regions), self);
    actorData->visibleId = handleone;
    actorData->allocationId = handletwo;
    actorData->actor = actor;
    actorData->flags = flags;
    priv->_trackedActors = g_list_append(priv->_trackedActors, actorData);
    cdos_layout_manager_queue_update_regions (actor, NULL, self);
}


/**
 * cdos_layout_manager_remove_chrome:
 * @self: #CdosLayoutManager 对象
 * @actor: #ClutterActor 对象
 *
 * 当这个 actor 不需要再现实，将它从 chrome 中移除，不再接受输入事件
 *
 * Returns: none
 */
void
cdos_layout_manager_remove_chrome(CdosLayoutManager *self,ClutterActor *actor){
    g_return_if_fail (CDOS_IS_LAYOUT_MANAGER (self));

	GList *actors = NULL;
	CdosLayoutManagerPrivate * priv;
//    ClutterActor *uiGroup = NULL;

    priv = self->priv = CDOS_LAYOUT_MANAGER_GET_PRIVATE (self);
	actors = priv->_trackedActors;
    GList *recorder = priv->_trackedActors;

	while(actors != NULL){
        ActorStruct *dataitem = actors->data;
		if(dataitem->actor == actor){
            recorder = g_list_remove (recorder, dataitem);
            g_signal_handler_disconnect (dataitem->actor, dataitem->visibleId);
            g_signal_handler_disconnect (dataitem->actor, dataitem->allocationId);
            g_free (dataitem);
            break ;
		}
        actors = actors->next;
	}

	priv->_trackedActors = recorder;
	cdos_layout_manager_queue_update_regions(actor,  NULL, self);
}

/**
 * cdos_layout_manager_update_regions:
 * @self: a #CdosLayoutManager
 *
 * 设置输入区域，这里先获取所有添加到 #CdosLayoutManager 中的 actors，再一一设置
 *
 * Returns: success
 */
gboolean
cdos_layout_manager_update_regions(CdosLayoutManager *self)
{
    g_return_val_if_fail (CDOS_IS_LAYOUT_MANAGER (self), FALSE);
	CdosLayoutManagerPrivate * priv;
	priv = self->priv;

	MetaRectangle *rect = priv->primarymonitor;
	if(!rect)
		return FALSE;
	GList *actors = NULL;
	GSList *rects = NULL; // FOR INPUT REGION SET
    GSList *struts = NULL;
	MetaRectangle *newRect;

    if (priv->_updateRegionIdle != 0ul){
        g_source_remove (priv->_updateRegionIdle);
        priv->_updateRegionIdle = 0ul;
    }

    actors = priv->_trackedActors;
    while(actors != NULL){
        GList *next = actors->next;
        ActorStruct *dataitem;
        dataitem = actors->data;
        actors = next;
        gint affect_struts = dataitem->flags & AFFECT_STRUTS;
        gint affect_input_region = dataitem->flags & AFFECT_INPUT_REGION;
        if (!affect_input_region && !affect_struts)
            continue;

        gfloat box_x, box_y, width, height;
        gint x, y, w, h;
        gint x1, x2, y1, y2;

        clutter_actor_get_transformed_position (dataitem->actor, &box_x, &box_y);
        clutter_actor_get_transformed_size (dataitem->actor, &width, &height);
        x = round(box_x);
        y = round(box_y);
        w = round(width);
        h = round(height);

        newRect = g_malloc0(sizeof(MetaRectangle));
        newRect->x = x;
        newRect->y = y;
        newRect->width = w;
        newRect->height = h;

        //Limit struts to the size of the screen
        gint screen_width, screen_height;
        MetaScreen *screen = cdos_global_get_screen (cdos_global_get());
        meta_screen_get_size (screen, &screen_width, &screen_height);

        // linjj begin: 当有popup menu显示的时候，指定整个屏幕可点击，否则popupmenumanager中无法收到
        // stage的captured-event信号。
        /* leit 不需要这部分
        gpointer delegate_obj = obj_get_delegate(dataitem->actor);

        if ((NULL != delegate_obj) && CDOS_IS_POPUP_MENU_BASE(delegate_obj))
        {
            newRect->x = 0;
            newRect->y = 0;
            newRect->width = screen_width;
            newRect->height = screen_height;
        }
        */
        rects = g_slist_append(rects, newRect); 
        // linjj end
        if (affect_struts)
        {
            x1 = MAX (x, 0);
            x2 = MIN (x+w, screen_width);
            y1 = MAX (y, 0);
            y2 = MIN (y+h, screen_height);

            MetaSide side;
            if ( x1 <= rect->x && x2 >= rect->x + rect->width ){
                if (y1 <= rect->y){
                    side = META_SIDE_TOP;
                }else if (y2 >= rect->y + rect->height){
                    side = META_SIDE_BOTTOM;
                }else {
                    continue;
                }
            }else if (y1 <= rect->y && y2 >= rect->y + rect->height){
                if (x1 < 0){
                    side = META_SIDE_LEFT;
                }else if (x2 >= screen_width){
                    side = META_SIDE_RIGHT;
                }else {
                    continue;
                }
            }else if (x1 <= 0){
                side = META_SIDE_LEFT;
            }else if (y1 <= 0){
                side = META_SIDE_TOP;
            }else if (x2 >= screen_width){
                side = META_SIDE_RIGHT;
            }else if (y2 >= screen_height){
                side = META_SIDE_BOTTOM;
            }else {
                continue;
            }

            switch (side) {
                case META_SIDE_TOP:
                    y1 = 0;
                    break;
                case META_SIDE_BOTTOM:
                    y2 = screen_height;
                    break;
                case META_SIDE_LEFT:
                    x1 = 0;
                    break;
                case META_SIDE_RIGHT:
                    x2 = screen_width;
                    break;
            }

            MetaRectangle strutRect;
            strutRect.x = x1;
            strutRect.y = y1;
            strutRect.width = x2 - x1;
            strutRect.height = y2 - y1;

            MetaStrut *strut = g_malloc0 (sizeof(MetaStrut));
            strut->rect = strutRect;
            strut->side = side;
            struts = g_slist_append (struts, strut); 
        }
    }
    /*
    gboolean enable_stage = TRUE;
    GList *top_windows = NULL;
    ClutterActor *top_window_group = cdos_global_get_top_window_group (global);
    top_windows = clutter_actor_get_children (top_window_group);

    GList *iter = NULL;
    for (iter = top_windows; iter; iter = iter->next){
    MetaWindowActor *window_actor = META_WINDOW_ACTOR (iter->data);
    MetaWindow *window = meta_window_actor_get_meta_window(window_actor);
    MetaWindowType type = meta_window_get_window_type(window);

    if (type != META_WINDOW_TOOLTIP){
        enable_stage = FALSE;
        break;
    }
    */

    cdos_global_set_stage_input_region(global, rects);

    // FIXME
    MetaScreen *screen = cdos_global_get_screen (global);
    gint n_spaces = meta_screen_get_n_workspaces (screen);
    //printf ("workspace num is %d\n ", n_spaces);
    gint i = 0;

    while (i < n_spaces){
        MetaWorkspace *space;

        space = meta_screen_get_workspace_by_index (screen, i);
        meta_workspace_set_builtin_struts (space, struts);
        ++i;
    }
    if (struts){
        g_slist_free_full (struts, (GDestroyNotify ) g_free);
    }
    if (rects){
        // 因为里面的成员也是动态分配的内存，所以需要逐个释放，好麻烦……
        g_slist_free_full (rects, (GDestroyNotify ) g_free);
    }
	return FALSE;
}

/**
 * cdos_layout_manager_new:
 *
 * Allocates a new #CdosLayoutManager.
 *
 * Return value: a new #CdosLayoutManager.
 */
CdosLayoutManager *
cdos_layout_manager_new (void)
{
    CdosLayoutManager *layout_manager ;
    layout_manager =  g_object_new (CDOS_TYPE_LAYOUT_MANAGER, NULL); //创建我们自定义的类对象
    // clutter_actor_set_name(CLUTTER_ACTOR(panel), "panel");
    //  clutter_actor_set_reactive(CLUTTER_ACTOR(panel), TRUE);
    if (layoutmanager == NULL)
	layoutmanager = layout_manager;
    return layout_manager;
}

CdosLayoutManager *
cdos_layout_manager_get_default (void)
{
  if (layoutmanager == NULL)
    layoutmanager = cdos_layout_manager_new ();
  return layoutmanager;
}

