#include <meta/util.h>

#include "cdos-lightbox.h"
#include "cdos-start.h"

/**
 * SECTION:cdos-lightbox
 * @title: CdosLightbox
 * @short_description: 设置lightbox
 *
 * 对lightbox显示,隐藏和创建,设置lightbox的变暗时间
 */
static void
destroy (ClutterActor *actor, CdosLightbox *box)
{
    if(box->allocation_changed_handler > 0){
        g_signal_handler_disconnect (box->container, box->allocation_changed_handler);
        box->allocation_changed_handler = 0;
    }
    if(box->actor_added_handler > 0){
        g_signal_handler_disconnect (box->container, box->actor_added_handler);
        box->actor_added_handler = 0;
    }
    if(box->actor_removed_handler > 0){
        g_signal_handler_disconnect (box->container, box->actor_removed_handler);
        box->actor_removed_handler = 0;
    }
    g_object_unref(box->container);
    box->highlighted = NULL;
    clutter_actor_destroy (CLUTTER_ACTOR (actor));
    g_list_free (box->children);
    g_free (box);
}

static void
lightbox_on_destroy (CdosLightbox *box)
{
    cdos_lightbox_highlight (box, NULL);
}
static void
actor_added (ClutterContainer *container,
             ClutterActor *child,
             CdosLightbox *lightbox)
{
    CdosLightbox *box = (CdosLightbox *)lightbox;
    GList *children = st_container_get_children_list ((StContainer *)container);
    gint my_index = -1, new_child_index = -1;
    GList *llist = NULL, *iter = children;
    gint i = 0;
    while(iter)
    {
        if (iter->data == child)
        { 
            new_child_index = i;
            llist = iter;
        }
        else if (iter->data == box->actor)
            my_index = i;
        if (my_index >= 0 && new_child_index >= 0)
            break;
        iter = iter->next;
        i++;
    }
    if(new_child_index > my_index)
    {
        st_container_move_before ((StContainer *)container, (ClutterActor *)child, box->actor);
        box->children = g_list_append (box->children, child);
    }
    else if(new_child_index == 0)
        box->children = g_list_prepend (box->children, child);
    else {
        /* 插入到正确位置 */
        gint pos = g_list_index (box->children, llist->prev); 
        if(pos != -1)
           box->children = g_list_insert (box->children, child, pos + 1); 
    }
}

static void
actor_removed (ClutterContainer *container,
               ClutterActor *child,
               CdosLightbox *lightbox)
{
    CdosLightbox *box = lightbox;
    gint index = g_list_index (box->children, child);
    if(index > -1)
        box->children = g_list_remove (box->children, child);
    if(child == box->highlighted)
        box->highlighted = NULL;
}
static gboolean
change_width_height (CdosLightbox *data)
{
    CdosLightbox *box = data;
    ClutterActor *actor = box->actor;
    clutter_actor_set_width (actor, box->width);
    clutter_actor_set_height (actor, box->height);
    return 0;
}
static void
allocation_changed (ClutterActor *container,
                    ClutterActorBox *clutter_box,
                    guint    flags,
                    CdosLightbox *lightbox)
{
    CdosLightbox *box = lightbox;
    meta_later_add (META_LATER_BEFORE_REDRAW, (GSourceFunc)change_width_height, box, NULL);

    ClutterActor *c = container;
    float width = clutter_actor_get_width (c);
    float height = clutter_actor_get_height (c);
    box->width = width;
    box->height = height;
}

static void
transition_complete_show (ClutterTimeline *tml, CdosLightbox *box)
{
    clutter_actor_remove_transition (box->actor, "animate-box-opacity");
    clutter_actor_show (box->actor);
}

/**
 * cdos_lightbox_show:
 * @box: 作为其他actor的一个容器
 *
 * 显示lightbox
 *
 * Returns: void
 */
void
cdos_lightbox_show (CdosLightbox *box)
{
    if (box->fade_time > 0)
    {
        clutter_actor_set_opacity (box->actor, 0x0);
        clutter_actor_show (box->actor);
        /*
        ClutterAnimation *anim = clutter_actor_animate (box->actor,
                                        CLUTTER_EASE_OUT_QUAD,
                                        box->fade_time,
                                        "opacity", 0xff,
                                        NULL);
        //ClutterTimeline *tml = clutter_animation_get_timeline (anim);
        g_signal_connect (anim, "completed",
                          G_CALLBACK (clutter_actor_show),
                          box->actor);
                          */
        ClutterTransition *transition = clutter_property_transition_new ("opacity");
        clutter_actor_set_easing_mode (box->actor, CLUTTER_EASE_OUT_QUAD);
        clutter_timeline_set_duration (CLUTTER_TIMELINE (transition), box->fade_time);
        clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (transition), 0);
        clutter_transition_set_from (transition, G_TYPE_UCHAR, 0x0);
        clutter_transition_set_to (transition, G_TYPE_UCHAR, 255);
        clutter_actor_add_transition (box->actor, "animate-box-opacity", transition);
        g_signal_connect (transition, "completed", G_CALLBACK (transition_complete_show), box);
    }
    else{
        clutter_actor_set_opacity (box->actor, 0xff);
        clutter_actor_show (box->actor);
    }
}

static void
transition_complete_hide (ClutterTimeline *tml, CdosLightbox *box)
{
    clutter_actor_remove_transition (box->actor, "animate-box-opacity");
    clutter_actor_hide (box->actor);
}


/**
 * cdos_lightbox_hide:
 * @box: 作为其他actor的一个容器
 *
 * 隐藏lightbox
 *
 * Returns: void
 */
void
cdos_lightbox_hide (CdosLightbox *box)
{
    if(box->fade_time > 0)
    {
        /*
        ClutterAnimation *anim = clutter_actor_animate (box->actor,
                                        CLUTTER_EASE_OUT_QUAD,
                                        box->fade_time,
                                        "opacity", 0x0,
                                        NULL);
        //ClutterTimeline *tml = clutter_animation_get_timeline (anim);
        g_signal_connect (anim, "completed",
                          G_CALLBACK (clutter_actor_hide),
                          box->actor);
        */
        ClutterTransition *transition = clutter_property_transition_new ("opacity");
        clutter_actor_set_easing_mode (box->actor, CLUTTER_EASE_OUT_QUAD);
        clutter_timeline_set_duration (CLUTTER_TIMELINE (transition), box->fade_time);
        clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (transition), 0);
        clutter_transition_set_from (transition, G_TYPE_UCHAR, 0xff);
        clutter_transition_set_to (transition, G_TYPE_UCHAR, 0x0);
        clutter_actor_add_transition (box->actor, "animate-box-opacity", transition);
        g_signal_connect (transition, "completed", G_CALLBACK (transition_complete_hide), box);
   
    }else{
        clutter_actor_hide (box->actor);
    }
}


/**
 * cdos_lightbox_highlight:
 * @box: 作为其他actor的一个容器
 * @window: 一个clutter的actor
 *
 * 设置lightbox的highlight
 *
 * Returns: void
 */
void
cdos_lightbox_highlight (CdosLightbox *box, ClutterActor *window)
{
    if(box->highlighted == window)
        return;
    ClutterActor *below = box->actor;
    GList *iter = NULL;
    for(iter = g_list_last (box->children);iter; iter = iter->prev)
    {
        if(iter->data == window)
        {
            //clutter_actor_raise_top (window);
            st_container_raise ((ClutterContainer *)box->container, window, NULL);
            return;
        }
        else if(iter->data == box->highlighted)
            st_container_move_before (box->container, CLUTTER_ACTOR (iter->data), below);
        else
            below = iter->data;
    }
    box->highlighted = window;
}

static ClutterActor *create (CdosLightbox *lightbox)
{
    StWidget *widget = st_bin_new ();
    lightbox->actor = (ClutterActor *)widget;
    //st_bin_set_alignment ((StBin *)widget, ST_ALIGH_START);
    st_widget_set_style_class_name (widget, "lightbox");
    clutter_actor_set_reactive ((ClutterActor *)widget, lightbox->inhibit_events);

    st_container_add_actor ((ClutterContainer *)(lightbox->container), (ClutterActor *)widget);
    clutter_actor_set_position ((ClutterActor *)widget, 0.0, 0.0);
    //clutter_actor_raise_top ((ClutterActor *)widget);
    st_container_raise ((ClutterContainer *)lightbox->container, (ClutterActor *)widget, NULL);
    clutter_actor_hide ((ClutterActor *)widget);

    g_signal_connect ((ClutterActor *)widget, "destroy", G_CALLBACK(destroy), lightbox);
    lightbox->actor_added_handler = 
        g_signal_connect (lightbox->container, "actor-added", 
                G_CALLBACK(actor_added), lightbox);

    lightbox->actor_removed_handler = 
        g_signal_connect (lightbox->container, "actor-removed", 
                G_CALLBACK(actor_removed), lightbox);

    lightbox->allocation_changed_handler = 
        g_signal_connect (lightbox->container, "allocation-changed",
                G_CALLBACK(allocation_changed), lightbox);

    lightbox->highlighted = NULL;
    return (ClutterActor *)widget;
}


/**
 * cdos_lightbox_set_fade_time:
 * @box: 作为其他actor的一个容器
 * @msec: 设置的时间
 *
 * 设置lightbox的变暗时间
 *
 * Returns: void
 */
void
cdos_lightbox_set_fade_time(CdosLightbox *box, guint msec)
{
    if(box == NULL)
        return;
    box->fade_time = msec;
}


/**
 * cdos_lightbox_new:
 * @container: 作为其他actor的一个容器
 * @width: lightbox的宽度
 * @height: lightbox的高度
 * @inhibit_events: 是否允许容器的事件发生
 *
 * 新建lightbox
 *
 * Returns: lightbox
 */
CdosLightbox *
cdos_lightbox_new (StContainer     *container,
                   gfloat           width,
                   gfloat           height,
                   gboolean         inhibit_events)
{
    container = g_object_ref(container);
    CdosLightbox *lightbox = g_new0 (CdosLightbox, 1);
    lightbox->width          = (width > 0.f) ? width : clutter_actor_get_width ((ClutterActor *)container);
    lightbox->height         = (height > 0.f) ? height : clutter_actor_get_height ((ClutterActor *)container);
    lightbox->inhibit_events = inhibit_events;

    lightbox->container      = container;
    lightbox->children       = g_list_copy (st_container_get_children_list (container));

    //lightbox->fade_time      = fade_time;
    create (lightbox);
    return lightbox;
}
