#include "cdos-modal-dialog.h"
#include "cdos.h"
#include "cdos-debug.h"
#include "cdos-base-util.h"

/**
 * SECTION: cdos-modal-dialog
 * @Title: ModalDialog
 * @short_description: 模态对话框
 *
 * 遮挡 #ClutterStage 的对话框，只有该对话框可以接收事件
 */

#define OPEN_AND_CLOSE_TIME 100
#define FADE_IN_BUTTONS_TIME 330
#define FADE_OUT_DIALOG_TIME 1000

/**
 * cdos_modal_dialog_destroy:
 * @self: #ModalDialog 
 *
 * 销毁 @self
 *
 */
void cdos_modal_dialog_destroy (ModalDialog *self)
{
    if(self != NULL)
    {
        if(self->buttons){
            GList *buttons = self->buttons;
            while(buttons){
                if(buttons->data){
                    g_clear_object(&buttons->data);
                }
                buttons = buttons->next;
            }
        }
        g_clear_object(&self->content_layout); // in dialog_layout;
        g_clear_object(&self->button_layout); // in dialog_layout;
        g_clear_object(&self->dialog_layout); // in bcakground_bin or stack
        g_clear_object(&self->constraint); // in group;
        g_clear_object(&self->background_bin); // ingroup;
        g_clear_object(&self->lightbox);
        g_clear_object(&self->event_blocker); // in stack
        g_clear_object(&self->stack); // in stack
        clutter_actor_destroy ((ClutterActor *)self->group);
        g_free (self);
    }
}

static void
on_group_destroy (ClutterActor *group, ModalDialog *modal_dialog) {

}

static gboolean
on_key_press_event (ClutterActor *group,
                    ClutterEvent *key_press_event,
                    ModalDialog *modal_dialog)
{
    ModalDialog *modal = (ModalDialog *)modal_dialog;
    ClutterEvent *ev = (ClutterEvent *)key_press_event;
    ClutterModifierType modifiers = cdos_base_util_get_event_state (ev);
    ClutterModifierType ctrl_alt_mask = CLUTTER_CONTROL_MASK | CLUTTER_MOD1_MASK;
    guint symbol = clutter_event_get_key_symbol (ev);

    printf("%s %s %d: key value: %d\n", __FILE__, __func__, __LINE__, symbol);

    if(symbol == CLUTTER_Escape && !(modifiers &ctrl_alt_mask))
    {
        cdos_modal_dialog_close(modal, 0);
        return TRUE;
    }
    KeyPressAction action = modal->action_keys[symbol % MAX_KEY_SYM];
    if(action)
        action(NULL, modal->action_data[symbol % MAX_KEY_SYM]);
    return FALSE;
}

static void
on_fade_open_completed (ClutterTimeline *tml, ModalDialog *modal)
{
    modal->state = MODAL_OPENED;
    clutter_actor_remove_transition ((ClutterActor *)modal->group, "animate-group-opacity");

    /* TODO */
    /* this->emit("opened") */
}
static void
modal_dialog_fade_open (ModalDialog *modal)
{
    /* TODO */
    /* 需要layoutmanager */
    MetaRectangle monitor;
    meta_screen_get_monitor_geometry (cdos_global_get_screen (global), 0, &monitor);
    ClutterActor *background_bin = CLUTTER_ACTOR (modal->background_bin);
    clutter_actor_set_position (background_bin, monitor.x, monitor.y);
    clutter_actor_set_size (background_bin, monitor.width, monitor.height);

    modal->state = MODAL_OPENING;

    clutter_actor_set_opacity (CLUTTER_ACTOR (modal->dialog_layout),
            255);
    if(modal->lightbox != NULL)
        cdos_lightbox_show (modal->lightbox);

    ClutterActor *group = CLUTTER_ACTOR(modal->group);
    clutter_actor_set_opacity (group, 0);
    clutter_actor_show (group);
    /*
    ClutterAnimation *animation = clutter_actor_animate (group,
            CLUTTER_EASE_OUT_QUAD, OPEN_AND_CLOSE_TIME,
            "opacity", 255,
            "signal::completed", on_fade_open_completed, modal,
            NULL);
    */
    clutter_actor_save_easing_state (group);
    ClutterTransition *transition = clutter_property_transition_new ("opacity");
    clutter_actor_set_easing_mode (group, CLUTTER_EASE_OUT_QUAD);
    clutter_timeline_set_duration (CLUTTER_TIMELINE (transition), OPEN_AND_CLOSE_TIME);
    clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (transition), 0);
    clutter_transition_set_from (transition, G_TYPE_UCHAR, 0x0);
    clutter_transition_set_to (transition, G_TYPE_UCHAR, 0xff);
    clutter_actor_add_transition (group, "animate-group-opacity", transition);
    g_signal_connect (transition, "completed", G_CALLBACK (on_fade_open_completed), modal);
    clutter_actor_restore_easing_state (group);
}

void
modal_dialog_set_initial_key_focus (ModalDialog *modal_dialog, ClutterActor *actor)
{
    modal_dialog->initial_key_focus = ST_WIDGET (actor);
}

gboolean
modal_dialog_push_modal (ModalDialog *modal_dialog,
                         guint32       timestamp)
{
    if(modal_dialog->has_modal)
        return 1;
    
    if(!cdos_push_modal ((ClutterActor *)modal_dialog->group, timestamp, 0))
    {
        cdos_debug ();
        return 0;
    }

    modal_dialog->has_modal = 1;
    if(modal_dialog->saved_key_focus)
    {
        clutter_actor_grab_key_focus ((ClutterActor *)modal_dialog->saved_key_focus);
        modal_dialog->saved_key_focus = NULL;
    }
    else
        clutter_actor_grab_key_focus ((ClutterActor *)modal_dialog->initial_key_focus);

    if(!modal_dialog->cdos_reactive)
        st_container_lower ((ClutterContainer *)modal_dialog->stack,
                    modal_dialog->event_blocker, NULL);

    return 1;
}

gboolean
modal_dialog_pop_modal (ModalDialog *modal_dialog,
                        guint32       timestamp)
{
    if(modal_dialog->has_modal == 0)
        return 0;
    ClutterActor *focus = NULL;

    focus = clutter_stage_get_key_focus ((ClutterStage *)cdos_global_get_stage (global));

    if(focus && clutter_actor_contains ((ClutterActor *)modal_dialog->group, focus))
        modal_dialog->saved_key_focus = (StWidget *)focus;
    else
        modal_dialog->saved_key_focus = NULL;

    cdos_pop_modal((ClutterActor *)modal_dialog->group, timestamp);
    GdkScreen *screen = cdos_global_get_gdk_screen (global);
    GdkDisplay *dpy = gdk_screen_get_display (screen);
    gdk_display_sync (dpy);

    modal_dialog->has_modal = 0;

    if(!modal_dialog->cdos_reactive)
        st_container_raise ((ClutterContainer *)modal_dialog->stack,
                    modal_dialog->event_blocker, NULL);
    return 1;
}

/**
 * cdos_modal_dialog_open:
 * @modal_dialog: #ModalDialog
 * @timestamp: time 
 *
 * Returns: 成功返回 1, 否则返回 0
 */
gboolean
cdos_modal_dialog_open (ModalDialog *modal_dialog, guint32 timestamp)
{
    if (timestamp == 0)
        timestamp = cdos_global_get_current_time (global);
    if(modal_dialog->state == MODAL_OPENED || modal_dialog->state == MODAL_OPENING)
        return 1;
    if(!modal_dialog_push_modal(modal_dialog, timestamp))
        return 0;
    modal_dialog_fade_open (modal_dialog);
    return 1;
}

static void
on_close_complete (ClutterTimeline *tml,
                   ModalDialog *modal)
{
    modal->state = MODAL_CLOSED;
    clutter_actor_remove_transition ((ClutterActor *)modal->group, "animate-group-opacity");

    clutter_actor_hide ((ClutterActor *)modal->group);
    //clutter_actor_detach_animation ((ClutterActor *)modal_dialog->group);
}

/**
 * @modal: #ModalDialog
 * @timestamp: 时间
 *
 * 关闭 @modal.
 *
 */
void
cdos_modal_dialog_close (ModalDialog *modal, guint32 timestamp)
{
    if(modal->state == MODAL_CLOSED || modal->state == MODAL_CLOSING)
        return;
    modal->state = MODAL_CLOSING;
    modal_dialog_pop_modal (modal, timestamp);
    modal->saved_key_focus = NULL;

    /*
    ClutterAnimation *animation = clutter_actor_animate (
            (ClutterActor *)(modal->group),
            CLUTTER_EASE_OUT_QUAD, OPEN_AND_CLOSE_TIME,
            "opacity", 0,
            "signal::completed", on_close_complete, modal,
            NULL);
    */
    clutter_actor_save_easing_state (CLUTTER_ACTOR (modal->group));
    ClutterTransition *transition = clutter_property_transition_new ("opacity");
    clutter_actor_set_easing_mode ((ClutterActor *)modal->group, CLUTTER_EASE_OUT_QUAD);
    clutter_timeline_set_duration (CLUTTER_TIMELINE (transition), OPEN_AND_CLOSE_TIME);
    clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (transition), 0);
    clutter_transition_set_from (transition, G_TYPE_UCHAR, 255);
    clutter_transition_set_to (transition, G_TYPE_UCHAR, 0);
    clutter_actor_add_transition ((ClutterActor *)modal->group, "animate-group-opacity", transition);
    g_signal_connect (transition, "completed", G_CALLBACK (on_close_complete), modal);
    clutter_actor_restore_easing_state ((ClutterActor *)modal->group);
}

/* *
 * cdos_modal_dialog_set_buttons
 * @modal_dialog: #ModalDialog
 * @buttons : 为 #ButtonInfo GList
 *
 * 设置 @modal_dialog 的按键
 */
void
cdos_modal_dialog_set_buttons (ModalDialog *modal_dialog, GList *buttons)
{
    modal_dialog->buttons = buttons;
    //gint had_children = st_container_get_children_list (
    //        (StContainer *)modal_dialog->button_layout) != NULL;
    st_container_destroy_children (ST_CONTAINER (modal_dialog->button_layout));
    //clutter_actor_destroy_all_children (modal_dialog->button_layout);
    gint i;
    for(i = 0; i < MAX_KEY_SYM; i++)
        modal_dialog->action_keys[i] = NULL;

    gint focus_set_explicity = 0;

    GList *iter = NULL;
    for (iter = buttons; iter; iter = iter->next)
    {
        ButtonInfo *button_info = iter->data;
        
        int key = button_info->key;
        KeyPressAction action = (KeyPressAction)button_info->action;
        StWidget *button = st_button_new_with_label (button_info->label);

        st_widget_set_style_class_name (button, "modal-dialog-button");
        clutter_actor_set_reactive ((ClutterActor *)button, 1);
        st_widget_set_can_focus (button, 1);

        StAlign x_align;
        if(iter == buttons && iter->next == NULL) /* 1个button */
            x_align = ST_ALIGN_END;
        else if(iter == buttons)
            x_align = ST_ALIGN_START;
        else if(iter->next == NULL)
            x_align = ST_ALIGN_END;
        else
            x_align = ST_ALIGN_MIDDLE;

        if(button_info->focused)
        {
            modal_dialog->initial_key_focus = button;
            focus_set_explicity = 1;
        }

        if(!focus_set_explicity && button_info->focused &&
                ((modal_dialog->initial_key_focus == modal_dialog ->dialog_layout) ||
                 (clutter_actor_contains ((ClutterActor *)modal_dialog->button_layout,
                                          CLUTTER_ACTOR (modal_dialog->initial_key_focus)))))
            modal_dialog->initial_key_focus = button;
        st_container_add_actor ((ClutterContainer *)modal_dialog->button_layout,
                (ClutterActor *)button);

        StBoxLayoutChild *button_meta = 
        (StBoxLayoutChild *)clutter_container_get_child_meta (
                CLUTTER_CONTAINER (modal_dialog->button_layout), (ClutterActor *)button);
        button_meta->x_fill  = 0;
        button_meta->y_fill  = 0;
        button_meta->expand  = 1;
        button_meta->x_align = x_align;
        button_meta->y_align = ST_ALIGN_MIDDLE;
        clutter_actor_queue_relayout ((ClutterActor *)modal_dialog->button_layout);
        /* TODO */
        if (button_info->callback)
        {
            g_signal_connect (ST_BUTTON (button), "clicked", 
                    G_CALLBACK (button_info->callback), button_info->userdata);
        }
        if(key)
        {
            modal_dialog->action_keys[key % MAX_KEY_SYM] = action;
            if(button_info->userdata)
                modal_dialog->action_data[key % MAX_KEY_SYM] = button_info->userdata;
        }

    }
}

/**
 * cdos_modal_dialog_init
 * @modal: #ModalDialog
 * @cdos_reactive: reactive 属性
 * @style_class: style class name
 * 
 * 对 @modal 初始化
 *
 * Returns: @modal
 *
 */
ModalDialog *
cdos_modal_dialog_init (ModalDialog *modal,
                        gboolean cdos_reactive,
                        gchar    *style_class)
{
    /* g_new0() 调用 g_malloc0(), 申请内存且初始化为 0 */
    //ModalDialog *modal_dialog = g_new0 (ModalDialog, 1);
    g_return_val_if_fail ((modal != NULL), NULL);
    modal->state = MODAL_CLOSED;
    modal->has_modal = 0;
    modal->cdos_reactive = cdos_reactive;

    StWidget *group = st_group_new ();
    modal->group = group;
    /* 设置 visible 为 false 时，先设置 parent，再设置 VISIBLE flag = 0，
     * 这样可以不必调用 clutter_actor_hide() 
     */
    ClutterActor *ui_group =  monkey_plugin_get_ui_group (monkey_plugin_get ());
    st_container_add_actor ((ClutterContainer *) ui_group,
            (ClutterActor *)group);
    clutter_actor_set_position ((ClutterActor *)group, 0.0, 0.0);
    ((ClutterActor *)group)->flags &= ~CLUTTER_ACTOR_VISIBLE;
    //clutter_actor_hide ((ClutterActor *)group);

    modal->constraint = clutter_bind_constraint_new (
            (ClutterActor *)cdos_global_get_stage (global), CLUTTER_BIND_ALL, 0.0);
    clutter_actor_add_constraint ((ClutterActor *)group, modal->constraint);
    g_signal_connect (group, "destroy", G_CALLBACK (on_group_destroy), modal);

    g_signal_connect (group, "key-press-event",
            G_CALLBACK (on_key_press_event), modal);

    StWidget *background_bin = st_bin_new ();
    modal->background_bin = background_bin;
    st_container_add_actor ((ClutterContainer *)group, (ClutterActor *)background_bin);

    StWidget *dialog_layout = st_box_layout_new ();
    modal->dialog_layout = dialog_layout;
    st_widget_set_style_class_name (dialog_layout, "modal-dialog");
    st_box_layout_set_vertical ((StBoxLayout *)dialog_layout, 1);
    if(style_class != NULL)
        st_widget_add_style_class_name (dialog_layout, style_class);

    if(!cdos_reactive)
    {
        CdosLightbox *box = cdos_lightbox_new((StContainer *)group, 0, 0, 1);
        cdos_lightbox_highlight (box, (ClutterActor *)background_bin);
        modal->lightbox = box;

        CdosStack *stack = g_object_new (CDOS_TYPE_STACK, NULL);
        modal->stack = stack;
        st_bin_set_child ((StBin *)background_bin, (ClutterActor *)stack);

        //ClutterActor *event_blocker = clutter_group_new ();
        ClutterActor *event_blocker = clutter_actor_new ();
        modal->event_blocker = event_blocker;
        clutter_actor_set_reactive (event_blocker, 1);
        st_container_add_actor ((ClutterContainer *)stack, event_blocker);
        st_container_add_actor ((ClutterContainer *)stack, (ClutterActor *)dialog_layout);
    }else
        st_bin_set_child ((StBin *)background_bin, (ClutterActor *)dialog_layout);

    StWidget *content_layout = st_box_layout_new ();
    modal->content_layout = content_layout;
    st_box_layout_set_vertical ((StBoxLayout *)content_layout, 1);

    st_container_add_actor ((ClutterContainer *)dialog_layout, (ClutterActor *)content_layout);
    StBoxLayoutChild *meta = 
        (StBoxLayoutChild *)clutter_container_get_child_meta (
                CLUTTER_CONTAINER (dialog_layout), (ClutterActor *)content_layout);
    meta->x_fill = 1;
    meta->y_fill = 1;
    meta->x_align = ST_ALIGN_MIDDLE;
    meta->y_align = ST_ALIGN_START;
    clutter_actor_queue_relayout ((ClutterActor *)dialog_layout);

    StWidget *button_layout = st_box_layout_new ();
    modal->button_layout = button_layout;
    st_widget_set_style_class_name (button_layout, "modal-dialog-button-box");
    st_box_layout_set_vertical ((StBoxLayout *)button_layout, 0);
    st_container_add_actor ((ClutterContainer *)dialog_layout,
                            (ClutterActor *)button_layout);
    StBoxLayoutChild *button_meta = 
        (StBoxLayoutChild *)clutter_container_get_child_meta (
                CLUTTER_CONTAINER (dialog_layout), (ClutterActor *)button_layout);
    button_meta->x_align = ST_ALIGN_MIDDLE;
    button_meta->y_align = ST_ALIGN_END;
    button_meta->expand  = 1;
    clutter_actor_queue_relayout ((ClutterActor *)dialog_layout);

    st_focus_manager_add_group (cdos_global_get_focus_manager (global),
            dialog_layout);

    modal->initial_key_focus = dialog_layout;
    modal->saved_key_focus = NULL;
    return modal;
}
