/**
 * SECTION: cdos-notification
 * @short_description: 通知结构体
 * @include: cdos-notification.h
 * 
 * CdosNotificationClass 类用来处理新建,转发,消除通知
 */
#define CLUTTER_DISABLE_DEPRECATION_WARNINGS

#include <math.h>
#include <string.h>
#include <meta/util.h>

#include "cdos-notification.h"
#include "cdos-global-private.h"
#include "cdos-generic-container.h"
#include "cdos-base-util.h"
#include "cdos-util.h"
#include "cdos-start.h"
#include "cdos-debug.h"

#define NOTIFICATION_TITLE_LENGTH 31
#define ANIMATION_TIME 200

static const gchar *DATE_FORMAT = "%l:%M %p"; // _("%l:%M %p")

G_DEFINE_TYPE (CdosNotification, cdos_notification, G_TYPE_OBJECT);

/* signal */
enum
{
    EXPANDED,    //expanded
    COLLAPSED,      //collapsed
    ACTION_INVOKED,         //action-invoked
    DONE_DISPLAYING,        //done-displaying
    CLICKED,        //clicked
    DESTROY,        //destroy
    LAST_SIGNAL
};

static guint cdos_notification_signals [LAST_SIGNAL] = { 0 };

struct _CStTableChildProps {
  gint col;
  gint row;
  gint col_span;
  gint row_span;
  StAlign x_align;
  StAlign y_align;
  guint x_expand : 1;
  guint y_expand : 1;
  guint x_fill : 1;
  guint y_fill : 1;
};

void cst_table_child_init (CStTableChildProps *props)
{
    props->col = 0;
    props->row = 0;
    props->col_span = 1;
    props->row_span = 1;
    props->x_align = ST_ALIGN_MIDDLE;
    props->y_align = ST_ALIGN_MIDDLE;
    props->x_expand = TRUE;
    props->y_expand = TRUE;
    props->x_fill = TRUE;
    props->y_fill = TRUE;
}

static void updated (CdosNotification *self) {
    if (self->expanded)
        cdos_notification_expand (self, FALSE);
}

static void onSourceDestroy (GObject *gobject, guint reason, gpointer user_data) {
    
    CdosNotification *self = CDOS_NOTIFICATION(user_data); 
    if (gobject == NULL || self == NULL) {
        return ;
    }
    
    cdos_notification_destroy (self, reason);
}

//handle self->actor:"clicked"
static void onClicked (StButton *object, gint button, gpointer user_data) {

    CdosNotification *self = CDOS_NOTIFICATION(user_data);

    g_signal_emit (self, cdos_notification_signals[CLICKED], 0);
    // We hide all types of notifications once the user clicks on them because the common
    // outcome of clicking should be the relevant window being brought forward and the user's
    // attention switching to the window.

    g_signal_emit (self, cdos_notification_signals[DONE_DISPLAYING], 0);

    if (!self->resident)
        cdos_notification_destroy(self, NOREASON);
}


//handle self->actor:"destroy"
static void onDestroy (ClutterActor *actor, gpointer user_data) {

    CdosNotification *self = CDOS_NOTIFICATION(user_data);

    if (self->destroyed) {
        return ;
    }

    self->destroyed = TRUE;

    if (!self->destroyedReason) {
        self->destroyedReason = DISMISSED;
    }

    g_signal_emit (self, cdos_notification_signals[DESTROY], 0, self->destroyedReason);
}

static void styleChanged (StWidget *object, gpointer user_data) {
    CdosNotification *self = CDOS_NOTIFICATION(user_data);
    self->spacing = st_theme_node_get_length (st_widget_get_theme_node (ST_WIDGET(self->table)) ,
                            "spacing-columns");
}

static void bannerBoxGetPreferredWidth (GObject *gobject, gfloat for_height, CdosGenericContainerAllocation *alloc, gpointer user_data) {

    CdosNotification *self = CDOS_NOTIFICATION(user_data);

    gfloat title_min;
    gfloat title_natural;
    gfloat banner_min;
    gfloat banner_natural;
    gfloat time_min;
    gfloat time_natural;

    clutter_actor_get_preferred_width (CLUTTER_ACTOR(self->titleLabel), 
                                       for_height,
                                       &title_min,
                                       &title_natural);

    clutter_actor_get_preferred_width (CLUTTER_ACTOR(self->bannerLabel),
                                       for_height,
                                       &banner_min,
                                       &banner_natural);

    clutter_actor_get_preferred_width (CLUTTER_ACTOR(self->timeLabel),
                                       for_height,
                                       &time_min,
                                       &time_natural);

    if (self->inNotificationBin) {
        alloc->min_size = MAX (title_min, time_min);
        alloc->natural_size = MAX (title_natural, time_natural) + self->spacing + banner_natural;
    } else {
        alloc->min_size = title_min;
        alloc->natural_size = title_natural + self->spacing + banner_natural;
    }
}

static void bannerBoxGetPreferredHeight (GObject *gobject, gfloat for_width, CdosGenericContainerAllocation *alloc, gpointer user_data) {

    CdosNotification *self = CDOS_NOTIFICATION(user_data);

    gfloat title_min;
    gfloat title_natural;
    gfloat time_min;
    gfloat time_natural;

    if (self->inNotificationBin) {
        clutter_actor_get_preferred_height (CLUTTER_ACTOR(self->titleLabel), 
                                            for_width,
                                            &title_min,
                                            &title_natural);

        clutter_actor_get_preferred_height (CLUTTER_ACTOR(self->timeLabel), 
                                            for_width,
                                            &time_min,
                                            &time_natural);

        alloc->min_size = title_min + time_min;

        alloc->natural_size = title_natural + time_natural;
    } 
    else 
    {
        clutter_actor_get_preferred_height (CLUTTER_ACTOR(self->timeLabel), 
                                            for_width,
                                            &time_min,
                                            &time_natural);
        alloc->min_size = time_min;
        alloc->natural_size = time_natural; 
    }
}

static void addBannerBody (CdosNotification *self) {
    if (self->bannerBodyText) {
        gchar *text = g_strdup (self->bannerBodyText);
        g_free (self->bannerBodyText);
        self->bannerBodyText = NULL;
        cdos_notification_add_body (self, text, self->bannerBodyMarkup, NULL);
        g_free (text);
    }
}

static gboolean canExpandContent (CdosNotification *self) {
        return self->bannerBodyText ||  
        (!self->titleFitsInBannerMode && 
            !st_widget_has_style_class_name (self->table, "multi-line-notification"));
}

static gboolean onMetaBeforeRedraw (gpointer user_data) {

    CdosNotification *self = CDOS_NOTIFICATION(user_data);

    if (canExpandContent(self)) {
       addBannerBody(self);
       st_widget_add_style_class_name (self->table, "multi-line-notification");    
       updated(self);
    }
    return FALSE;
}

static void bannerBoxAllocate (GObject *gobject, ClutterActorBox *box, ClutterAllocationFlags flags, gpointer user_data) {
    CdosNotification *self = CDOS_NOTIFICATION(user_data);

    gfloat title_min_width;
    gfloat title_natural_width;
    gfloat title_min_height;
    gfloat title_natural_height;
    gfloat time_min_width;
    gfloat time_natural_width;
    gfloat time_min_height;
    gfloat time_natural_height;
    gfloat banner_min_width;
    gfloat banner_natural_width;

    gfloat titleBoxW, timeBoxW;

    gboolean bannerFits = TRUE;

    gfloat availWidth = box->x2 - box->x1;

    //let [titleMinW, titleNatW] = this._titleLabel.get_preferred_width(-1);
    clutter_actor_get_preferred_width (CLUTTER_ACTOR(self->titleLabel),
            -1,
            &title_min_width,
            &title_natural_width);

    //let [titleMinH, titleNatH] = this._titleLabel.get_preferred_height(availWidth);    
    clutter_actor_get_preferred_height (CLUTTER_ACTOR(self->titleLabel), 
            availWidth,
            &title_min_height,
            &title_natural_height);

    //let [timeMinW, timeNatW] = this._timeLabel.get_preferred_width(-1);
    clutter_actor_get_preferred_width (CLUTTER_ACTOR(self->timeLabel),
            -1,
            &time_min_width,
            &time_natural_width);

    //let [timeMinH, timeNatH] = this._timeLabel.get_preferred_height(availWidth);
    clutter_actor_get_preferred_height (CLUTTER_ACTOR(self->timeLabel), 
            availWidth,
            &time_min_height,
            &time_natural_height);

    //let [bannerMinW, bannerNatW] = this._bannerLabel.get_preferred_width(availWidth);
    clutter_actor_get_preferred_width (CLUTTER_ACTOR(self->bannerLabel),
            availWidth,
            &banner_min_width,
            &banner_natural_width);


    ClutterActorBox titleBox, timeBox;

    titleBoxW = MIN (title_natural_width, availWidth); 
    timeBoxW = MIN (time_natural_width, availWidth); 

    if (self->titleDirection == ST_TEXT_DIRECTION_RTL) {

        titleBox.x1 = availWidth - titleBoxW;
        titleBox.x2 = availWidth;
        timeBox.x1 = availWidth - timeBoxW;
        timeBox.x2 = availWidth;
    } 
    else 
    {
        titleBox.x1 = 0;
        timeBox.x1 = 0;
        titleBox.x2 = titleBoxW;
        timeBox.x2 = timeBoxW;
    }

    if (self->inNotificationBin) {

        timeBox.y1 = 0;
        timeBox.y2 = time_natural_height;
        titleBox.y1 = time_natural_height;
        titleBox.y2 = time_natural_height + title_natural_height;
    } 
    else 
    {
        titleBox.y1 = 0;
        titleBox.y2 = title_natural_height;
    }

    clutter_actor_allocate (CLUTTER_ACTOR(self->titleLabel), &titleBox, flags);

    if (self->inNotificationBin)
        clutter_actor_allocate (CLUTTER_ACTOR(self->timeLabel), &timeBox, flags);

    self->titleFitsInBannerMode = (title_natural_width <= availWidth);

    if (titleBoxW + self->spacing > availWidth) {
        clutter_actor_set_opacity (CLUTTER_ACTOR(self->bannerLabel), 0);
        bannerFits = FALSE;
    } 
    else 
    {
        ClutterActorBox bannerBox;

        if (self->titleDirection == ST_TEXT_DIRECTION_RTL) {
            bannerBox.x1 = 0;
            bannerBox.x2 = titleBox.x1 - (gfloat)self->spacing;

            bannerFits = ((bannerBox.x2 - banner_natural_width) >= 0);
        }
        else
        {
            bannerBox.x1 = titleBox.x2 + (gfloat)self->spacing;
            bannerBox.x2 = availWidth;

            bannerFits = (bannerBox.x1 + banner_natural_width <= availWidth);
        }

        if (self->inNotificationBin) {
            bannerBox.y1 = time_natural_height;
            bannerBox.y2 = time_natural_height + title_natural_height;
        }
        else
        {
            bannerBox.y1 = 0;
            bannerBox.y2 = title_natural_height;
        }

        clutter_actor_allocate (CLUTTER_ACTOR(self->bannerLabel), &bannerBox, flags);

        // Make _bannerLabel visible if the entire notification
        // fits on one line, or if the notification is currently
        // unexpanded and only showing one line anyway.
        if (!self->expanded || (bannerFits && st_table_get_row_count ((StTable *)self->table) == 1))
            clutter_actor_set_opacity (CLUTTER_ACTOR(self->bannerLabel), 255);
    }

    // If the banner doesn't fully fit in the banner box, we possibly need to add the
    // banner to the body. We can't do that from here though since that will force a
    // relayout, so we add it to the main loop.
    if (!bannerFits && canExpandContent(self)){
        meta_later_add (META_LATER_BEFORE_REDRAW, onMetaBeforeRedraw, self, NULL);
    }
}


static void cdos_notification_finalize (GObject *obj)
{
    CdosNotification *self = CDOS_NOTIFICATION (obj);

    g_free (self->title);
    g_free (self->banner);
    g_free (self->bannerBodyText);

    //g_free (self->actor_delegate); //js中有动态建立的delegate和parent_container成员
    //g_free (self->actor_parent_container); //同上

    G_OBJECT_CLASS(cdos_notification_parent_class)->finalize(obj);
}
static void cdos_notification_dispose (GObject *obj) {
    CdosNotification * self = CDOS_NOTIFICATION (obj);
    g_date_time_unref (self->timestamp);

    //以下顺序不可乱动；
    g_clear_object(&self->icon);
    g_clear_object(&self->imageBin);
    g_clear_object(&self->buttonBox);

    g_clear_object(&self->buttonFocusManager);
    g_clear_object(&self->timeLabel);
    g_clear_object(&self->titleUrlHighlighter);
    g_clear_object(&self->bannerUrlHighlighter);

    g_clear_object(&self->scrollArea);

    g_clear_object(&self->table);

    g_clear_object(&self->actor);

    G_OBJECT_CLASS(cdos_notification_parent_class)->dispose(obj);
}

static void
cdos_notification_class_init (CdosNotificationClass *klass) {

    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

    gobject_class->dispose = cdos_notification_dispose;
    gobject_class->finalize = cdos_notification_finalize;


    /**
     * CdosNotification::expanded:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * notification能根据内容来自适应大小,当默认大小不满足时,就会触发该信号.
     */    
    cdos_notification_signals [EXPANDED] =    
        g_signal_new ("expanded",
                      G_TYPE_FROM_CLASS (klass),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL, NULL,
                      g_cclosure_marshal_VOID__VOID,
                      G_TYPE_NONE, 0);

    /**
     * CdosNotification::collapsed:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 信号发送条件:
     * 若notification的destroyed成员变量为 %TRUE, 则返回;
     * 否则,不调整大小,不对文本进行自动换行,然后发送该信号.
     */    
    cdos_notification_signals [COLLAPSED] =
            g_signal_new ("collapsed",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__VOID,
                    G_TYPE_NONE, 0);

    /**
     * CdosNotification::action-invoked:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 当触发notification的"clicked"信号时,会在回调函数时发送.
     */
    cdos_notification_signals [ACTION_INVOKED] =
            g_signal_new ("action-invoked",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0, 
                    NULL, NULL,
                    g_cclosure_marshal_VOID__CHAR,
                    G_TYPE_NONE, 1,
        G_TYPE_STRING);

    /**
     * CdosNotification::done-displaying:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 当完成notification的展示的时候,发送该信号
     */     
    cdos_notification_signals [DONE_DISPLAYING] =
            g_signal_new ("done-displaying",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__VOID,
                    G_TYPE_NONE, 0);

    /**
     * CdosNotification::clicked:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 通过StButton的"clicked"信号通过回调函数转换成notification的"clicked"信号.
     */
    cdos_notification_signals [CLICKED] =
            g_signal_new ("clicked",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__VOID,
                    G_TYPE_NONE, 0);

    /**
     * CdosNotification::destroy:
     * @object: 绑定信号的对象
     * @reason: 删除的原因
     * @user_data: 真对回调函数所传入的参数
     *
     * 将ClutterActor的"destory"信号通过回调函数转换成notification的"destroy"
     */
    cdos_notification_signals [DESTROY] =
            g_signal_new ("destroy",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__UINT,
                    G_TYPE_NONE, 1,
           G_TYPE_UINT);

}


static void
cdos_notification_init (CdosNotification *self) {

    self->urgency = NORMAL;
    self->resident = FALSE;
    self->isTransient = FALSE;
    self->expanded = FALSE;
    self->destroyed = FALSE;
    self->useActionIcons = FALSE;
    self->customContent = FALSE;
    self->bannerBodyText = NULL;
    self->bannerBodyMarkup = FALSE;
    self->titleFitsInBannerMode = TRUE;
    self->titleDirection =  ST_TEXT_DIRECTION_NONE;
    self->spacing = 0;

    self->imageBin = NULL;

    self->timestamp = g_date_time_new_now_local();

    self->inNotificationBin = FALSE;

    self->actor = st_button_new ();
    //cdos_draggable_set_delegate (self->actor_delegate, self);  /*找不到delegate及parent_container,因為是动态建立加入到物件中*/
    self->actor_parent_container = NULL;

    clutter_actor_set_reactive(CLUTTER_ACTOR(self->actor), TRUE);

    g_signal_connect (self->actor, "clicked", G_CALLBACK (onClicked), self);
    g_signal_connect (self->actor, "destroy", G_CALLBACK (onDestroy), self);

    self->buttonFocusManager = st_focus_manager_get_for_stage (cdos_global_get_stage(global));

    self->table = st_table_new();
    st_widget_set_style_class_name(self->table,"notification");
    clutter_actor_set_name((ClutterActor *)self->table,"notification");
    clutter_actor_set_reactive (CLUTTER_ACTOR(self->table), TRUE);

    g_signal_connect (self->table, "style-changed", G_CALLBACK (styleChanged), self);

    st_bin_set_child (ST_BIN (self->actor), CLUTTER_ACTOR(self->table));

    // The first line should have the title, followed by the
    // banner text, but ellipsized if they won't both fit. We can't
    // make St.Table or St.BoxLayout do this the way we want (don't
    // show banner at all if title needs to be ellipsized), so we
    // use Cdos.GenericContainer.
    //this._bannerBox = new Cdos.GenericContainer();

    /*CdosGenericContainer继承自StContainer*/
    self->bannerBox = cdos_generic_container_new ();

    g_signal_connect (self->bannerBox, "get-preferred-width", G_CALLBACK (bannerBoxGetPreferredWidth), self);
    g_signal_connect (self->bannerBox, "get-preferred-height", G_CALLBACK (bannerBoxGetPreferredHeight), self);
    g_signal_connect (self->bannerBox, "allocate", G_CALLBACK (bannerBoxAllocate), self);

    st_container_add_actor ((ClutterContainer *)self->table, self->bannerBox);
    ClutterChildMeta *bannerbox_meta = clutter_container_get_child_meta ((ClutterContainer *)self->table, (ClutterActor *)self->bannerBox);
    g_object_set (bannerbox_meta, 
                  "row", 0,
                  "col", 1,
                  "col-span", 2,
                  "x-expand", FALSE,
                  "y-expand", FALSE,
                  "y-fill", FALSE,
                  NULL);

    // This is an empty cell that overlaps with this._bannerBox cell to ensure
    // that this._bannerBox cell expands horizontally, while not forcing the
    // this._imageBin that is also in col: 2 to expand horizontally.
    StWidget *empty_cell = st_bin_new ();
    st_container_add_actor ((ClutterContainer *)self->table, CLUTTER_ACTOR(empty_cell));
    ClutterChildMeta *cell_meta = clutter_container_get_child_meta ((ClutterContainer *)self->table, (ClutterActor *)empty_cell);
    g_object_set (cell_meta, 
                  "row", 0,
                  "col", 2,
                  "y-expand", FALSE,
                  "y-fill", FALSE,
                  NULL);

    self->timeLabel = (StLabel *)st_label_new("");

    self->titleLabel = (StLabel *)st_label_new ("");
 
    st_container_add_actor ((ClutterContainer *)self->bannerBox, (ClutterActor *)self->timeLabel);
    st_container_add_actor ((ClutterContainer *)self->bannerBox, (ClutterActor *)self->titleLabel);

    clutter_actor_hide ((ClutterActor *)self->timeLabel);

    self->bannerUrlHighlighter = cdos_url_highlighter_new (NULL, TRUE, FALSE);

    self->bannerLabel = self->bannerUrlHighlighter->actor;

    st_container_add_actor ((ClutterContainer *)self->bannerBox,  (ClutterActor *)self->bannerLabel);
}

static void actor_text_relayout (StLabel *actor,gfloat  width, guint max_lines)
{
    if(width  <= 0 )
        return ;

    ClutterActor *clutter_text = st_label_get_clutter_text (actor);
    PangoLayout *layout = clutter_text_get_layout(CLUTTER_TEXT (clutter_text));
    GSList *logical_list = pango_layout_get_lines(layout);
    
    gfloat pixel_height = 0;
    guint visible_lines = 0;
    while (logical_list) {
        PangoLayoutLine *line = logical_list->data;
        PangoRectangle logical_rect;
        pango_layout_line_get_extents(line,NULL, &logical_rect);
        gfloat pixel_size = (logical_rect.width/1024.0f) ;
        if(pixel_height == 0)
            pixel_height = (logical_rect.height/1024.0f);

        visible_lines = visible_lines + (guint)ceil(pixel_size/width);
        logical_list = logical_list->next;
    }
    if(pixel_height > 0){
        if (max_lines <= visible_lines)
            visible_lines = max_lines;
        clutter_actor_set_height(CLUTTER_ACTOR (actor), visible_lines  * pixel_height);
    }   
}

/**
 * cdos_notification_new:
 * @source: A #CdosSource 
 * @title: 设置 #CdosNotification 实例的标题 
 * @banner: 设置 #CdosNotification 实例的内容
 * @params: 必要参数,参考 #Params
 * 
 * 新建一个 #CdosNotification 类的实例。
 *
 * Returns: 返回一个 #CdosNotification 的实例。
 */
CdosNotification *
cdos_notification_new (CdosSource *source, gchar *title, gchar *banner, Params *params)
{
    CdosNotification *self = g_object_new (CDOS_TYPE_NOTIFICATION, NULL);
    self->source = source;
    self->title = g_strdup (title);
    g_signal_connect (source, "destroy", G_CALLBACK (onSourceDestroy), self);

    cdos_notification_update (self, self->title, banner, params);

    actor_text_relayout(self->titleLabel,344, 3);
    actor_text_relayout(self->bannerLabel,344, 10);

    return self;
}

static void updateLastColumnSettings (CdosNotification *self) {
    if (self->scrollArea){
        ClutterChildMeta *meta = clutter_container_get_child_meta ((ClutterContainer *)self->table, (ClutterActor *)self->scrollArea);
        g_object_set (meta,
                      "col", self->imageBin ? 2 : 1,
                      "col_span", self->imageBin ? 1 : 2,
                      NULL);
    }

    if (self->actionArea){
        ClutterChildMeta *meta = clutter_container_get_child_meta ((ClutterContainer *)self->table, (ClutterActor *)self->actionArea);
        g_object_set (meta,
                      "col", self->imageBin ? 2 : 1,
                      "col_span", self->imageBin ? 1 : 2,
                      NULL);
    }
}

static void createScrollArea (CdosNotification *self) {
    st_widget_add_style_class_name (self->table, "multi-line-notification");
    /**
     * scrollArea 不知道有什么用，且会影响 Notification Applet里面的鼠标滚动事件，注释掉
     */
    /*
    self->scrollArea = ST_SCROLL_VIEW (st_scroll_view_new ());

    clutter_actor_set_name((ClutterActor *)self->scrollArea, "notification-scrollview");

    st_scroll_view_set_policy (self->scrollArea, GTK_POLICY_NEVER, GTK_POLICY_NEVER);
    st_widget_set_style_class_name ((StWidget *)self->scrollArea, "vfade");

    clutter_container_add_actor ((ClutterContainer *)self->table, (ClutterActor *)self->scrollArea);
    ClutterChildMeta *meta = clutter_container_get_child_meta ((ClutterContainer *)self->table, (ClutterActor *)self->scrollArea);
    g_object_set (meta, "row", 1, "col", 2, NULL);

    updateLastColumnSettings (self);
    */

    self->contentArea = ST_BOX_LAYOUT (st_box_layout_new ());
    clutter_actor_set_name((ClutterActor *)self->contentArea, "notification-body");
    st_box_layout_set_vertical (self->contentArea, TRUE);
    
    self->scrollArea = (StWidget *)self->contentArea;
    clutter_container_add_actor ((ClutterContainer *)self->table, (ClutterActor *)self->scrollArea);
    ClutterChildMeta *meta = clutter_container_get_child_meta ((ClutterContainer *)self->table, (ClutterActor *)self->scrollArea);
    g_object_set (meta, "row", 1, "col", 2, NULL);

    updateLastColumnSettings (self);
    //clutter_container_add_actor ((ClutterContainer *)self->scrollArea, (ClutterActor *)self->contentArea);

    // If we know the notification will be expandable, we need to add
    // the banner text to the body as the first element.
    addBannerBody (self);
            
}


// 从StButton那边传过来两个参数 (0, int clicked_button)
//_onActionInvoked: function(actor, mouseButtonClicked, id)
static void onActionInvoked (StButton *button, gint mouseButtonClicked, CdosNotification *self) {
    const gchar *id = clutter_actor_get_name ((ClutterActor *)button);
    g_signal_emit (self, cdos_notification_signals[ACTION_INVOKED], 0, id);

    if (!self->resident) {
        // We don't hide a resident notification when the user invokes one of its actions,
        // because it is common for such notifications to update themselves with new
        // information based on the action. We'd like to display the updated information
        // in place, rather than pop-up a new notification.
        g_signal_emit (self, cdos_notification_signals[DONE_DISPLAYING], 0);

        cdos_notification_destroy (self, NOREASON);
    }
}

/*公有函数定义*/

/**
 * cdos_notification_update:
 * @title: the new title
 * @banner: the new banner
 * @params: as in the CdosNotification constructor
 *
 * Updates the notification by regenerating its icon and updating
 * the title/banner. If @params.clear is %TRUE, it will also
 * remove any additional actors/action buttons previously added.
 */

void cdos_notification_update (CdosNotification *self, const gchar *title, const gchar *banner, const Params *params) {
    ClutterActor *oldFocus; 

    if (NULL != self->timestamp) {
        g_date_time_unref(self->timestamp);
    }

    self->timestamp = g_date_time_new_now_local ();
    self->inNotificationBin = FALSE;
        
    self->customContent = params->customContent;
    oldFocus = clutter_stage_get_key_focus (cdos_global_get_stage(global));

    if (self->icon && (params->icon || params->clear)) {

        clutter_actor_destroy ((ClutterActor *)self->icon);
        self->icon = NULL;
    }
    
    // We always clear the content area if we don't have custom
    // content because it might contain the @banner that didn't
    // fit in the banner mode.
    if (self->scrollArea && (!self->customContent || params->clear)) {

        if (oldFocus && clutter_actor_contains ((ClutterActor *)self->scrollArea, oldFocus))
            clutter_actor_grab_key_focus (CLUTTER_ACTOR(self->actor));

        clutter_actor_destroy (CLUTTER_ACTOR (self->scrollArea));
        self->scrollArea = NULL;
        self->contentArea = NULL;
    }

    if (self->actionArea && params->clear) {

        if (oldFocus && clutter_actor_contains (CLUTTER_ACTOR (self->actionArea), oldFocus))
            clutter_actor_grab_key_focus (CLUTTER_ACTOR (self->actor));

        clutter_actor_destroy (CLUTTER_ACTOR (self->actionArea));

        self->actionArea = NULL;
        self->buttonBox = NULL;
    }

    if (self->imageBin && params->clear) {
        cdos_notification_unset_image (self);
    }

    if (!self->scrollArea && !self->actionArea && !self->imageBin)
        st_widget_remove_style_class_name (self->table, "multi-line-notification");
    
    if (!self->icon) {
        //self->icon = params->icon || this.source.createNotificationIcon();
        // 先不考虑从source, cdos_source_create_notification_icon (self->source);
        self->icon = params->icon;
        clutter_container_add_actor ((ClutterContainer *)self->table, (ClutterActor *)self->icon);
        ClutterChildMeta *meta = clutter_container_get_child_meta ((ClutterContainer *)self->table, (ClutterActor *)self->icon);
        g_object_set (meta, "row", 0,
                            "col", 0,
                            "x-expand", FALSE,
                            "y-expand", FALSE,
                            "y-fill", FALSE,
                            "y-align", ST_ALIGN_START,
                            NULL);
    }

    gchar *new_title = g_strdup (title);
    //title = title ? _fixMarkup(title.replace(/\n/g, ' '), params.titleMarkup) : '';
    if (new_title) {
        GRegex *regex = g_regex_new ("\n", 0, 0, NULL);
        self->title = g_regex_replace (regex, new_title, -1, 0, " ", 0, NULL);
        g_regex_unref (regex);
        g_free (new_title);
    }
    else
        self->title = g_strdup ("");

    gchar *strong_str = g_strconcat ("<b>", self->title, "</b>", NULL);
    clutter_text_set_markup ((ClutterText *)st_label_get_clutter_text (self->titleLabel), strong_str);
    g_free (strong_str);
    
    gchar *time = g_date_time_format (self->timestamp, DATE_FORMAT);
    clutter_text_set_markup (CLUTTER_TEXT (st_label_get_clutter_text(self->timeLabel)), time);
    g_free (time);
    clutter_actor_hide (CLUTTER_ACTOR (self->timeLabel));

    if (pango_find_base_dir (new_title, -1) == PANGO_DIRECTION_RTL)
        self->titleDirection = ST_TEXT_DIRECTION_RTL;
    else 
        self->titleDirection = ST_TEXT_DIRECTION_LTR;
    
    // Let the title's text direction control the overall direction
    // of the notification - in case where different scripts are used
    // in the notification, this is the right thing for the icon, and
    // arguably for action buttons as well. Labels other than the title
    // will be allocated at the available width, so that their alignment
    // is done correctly automatically.
    st_widget_set_direction (self->table, self->titleDirection);

    // Unless the notification has custom content, we save this._bannerBodyText
    // to add it to the content of the notification if the notification is
    // expandable due to other elements in its content area or due to the banner
    // not fitting fully in the single-line mode.
    self->bannerBodyText = self->customContent? NULL : g_strdup (banner);
    self->bannerBodyMarkup = params->bannerMarkup;
    gchar *new_banner = NULL;
    if (banner) {
        GRegex *regex = g_regex_new ("\n", 0, 0, NULL);
        new_banner = g_regex_replace (regex, banner, -1, 0, "  ", 0, NULL);
        g_regex_unref (regex);
    }
    cdos_url_highlighter_set_markup (self->bannerUrlHighlighter, new_banner , params->bannerMarkup);
    g_free (new_banner);
    clutter_actor_queue_relayout (CLUTTER_ACTOR (self->bannerLabel));
        
    if (self->bannerBodyText && index (self->bannerBodyText, '\n') != NULL)
        addBannerBody (self);
    if (params->body) 
        cdos_notification_add_body (self,  params->body, params->bodyMarkup, NULL);
    updated (self);
}

/**
 * cdos_notification_set_icon_visible:
 * @self: A #CdosNotification
 * @visible: 是否可见
 *
 * 设置notification的icon是否可见
 * 
 */
void cdos_notification_set_icon_visible (CdosNotification *self, gboolean visible){
    if (visible)
        clutter_actor_show ((ClutterActor *)self->icon);
    else
        clutter_actor_hide ((ClutterActor *)self->icon);
}

/**
 * cdos_notification_add_actor:
 * @actor: actor to add to the body of the notification
 *
 * 把actor添加到消息的body里面
 */
void cdos_notification_add_actor (CdosNotification *self, ClutterActor *actor, const char *style){

    if (!self->scrollArea)
        createScrollArea (self);

    clutter_container_add_actor ((ClutterContainer *)self->contentArea, actor);
    if (style)
        st_widget_set_style ((StWidget *)actor, style);

    updated (self);
}

/**
 * cdos_notification_add_body:
 * @text: the text
 * @markup: %TRUE if @text contains pango markup
 * @style: style to use when adding the actor containing the text
 *
 * 往通知中加入多行label, 包括text.
 *
 * Return value: 新加入的 label
 */
StLabel *cdos_notification_add_body (CdosNotification *self, const gchar *text, gboolean markup, const  gchar *style){
    CdosUrlHighlighter *label = cdos_url_highlighter_new (text, TRUE, markup);
    //g_print("\n\n INFO: text %s \n\n", text);
    cdos_notification_add_actor (self, CLUTTER_ACTOR (label->actor), style);
    return label->actor;
}

/**
 * cdos_notification_scroll_to:
 * @side类型為 StSide, 定义在st-theme-node.h中
 *
 * Scrolls the content area (if scrollable) to the indicated edge
 */
void cdos_notification_scroll_to (CdosNotification *self , StSide side) {  
    /*
    StScrollBar *vscroll = ST_SCROLL_BAR(st_scroll_view_get_vscroll_bar (self->scrollArea));
    
    StAdjustment *adjustment = st_scroll_bar_get_adjustment (vscroll);    

    gdouble value, lower, upper, page_size;

    st_adjustment_get_values (adjustment, &value, &lower, &upper, NULL, NULL, &page_size);

    if (side ==  ST_SIDE_TOP)
        st_adjustment_set_value (adjustment , lower);
    else if (side ==  ST_SIDE_BOTTOM)
        st_adjustment_set_value (adjustment , upper);
    */
}

/**
 * cdos_notification_set_action_area:
 * @actor: the actor
 * @props: (option) St.Table child properties
 *
 * 把actor加入action area,代替之前其中的内容
 */
void cdos_notification_set_action_area (CdosNotification *self , StBoxLayout *actor , CStTableChildProps *props) {
    if (self->actionArea) {

        clutter_actor_destroy (CLUTTER_ACTOR(self->actionArea));
        self->actionArea = NULL;

        if (self->buttonBox) {
            clutter_actor_destroy((ClutterActor *)self->buttonBox);
            self->buttonBox = NULL;
        }

    } else {
        addBannerBody (self);
    }

    self->actionArea = actor;

    st_widget_add_style_class_name (self->table, "multi-line-notification");

    clutter_container_add_actor ((ClutterContainer *)self->table, (ClutterActor *)self->actionArea);
    StTableChild *meta = (StTableChild *)clutter_container_get_child_meta ((ClutterContainer *)self->table, (ClutterActor *)self->actionArea);
    if (props)
    {
        meta->col_span = props->col_span;
        meta->row_span = props->row_span;
        meta->x_align  = props->x_align;
        meta->y_align  = props->y_align;
        meta->x_expand = props->x_expand;
        meta->y_expand = props->y_expand;
        meta->x_fill   = props->x_fill;
        meta->y_fill   = props->y_fill;
    }
    g_object_set (meta, "col", 2, "row", 2, NULL);
    updateLastColumnSettings (self);
    updated (self);

}

/**
 * cdos_notification_set_image:
 * @self: A #CdosNotification
 * @Notification: 图片元素
 *
 * 设置带图片的通知
 *
 */
void cdos_notification_set_image (CdosNotification *self , ClutterActor *image) {

    if (self->imageBin) {
        cdos_notification_unset_image (self);
    }

    self->imageBin = ST_BIN (st_bin_new ());
    
    st_bin_set_child (self->imageBin, image);

    clutter_actor_set_opacity (CLUTTER_ACTOR (self->imageBin) , 230);

    st_widget_add_style_class_name (self->table, "multi-line-notification");

    st_widget_add_style_class_name (self->table, "notification-with-image");

    addBannerBody (self);

    updateLastColumnSettings (self);

    clutter_container_add_actor ((ClutterContainer *)self->table, (ClutterActor *)self->imageBin);
    /* StTableChild */
    ClutterChildMeta *meta = clutter_container_get_child_meta ((ClutterContainer *)self->table, (ClutterActor *)self->imageBin);
    g_object_set (meta, "x-fill", FALSE,
                        "y-fill", FALSE,
                        "row", 1,
                        "col", 1,
                        "row-span", 2,
                        "x-expand", FALSE,
                        "y-expand", FALSE,
                        NULL);
}

/**
 * cdos_notification_unset_image:
 * @self: A #CdosNotification
 *
 * 移除通知所带的图片
 *
 */
void cdos_notification_unset_image (CdosNotification *self) {

    if (self->imageBin) {
        st_widget_remove_style_class_name (self->table, "notification-with-image");
        st_container_remove_actor ((ClutterContainer *)self->table, (ClutterActor *)self->imageBin);
        self->imageBin = NULL;
        updateLastColumnSettings (self);

        if (!self->scrollArea && !self->actionArea){
            st_widget_remove_style_class_name (self->table, "multi-line-notification");
        }
    }
}

/**
 * cdos_notification_add_button:
 * @id: the action ID
 * @label: the label for the action's button
 *
 * 给通知添加一个按钮并且设置标签, 所有的通知按钮都出现在底部,单独占一行.
 *
 * 如果按钮被点击,通知则会发送 %action-invoked 信号,并且把id作为一个参数一齐发送.
 */
void cdos_notification_add_button (CdosNotification *self, gchar *id, gchar *label) {
    if (!self->buttonBox) {
        StBoxLayout *box = ST_BOX_LAYOUT(st_box_layout_new ());
        clutter_actor_set_name((ClutterActor *)box, "notification-actions");
        CStTableChildProps props;
        cst_table_child_init (&props);
        props.x_expand = TRUE;
        props.y_expand = FALSE;
        props.x_fill = TRUE;
        props.y_fill = FALSE;
        props.x_align = ST_ALIGN_START;
        cdos_notification_set_action_area (self, box, &props);
        self->buttonBox = box;
    }

    StWidget *button;
    button = st_button_new ();
    st_widget_set_can_focus (button, TRUE); 
    
    if (self->useActionIcons &&  gtk_icon_theme_has_icon (gtk_icon_theme_get_default (), id)) {
    
        StIcon *icon = ST_ICON (st_icon_new ());

        st_icon_set_icon_name (icon, id);

        st_widget_add_style_class_name (button, "notification-icon-button");

        st_bin_set_child (ST_BIN(button), CLUTTER_ACTOR(icon));        
    } else {
        st_widget_add_style_class_name (button, "notification-button");
        st_button_set_label ((StButton *)button, label);
    }

    if (clutter_actor_get_n_children (CLUTTER_ACTOR(self->buttonBox)) > 0)
        st_focus_manager_remove_group (self->buttonFocusManager , ST_WIDGET(self->buttonBox));

    clutter_container_add_actor ((ClutterContainer *)self->buttonBox, CLUTTER_ACTOR (button));
    st_focus_manager_add_group (self->buttonFocusManager , ST_WIDGET(self->buttonBox));

    clutter_actor_set_name ((ClutterActor *)button, id);
    g_signal_connect (button, "clicked", G_CALLBACK (onActionInvoked), self);

    updated (self);
}

/**
 * cdos_notification_set_urgency:
 * @self: A #CdosNotification
 * @urgency: 通知的紧急程度 
 *
 * 设置通知的紧急程度
 *
 */
void cdos_notification_set_urgency (CdosNotification *self, guint urgency) {
    self->urgency = urgency;
}

/**
 * cdos_notification_set_resident:
 * @self: A #CdosNotification
 * @resident: 是否是常驻通知
 *
 * 设置通知是否是常驻
 *
 */
void cdos_notification_set_resident (CdosNotification *self, gboolean resident) {
    self->resident = resident;
}

/**
 * cdos_notification_set_transient:
 * @self: A #CdosNotification
 * @isTransient: 是否是短暂的
 *
 * 设置通知是否是短暂的
 *
 */
void cdos_notification_set_transient (CdosNotification *self, gboolean isTransient) {
    self->isTransient = isTransient;
}

/**
 * cdos_notification_set_use_action_icons:
 * @self: A #CdosNotification
 * @useIcons: 是否使用图标
 *
 * 设置通知是否使用图标
 *
 */
void cdos_notification_set_use_action_icons (CdosNotification *self, gboolean useIcons) {
    self->useActionIcons = useIcons;
}

/**
 * cdos_notification_expand:
 * @self: A #CdosNotification
 * @animate: 默认大小能完全显示时,是否扩展通知
 *
 * 设置通知是否是自动调节大小
 *
 */
void cdos_notification_expand (CdosNotification *self , gboolean animate) {
    self->expanded = TRUE;
    // The banner is never shown when the title did not fit, so this
    // can be an if-else statement.

    if (!self->titleFitsInBannerMode) {
        // Remove ellipsization from the title label and make it wrap so that
        // we show the full title when the notification is expanded.
        ClutterText *text = (ClutterText *)st_label_get_clutter_text (self->titleLabel);
        clutter_text_set_line_wrap (text, TRUE);
        clutter_text_set_line_wrap_mode (text, PANGO_WRAP_WORD_CHAR);
        clutter_text_set_ellipsize (text , PANGO_ELLIPSIZE_END);
    } 
    else if (st_table_get_row_count ((StTable *)self->table) > 1 && clutter_actor_get_opacity (CLUTTER_ACTOR(self->bannerLabel)) != 0) {
        // We always hide the banner if the notification has additional content.
        //
        // We don't need to wrap the banner that doesn't fit the way we wrap the
        // title that doesn't fit because we won't have a notification with
        // row_count=1 that has a banner that doesn't fully fit. We'll either add
        // that banner to the content of the notification in _bannerBoxAllocate()
        // or the notification will have custom content.
        if (animate) {
            clutter_actor_save_easing_state ((ClutterActor *)self->bannerLabel);
            clutter_actor_set_easing_mode ((ClutterActor *)self->bannerLabel, CLUTTER_EASE_OUT_QUAD);
            clutter_actor_set_easing_duration ((ClutterActor *)self->bannerLabel, ANIMATION_TIME);
            clutter_actor_set_opacity ((ClutterActor *)self->bannerLabel, 0);
            clutter_actor_restore_easing_state ((ClutterActor *)self->bannerLabel);
        }            
        else
            clutter_actor_set_opacity (CLUTTER_ACTOR(self->bannerLabel) , 0);
    }
    g_signal_emit (self, cdos_notification_signals[EXPANDED], 0);  //发向SummaryItem

}

/**
 * cdos_notification_collapseCompleted:
 * @self: A #CdosNotification
 *
 * 收缩通知:如果可删除,什么都不做;否则,不自动换行.
 *
 */
void cdos_notification_collapseCompleted (CdosNotification *self) {
    if (self->destroyed)
        return;

    self->expanded = FALSE;

    // Make sure we don't line wrap the title, and ellipsize it instead.
    clutter_text_set_line_wrap (CLUTTER_TEXT(st_label_get_clutter_text(self->titleLabel)), TRUE);

    clutter_text_set_ellipsize (CLUTTER_TEXT(st_label_get_clutter_text(self->titleLabel)),
                                                    PANGO_ELLIPSIZE_END);

    // Restore banner opacity in case the notification is shown in the
    // banner mode again on update.
    clutter_actor_set_opacity (CLUTTER_ACTOR(self->bannerLabel) , 255);

    g_signal_emit (self, cdos_notification_signals[COLLAPSED], 0);  //信号发给SummaryItem

}

/**
 * cdos_notification_destroy:
 * @self: A #CdosNotification
 * @reason: 删除的原因 
 *
 * 删除通知,并给出原因.
 *
 */
void cdos_notification_destroy (CdosNotification *self, guint reason) {
    //g_print("\nLINE: %d, FUNC: %s -----> is self valid?(%d)\n", __LINE__, __FUNCTION__, (NULL != self && CDOS_IS_NOTIFICATION(self)));
    g_return_if_fail(CDOS_IS_NOTIFICATION(self));
    self->destroyedReason = reason;
    if (self->actor) {
        clutter_actor_destroy ((ClutterActor *)self->actor);
        self->actor = NULL;
    }
    //FIXME:函数不允许参数为空
    //cdos_draggable_set_delegate (self->actor_delegate, NULL); 
}


gboolean cdos_notification_get_resident (CdosNotification *self) {
    return self->resident;    
}
