//
// Created by zhangtao on 16-3-17.
// source:overview.js
/**
 * SECTION: cdos-overview
 * @short_description:用于工作起切换时，生成各工作区的概观图示。
 * @title: CdosOverview
 * @include: cdos-overview.h
 *
 * #CdosOverview在激活工作区切换时，生成各工作区相应可视化缩略图，辅助标识当前各个工作区的桌面视图状态。
 *
 * CdosOverview结构关系:
 *
 * |________ cover_pane:ClutterActor
 *
 * |____ background.actor:ClutterActor
 *
 * |____ background_shade:StWidget
 *
 * |____ workspaces_view_actor:ClutterActor
 */
//添加依赖库;

#define CLUTTER_DISABLE_DEPRECATION_WARNINGS

#include <math.h>
#include <meta/boxes.h>
#include <meta/meta-background-group.h>

#include "cdos-overview.h"
#include "cdos-overview-thumbnails-box.h"
#include "cdos-workspaces-view.h"

#include "cdos-layout-manager.h"
#include "cdos-global.h"
#include "cdos-start.h"
#include "cdos-background.h"

G_DEFINE_TYPE (CdosOverview, cdos_overview, ST_TYPE_WIDGET);

#define CDOS_OVERVIEW_GET_PRIVATE(obj) \
    (G_TYPE_INSTANCE_GET_PRIVATE ((obj), CDOS_TYPE_OVERVIEW, CdosOverviewPrivate))

#define ANIMATION_TIME   200

enum {
    OVERVIEW_DESTROY,
    OVERVIEW_HIDING,
    OVERVIEW_HIDDEN,
    OVERVIEW_SHOWING,
    OVERVIEW_SHOWN,
    LAST_SIGNAL
};

static CdosOverview *overview = NULL;

struct _CdosOverviewPrivate{
    CdosWorkspacesView *workspaces_view;
    ClutterActor *thumbnails_box;

    gdouble spacing;
    gboolean visible;
    gboolean shown;
    gboolean show_temporarily;
    gboolean modal;
    gboolean hide_in_progress;

    gfloat drag_start_x;
    gfloat drag_start_y;

    gfloat drag_x;
    gfloat drag_y;

    gdouble drag_start_value;
    gint last_motion_time;

    ClutterActor *cover_pane;
    ClutterActor *background_group;
    CdosBackgroundManager *background;

    guint captured_event_id;
    guint button_press_id;
    guint button_release_id;
    guint32 window_switch_timestamp;
    guint window_switch_timeout_id;
    guint last_active_workspace_index;
    MetaWindow *last_hovered_window;
    
    gint result; // 为了传递动画complete 回调参数

};


static void cdos_overview_dispose(GObject *object);
static void cdos_overview_finalize(GObject *object);

static gint cdos_overview_signals[LAST_SIGNAL] = { 0 };

static void  cdos_overview_group_style_changed_callback(StWidget *actor,CdosOverview *self);

static void  cdos_overview_layout_manager_monitors_changed_callback(CdosLayoutManager *cdos_layout_manager,CdosOverview *self);

static gboolean cdos_overview_on_captured_event(ClutterStage *stage, ClutterActor *actor,ClutterEvent *event, gpointer data);

static void cdos_overview_animate_visible(CdosOverview *self);

static gboolean cdos_overview_sync_input_mode(CdosOverview *self);

static void cdos_overview_animate_not_visible(CdosOverview *self);

static void cdos_overview_show_done(ClutterTimeline *tml, CdosOverview *self);

static void cdos_overview_hide_done(ClutterTimeline *tml, CdosOverview *self);

static gboolean cdos_overview_cover_pane_event_callback(ClutterActor *actor,ClutterEvent *event,CdosOverview *self);

static void
cdos_overview_init (CdosOverview *self)
{
    CdosOverviewPrivate *priv;
    priv = self->priv = CDOS_OVERVIEW_GET_PRIVATE (self);

    priv->spacing = 0;
    priv->captured_event_id = 0;
    priv->button_press_id = 0;
    priv->button_release_id = 0;

    priv->shown = FALSE;
    priv->modal = FALSE;
    self->animation_in_progress = FALSE;
    priv->hide_in_progress = FALSE;
    priv->visible = FALSE;

    priv->window_switch_timeout_id = 0;
    priv->window_switch_timestamp = 0;
    priv->last_hovered_window = NULL;
}

static void
cdos_overview_group_style_changed_callback(StWidget *actor,CdosOverview *self)
{
    StThemeNode *node ;
    node = st_widget_get_theme_node (actor);
    gdouble spacing = st_theme_node_get_length (node,"spacing");

    CdosOverviewPrivate *priv = self->priv;

    if(spacing != priv->spacing)
        priv->spacing = spacing;
}

static void
cdos_overview_layout_manager_monitors_changed_callback(CdosLayoutManager *cdos_layout_manager,
                                                       CdosOverview *self)
{
    clutter_actor_hide ((ClutterActor *)self);
}

static gboolean 
cdos_overview_on_button_press(ClutterActor *actor,ClutterButtonEvent *event)
{
    return TRUE;
}

static gboolean
cdos_overview_on_button_release(ClutterActor *actor,ClutterButtonEvent *event)
{
    clutter_actor_hide (actor);
    return TRUE;
}

static gboolean
cdos_overview_on_key_release (ClutterActor *actor,
                              ClutterKeyEvent *event)
{
    guint symbol = clutter_event_get_key_symbol ((ClutterEvent *)event);
    if (symbol == CLUTTER_KEY_Escape)
    {
        clutter_actor_hide (actor);
        return TRUE;
    }
    return FALSE;
}

/*
 * 在 DND 中使用
 */
static gboolean 
cdos_overview_on_captured_event(ClutterStage *stage,ClutterActor *actor,ClutterEvent *event, gpointer data)
{
    /*
    gfloat stage_x,stage_y;
    GtkSettings *gtk_settings;
    gtk_settings = gtk_settings_get_default ();
    gint threshold;
    g_object_get(gtk_settings,"gtk-dnd-drag-threshold",&threshold,NULL);

    gint event_type = clutter_event_type (event);
    CdosOverview *self = (CdosOverview *)data;
    CdosOverviewPrivate *priv = CDOS_OVERVIEW_GET_PRIVATE (self);

    clutter_event_get_coords (event,&stage_x,&stage_y);
    gfloat new_value = priv->drag_start_value;
    gdouble min_value,  max_value,  page_size, value;
    cdos_debug ();
    */
    return FALSE;
}

static gboolean
cdos_overview_sync_input_mode(CdosOverview *self)
{
    CdosOverviewPrivate *priv;
    priv = self->priv;

    if(self->animation_in_progress){
        return TRUE;
    }

    if(priv->shown){
        if(!priv->modal){
            if(cdos_push_modal((ClutterActor *)self, 0, 0)){
                priv->modal = TRUE;
            }else{
                clutter_actor_hide ((ClutterActor *)self);
                return FALSE;
            }
        }
    }else{
        if(priv->modal){
            cdos_pop_modal ((ClutterActor *)self, 0);
            priv->modal = FALSE;
        }
    }
    return TRUE;
}

static void 
cdos_overview_animate_visible(CdosOverview *self)
{
    CdosOverviewPrivate *priv;
    priv = self->priv;

    if(priv->visible || self->animation_in_progress)
        return ;
    MetaScreen *screen = cdos_global_get_screen (global);
    gint screen_width,screen_height;
    meta_screen_get_size(screen,&screen_width,&screen_height);

    priv->background_group = meta_background_group_new ();
    clutter_actor_add_child ((ClutterActor *)self, priv->background_group);
    priv->background = cdos_background_manager_new (priv->background_group, 1, 0);
    clutter_actor_set_reactive (priv->background_group, TRUE);
    clutter_actor_show (priv->background_group);

    priv->visible = TRUE;
    self->animation_in_progress = TRUE;

    priv->cover_pane = clutter_rectangle_new();
    ClutterColor color = {0, 0, 0, 0};
    clutter_rectangle_set_color ((ClutterRectangle *)priv->cover_pane, &color);
    //clutter_actor_set_opacity(priv->cover_pane, 0);
    clutter_actor_set_reactive(priv->cover_pane,TRUE);

    clutter_actor_add_child ((ClutterActor *)self, priv->cover_pane);
    clutter_actor_set_position(priv->cover_pane, 0.f, 0.f);
    clutter_actor_set_size(priv->cover_pane, (gfloat)screen_width, (gfloat)screen_height);
    g_signal_connect(priv->cover_pane,"event",G_CALLBACK(cdos_overview_cover_pane_event_callback),self);

    meta_disable_unredirect_for_screen(screen);
    ClutterActor *window_group;
    window_group = cdos_global_get_window_group (global);
    clutter_actor_hide (window_group);
    
    CdosOverviewThumbnailsBox *thumbnails_box = cdos_overview_thumbnails_box_new ();
    clutter_actor_add_child ((ClutterActor *)self, (ClutterActor *)thumbnails_box);
    priv->thumbnails_box = (ClutterActor *)thumbnails_box;
    clutter_actor_set_position (priv->thumbnails_box, 0.f, 0.f);

    CdosWorkspacesView *workspaces_view = cdos_workspaces_view_new();
    ClutterActor *workspaces_view_actor = (ClutterActor *)workspaces_view;
    clutter_actor_add_child ((ClutterActor *)self, workspaces_view_actor);
    priv->workspaces_view = workspaces_view;

    //clutter_actor_set_opacity ((ClutterActor *)self, 0);
    CLUTTER_ACTOR_CLASS (cdos_overview_parent_class)->show ((ClutterActor *)self);

    clutter_container_raise_child ((ClutterContainer *)ui_group, (ClutterActor *)self, NULL);

    ClutterActor *bg = clutter_actor_get_first_child (priv->background_group);
    while (bg)
    {
        cdos_transition_simple (bg,
                                CLUTTER_EASE_OUT_QUAD,
                                ANIMATION_TIME,
                                "brightness", 0.55,
                                NULL);
        bg = clutter_actor_get_next_sibling (bg);
    }

    clutter_actor_set_child_above_sibling ((ClutterActor *)self, priv->cover_pane, NULL);
    cdos_transition_to ((ClutterActor *)self,
                        CLUTTER_EASE_OUT_QUAD,
                        ANIMATION_TIME,
                        0,
                        NULL, NULL,
                        "show-overview",
                        "opacity", TRANS_PROP_UCHAR, 255,
                        "signal::completed", cdos_overview_show_done, self,
                        NULL);
    clutter_actor_show(priv->cover_pane);
    g_signal_emit(self,cdos_overview_signals[OVERVIEW_SHOWING],0,NULL);
}

static void
cdos_overview_animate_not_visible(CdosOverview *self){
    CdosOverviewPrivate *priv;
    priv = CDOS_OVERVIEW_GET_PRIVATE (self);

    if(!priv->visible || self->animation_in_progress){
        return ;
    }

    self->animation_in_progress = TRUE;
    priv->hide_in_progress = TRUE;

    //cdos_start_enable_panels();

    clutter_actor_hide((ClutterActor *)priv->workspaces_view);
    clutter_actor_hide (priv->thumbnails_box);

        clutter_actor_save_easing_state ((ClutterActor *)self);
        ClutterTransition *transition = clutter_property_transition_new ("opacity");
        clutter_actor_set_easing_mode ((ClutterActor *)self, CLUTTER_EASE_OUT_QUAD);
        clutter_timeline_set_duration (CLUTTER_TIMELINE (transition), ANIMATION_TIME);
        clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (transition), 0);
        clutter_transition_set_to (transition, G_TYPE_UCHAR, 255);
        clutter_actor_add_transition ((ClutterActor *)self, "hide-overview", transition);
        g_signal_connect (transition, "completed", G_CALLBACK (cdos_overview_hide_done), self);
        clutter_actor_restore_easing_state ((ClutterActor *)self);

    clutter_actor_set_child_above_sibling ((ClutterActor *)self, priv->cover_pane, NULL);
    clutter_actor_show(priv->cover_pane);
    g_signal_emit(self,cdos_overview_signals[OVERVIEW_HIDING],0,NULL);
}

static void cdos_overview_show_done(ClutterTimeline *tml, CdosOverview *self)
{
    CdosOverviewPrivate *priv;
    priv = self->priv;
    clutter_actor_remove_transition ((ClutterActor *)self, "show-overview");
    self->animation_in_progress = FALSE;
    clutter_actor_hide(priv->cover_pane);

    g_signal_emit(self, cdos_overview_signals[OVERVIEW_SHOWN],0,NULL);

    if(!priv->shown){
        cdos_overview_animate_not_visible(self);
    }
    cdos_overview_sync_input_mode (self);
    cdos_global_sync_pointer (global);
}

static void
cdos_overview_hide_done(ClutterTimeline *tml, CdosOverview *self)
{
    CdosOverviewPrivate *priv;
    priv = self->priv;
    
    clutter_actor_remove_transition ((ClutterActor *)self, "hide-overview");
    clutter_actor_remove_child ((ClutterActor *)self, priv->cover_pane);
    if (CLUTTER_IS_ACTOR (priv->cover_pane))
        clutter_actor_destroy (priv->cover_pane);
    priv->cover_pane = NULL;

    meta_enable_unredirect_for_screen(cdos_global_get_screen (global));
    ClutterActor *window_group = cdos_global_get_window_group (global);
    clutter_actor_show (window_group);

    cdos_background_manager_destroy (priv->background);
    //clutter_actor_remove_child ((ClutterActor *)self, priv->background_group);
    clutter_actor_destroy (priv->background_group);
    priv->background_group = NULL;

    //clutter_actor_remove_child ((ClutterActor *)self, (ClutterActor *)priv->workspaces_view);
    if (priv->workspaces_view)
        clutter_actor_destroy((ClutterActor *)priv->workspaces_view);
    priv->workspaces_view = NULL;
    //clutter_actor_remove_child ((ClutterActor *)self, (ClutterActor *)priv->thumbnails_box);
    if (priv->thumbnails_box)
        clutter_actor_destroy((ClutterActor *)priv->thumbnails_box);
    priv->thumbnails_box = NULL;

    priv->visible = FALSE;
    self->animation_in_progress = FALSE;
    priv->hide_in_progress = FALSE;
    priv->shown = FALSE;

    CLUTTER_ACTOR_CLASS (cdos_overview_parent_class)->hide ((ClutterActor *)self);

    //clutter_container_remove_actor ((ClutterContainer *)ui_group, (ClutterActor *)self);

    g_signal_emit(self,cdos_overview_signals[OVERVIEW_HIDDEN],0,NULL);

    if(priv->shown || self->animation_in_progress){
        cdos_overview_animate_visible(self);
    }
    cdos_overview_sync_input_mode(self);

    CdosLayoutManager *layout_manager = cdos_layout_manager_get_default();
    cdos_layout_manager_update_regions (layout_manager);
}

static gboolean
cdos_overview_cover_pane_event_callback(ClutterActor *actor,ClutterEvent *event,CdosOverview *self){
    return TRUE;
}


/**
 * cdos_overview_new:
 *
 * 创建一个新的实例对象；
 *
 * Returns:返回一个#CdosOverview的实例对象
 */
CdosOverview *
cdos_overview_new(void)
{
    if (overview)
        return overview;

    CdosOverview *self;
    self = g_object_new (CDOS_TYPE_OVERVIEW,  
                         "name", "overview",
                         "style-class", "overview-group",
                         "reactive", TRUE,
                         "visible", FALSE,
                         NULL);

    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    MetaRectangle *primary = cdos_layout_manager_get_primary_monitor (layout);
    clutter_actor_set_position ((ClutterActor *)self, primary->x, primary->y);
    clutter_actor_set_size ((ClutterActor *)self, primary->width, primary->height);

    clutter_container_add_actor ((ClutterContainer *)ui_group, (ClutterActor *)self);

    g_signal_connect (layout, "monitors-changed", 
            G_CALLBACK (cdos_overview_layout_manager_monitors_changed_callback), self);

    overview = self;
    return self;
}

CdosOverview *
cdos_overview_get ()
{
    return cdos_overview_new ();
}
/**
 * cdos_overview_toggle:
 * @self: 当前实例对象
 *
 * 来回切换当前实例元素的显示和隐藏。
 */
void  cdos_overview_toggle(CdosOverview *self){
    g_return_if_fail(CDOS_IS_OVERVIEW(self));
    CdosOverviewPrivate *priv;
    priv = CDOS_OVERVIEW_GET_PRIVATE (self);

    if(priv->shown){
        clutter_actor_hide((ClutterActor *)self);
    }else{
        clutter_actor_show ((ClutterActor *)self);
    }
}

static void
cdos_overview_show (ClutterActor *actor){
    g_return_if_fail(CDOS_IS_OVERVIEW(actor));
    CdosOverview *self = (CdosOverview *)actor;
    CdosOverviewPrivate *priv;
    priv = CDOS_OVERVIEW_GET_PRIVATE (self);
    if(priv->shown){
        return ;
    }
    priv->shown = TRUE;
    //clutter_container_add_actor ((ClutterContainer *)ui_group, (ClutterActor *)self);

    cdos_overview_sync_input_mode (self);

    cdos_overview_animate_visible(self);
}

static void
cdos_overview_hide (ClutterActor *actor)
{
    CdosOverview *self = (gpointer)actor;
    g_return_if_fail (CDOS_IS_OVERVIEW (self));

    CdosOverviewPrivate *priv;
    priv = self->priv;

    if(!priv->shown){
        return ;
    }
    priv->shown = FALSE;
    cdos_overview_animate_not_visible(self);
    cdos_overview_sync_input_mode(self);
}

gboolean
cdos_overview_get_hidding(CdosOverview *self){
    g_return_val_if_fail(CDOS_IS_OVERVIEW(self),FALSE);
    CdosOverviewPrivate *priv;
    priv = self->priv;

    return priv->hide_in_progress;
}
/**
 * cdos_overview_get_visible:
 * @self: 当前实例对象
 *
 * 获取当前实例元素是否为可见状态。
 *
 * Returns: TRUE-可见，FALSE-不可见
 */
gboolean cdos_overview_get_visible(CdosOverview *self){
    g_return_val_if_fail(CDOS_IS_OVERVIEW(self),FALSE);
    CdosOverviewPrivate *priv;
    priv = self->priv;

    return priv->visible;
}

static void cdos_overview_dispose(GObject *object){
    CdosOverview *self = CDOS_OVERVIEW(object);
    CdosOverviewPrivate *priv = self->priv;
    if(priv->captured_event_id > 0){
        g_signal_handler_disconnect(CLUTTER_ACTOR(cdos_global_get_stage (global)),priv->captured_event_id);
        priv->captured_event_id = 0;
    }
    if(priv->visible)
        clutter_actor_hide ((ClutterActor *)self);

    G_OBJECT_CLASS(cdos_overview_parent_class)->dispose(object);
}

static void
cdos_overview_finalize(GObject *object)
{
    //CdosOverview *self = CDOS_OVERVIEW(object);
    overview = NULL;
    G_OBJECT_CLASS(cdos_overview_parent_class)->finalize(object);
}

static void
cdos_overview_allocate (ClutterActor *actor, const ClutterActorBox *box, ClutterAllocationFlags flags)
{
    CLUTTER_ACTOR_CLASS (cdos_overview_parent_class)->allocate (actor, box, flags);

    CdosOverview *self = (gpointer)actor;
    CdosOverviewPrivate *priv = self->priv;
    if (priv->background_group)
        clutter_actor_allocate_preferred_size (priv->background_group, flags);
    if (priv->thumbnails_box)
    {
        CdosLayoutManager *layout = cdos_layout_manager_get_default ();
        MetaRectangle *rect = cdos_layout_manager_get_primary_monitor (layout);
        ClutterActorBox box1;
        box1.x1 = rect->x;
        box1.x2 = rect->x + rect->width;
        box1.y1 = rect->y;
        box1.y2 = rect->y + rect->height;

        //clutter_actor_allocate (priv->background->actor, &box1, flags);
        clutter_actor_allocate (priv->cover_pane, &box1, flags);
        clutter_actor_allocate ((ClutterActor *)priv->workspaces_view, &box1, flags);

        gfloat height = clutter_actor_get_height (priv->thumbnails_box);
        box1.x1 = rect->x;
        box1.x2 = rect->x + rect->width;
        box1.y1 = rect->y + rect->height - height;
        box1.y2 = rect->y + rect->height;
        StThemeNode *theme = st_widget_get_theme_node ((StWidget *)priv->thumbnails_box);
        ClutterActorBox content_box;
        st_theme_node_get_content_box (theme, &box1, &content_box);
        /* FIXME */
        content_box.y1 = rect->y + rect->height - height;
        content_box.y2 = rect->y + rect->height;
        
        clutter_actor_allocate (priv->thumbnails_box, &content_box, flags);

        height = max (height, rect->height * thumbnail_box_scale);
        box1.x1 = rect->x;
        box1.x2 = rect->x + rect->width;
        box1.y1 = rect->y;
        box1.y2 = rect->y + rect->height;
        gfloat scale = (rect->height - height * 2.0) / (gfloat)rect->height;
        clutter_actor_set_pivot_point ((ClutterActor *)priv->workspaces_view, 0.5, 0.5);
        clutter_actor_set_scale ((ClutterActor *)priv->workspaces_view, scale, scale);
        gfloat x = rect->width * (1 - scale) / 2.0;
        box1.x1 += x;
    }
}

static void 
cdos_overview_paint (ClutterActor *actor)
{
    CLUTTER_ACTOR_CLASS (cdos_overview_parent_class)->paint (actor);
    ClutterActor *child;
    for (child = clutter_actor_get_first_child (actor);
         child;
         child = clutter_actor_get_next_sibling (child))
        clutter_actor_paint (child);
}
static void
cdos_overview_class_init(CdosOverviewClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->dispose = cdos_overview_dispose;
    gobject_class->finalize = cdos_overview_finalize;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->allocate = cdos_overview_allocate;
    actor_class->paint = cdos_overview_paint;
    /*
    actor_class->get_prefered_width = cdos_overview_get_preferred_width;
    actor_class->get_prefered_height = cdos_overview_get_preferred_width;
    */
    actor_class->show = cdos_overview_show;
    actor_class->hide = cdos_overview_hide;

    actor_class->button_press_event = cdos_overview_on_button_press;
    actor_class->button_release_event = cdos_overview_on_button_release;
    actor_class->key_release_event = cdos_overview_on_key_release;

    g_type_class_add_private (klass, sizeof (CdosOverviewPrivate));
    /**
     * CdosOverview::hidden
     * @self: 信号发射实例
     *
     * 实例元素动画隐藏已经完成。
     */
    cdos_overview_signals[OVERVIEW_HIDDEN]  =
            g_signal_new(
                    "hidden",
                    G_TYPE_FROM_CLASS(klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL,NULL,
                    g_cclosure_marshal_VOID__VOID,
                    G_TYPE_NONE,
                    0
            );
    /**
     * CdosOverview::hiding
     * @self: 信号发射实例
     *
     * 实例元素动画隐藏正在进行中。
     */
    cdos_overview_signals[OVERVIEW_HIDING]  =
            g_signal_new(
                    "hiding",
                    G_TYPE_FROM_CLASS(klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL,NULL,
                    g_cclosure_marshal_VOID__VOID,
                    G_TYPE_NONE,
                    0
            );
    /**
     * CdosOverview::showing
     * @self: 信号发射实例
     *
     * 实例元素动画显示正在进行中。
     */
    cdos_overview_signals[OVERVIEW_SHOWING]  =
            g_signal_new(
                    "showing",
                    G_TYPE_FROM_CLASS(klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL,NULL,
                    g_cclosure_marshal_VOID__VOID,
                    G_TYPE_NONE,
                    0
            );
    /**
     * CdosOverview::shown
     * @self: 信号发射实例
     *
     * 实例元素动画显示已经完成。
     */
    cdos_overview_signals[OVERVIEW_SHOWN]  =
            g_signal_new(
                    "shown",
                    G_TYPE_FROM_CLASS(klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL,NULL,
                    g_cclosure_marshal_VOID__VOID,
                    G_TYPE_NONE,
                    0
            );
}
