#include <stdlib.h>
#include "cdos-provider.h"
#include "cdos-setting.h"
#include "cdos-setting-obj.h"
#include "cdos-settings-manager.h"
#include "cdos-base-util.h"
#include "cdos-global.h"
/**
 * SECTION: cdos-provider.c
 * @Title: CdosProvider
 * @short_description: 当我们修改桌面显示的配置时，我们会将每个applet配置
 *　 新建一个对象
 *
 * CdosProvider　为监听桌面config改变而进行画面更新的功能。 
 */

#define SETTING_SCHEMA_FILE "setting_schema.ini"

struct _CdosProviderPrivate
{
        GObject                 *xlet;
        gchar                   *uuid;
        gchar                   *instancedId;
        CdosExtensionType       *ext_type;
        gchar                   *xlet_str;
        gboolean                 valid;
        GFile                   *settings_file;
        CdosSettingObj          *settings_obj;
        GFile                   *applet_dir;    
        gboolean                 multi_instace;
        GHashTable              *metaBindings;
        GKeyFile                *key;
};

#define CDOS_PROVIDER_GET_PRIVATE(obj)    \
        (G_TYPE_INSTANCE_GET_PRIVATE((obj), CDOS_TYPE_PROVIDER, CdosProviderPrivate))

G_DEFINE_TYPE(CdosProvider, cdos_provider, G_TYPE_OBJECT);

static gboolean
cdos_provider_sanity_check(GKeyFile *key, const gchar *val, const gchar *group);

static gboolean 
cdos_provider_maybe_update_settings_file(CdosProvider *provider);

static void
cdos_provider_invalid_setting_type_error(const gchar *key_name,
                                         const gchar *uuid, 
                                         const gchar *type);

static void
cdos_provider_key_not_found_error(const gchar *key_name, 
                                  const gchar *uuid);

static void
cdos_provider_setting_not_initialized_error(const gchar *uuid);

void
cdos_provider_value_changed_notify(CdosProvider *provider, 
                                   const gchar *key, 
                                   const gchar* oldval, 
                                   const gchar *newval);

static gboolean
cdos_provider_do_upgrade(CdosProvider *provider, 
                         GKeyFile *new_key, 
                         GKeyFile *old_key, 
                         const gchar *checksum);

static gboolean 
search(const gchar *src, gchar **des, gint length);

static gboolean
check(CdosProvider *provider, GKeyFile *key, const gchar *group, const gchar *types);

static gboolean
cdos_provider_check_for_min_props(CdosProvider *provider, GKeyFile *key, const gchar *group);

static gboolean
cdos_provider_keyfile_validity_check(CdosProvider *provider, GKeyFile *key);

static gboolean 
cdos_provider_create_settings_file(CdosProvider *provider);

static gboolean
cdos_provider_get_is_multi_instance_xlet(const gchar *uuid);

static void 
cdos_provider_init(CdosProvider  *provider);

static void 
cdos_provider_class_init(CdosProviderClass  *klass);

static void
cdos_provider_finalize(GObject *gobject);
static void
cdos_provider_dispose(GObject *gobject);


/**
 * cdos_provider_new_with_params:
 * @xlet: xlet 对象 
 * @uuid: 对象的uuid
 * @instancedId: 对象的实例id
 * @type: 类型
 * @string: 描述的字符串
 * 
 * 新建一个 #CdosProvider 类的实例。
 *
 * Returns: 返回一个 #CdosProvider 的实例。
 */
CdosProvider *
cdos_provider_new_with_params(GObject *xlet, 
                            gchar *uuid, 
                            gchar *instancedId, 
                            CdosExtensionType *type, 
                            gchar *string)
{
        GObject *object;
        object = g_object_new(CDOS_TYPE_PROVIDER, NULL);
        cdos_provider_set_params(CDOS_PROVIDER(object), xlet, uuid, instancedId, type, string);
        return CDOS_PROVIDER(object);
}

/**
 * cdos_provider_set_params:
 * @provider: A #CdosProvider
 * @xlet: xlet对象
 * @uuid: uuid
 * @instancedId: 对象的id
 * @type: 类型
 * @string: 描述对象的字符串
 *
 * 设置provider对象的属性，当我们用cdos_provider_new创建一个对象时，还需要用此方法设置属性
 *
 */
void 
cdos_provider_set_params(CdosProvider *provider,
                         GObject *xlet, 
                         gchar *uuid,
                         gchar *instancedId, 
                         CdosExtensionType *type, 
                         gchar *string)
{
        g_return_if_fail(CDOS_IS_PROVIDER(provider));
        //notice:此处传过来的值不知道是什么？但是在cdos_setting_set_params中需要是GHashTable.
        CdosProviderPrivate *priv = provider->priv;

        if ((type != NULL) && (string != NULL)) {
                priv->ext_type = type;
                g_object_ref(priv->ext_type);
                priv->xlet_str =  g_strdup(string);
        }

        if (!xlet) {
                g_print("%s constructor arguments invalid\n", priv->xlet_str);
                g_print("First argument MUST be a(n) %s  object (use 'this' as the first argument)", priv->xlet_str);
                 return;        
        }        

        if (!uuid) {
                g_print("%s Settings constructor arguments invalid", priv->xlet_str);
                g_print("Missing required UUID as second argument - should be a string:");
                g_print("<xlet-name>@<your-id>.org or something similar");
                return;
        }

        priv->uuid = g_strdup(uuid);
        priv->xlet = xlet;
        if ((!instancedId) && (g_strcmp0(priv->ext_type->name, "Extension")) ) {
                g_print("Settings constructor arguments warning");
                g_print("Missing instanced ID as third argument");
                g_print("The UUID is %s", priv->uuid);
        }

        priv->instancedId = g_strdup(instancedId);
        priv->valid = FALSE;

        priv->applet_dir = cdos_extension_find_extension_directory(priv->uuid, priv->ext_type);
        if (!priv->applet_dir) {
                g_print("Could not find installation directory for %s", priv->uuid);
                return;
        }

        priv->multi_instace = cdos_provider_get_is_multi_instance_xlet(priv->uuid);
        if (priv->multi_instace && (!priv->instancedId)) {
                g_print("%s Settings fatal error!", priv->xlet_str);
                g_print("Multi-instanciable xlet with no instance ID supplied");
                g_print("The UUID is  %s", priv->uuid);
                return;
        }

        gchar *fn = priv->multi_instace ? instancedId : uuid;
        const gchar *home_dir = g_get_home_dir();
        if(home_dir == NULL){
            g_warning("%s:home directory is not exist!",__func__);
            return ;
        }
        gchar *settings_path = g_strconcat(home_dir, "/",
                                        ".cdos", "/", 
                                        "configs", "/",
                                         priv->uuid, "/",
                                           fn, ".ini", NULL);
        priv->settings_file = g_file_new_for_path(settings_path);
        g_free(settings_path);

        if(!g_file_query_exists(priv->settings_file, NULL)) {
                if (!cdos_provider_create_settings_file(provider)) {
                        g_print("Problem initializing setting for %s", priv->uuid);
                        return;
                }
        } else {
                if (!cdos_provider_maybe_update_settings_file(provider)) {
                        g_print("Problem updating setting for  %s", priv->uuid);
                        return;
                }
        }

        priv->metaBindings = g_hash_table_new(g_direct_hash, g_direct_equal);
        if (!priv->multi_instace) {
                priv->instancedId = priv->uuid;
        }

        priv->settings_obj = cdos_setting_obj_new(provider, 
                                                priv->settings_file, 
                                                priv->uuid, 
                                                priv->instancedId);
        priv->valid =TRUE;

        CdosSettingsManager *manager = CDOS_SETTINGS_MANAGER(cdos_settings_manager_get_default());
        cdos_settings_manager_register(manager, priv->uuid, priv->instancedId, G_OBJECT(provider));
}

/**
 * cdos_provider_new:
 * 
 * 新建一个 #CdosProvider 类的实例，还需要使用cdos_provider_set_params设置属性。
 *
 * Returns: 返回一个 #CdosProvider 的实例。
 */
CdosProvider *
cdos_provider_new(void)
{
    return g_object_new(CDOS_TYPE_PROVIDER, NULL);
}

static void 
cdos_provider_init(CdosProvider  *provider)
{
    CdosProviderPrivate *priv;
	priv = provider->priv = CDOS_PROVIDER_GET_PRIVATE(provider);
	
    priv->key = g_key_file_new();
	if (!g_key_file_load_from_file(priv->key, "key.ini", G_KEY_FILE_NONE, NULL)) {
        g_print("can not load file\n");
    }
}

static void 
cdos_provider_class_init(CdosProviderClass  *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
//    GParamSpec *pspec;
    gobject_class->dispose = cdos_provider_dispose;
    gobject_class->finalize = cdos_provider_finalize;
    g_type_class_add_private(klass, sizeof(CdosProviderPrivate));
}

static gboolean
cdos_provider_get_is_multi_instance_xlet(const gchar *uuid)
{       
        return FALSE;
}

static gboolean 
cdos_provider_create_settings_file(CdosProvider *provider)
{
    CdosProviderPrivate *priv = provider->priv;

    GFile *file = g_file_get_parent(priv->settings_file);
    if (!g_file_query_exists(file, NULL)) {
            g_file_make_directory_with_parents(file, NULL, NULL);
    }
    if(file){
        g_object_unref(file);
    }
    GFile *orig_file = g_file_get_child(priv->applet_dir, SETTING_SCHEMA_FILE);

    if (!g_file_query_exists(orig_file, NULL)) {
        g_print("Failed to locate setting schema file for %s\n", priv->uuid);
        if(orig_file){
            g_object_unref(orig_file);
        }
        return FALSE;
    }

    gchar *path = g_file_get_path(orig_file);
    gchar *init_file_content = cdos_base_util_get_file_contents_utf8_sync(path, NULL);
    gchar *checksum = cdos_global_get_md5_for_string(init_file_content);

    GKeyFile *key = g_key_file_new();
    gboolean load_from_file = g_key_file_load_from_file(key, path, G_KEY_FILE_NONE, NULL);
    gboolean keyfile_validity_check = cdos_provider_keyfile_validity_check(provider, key);
    if(orig_file){
        g_object_unref(orig_file);
    }
    g_free(path);
    g_free(init_file_content);

    if (!load_from_file) {
            g_print("Cannot parse setting schema file for %s . Check teh sructer for missing commas, etc... Error \n", priv->uuid);
            g_free(checksum);
            g_key_file_free(key);
            return FALSE;
    }

    if (!keyfile_validity_check) {
        g_print("Initial settings schema file is not valid for %s\n", priv->uuid);
        g_free(checksum);
        g_key_file_free(key);
       return FALSE;
    }

    gsize length;
    gint i;
    gchar *value;
    gchar **key_group = g_key_file_get_groups(key, &length);

    for (i = 0; i < length; i++)
    {
       value = g_key_file_get_string(key, *(key_group+i), "default", NULL);
       g_key_file_set_string(key, *(key_group+i), "value", value);
    }

    g_key_file_set_string(key, "Md5", "__md5__", checksum);
    gchar *file_name = g_file_get_path(priv->settings_file);
    g_key_file_save_to_file(key, file_name, NULL);

    g_free(file_name);
    g_strfreev(key_group);
    g_free(checksum);
    g_key_file_free(key);
    return TRUE;
}

static gboolean
cdos_provider_keyfile_validity_check(CdosProvider *provider, GKeyFile *key)
{
        gboolean valid = FALSE;
        gsize len;
        int i;
        gchar **group = g_key_file_get_groups(key, &len);	
        for (i = 0; i < len; i++)
        {
	       valid = cdos_provider_check_for_min_props(provider, key, *(group+i));
	       if (!valid)
                break;
        }
        g_strfreev(group);
        return valid;	
}

static gboolean
cdos_provider_check_for_min_props(CdosProvider *provider, GKeyFile *key, const gchar *group)
{
	gsize key_len;
	gsize group_len;
	
	gint i;
    gint j;
	gchar **value;
    CdosProviderPrivate *priv = provider->priv;
	gchar *node_type = g_key_file_get_string(key, group, "type", NULL);
	gchar **groups = g_key_file_get_groups(priv->key, &group_len);
	gboolean flag = FALSE;
	for (i = 0; i < group_len; i++)
	{
		value = g_key_file_get_keys(key, *(groups+i), &key_len, NULL);
		for (j = 0; j < key_len; j++) 
		{
			if (g_strcmp0(node_type, *(value+j)))
			{
                flag = TRUE;
                break;
			}
		}
        g_strfreev(value);
        value = NULL;
        if(flag){
            break ;
        }
	}
    g_free(node_type);
    g_strfreev(groups);

    if(flag == TRUE){
        return  check(provider, key, group, *(groups));
    }
	return FALSE; 
}

static gboolean
check(CdosProvider *provider, GKeyFile *key, const gchar *group, const gchar *types)
{
	gsize key_len = 0;   
	gsize group_len = 0;   
	gsize value_len = 0;   
	gint i;
	gint j;
    CdosProviderPrivate *priv = provider->priv;

	gchar *node_type = g_key_file_get_string(key, group, "type", NULL);
	gchar **boolean_types = g_key_file_get_keys(priv->key, types, &key_len, NULL);
	gchar **node = g_key_file_get_groups(key, &group_len);
	gchar **value = g_key_file_get_string_list(key, group, node_type, &value_len, NULL);
	gboolean flag = TRUE;
	for (i = 0; i < value_len; i++) 
	{
		for (j = 0; j < group_len; j++)
		{
			if(g_strcmp0(*(value+i),*(node+j))) {
                flag = FALSE;
                break ;
            }
		}
        if(flag){
            break ;
        }
	}
    g_strfreev(value);
    g_strfreev(node);
    g_strfreev(boolean_types);
    g_free(node_type);
	return flag;
}

static gboolean 
search(const gchar *src, gchar **des, gint length)
{
	int i;
	for(i = 0; i < length; i++)
	{
		if (!g_strcmp0(src, *(des+i)))
			return TRUE;
	}

	return FALSE;
}

static gboolean
cdos_provider_do_upgrade(CdosProvider *provider, 
                         GKeyFile *new_key, 
                         GKeyFile *old_key, 
                         const gchar *checksum)
{
	
	CdosProviderPrivate *priv = provider->priv;

	if (!cdos_provider_keyfile_validity_check(provider, new_key))
	{
		g_print("Upgrade settings schema file is NOT valid for %s\n", priv->uuid);
		g_print("Aborting settings upgrade.\n");
		return FALSE;
	}
	
	gsize new_key_group_len;
	gsize old_key_group_len;
	gchar **new_group = g_key_file_get_groups(new_key, &new_key_group_len);
	gchar **old_group = g_key_file_get_groups(old_key, &old_key_group_len);

	int i;
	gchar *new_key_value = NULL;
	gchar *old_key_value = NULL;
	for (i = 0; i < new_key_group_len; i++)
	{
		if (search(*(new_group+i), old_group, old_key_group_len))
		{
			new_key_value = g_key_file_get_string(new_key, *(new_group + i), "type", NULL);
			old_key_value = g_key_file_get_string(old_key, *(old_group + i), "type", NULL);
			if (g_strcmp0(new_key_value, old_key_value))
			{				
				if (cdos_provider_sanity_check(new_key, old_key_value, *(new_group+i)))
				{
                    g_free(new_key_value);
                    new_key_value = g_key_file_get_string(old_key, *(new_group + i), "value", NULL);
                    g_key_file_set_string(new_key, *(new_group), "value", new_key_value);
		              
				} else {
                    g_free(new_key_value);
					new_key_value = g_key_file_get_string(new_key, *(new_group + i), "default", NULL);
					g_key_file_set_string(new_key, *(new_group), "value", new_key_value);
				}
			} else {
                g_free(new_key_value);
			    new_key_value = g_key_file_get_string(new_key, *(new_group + i), "default", NULL);
			    g_key_file_set_string(new_key, *(new_group), "value", new_key_value);
			}

		} else {
			new_key_value = g_key_file_get_string(new_key, *(new_group + i), "default", NULL);
			g_key_file_set_string(new_key, *(new_group), "value", new_key_value);
		}
        if(new_key_value != NULL){
            g_free(new_key_value);
        }
        if(old_key_value != NULL){
            g_free(old_key_value);
        }

	}
    g_strfreev(new_group);
    g_strfreev(old_group);

	g_key_file_set_string(new_key, "Md5","md5", checksum);
    if (g_file_delete(priv->settings_file, NULL, NULL)) {
        gchar *file_name = g_file_get_path(priv->settings_file);
        g_key_file_save_to_file(new_key, file_name, NULL);

        g_free(file_name);
        g_print("Upgrade complete\n");
        return TRUE;
    } else {
        g_print("Failed to agin write access to save update settings for %s ... %s", priv->uuid, priv->instancedId);  
        return FALSE;
    }
}

static void
cdos_provider_remote_set(CdosProvider *provider,
                        const gchar *key, 
                        const gchar *payload)
{
//        cdos_setting_obj_set_node_from_dbus(priv->settings_obj, key, payload);
}       

/**
 * cdos_provider_setting_file_changed_notify:
 * @provider: A #CdosProvider
 *
 * 当配置文件发生改变时，调用该函数发出信号
 *
 */
void
cdos_provider_setting_file_changed_notify(CdosProvider *provider)
{
    g_return_if_fail(CDOS_IS_PROVIDER(provider));
    g_signal_emit_by_name(provider, "setting-changed");
}

/**
 * cdos_provider_value_changed_notify:
 * @provider: A #CdosProvider
 * @key: key字符串
 * @oldval: key对应的修改之前的值
 * @newval: key对应的修改之后的值
 *
 * 当我们对某个key的值进行更改之后可以发出信号
 *
 */
void
cdos_provider_value_changed_notify(CdosProvider *provider, 
				                   const gchar *key, 
				                   const gchar* oldval, 
				                   const gchar *newval)
{
    g_return_if_fail(CDOS_IS_PROVIDER(provider));
    gchar *signals = g_strconcat("notify::", key, key, oldval, newval, NULL);
    g_signal_emit_by_name(provider, signals);
}

static gboolean 
compare(gchar **group1, gchar **group2, gchar **group3, gsize len1, gsize len2, gsize len3, gchar *type)
{
                int i;
                for (i = 0; i < len1; i++) {
                        if(g_strcmp0(*(group1+i), type) == 0) 
                                return TRUE;
                }

                for (i = 0; i < len2; i++) {
                        if(g_strcmp0(*(group2+i), type) == 0) 
                                return TRUE;
                }

                for (i = 0; i < len3; i++) {
                        if(g_strcmp0(*(group3+i), type) == 0) 
                                return TRUE;
                }

                return FALSE;
}

/**
 * cdos_provider_bind_property:
 * @provider: A #CdosProvider
 * @sync_type: 绑定目录类型 
　* @key_name: key名字 
　* @applet_var: applet变量 
　* @applet_callback: 回调函数 
　* @user_data: 传递的参数
 *
 * 为applet绑定，当自定义的配置改变可以进行更新
 *
 * Returns: $goboolean 是否绑定成功
 */
gboolean
cdos_provider_bind_property(CdosProvider *provider,
                            BINDINGDIRECTION sync_type,
                            gchar   *key_name,
                            gchar   *applet_var,
                            func     applet_callback,
                            gpointer user_data)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),FALSE);

    CdosProviderPrivate *priv = provider->priv;

    if (!priv->valid) {
        cdos_provider_setting_not_initialized_error(priv->uuid);
        return FALSE;
    }

    gsize group_len1;
    gsize group_len2;
    gsize group_len3;
    //gboolean flag;
    gchar* type = cdos_setting_obj_get_key_exists_and_type(priv->settings_obj, key_name);
    if (type) {
        gchar **group1 = g_key_file_get_groups(priv->key, &group_len1);
        gchar **group2 = g_key_file_get_groups(priv->key, &group_len2);
        gchar **group3 = g_key_file_get_groups(priv->key, &group_len3);

        gboolean c = compare(group1, group2, group3, group_len1, group_len2, group_len3, type);
        g_strfreev(group1);
        g_strfreev(group2);
        g_strfreev(group3);
        if (c) {
            CdosSetting *setting = cdos_setting_new_with_params(sync_type,
                                                            priv->xlet,
                                                            key_name,
                                                            priv->settings_obj,
                                                            applet_var,
                                                            applet_callback,
                                                            user_data);
            g_hash_table_insert(priv->metaBindings, key_name, setting);
            g_free(type);
            return TRUE;
        } else {
            cdos_provider_invalid_setting_type_error(key_name, priv->uuid, type);
            g_free(type);
            return FALSE;
        }
        g_free(type);
    } else {
        cdos_provider_key_not_found_error(key_name, priv->uuid);
        return FALSE;
    }

}

/**
 * cdos_provider_unbind_property:
 * @provider: A #CdosProvider
 * @key_name: 解绑的key名字 
 *
 * 通过key_name对applet进行解绑
 *
 * Returns: $goboolean 是解绑定成功
 */
gboolean
cdos_provider_unbind_property(CdosProvider *provider, const gchar *key_name)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),FALSE);
    if (key_name == NULL)
        return FALSE;
    CdosProviderPrivate *priv = provider->priv;

    CdosSetting *value = g_hash_table_lookup(priv->metaBindings, key_name);
    if (value) {
        g_object_unref(value);
        g_hash_table_remove(priv->metaBindings,key_name);
        return TRUE;
    } else {
        g_print("unbindProperty failed for %s. Key name  %s  did not exist", priv->uuid, key_name);
        return FALSE;
    }
}

static void
cdos_provider_dispose(GObject *gobject){

    CdosProvider *provider = CDOS_PROVIDER(gobject);
    CdosProviderPrivate *priv = provider->priv;

    cdos_settings_manager_unregister(cdos_settings_manager_get_default(), priv->uuid, priv->instancedId);

    GList *list = g_hash_table_get_values(priv->metaBindings);
    GList *l;
    for (l = list; l != NULL; l = l->next)
    {
        if(l->data != NULL){
            g_object_unref(l->data);
            l->data = NULL;
        }
    }
    g_hash_table_unref(priv->metaBindings);

    g_clear_object(&priv->settings_obj);
    g_clear_object(&priv->settings_file);
    g_object_unref(priv->applet_dir);
    g_key_file_free(priv->key);

    g_object_unref(priv->ext_type);

    G_OBJECT_CLASS(cdos_provider_parent_class)->dispose(gobject);
}

static void
cdos_provider_finalize(GObject *gobject)
{
    CdosProvider *provider = CDOS_PROVIDER(gobject);
    CdosProviderPrivate *priv = provider->priv;

    g_free(priv->xlet_str);
    g_free(priv->uuid);
    g_free(priv->instancedId);

    G_OBJECT_CLASS(cdos_provider_parent_class)->finalize(gobject);
}

/**
 * cdos_provider_get_value:
 * @provider: A #CdosProvider
 * @key_name: key名字
 *
 * 获取provider的key_name对应的值
 *
 * Returns: 当返回值为空时返回NULL，使用完成之后调用g_free()予以释放。
 */
gchar *
cdos_provider_get_value(CdosProvider *provider, 
			            const gchar *key_name)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),NULL);
    CdosProviderPrivate *priv = provider->priv;

    gchar *result = cdos_setting_obj_get_data(priv->settings_obj, key_name, "value");
    if (!result) {
        cdos_provider_key_not_found_error(key_name, priv->uuid);
    }
    return result;
}

/**
 * cdos_provider_set_value:
 * @provider: A #CdosProvider
 * @key_name: 配置key的名字
　* @value: 需要设置的值 
 *
 * 设置provider对应的配置的值
 *
 */
void 
cdos_provider_set_value(CdosProvider *provider,
                        const gchar  *key_name,
                        const gchar  *value)
{
    g_return_if_fail(CDOS_IS_PROVIDER(provider));
    CdosProviderPrivate *priv = provider->priv;
    gchar *result = cdos_setting_obj_get_data(priv->settings_obj, key_name, "value");
    if (!result) {
        cdos_provider_key_not_found_error(key_name, priv->uuid);
    } else {
        cdos_setting_obj_set_value(priv->settings_obj, key_name, value);
    }
}

static void
cdos_provider_setting_not_initialized_error(const gchar *uuid)
{
        g_print("Could not set up binding - setting object was not initialized successfully for %s", uuid);
}

static void
cdos_provider_key_not_found_error(const gchar *key_name, const gchar *uuid)
{
        g_print("Could not find setting key  %s for applet/desklet uuid %s", key_name, uuid);
}

static void
cdos_provider_invalid_setting_type_error(const gchar *key_name, const gchar *uuid, const gchar *type)
{
        g_print("Invalid setting type %s for setting key %s of applet/desklet uuid %s", key_name, uuid, type);
}

static gboolean 
cdos_provider_maybe_update_settings_file(CdosProvider *provider)
{
	CdosProviderPrivate *priv;
	priv = provider->priv;

	GFile *orig_file = g_file_get_child(priv->applet_dir, SETTING_SCHEMA_FILE);
	if(orig_file && !g_file_query_exists(orig_file, NULL)) {
		g_print("Failed to locate settings schema file to check for update: %s", priv->uuid);
		g_print("Something may be not right");
        g_object_unref(orig_file);
		return FALSE;
	}
	gchar *init_file_contents = cdos_base_util_get_file_contents_utf8_sync(g_file_get_path(orig_file),NULL);
	gchar *checksum = cdos_global_get_md5_for_string(init_file_contents);
	gchar *existing_setting_file = cdos_base_util_get_file_contents_utf8_sync(g_file_get_path(priv->settings_file),NULL);

	GKeyFile *new_key = g_key_file_new();
    gchar *orig_file_path = g_file_get_path(orig_file);

    if (!g_key_file_load_from_file(new_key, orig_file_path, G_KEY_FILE_NONE, NULL)) {
        g_print("Problem parsing %s while preparing to perform upgrade", orig_file_path);
        g_print("Skipping upgrade for now - something may be wrong with the new settings schema file.");

        g_free(orig_file_path);
        g_key_file_free(new_key);
        g_free(existing_setting_file);
        g_free(checksum);
        g_free(init_file_contents);

        g_object_unref(orig_file);

        return FALSE;
    }

	GKeyFile *existing_key = g_key_file_new();
    gchar *settings_file_path = g_file_get_path(priv->settings_file);

	if (!g_key_file_load_from_file(existing_key, settings_file_path, G_KEY_FILE_NONE, NULL)) {
		g_print("Problem  parsing %s  while preparing to perform upgrade.",settings_file_path);
		g_print("Re-creating setting file.");
		g_file_delete(priv->settings_file, NULL, NULL);

        g_free(settings_file_path);
        g_key_file_free(existing_key);

        g_free(orig_file_path);
        g_key_file_free(new_key);
        g_free(existing_setting_file);
        g_free(checksum);
        g_free(init_file_contents);

        g_object_unref(orig_file);

		return cdos_provider_create_settings_file(provider);
	}

    gchar *md5 = g_key_file_get_string(existing_key, "MD5", "md5", NULL);

    gboolean result = TRUE;
	if (!g_strcmp0(md5, checksum)) {
		g_print("Updated setting file detected. Beginning upgrade of existing settings");
        result = cdos_provider_do_upgrade(provider, new_key, existing_key, checksum);
	}
    g_free(md5);

    g_free(settings_file_path);
    g_key_file_free(existing_key);

    g_free(orig_file_path);
    g_key_file_free(new_key);
    g_free(existing_setting_file);
    g_free(checksum);
    g_free(init_file_contents);

    g_object_unref(orig_file);

    return result;
}

static gboolean
cdos_provider_sanity_check(GKeyFile *key, const gchar *val, const gchar *group)
{
	gboolean found,result;

	gchar *value = g_key_file_get_string(key, group, "type", NULL);
	if (g_strcmp0(value, "spinbutton") || g_strcmp0(value, "scale")) {
		int max = g_key_file_get_integer(key, group, "max", NULL);
		int min = g_key_file_get_integer(key, group, "min", NULL);
        int va = atoi(val);
		result = (va < max && va >min);
	} else if (g_strcmp0(value, "combobox")) {
		found = FALSE;
        gsize len;
        gchar **keys = g_key_file_get_keys(key, group, &len, NULL);
        int i;
        for (i = 0; i < len; i++) 
        {
            if(!g_strcmp0(*(keys+i), val)) {
                found = TRUE;
                break;
            }   
        }
        g_strfreev(keys);
		result = found;
	} else if (g_strcmp0(value, "radiogroup")) {
		found = FALSE;
        gsize len;
        gchar **keys = g_key_file_get_keys(key, group, &len, NULL);
        int i;
        for (i = 0; i < len; i++) 
        {
            if((!g_strcmp0(*(keys+i),val)) ||(!g_strcmp0(*(keys+i), "custom"))) {
                found = TRUE;
                break;
            }   
        }
        g_strfreev(keys);
        result = found;
	} else {
        result = TRUE;
	}

    g_free(value);

    return result;
}

/**
 * cdos_provider_get_string:
 * @provider: A #CdosProvider
 * @group_name: 配置.ini中的[group-name]
　* @key: key key=value  
 *
 * 通过group名字和key的名字获取对应的value,如果key值为空，默认key="value"
 *
 * Returns: $gchar* 使用完成之后，调用g_free()予以释放。
 */
gchar* 
cdos_provider_get_string(CdosProvider *provider,
                         gchar        *group_name,
                         gchar        *key)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),NULL);
    CdosProviderPrivate *priv = provider->priv;
    gchar *result = NULL;
    if(group_name == NULL){
        g_warning("%s:I need param of 'group_name',but it's NULL!",__func__);
        return  result;
    }
    if (NULL == key)
        key = "value";
    if(priv->settings_obj){
        GKeyFile *keyFile = cdos_setting_obj_get_key(provider->priv->settings_obj);
        result = g_key_file_get_string(keyFile, group_name, key, NULL);
    }
    return  result;
}

/**
 * cdos_provider_get_string:
 * @provider: A #CdosProvider
 * @group_name: 配置.ini中的[group-name]
　* @key: key key=value  
 *
 * 通过group名字和key的名字获取对应的value,如果key值为空，默认key="value"
 *
 * Returns: $gboolean
 */
gboolean 
cdos_provider_get_boolean(CdosProvider *provider,
                          gchar        *group_name,
                          gchar        *key)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),FALSE);
    CdosProviderPrivate *priv = provider->priv;
    gboolean result = FALSE;
    if (NULL == key)
        key = "value";
    if(group_name == NULL){
        g_warning("%s:I need param of 'group_name',but it's NULL!",__func__);
        return  result;
    }
    if(priv->settings_obj){
        GKeyFile *keyFile = cdos_setting_obj_get_key(provider->priv->settings_obj);
        result = g_key_file_get_boolean(keyFile, group_name, key, NULL);
    }
    return  result;
}

/**
 * cdos_provider_get_string:
 * @provider: A #CdosProvider
 * @group_name: 配置.ini中的[group-name]
　* @key: key key=value  
 *
 * 通过group名字和key的名字获取对应的value,如果key值为空，默认key="value"
 *
 * Returns: $gint
 */
gint 
cdos_provider_get_int(CdosProvider *provider,
                      gchar        *group_name,
                      gchar        *key)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),0);
    CdosProviderPrivate *priv = provider->priv;
    gint result = 0;
    if(group_name == NULL){
        g_warning("%s:I need param of 'group_name',but it's NULL!",__func__);
        return  result;
    }
    if (NULL == key)
        key = "value";
    if(priv->settings_obj){
        GKeyFile *keyFile = cdos_setting_obj_get_key(provider->priv->settings_obj);
        result = g_key_file_get_integer(keyFile, group_name, key, NULL);
    }

    return  result;
}

/**
 * cdos_provider_get_string:
 * @provider: A #CdosProvider
 * @group_name: 配置.ini中的[group-name]
　* @key: key key=value  
 *
 * 通过group名字和key的名字获取对应的value,如果key值为空，默认key="value"
 *
 * Returns: $gchar**,使用完成之后，调用g_free(),予以销毁。
 */
gchar** 
cdos_provider_get_array(CdosProvider *provider,
                        gchar *group_name,
                        gchar *key,
                        gsize length)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),NULL);
    CdosProviderPrivate *priv = provider->priv;
    gchar **result = NULL;
    if(group_name == NULL){
        g_warning("%s:I need param of 'group_name',but it's NULL!",__func__);
        return  result;
    }
    if (NULL == key)
        key = "value";
    if(priv->settings_obj){
        GKeyFile *keyFile = cdos_setting_obj_get_key(provider->priv->settings_obj);
        result = g_key_file_get_string_list(keyFile, group_name, key, &length, NULL);
    }
    return  result;
}

/**
 * cdos_provider_set_string:
 * @provider: A #CdosProvider
 * @group_name: 配置.ini中的[group-name]
　* @key: key key=value  
 *
 * 通过group名字和key的名字设置对应的value,如果key值为空，默认key="value"
 *
 */
void 
cdos_provider_set_string(CdosProvider *provider,
                         gchar *group_name,
                         gchar *key,
                         gchar *value)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),NULL);
    CdosProviderPrivate *priv = provider->priv;
    if(group_name == NULL){
        g_warning("%s:I need param of 'group_name',but it's NULL!",__func__);
        return  ;
    }
    if (NULL == key)
        key = "value";
    if(priv->settings_obj){
        GKeyFile *keyFile = cdos_setting_obj_get_key(provider->priv->settings_obj);
        g_key_file_set_string(keyFile, group_name, key, value);
    }
}

/**
 * cdos_provider_set_string:
 * @provider: A #CdosProvider
 * @group_name: 配置.ini中的[group-name]
　* @key: key key=value  
 *
 * 通过group名字和key的名字设置对应的value,如果key值为空，默认key="value"
 *
 */
void 
cdos_provider_set_boolean(CdosProvider *provider,
                          gchar *group_name,
                          gchar *key,
                          gboolean value)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),NULL);
    CdosProviderPrivate *priv = provider->priv;
    if(group_name == NULL){
        g_warning("%s:I need param of 'group_name',but it's NULL!",__func__);
        return  ;
    }
    if (NULL == key)
        key = "value";
    if(priv->settings_obj){
        GKeyFile *keyFile = cdos_setting_obj_get_key(provider->priv->settings_obj);
        g_key_file_set_boolean(keyFile, group_name, key, value);
    }


}

/**
 * cdos_provider_set_string:
 * @provider: A #CdosProvider
 * @group_name: 配置.ini中的[group-name]
　* @key: key key=value  
 *
 * 通过group名字和key的名字设置对应的value,如果key值为空，默认key="value"
 *
 */
void 
cdos_provider_set_int(CdosProvider *provider,
                      gchar *group_name,
                      gchar *key,
                      gint value)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),NULL);
    CdosProviderPrivate *priv = provider->priv;
    if(group_name == NULL){
        g_warning("%s:I need param of 'group_name',but it's NULL!",__func__);
        return  ;
    }
    if (NULL == key)
        key = "value";
    if(priv->settings_obj){
        GKeyFile *keyFile = cdos_setting_obj_get_key(provider->priv->settings_obj);
        g_key_file_set_integer(keyFile, group_name, key, value);
    }
}

/**
 * cdos_provider_set_string:
 * @provider: A #CdosProvider
 * @group_name: 配置.ini中的[group-name]
　* @key: key key=value 
　* @value: GList  
 *
 * 通过group名字和key的名字设置对应的value Glist,如果key值为空，默认key="value"
 *
 */
void 
cdos_provider_set_array(CdosProvider *provider,
                        gchar *group_name,
                        gchar *key,
                        GList *value)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),NULL);
    CdosProviderPrivate *priv = provider->priv;
    if(group_name == NULL){
        g_warning("%s:I need param of 'group_name',but it's NULL!",__func__);
        return  ;
    }
    if (NULL == key)
        key = "value";
    if(priv->settings_obj){
        GKeyFile *keyFile = cdos_setting_obj_get_key(provider->priv->settings_obj);
        gint vlen = g_list_length(value);
        gchar **arr = g_malloc_n(sizeof(gchar *),vlen);
        gsize length = 0;
        while(value != NULL) {
            if(value->data != NULL){
                arr[length] = value->data;
                length++;
            }
            value = value->next;
        }
        g_key_file_set_string_list(keyFile, group_name, key,(const gchar **) arr, length);

        g_strfreev(arr);
    }
}


