/**
 * SECTION: cdos-session
 * @short_description: 登入登出会话进程
 * @include: cdos-session.h
 * 
 * CdosSessionPresenceClass, CdosSessionInhibitor 类
 * 以及相关的dbus来处理登入登出会话进程.
 */
#include "cdos-session.h"



/* Presence */

struct _CdosSessionPresence
{
    GObject parent_instance;

	GDBusProxy *proxy;
};


G_DEFINE_TYPE ( CdosSessionPresence, cdos_session_presence, G_TYPE_OBJECT );

static void
cdos_session_presence_finalize ( GObject *gobject ) {

	//CdosSessionPresence *self = CDOS_SESSION_PRESENCE (gobject);

	/* 调用父类的 finalize() 方法 */
    G_OBJECT_CLASS(cdos_session_presence_parent_class)->finalize(gobject);
}

static void
cdos_session_presence_dispose ( GObject *gobject ) {

    CdosSessionPresence *self = CDOS_SESSION_PRESENCE (gobject);

    g_object_unref(self->proxy);

    G_OBJECT_CLASS(cdos_session_presence_parent_class)->dispose(gobject);
}

static void
cdos_session_presence_class_init ( CdosSessionPresenceClass *klass ) {

    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

	gobject_class->finalize = cdos_session_presence_finalize;
	gobject_class->dispose = cdos_session_presence_dispose;
}


static void
cdos_session_presence_init ( CdosSessionPresence *self ) {

	GError *error = NULL;

	self->proxy = NULL;

	//建立连接到 /org.gnome/SessionManager/Presence的代理
    self->proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                           G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                           NULL,
                                           "org.gnome.SessionManager",
                                           "/org/gnome/SessionManager/Presence",
                                           "org.gnome.SessionManager.Presence",
                                           NULL, &error);

    if (self->proxy == NULL) {
            g_warning ("Couldn't get org.gnome.SessionManager.Presence proxy: %s", error->message);
            g_error_free (error);
            return;
    }
}


/**
 * cdos_session_presence_new: 
 * 
 * 新建一个 #CdosSessionPresence 类的实例。
 *
 * Returns: 返回一个 #CdosSessionPresence 的实例。
 */
CdosSessionPresence *cdos_session_presence_new (void) {

	return g_object_new ( CDOS_TYPE_SESSION_PRESENCE, NULL );
}


/**
 * cdos_session_presence_get_status:
 * @self: A #CdosSessionPresence
 * @callback: 回调函数 
 *
 * 这里通过dbus同步Proxy直接获取类返回值,因此没有使用callback
 *
 * Returns (guint): status
 */
guint cdos_session_presence_get_status ( CdosSessionPresence *self, GFunc callback ) {

    g_return_val_if_fail(CDOS_IS_SESSION_PRESENCE(self),0);
	//GError *error = NULL;
	guint status = 0;
    GVariant *result = NULL;

    /* not yet connected to cdos-session */
    if (self->proxy == NULL) {
        g_warning ("session is not available");
    }

    /* get the session status */
    result = g_dbus_proxy_get_cached_property ( self->proxy, "status");

    if (result == NULL) {
        g_warning ("no readable status property on %s", g_dbus_proxy_get_interface_name (self->proxy));
    }
    else {
        g_variant_get (result, "u", &status);

        g_variant_unref (result);
    }

	return status;
}


/**
 * cdos_session_presence_set_status:
 * @self: A #CdosSessionPresence
 * @status: 状态
 *
 * 这里通过dbus同步Proxy设置获取类返回值,因此没有使用callback
 *
 * Returns: 设置成功返回 %TRUE, 否则返回 %FALSE 
 */
gboolean cdos_session_presence_set_status ( CdosSessionPresence *self, guint status ) {

    g_return_val_if_fail(CDOS_IS_SESSION_PRESENCE(self),FALSE);
    GError *error = NULL;
	GVariant *retval;

    retval = g_dbus_proxy_call_sync ( self->proxy,
                            "SetStatus",
                            g_variant_new ("(u)", status),
                            G_DBUS_CALL_FLAGS_NONE,
                            -1, NULL,
                            &error );
	
	//发送一个同步dbus设置status
	if (retval == NULL){
        g_warning ("%s:Couldn't call org.gnome.SessionManager: %s",__func__,error->message);
        g_error_free (error);
        return FALSE;
    }
	g_dbus_proxy_set_cached_property (self->proxy, "status", retval);
	g_variant_unref (retval);
	return TRUE;
}


/* Inhibitor */


struct _CdosSessionInhibitor
{
	GObject parent_instance;
	gboolean isLoaded;	
	guint loadingPropertiesCount;
};

gchar *InhibitorIface_properties_name [] = 	{ "app_id", "client_id", "reason", "flags", "toplevel_xid", "cookie" };
gchar *InhibitorIface_properties_signature [] =	{ "s"		,"s"	,"s",		"u",	"u",		"u"};	
gchar *InhibitorIface_properties_access [] =	{ "readonly", "readonly", "readonly", "readonly", "readonly", "readonly"};

#define InhibitorIface_properties_length 6


G_DEFINE_TYPE ( CdosSessionInhibitor, cdos_session_inhibitor, G_TYPE_OBJECT);


/* signal */
enum
{
        IS_LOADED,       //expanded
        LAST_SIGNAL
};


static guint cdos_session_inhibitor_signals [LAST_SIGNAL] = { 0 };


static void 
cdos_session_inhibitor_class_init (CdosSessionInhibitorClass *klass) {

    //GObjectClass *gobject_class = G_OBJECT_CLASS(klass);


    /**
    * CdosSessionInhibitor::is-loaded:
    * @object: 绑定信号的对象
    * @user_data: 真对回调函数所传入的参数
    *
    * 该信号对 #CdosSessionInhibitor 的载入进行监听
    */
	cdos_session_inhibitor_signals [IS_LOADED] =
                g_signal_new ("is-loaded",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__VOID,
                        G_TYPE_NONE, 0 );

}


static void
cdos_session_inhibitor_init ( CdosSessionInhibitor *self ) {


}


static void cdos_session_inhibitor_post_init ( CdosSessionInhibitor *self, gchar* objectPath ) {
    g_return_if_fail(CDOS_IS_SESSION_INHIBITOR(self));
    //DBus.session.proxifyObject(this, 'org.gnome.SessionManager', objectPath);
    GError *error = NULL;
    GDBusProxy *proxy;
    //GVariant *parameters = NULL;
    GVariant *v;
    //guint mode;
    gint i;

    proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                           G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                           NULL,
                                           "org.gnome.SessionManager",
                                           objectPath,
                                           "org.gnome.SessionManager.Inhibitor",
                                           NULL, &error);

    self->isLoaded = FALSE;

    self->loadingPropertiesCount = InhibitorIface_properties_length;


    for ( i = 0 ; i < InhibitorIface_properties_length ; i++ ) {

        //this.GetRemote(propertyName, Lang.bind(this, property_is_loaded ));
        gchar *propertyName = InhibitorIface_properties_name[i];
        GError *error = NULL;
        GVariant *result;

        result = g_dbus_proxy_call_sync ( proxy,
                                    propertyName,
                                    NULL,
                                    G_DBUS_CALL_FLAGS_NONE,
                                    -1,
                                    NULL,
                                    &error );

        if (result == NULL) {
            g_warning ("can not get CanShutdown flag");
            g_error_free (error);
        }
        else
        {
            g_variant_get (result, "(v)", &v);
            g_dbus_proxy_set_cached_property (proxy, InhibitorIface_properties_name[i], v);
            g_variant_unref (result);
        }

        //property_is_loaded
        self->loadingPropertiesCount--;

        if ( self->loadingPropertiesCount == 0 ) {
                self->isLoaded = TRUE;
                //this.emit('is-loaded');
                g_signal_emit_by_name ( self , "is-loaded" );
        }
    }
}

/**
 * cdos_session_inhibitor_new:
 * @objectPath: 对象路径 
 * 
 * 新建一个 #CdosSessionInhibitor 类的实例。
 *
 * Returns (CdosSessionInhibitor): 返回一个 #CdosSessionInhibitor 的实例。
 */
CdosSessionInhibitor *cdos_session_inhibitor_new ( gchar *objectPath ) {
	CdosSessionInhibitor *instance = g_object_new ( CDOS_TYPE_SESSION_INHIBITOR, NULL );
	cdos_session_inhibitor_post_init (instance, objectPath);
	return instance;
}




/* Session */

/*
与session manager建立dbus连线
*/

//        { name: 'Logout', inSignature: 'u', outSignature: '' },
//        { name: 'Shutdown', inSignature: '', outSignature: '' },
 //       { name: 'CanShutdown', inSignature: '', outSignature: 'b' }

static void session_logout_cb ( GObject *source_object,
                                GAsyncResult *res,
                                gpointer user_data) {
    GVariant *result;
    GError *error = NULL;

    result = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object),
                                       res,
                                       &error);
    if (result == NULL) {
        g_warning ("couldn't logout using cdos-session: %s", error->message);
        g_error_free (error);
    } else {
        g_variant_unref (result);
    }
}


static void session_shutdown_cb ( GObject *source_object,
                                GAsyncResult *res,
                                gpointer user_data) {

    GVariant *result;
    GError *error = NULL;

    result = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object),
                                       res,
                                       &error);
    if (result == NULL) {
        g_warning ("couldn't shutdown using cdos-session: %s", error->message);
        g_error_free (error);
    } else
        g_variant_unref (result);
}

/**
 * cdos_session_manager_new:
 * 
 * 新建一个 #GDBusProxy 类的实例。
 *
 * Returns (GDBusProxy): 返回一个 名为"org.gnome.SessionManager"
 * 的 #GDBusProxy 的实例,使用完之后调用g_object_unref()进行释放。
 */
GDBusProxy *cdos_session_manager_new (void) {

	//DBus.session.proxifyObject(this, 'org.gnome.SessionManager', '/org.cdos/SessionManager');
	GError *error = NULL;
  	GDBusProxy *proxy;

    proxy = g_dbus_proxy_new_for_bus_sync (G_BUS_TYPE_SESSION,
                                           G_DBUS_PROXY_FLAGS_DO_NOT_LOAD_PROPERTIES,
                                           NULL,
                                           "org.gnome.SessionManager",
                                           "/org/gnome/SessionManager",
                                           "org.gnome.SessionManager",
                                           NULL, &error);

    if (proxy == NULL) {
        g_warning ("cannot connect to cdos-session: %s", error->message);
        g_error_free (error);
    }
	return proxy;
}

/**
 * cdos_session_logout:
 * @proxy: A #GDBusProxy
 * @mode: 模式 
 *
 * 通过DBus代理获取当前模式进行注销;
 *
 */
void cdos_session_logout (GDBusProxy *proxy, guint mode) {
    g_return_if_fail(G_IS_DBUS_PROXY(proxy));
    g_dbus_proxy_call (proxy,
                       "Logout",
                       g_variant_new ("(u)", mode),
                       G_DBUS_CALL_FLAGS_NONE,
                       -1, NULL,
                       session_logout_cb, NULL);
    g_object_unref (proxy);
}

/**
 * cdos_session_shutdown:
 * @proxy: A #GDBusProxy
 *
 * 通过session会话进行关机
 *
 */
void cdos_session_shutdown (GDBusProxy *proxy) {
    g_return_if_fail(G_IS_DBUS_PROXY(proxy));
    g_dbus_proxy_call (proxy,
                       "Shutdown",
                       NULL,
                       G_DBUS_CALL_FLAGS_NONE,
                       -1, NULL,
                       session_shutdown_cb, NULL);

    g_object_unref (proxy);
}

/**
 * cdos_session_can_shutdown:
 * @proxy: A #GDBusProxy
 *
 * 通过DBus获取当前代理的"Canshutdown", 查询是否可以关机
 *
 * Returns: %TRUE 可以关机; %FALSE 则不可以.
 */
gboolean cdos_session_can_shutdown (GDBusProxy *proxy) {
    g_return_val_if_fail(G_IS_DBUS_PROXY(proxy),FALSE);
    GError *error = NULL;
    GVariant *result;
	gboolean can_shutdown;

    result = g_dbus_proxy_call_sync ( proxy,
                                    "CanShutdown",
                                    NULL,
                                    G_DBUS_CALL_FLAGS_NONE,
                                    -1,
                                    NULL,
                                    &error );
    if (result == NULL) {
        g_warning ("can not get CanShutdown flag");
        g_error_free(error);
    }
	g_variant_get (result, "(b)", &can_shutdown);
	return can_shutdown;	
}


