#include "cdos-setting-obj.h"

static void
cdos_setting_obj_dispose(GObject *gobject);

static void
cdos_setting_obj_finalize(GObject *gobject);

static void
cdos_setting_obj_class_init(CdosSettingObjClass *klass);

static void 
cdos_setting_obj_init(CdosSettingObj *obj);

static void 
cdos_setting_obj_set_params(CdosSettingObj *obj,
                        CdosProvider  *provider,
                        GFile *file,
                        const gchar *uuid,
                        const gchar *instanceId);

static gboolean
cdos_setting_obj_on_file_changed_timeout(CdosSettingObj *obj);

static void
cdos_setting_obj_on_file_changed(CdosSettingObj *obj);



struct _CdosSettingObjPrivate
{
        CdosProvider            *provider;
        GFile                   *file;
        gchar                   *uuid;
        gchar                   *instanceId;
        GKeyFile                *key;
        GFileMonitor            *setting_file_monitor;
        guint                   file_changed_timeout;
        gint                    monitor_id;
};

#define CDOS_SETTING_OBJ_GET_PRIVATE(obj)    \
        (G_TYPE_INSTANCE_GET_PRIVATE((obj), CDOS_TYPE_SETTING_OBJ, CdosSettingObjPrivate))

G_DEFINE_TYPE(CdosSettingObj, cdos_setting_obj, G_TYPE_OBJECT);

/**
 * cdos_setting_obj_new:
 * @provider: 一个#CdosProvider对象
 * @file: 一个#GFile对象，一般由路径构成
 * @uuid: 通用唯一识别码。例如：/home目录下.cdos/configs/cdos-menu@cdos.org.cdos-menu@cdos.org是uuid。
 * @instanceId: 实例ID
 *
 * 创建#CdosSettingObj对象，同时根据传入的参数设置对应的属性。
 *
 * Returns: 一个#CdosSettingObj对象
 */

CdosSettingObj *
cdos_setting_obj_new(CdosProvider *provider,
                    GFile  *file,
                    const gchar *uuid,
                    const gchar *instanceId)
{
    g_return_val_if_fail(CDOS_IS_PROVIDER(provider),NULL);
    g_return_val_if_fail(G_IS_FILE(file),NULL);
    
    GObject  *gobject;
    gobject = g_object_new(CDOS_TYPE_SETTING_OBJ, NULL);
    cdos_setting_obj_set_params(CDOS_SETTING_OBJ(gobject), provider ,file, uuid, instanceId);
    return CDOS_SETTING_OBJ(gobject);
}

static void 
cdos_setting_obj_set_params(CdosSettingObj *obj,
                        CdosProvider  *provider,
                        GFile *file,
                        const gchar *uuid,
                        const gchar *instanceId)
{
        CdosSettingObjPrivate              *priv;
        priv = obj->priv;

        priv->provider = g_object_ref(provider);
        priv->file = g_object_ref(file);
        priv->uuid = g_strdup(uuid);
        priv->instanceId = g_strdup(instanceId);
        priv->key = g_key_file_new();

        gchar *file_path = g_file_get_path(priv->file);
        if (!g_key_file_load_from_file(priv->key, file_path, G_KEY_FILE_NONE, NULL)) {
            g_print("can not load file !\n");
            return;
        }

        priv->setting_file_monitor = g_file_monitor_file(priv->file, G_FILE_MONITOR_NONE, NULL, NULL);

        priv->monitor_id = g_signal_connect(priv->setting_file_monitor,
                                        "changed",
                                        G_CALLBACK(cdos_setting_obj_on_file_changed),
                                        provider);
}

static void 
cdos_setting_obj_init(CdosSettingObj *obj)
{
        CdosSettingObjPrivate        *priv;
        priv = obj->priv = CDOS_SETTING_OBJ_GET_PRIVATE(obj);

        priv->provider = NULL;
        priv->file = NULL;
        priv->uuid = NULL;
        priv->instanceId = NULL;
        priv->key = g_key_file_new();
        priv->setting_file_monitor = NULL;
        priv->file_changed_timeout = 0;
}

static void
cdos_setting_obj_class_init(CdosSettingObjClass *klass)
{
        GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
   //     GParamSpec *pspec;

        g_type_class_add_private(klass, sizeof(CdosSettingObjPrivate));

        gobject_class->dispose = cdos_setting_obj_dispose;
        gobject_class->finalize = cdos_setting_obj_finalize;
}

/**
 * cdos_setting_obj_save:
 * @obj: 一个#CdosSettingObj对象
 *
 * 通过调用g_key_file_save_to_file将文件路径内容写入到file属性。
 *
 */

void
cdos_setting_obj_save(CdosSettingObj *obj)
{
    g_return_if_fail(CDOS_IS_SETTING_OBJ(obj));
    CdosSettingObjPrivate *priv = obj->priv;

    g_signal_handler_disconnect(priv->setting_file_monitor, priv->monitor_id);
    priv->monitor_id = 0;

    if (g_file_delete(priv->file, NULL, NULL)) {
            gchar *file_path = g_file_get_path(priv->file);
            g_key_file_save_to_file(priv->key, file_path, NULL);
            g_free(file_path);
    } else {
            g_print("Failed gain write access to settings file for applet/desklet %s , instance %s",
                    priv->uuid,
                    priv->instanceId);
    }
    priv->monitor_id = g_signal_connect(priv->setting_file_monitor,
                                    "changed",
                                    G_CALLBACK(cdos_setting_obj_on_file_changed),
                                    priv->provider);
}

/**
 * cdos_setting_obj_get_data:
 * @obj: 一个#CdosSettingObj对象
 * @group: 一个组名
 * @key: 一个键
 *
 * 返回在组名下与键相关联的字符串的值，该方法调用g_key_file_get_string。
 *
 * Returns: 一个新分配的字符串;如果指定的键没有找到，则返回NULL,使用之后调用g_free()予以释放。
 */

gchar *
cdos_setting_obj_get_data(CdosSettingObj *obj, 
                        const gchar *group, 
                        const gchar *key)
{
    g_return_val_if_fail(CDOS_IS_SETTING_OBJ(obj),NULL);
    g_return_val_if_fail(group != NULL,NULL);
    g_return_val_if_fail(key != NULL,NULL);

    return g_key_file_get_string(obj->priv->key, group, key, NULL);
}

/**
 * cdos_setting_obj_get_key_exists_and_type:
 * @obj: 一个#CdosSettingObj对象
 * @group: 一个组名
 *
 * 返回在组名下与键“type”相关联的字符串的值，该方法调用g_key_file_get_string。
 *
 * Returns: 一个新分配的字符串;如果指定的键没有找到，则返回NULL，使用之后调用g_free()予以释放。
 */

gchar *
cdos_setting_obj_get_key_exists_and_type(CdosSettingObj *obj, const gchar *group)
{
    g_return_val_if_fail(CDOS_IS_SETTING_OBJ(obj),NULL);
    g_return_val_if_fail(group != NULL,NULL);

    return g_key_file_get_string(obj->priv->key, group, "type", NULL);
}

/**
 * cdos_setting_obj_set_value:
 * @obj: 一个#CdosSettingObj对象
 * @group: 一个组名
 * @value: 一个新的字符串
 *
 * 将一个新的字符串值与组名下的键“value”相关联，该方法调用g_key_file_set_string。
 *
 */

void 
cdos_setting_obj_set_value(CdosSettingObj *obj,
                        const gchar *group, 
                        const gchar *value)
{
    g_return_if_fail(CDOS_IS_SETTING_OBJ(obj));
    g_return_if_fail(group != NULL);
    g_return_if_fail(value != NULL);

    g_key_file_set_string(obj->priv->key, group, "value", value);
    cdos_setting_obj_save(obj);
}

/**
 * cdos_setting_obj_set_node_from_dbus:
 * @obj: 一个#CdosSettingObj对象
 *
 * FIXME
 *
 */

void
cdos_setting_obj_set_node_from_dbus(CdosSettingObj *obj, 
                                const gchar *group_key,
                                const gchar *group_node)
{
    g_return_if_fail(CDOS_IS_SETTING_OBJ(obj));
    g_return_if_fail(group_key != NULL);
    g_return_if_fail(group_node != NULL);

    CdosSettingObjPrivate *priv = obj->priv;

    //gboolean different = FALSE;
    gchar *old_val = g_key_file_get_string(priv->key, group_key, "value", NULL);
    gchar *new_val = g_key_file_get_string(priv->key, group_node, "value", NULL);

    if (g_strcmp0(old_val, new_val)){
        cdos_setting_obj_save(obj);
    }
}
static void
cdos_setting_obj_finalize(GObject *gobject) {

    CdosSettingObjPrivate *priv = CDOS_SETTING_OBJ(gobject)->priv;
    if (priv->uuid) {
        g_free(priv->uuid);
    }
    if (priv->instanceId) {
        g_free(priv->instanceId);
    }
    if (priv->setting_file_monitor) {
        g_object_unref(priv->setting_file_monitor);
    }
    G_OBJECT_CLASS(cdos_setting_obj_parent_class)->finalize(gobject);
}

/**
 * cdos_setting_obj_on_file_changed:
 * @obj: 一个#CdosSettingObj对象
 *
 * 当GFileMonitor捕获到“changed”信号时，触发该函数
 *
 */

static void
cdos_setting_obj_on_file_changed(CdosSettingObj *obj)
{

    CdosSettingObjPrivate *priv = obj->priv;
    if (priv->file_changed_timeout) {
        //GSource *source = g_main_current_source();
        g_source_remove(priv->file_changed_timeout);
    }
    priv->file_changed_timeout = g_timeout_add(300, (GSourceFunc)cdos_setting_obj_on_file_changed_timeout, NULL);
}

/**
 * cdos_setting_obj_on_file_changed_timeout:
 * @obj: 一个#CdosSettingObj对象
 * @eventType: #GFileMonitorEvent事件
 *
 * 当定时器超时时，触发该函数
 *
 */

static gboolean
cdos_setting_obj_on_file_changed_timeout(CdosSettingObj *obj)
{
        CdosSettingObjPrivate *priv = obj->priv;

        if (g_file_query_exists(priv->file, NULL)) {
            //if (*eventType != G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT) {
                //return FALSE;
            //}
            GKeyFile *new_key = g_key_file_new();
            g_key_file_load_from_file(new_key, g_file_get_path(priv->file), G_KEY_FILE_NONE, NULL);
            gsize length;
            gchar **groups = g_key_file_get_groups(new_key, &length);
            int i;
            for (i = 0; i < length; i++) 
            {
                gchar *value1 = g_key_file_get_string(priv->key, *(groups + i), "value", NULL);
                gchar *value2 = g_key_file_get_string(new_key, *(groups + i), "value", NULL);
                if (g_strcmp0(value1, value2)) {
                    g_key_file_set_string(priv->key, *(groups + i), "value", value2);
                     cdos_provider_value_changed_notify(priv->provider, *(groups + i), value1, value2);
                }
            }
            g_signal_emit_by_name(obj, "setting-file-changed");     
            cdos_provider_setting_file_changed_notify(priv->provider);
        } else {
            g_signal_handler_disconnect(priv->setting_file_monitor, priv->monitor_id);
        }

        priv->file_changed_timeout = 0;
        return FALSE;
}


static void
cdos_setting_obj_dispose(GObject *gobject)
{
    CdosSettingObjPrivate *priv = CDOS_SETTING_OBJ(gobject)->priv;
    if (priv->file_changed_timeout) {
        //GSource *source = g_main_current_source();
        g_source_remove(priv->file_changed_timeout);
    }

    if (priv->monitor_id > 0) {
        g_signal_handler_disconnect(priv->setting_file_monitor, priv->monitor_id);
        priv->monitor_id = 0;
    }

    if (priv->provider) {
        g_object_unref(priv->provider);
    }
    if (priv->instanceId) {
        g_free(priv->instanceId);
        priv->instanceId = NULL;
    }
    if (priv->setting_file_monitor) {
        g_object_unref(priv->setting_file_monitor);
    }

    G_OBJECT_CLASS(cdos_setting_obj_parent_class)->dispose(gobject);
}

/**
 * cdos_setting_obj_get_key:
 * @obj: 一个#CdosSettingObj对象
 *
 * 获取#CdosSettingObj对象的key属性值
 *
 * Returns: key属性
 */

GKeyFile *
cdos_setting_obj_get_key(CdosSettingObj *obj)
{
    g_return_val_if_fail(CDOS_IS_SETTING_OBJ(obj),NULL);
    if (obj == NULL) 
        return NULL;

    return obj->priv->key;
}
