/**
 * SECTION: cdos-theme-manager
 * @short_description: 管理主题
 * @include: cdos-theme-manager.h
 * 
 * CdosThemeManagerClass 类用来处理主题相关操作.
 */
#include "cdos-theme-manager.h"
#include "cdos.h"
#include "cdos-debug.h"

#include <gio/gio.h>
#include <gtk/gtk.h>
#include <string.h>

G_DEFINE_TYPE (CdosThemeManager, cdos_theme_manager, G_TYPE_OBJECT);

/* 关于 GType 的这个宏，查看:
 * http://library.gnome.org/devel/gobject/2.46/gobject-Type-Information.html#G-DEFINE-TYPE:CAPS
 * char *strcat(char *destin, char *source); 
 */

#define SETTINGS_SCHEMA  "org.cdos.theme"
#define SETTINGS_KEY  "name"
#define HOME_THEMES  "/.themes/"
#define CINNAMON_CSS_FILE  "cdos.css"
#define CINNAMON_DIR "/cdos/"

enum
{
    THEME_SET,
    LAST_SIGNAL
};

static guint cdos_theme_manager_signals [LAST_SIGNAL] = { 0, };

/*声明*/

/*字符串拼接*/
static gchar* cdos_theme_manager_create_path(const gchar *originPath, const gchar *themeName);

/*寻找主题*/
static gchar* cdos_theme_manager_find_theme(const gchar *themeName);

/*改变主题*/
static void cdos_theme_manager_changed_theme(GSettings *settings,
                                             gchar *key,
                                             CdosThemeManager *tm);

/*finalize*/
static void
cdos_theme_manager_finalize(GObject *gobject)
{
    CdosThemeManager *tm = CDOS_THEME_MANAGER(gobject);
    g_free(tm->theme_path);

    G_OBJECT_CLASS (cdos_theme_manager_parent_class)->finalize(gobject);
};

static void
cdos_theme_manager_dispose(GObject *gobject)
{
    CdosThemeManager *tm = CDOS_THEME_MANAGER(gobject);

    g_signal_handler_disconnect(tm->settings,tm->changedId);

    g_clear_object(&tm->settings);


    G_OBJECT_CLASS (cdos_theme_manager_parent_class)->dispose(gobject);

};

/*
 * _init,_class_init初始化
 */
static void
cdos_theme_manager_class_init(CdosThemeManagerClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->finalize = cdos_theme_manager_finalize;
    gobject_class->dispose = cdos_theme_manager_dispose;
    /**
     * CdosThemeManager::theme-set:
     * @object: 绑定信号的对象
     * @user_data: 真对回调函数所传入的参数
     *
     * 该信号对#CdosThemeManager 的设置主题进行监听
     */
    cdos_theme_manager_signals[THEME_SET] =
        g_signal_new("theme-set",
                G_TYPE_FROM_CLASS(klass),
                G_SIGNAL_RUN_LAST,
                G_STRUCT_OFFSET(CdosThemeManagerClass, theme_set),
                NULL,
                NULL,
                g_cclosure_marshal_VOID__VOID,
                /*extension:137:Lang.bind(this, function(){...}*/
        G_TYPE_NONE, 0,
        NULL
    );
}

static void
cdos_theme_manager_init(CdosThemeManager *tm)
{
    tm->settings = g_settings_new (SETTINGS_SCHEMA);
//    cdos_load_theme ();
    tm->changedId = g_signal_connect(tm->settings, "changed::name", G_CALLBACK(cdos_theme_manager_changed_theme),tm);
    tm->theme_path = NULL;
    cdos_theme_manager_changed_theme(tm->settings, NULL, tm);
    //cdos_load_theme ();
}

/*
 * 字符串拼接函数
 */
static gchar*
cdos_theme_manager_create_path(const gchar *originPath, const gchar *themeName)
{
    gchar *data = NULL;
    gint len = strlen(originPath) + strlen(HOME_THEMES) + 
        strlen(themeName) + strlen (CINNAMON_DIR) + strlen(CINNAMON_CSS_FILE);
    data = g_malloc0 (len + 1);
    snprintf(data, len, "%s%s%s%s%s", 
            originPath, HOME_THEMES, themeName, CINNAMON_DIR, CINNAMON_CSS_FILE);

    return data;
}

static gchar*
cdos_theme_manager_find_theme(const gchar *themeName)
{
    gchar *theme_path = NULL;
    const gchar *homeDir = g_get_home_dir();

    /* g_strjoin (NULL, ...) */
    gchar *user_path = cdos_theme_manager_create_path(homeDir, themeName);

    GFile *file = g_file_new_for_path(user_path);
    /*释放GFile内存g_object_unref()*/

    if (g_file_query_exists(file, NULL))
    {
        /*g_file_query_exists 是gboolean类型.*/
        theme_path = g_strdup(user_path);
    } else {

        /*返回不可变,不能被free的,以NULL结尾的字符指针二维数组, (const gchar*(const *))类型 */
        const gchar *const *sysdirs = g_get_system_data_dirs();

        guint i;
        for(i = 0; sysdirs[i] != NULL; i++)
        {
            gchar *syspath = g_strjoin (NULL, sysdirs[i], "themes/", themeName, 
                    "/cdos/", CINNAMON_CSS_FILE, NULL);

            GFile *file2 = g_file_new_for_path(syspath);
            if(g_file_query_exists(file2, NULL)) {
                theme_path = syspath;
                g_object_unref(file2);
                break;
            }
            else {
                g_object_unref(file2);
                g_free (syspath);
            }
        }
    }
    g_free (user_path);
    cdos_debug ("theme directory: %s\n", theme_path);
    if (file){
        g_object_unref(file);
    }
    return theme_path;
}



/*_changedTheme()*/
static void
cdos_theme_manager_changed_theme(GSettings *settings,
                                 gchar     *key,
                                 CdosThemeManager *tm)
{
    GtkIconTheme *icon_theme = gtk_icon_theme_get_default();

    if (tm->theme_path)
    {
        int len = strlen(tm->theme_path) - strlen(CINNAMON_CSS_FILE);
        gchar *tmp_path = g_strndup (tm->theme_path, len);

        gchar **search_path = NULL;
        gint i, n_elements;
        /*后面两个参数是接收第一个参数传入后返回的两个结果*/
        gtk_icon_theme_get_search_path(icon_theme, &search_path, &n_elements);
        gboolean find_icon_dir = 0;
        for(i = 0; i < n_elements; i++) {
            if(g_utf8_collate(search_path[i],tmp_path)==0)
            {
                find_icon_dir = 1;
                break;
            }
        }
        g_free (tmp_path);
        gint j = 0, k = 0;
        gchar **tmp = g_new (gchar *, n_elements - find_icon_dir);
        
        while(j < n_elements)
        {
            if(i != j)
            {
                tmp[k] = g_strdup(search_path[j]);
                k++;
            }
            j++;
        }
        gtk_icon_theme_set_search_path(icon_theme, (const gchar **)tmp, n_elements-find_icon_dir);
        g_strfreev(search_path);
    }

    gchar *stylesheet = NULL;
    gchar *themeName = g_settings_get_string(tm->settings, SETTINGS_KEY);

    if(themeName) {
        g_free (tm->theme_path);
        tm->theme_path = NULL;
        tm->theme_path = cdos_theme_manager_find_theme(themeName);

        if(tm->theme_path != NULL)
        {
            stylesheet = tm->theme_path;
        }
    }

    if (stylesheet){
        printf("loading user theme: %s\n", stylesheet);
    }else
    {
        printf("loading default theme\n");
    }
    /*
       导入Main函数
       Main.setThemeStylesheet(_stylesheet);
       Main.loadTheme();
       */
    cdos_set_theme_style_sheet (stylesheet);
    cdos_load_theme ();
    if (tm->theme_path != NULL)
    {
        int len = strlen(tm->theme_path) - strlen(CINNAMON_CSS_FILE);
        gchar *tmp = g_strndup (tm->theme_path, len);
        gtk_icon_theme_append_search_path(icon_theme, tmp);
        printf("added icon directory: %s\n", tmp);
        g_free (tmp);
    }
    g_signal_emit_by_name(tm, "theme-set", NULL);
    g_free (themeName);
}

/*
   Signals.addSignalMethods(ThemeManager.prototype);
   不用管
   */
