/**
 * SECTION: cdos-url-highlighter
 * @short_description:  用于对通知的URL进行高亮显示
 * @include: ui/cdos-url-highlighter.h
 *
 * CdosUrlHighlighterClass 类用来处理通知高亮和换行
 */
#include <stdlib.h>
#include <string.h>
#include "cdos-url-highlighter.h"
#include "cdos-global.h"
#include "cdos-debug.h"

G_DEFINE_TYPE (CdosUrlHighlighter, cdos_url_highlighter, G_TYPE_OBJECT);

/* ========== declear ========== */

/* private func */
static void cdos_url_highlighter_highlight_urls (CdosUrlHighlighter *self);
static gint cdos_url_highlighter_find_url_at_pos (CdosUrlHighlighter *self, ClutterEvent *event);
/* callback func */
static void onStyleChanged (StWidget *actor, gpointer user_data);
static gboolean onButtonPressEvent (GObject *object, ClutterEvent *event, gpointer user_data);
static gboolean onButtonReleaseEvent (GObject *object, ClutterEvent *event, gpointer user_data);
static gboolean onMotionEvent (GObject *object , ClutterEvent *event, gpointer user_data);
static gboolean onLeaveEvent (GObject *object, ClutterEvent* event, gpointer user_data);

static char* cdos_url_sub_string(const char* src, int startpos, int substrlen);

/* ======== declear end ======== */
/* public func */

/**
 * cdos_url_highlighter_set_markup:
 * @self: A #CdosUrlHighlighter
 * @text: 待处理文本
 * @allowMarkup: 是否允许处理
 *
 * 判断是否要对传入文本进行处理
 *
 */
void cdos_url_highlighter_set_markup (CdosUrlHighlighter *self, gchar *text, gboolean allowMarkup)
{
    g_return_if_fail(CDOS_IS_URL_HIGHLIGHTER(self));
    gchar *new_text = NULL;
    if (text != NULL) {
        new_text = cdos_url_highlighter_fix_markup (text, allowMarkup);
    } else {
        new_text = g_strdup("");
    }
    if(self->text){
        g_free(self->text);
    }
    self->text = g_strdup (new_text);
    //StLabel *actor = self->actor;
    ClutterActor *clutter_text = st_label_get_clutter_text (self->actor);
    clutter_text_set_markup (CLUTTER_TEXT (clutter_text), (const gchar *)self->text);
    const gchar *t = clutter_text_get_text (CLUTTER_TEXT (clutter_text));
    self->urls = cdos_util_find_urls ((gchar *)t);
    cdos_url_highlighter_highlight_urls (self);
    //g_print("\n\n INFO: \nnew_text %s| \nt %s| \n\n", new_text, t);
    g_free(new_text);
}

static char* cdos_url_sub_string(const char* src, int startpos, int substrlen) {
    if(src == NULL){
        return g_strdup("");
    }

    int srclen = strlen(src);
    if (srclen <= startpos) {
        return g_strdup("");
    }

    char* ret = NULL;
    int mallocsize = 0;
    if (0 > substrlen) {
        mallocsize = sizeof(char) * (srclen - startpos) + 1;
    } else {
        mallocsize = sizeof(char) * (substrlen) + 1;
    }

    ret = (char*)malloc(mallocsize);
    memset(ret, 0, mallocsize);
    int i = 0;
    while ('\0' != *(src + startpos + i) && i < (mallocsize - 1) / sizeof(char)) {
        *(ret + i) = *(src + startpos + i);
        ++i;
    }
    return ret;
}

 /**
  * cdos_url_highlighter_fix_markup:
  * @text: 待处理文本
  * @allowMarkup: 是否允许进行标记
  *
  * 利用正则表达式处理文本内部的html转移字符，并保留加粗，斜体，下划线样式。
  *
  * Returns (gchar): 处理之后的文本
  */

gchar *cdos_url_highlighter_fix_markup (gchar *text, gboolean allowMarkup)
{
    g_return_val_if_fail(text != NULL,NULL);
    if (allowMarkup) {
        gchar *new_str = NULL, *text_str = NULL;
        GRegex *regex_1, *regex_2;

        regex_1 = g_regex_new ("(&quot;|&apos;|&lt;|&gt;)", 0, 0, NULL);
        new_str = g_regex_replace (regex_1, text, -1, 0, "&amp;", 0, NULL);
        g_regex_unref (regex_1);

        //regex_2 = g_regex_new ("<(?!\/?[biu]>)", 0, 0, NULL);
        regex_2 = g_regex_new ("<[/]?([^biu]|[a-zA-Z]{2,}[/]?)>", 0, 0, NULL);
        text_str = g_regex_replace (regex_2, new_str, -1, 0, "&lt;", 0, NULL);
        g_regex_unref (regex_2);

        pango_parse_markup (text_str , -1 , 0 , NULL , NULL , NULL , NULL);

        g_free(new_str);

        return text_str;
    }

    return g_markup_escape_text ( text , -1 );
}

/* private func */
static void cdos_url_highlighter_highlight_urls (CdosUrlHighlighter *self)
{
    gchar *markup = g_strdup("");
    gint pos = 0;
    //guint i = 0;
    GList *urls = cdos_util_find_urls (self->text);     //返回一个url的阵列，url类型未知
    gchar *temp = NULL;
    //g_print("\n ------------------------------ \n");
    while (urls != NULL){
        MatchedUrl *url = urls->data;
        gint start_pos = cdos_util_get_url_start_pos (url);
        char *str = cdos_url_sub_string (self->text, pos, start_pos - pos);

        //g_print("\n %s:start_pos=%d; \n",__func__,start_pos);

        temp = g_strconcat (markup, str, "<span foreground=\"", self->linkColor, "\"><u>", cdos_util_get_url (url), "</u></span>", NULL);
        //g_print("\n %s:str=%s; \n",__func__,str);
        //g_print("\n %s:temp=%s; \n",__func__,temp);
        pos = (gint)(start_pos + strlen (cdos_util_get_url (url)));
        g_free(markup);
        markup = temp;
        g_free(str);
        urls = urls->next;
    }

    //g_print("\n %s:pos=%d; \n",__func__,pos);
    //g_print("\n %s:len=%d; \n",__func__,strlen (self->text));

    char *substr = cdos_url_sub_string (self->text, pos, strlen (self->text));
    temp = g_strconcat (markup, substr , NULL );

    //g_print("\n %s:temp = %s; \n",__func__,temp);
    //g_print("\n %s:markup = %s; \n",__func__,markup);

    g_free(markup);
    g_free(substr);

    markup = temp;

    ClutterActor *clutter_text = st_label_get_clutter_text(self->actor);
    clutter_text_set_markup (CLUTTER_TEXT(clutter_text), markup);

    g_free(markup);
}

static gint cdos_url_highlighter_find_url_at_pos (CdosUrlHighlighter *self, ClutterEvent *event)
{
    gfloat x, y;
    gfloat x_out, y_out;
    gint i, j = 0;
    gint find_pos = -1;
    ClutterText *actor_clutter_text;


    clutter_event_get_coords ( event, &x, &y );
    clutter_actor_transform_stage_point ( CLUTTER_ACTOR(self->actor), x , y , &x_out, &y_out);

    actor_clutter_text = CLUTTER_TEXT (st_label_get_clutter_text (self->actor));

    const gchar *clutter_text = clutter_text_get_text ( actor_clutter_text );
    glong clutter_text_len = strlen (clutter_text);

    //g_print("LINE: %d, FUNC: %s -----> clutter_text_len: %d, clutter_text: %s!\n", __LINE__, __func__, clutter_text_len, clutter_text);
    for ( i = 0 ; i < clutter_text_len;  i++ ) {
        gfloat px, py, line_height;
        clutter_text_position_to_coords (actor_clutter_text, i, &px, &py, &line_height);
        //g_print("LINE: %d, FUNC: %s -----> (%d, %d), (%d, %d)_out, p(%d, %d)!\n",__LINE__, __func__, x, y, x_out, y_out, px, py);
        if ( py > y_out || py + line_height < y_out  || x_out < px ) {
            //g_print("LINE: %d, FUNC: %s -----> continue, donothing!\n", __LINE__, __func__);
            continue ;
        }
        find_pos = i;
    }

    if (find_pos != -1) {
        GList *urls = self->urls;	
        while(urls != NULL){
            MatchedUrl *url = urls->data;
				if(url){
                    gint url_pos = cdos_util_get_url_start_pos (url);
                    glong url_len = strlen (cdos_util_get_url (url));
                    if ((find_pos >= url_pos) && (url_pos + url_len > find_pos)) {
                        return j;
                    }
				}
				j++;
            urls = urls->next;
        }
    }

    return -1;
}

/* callback func */
static void onStyleChanged (StWidget *actor, gpointer user_data)
{

    CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (user_data);

    gboolean hasColor;
    ClutterColor *color = NULL;
    hasColor = st_theme_node_lookup_color ( st_widget_get_theme_node (actor),
                                            "link-color",
                                            FALSE,
                                            color );
    if (hasColor) {
        gchar *color_string = clutter_color_to_string (color);
        gchar *linkColor = g_utf8_substring ( color_string, 0, 7 );

        if (g_utf8_collate ((const gchar *)linkColor, (const gchar *)self->linkColor) != 0) {

            self->linkColor = g_strdup (linkColor);

            cdos_url_highlighter_highlight_urls (self);
        }
        if(color_string){
            g_free(color_string);
        }
        if (linkColor) {
            g_free (linkColor);
        }

        if (color) {
            g_free (color);
        }
    }
}


static gboolean onButtonPressEvent (GObject *object, ClutterEvent *event, gpointer user_data)
{
   CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (user_data);
   
    //g_print ("\n INFO: button-press-event! (%s) \n", __FUNCTION__);
    // Don't try to URL highlight when invisible.
    // The MessageTray doesn't actually hide us, so
    // we need to check for paint opacities as well.

    if ( !clutter_actor_get_paint_visibility ( CLUTTER_ACTOR(self->actor) )
         || clutter_actor_get_paint_opacity ( CLUTTER_ACTOR(self->actor) ) == 0 )
        return FALSE;

    // Keep Notification.actor from seeing this and taking
    // a pointer grab, which would block our button-release-event
    // handler, if an URL is clicked
    int urlId = cdos_url_highlighter_find_url_at_pos (self, event) != -1;
    //g_print("LINE: %d, FUNC: %s -----> urlid: %d!\n", __LINE__, __func__, urlId);
    if (0 > urlId) {
        return FALSE;
    }
    return TRUE;
}


static gboolean onButtonReleaseEvent (GObject *object, ClutterEvent *event, gpointer user_data)
{
   CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (user_data);
    gint urlId = -1;
    //g_print ("\n INFO: button-release-event! (%s) \n", __FUNCTION__);
    if ( !clutter_actor_get_paint_visibility ( CLUTTER_ACTOR(self->actor) )
         || clutter_actor_get_paint_opacity ( CLUTTER_ACTOR(self->actor) ) == 0 )
        return FALSE;

    urlId = cdos_url_highlighter_find_url_at_pos( self, event );
    //g_print("LINE: %d, FUNC: %s -----> urlid: %d!\n", __LINE__, __func__, urlId);

    if (urlId != -1) {

        GError *error = NULL;
        GAppLaunchContext *context;
        //gchar *url = self->urls[urlId].url;  //不确定
        MatchedUrl *matchedUrl = g_list_nth_data ( self->urls , (guint) urlId );
        gchar *url = cdos_util_get_url (matchedUrl);

        GRegex *regex_1;
        gchar *new_str;

        regex_1 = g_regex_new ("(\\s)", 0, 0, NULL);
        new_str = g_regex_replace (regex_1, url, -1, 0, "", 0, NULL);
        g_regex_unref (regex_1);

        url = new_str;

        if (index (url , ':') == NULL ) {
            gchar *temp = url;
            url = g_strconcat("http://", url, NULL);
            g_free(temp);
        }

        //g_print("\n %s:url = %s; \n",__func__,url);
        context = cdos_global_create_app_launch_context (cdos_global_get ());

        g_app_info_launch_default_for_uri ( url, context, &error );

        g_object_unref(context);
        //g_print("\n %s:error = %x; \n",__func__,error);

        if (error == NULL) {
            g_free(url);
            return TRUE;
        } else {
            g_warning ("\n Func:%s Error message: %s \n", __FUNCTION__, error->message);
            gint pid;
            gchar *cmd = g_strconcat ("gvfs-open ", url, NULL);
            const gchar  *new_cmd = (const gchar *)cmd;

            //g_print("\n %s:new_cmd = %s; \n",__func__,new_cmd);

            pid = cdos_util_spawn (&new_cmd);
            if (pid <= 0) {
                g_warning ("\n INFO: maybe error in %s:%d, becase: %s \n", __FUNCTION__, __LINE__, error->message);
            }
            g_free(url);
            g_error_free(error);
            g_free(cmd);
            return TRUE;
        }
    }
    return FALSE;
}


static gboolean onMotionEvent (GObject *object , ClutterEvent *event, gpointer user_data)
{

   CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (user_data);
    gint urlId;
    CdosGlobal *global = cdos_global_get();

    if ( !clutter_actor_get_paint_visibility ( CLUTTER_ACTOR(self->actor) )
         || clutter_actor_get_paint_opacity ( CLUTTER_ACTOR(self->actor) ) == 0 )
        return FALSE;

    urlId = cdos_url_highlighter_find_url_at_pos( self, event );
    MetaScreen *screen = cdos_global_get_screen (global);
    if (urlId != -1 && !self->cursorChanged) {
        meta_screen_set_cursor (screen, META_CURSOR_POINTING_HAND );
        self->cursorChanged = TRUE;
    } else if (urlId == -1) {
        meta_screen_set_cursor (screen, META_CURSOR_DEFAULT);
        self->cursorChanged = FALSE;
    }
    return FALSE;
}


static gboolean onLeaveEvent (GObject *object, ClutterEvent* event, gpointer user_data)
{

    CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (user_data);
    CdosGlobal *global = cdos_global_get();

    if ( !clutter_actor_get_paint_visibility ( CLUTTER_ACTOR(self->actor) )
         || clutter_actor_get_paint_opacity ( CLUTTER_ACTOR(self->actor) ) == 0 ){
        return FALSE;
    }

    if (self->cursorChanged) {
        self->cursorChanged = FALSE;
        meta_screen_set_cursor (cdos_global_get_screen (global), META_CURSOR_DEFAULT);
    }
    return FALSE;
}

/* GObject */
static void cdos_url_highlighter_finalize (GObject *object)
{
    CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (object);
    GList *urls = self->urls;
    while (urls != NULL){
        MatchedUrl *url = urls->data;
        if(url){
            gchar *u = cdos_util_get_url(url);
            g_free(u);
            g_free(url);
        }
        urls = urls->next;
    }
    g_list_free (self->urls);
    if(self->linkColor){
        g_free (self->linkColor);
    }
    G_OBJECT_CLASS(cdos_url_highlighter_parent_class)->finalize(object);
}

static void cdos_url_highlighter_dispose (GObject *object)
{
    CdosUrlHighlighter *self = CDOS_URL_HIGHLIGHTER (object);
    g_clear_object (&self->actor);
    G_OBJECT_CLASS(cdos_url_highlighter_parent_class)->dispose(object);
}

static void cdos_url_highlighter_class_init (CdosUrlHighlighterClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->finalize = cdos_url_highlighter_finalize;
    gobject_class->dispose = cdos_url_highlighter_dispose;
}

static void cdos_url_highlighter_init (CdosUrlHighlighter *self)
{
    self->actor = ST_LABEL (st_label_new ("default"));
    st_widget_set_style_class_name (ST_WIDGET (self->actor), "url-highlighter");
    clutter_actor_set_reactive (CLUTTER_ACTOR (self->actor), TRUE);

    self->linkColor = g_strdup("#ccccff");
    self->text = NULL;
    self->urls = NULL;
    self->cursorChanged = FALSE;
}

 /**
  * cdos_url_highlighter_new:
  * @text: 待处理文本
  * @lineWrap: 是否允许断行
  * @allowMarkup: 是否允许标记
  * 
  * 新建一个#CdosUrlHighlighter类的实例。
  *
  * Returns (CdosUrlHighlighter): 返回一个#CdosUrlHighlighter类的实例。
  */
CdosUrlHighlighter *cdos_url_highlighter_new (const gchar *text, gboolean lineWrap, gboolean allowMarkup)
{
    CdosUrlHighlighter *self = g_object_new (CDOS_TYPE_URL_HIGHLIGHTER, NULL);
    gchar *new_text = NULL;
    if (text) {
        new_text = g_strdup(text);
    }

    g_signal_connect (self->actor , "style-changed", G_CALLBACK (onStyleChanged), self);

    if (lineWrap) {

        ClutterText *actor_clutter_text = CLUTTER_TEXT(st_label_get_clutter_text (self->actor));

        clutter_text_set_line_wrap (actor_clutter_text, TRUE);
        clutter_text_set_line_wrap_mode (actor_clutter_text, PANGO_WRAP_CHAR);
        clutter_text_set_ellipsize (actor_clutter_text, PANGO_ELLIPSIZE_END);
    }

    cdos_url_highlighter_set_markup (self, new_text, allowMarkup);

    g_signal_connect (self->actor, "button-press-event", G_CALLBACK (onButtonPressEvent), self);
    g_signal_connect (self->actor, "button-release-event", G_CALLBACK (onButtonReleaseEvent), self);
    g_signal_connect (self->actor, "motion-event", G_CALLBACK (onMotionEvent), self);
    g_signal_connect (self->actor, "leave-event", G_CALLBACK (onLeaveEvent), self);
    if(new_text)
        g_free(new_text);
    return self;
}

/* ======== function end ======== */
