#define CLUTTER_DISABLE_DEPRECATION_WARNINGS

#include <math.h>

#include "cdos-window-clone.h"
#include "cdos-layout-manager.h"

#include "cdos-global.h"
#include "cdos-base-util.h"
#include "cdos-debug.h"

/**
 * SECTION:cdos-window-clone
 * @Title: CdosWindowClone
 * @Short_description: 窗口克隆体
 * 
 * CdosWindowClone 就是当前工作区内，窗口的克隆体。我们通过组合键进入 overview 缩略图模式，
 * 这个对象就是每个窗口的克隆体，由于是克隆体，退出 overview 模式时，需要将这些克隆体销毁，
 * 否则桌面上会多出窗口的复制体，不可使用。
 */
G_DEFINE_TYPE (CdosWindowClone, cdos_window_clone, CLUTTER_TYPE_ACTOR);

enum {
	WORKSPACE_CHANGED,
	SELECTED,
	ACTIVATED,
	CLOSED,
    DRAG_BEGIN,
    DRAG_CANCELLED,
    DRAG_END,
	LAST_SIGNAL
};

static guint cdos_window_clone_signals [LAST_SIGNAL] = {0};

static void
getInvisibleBorderPadding (CdosWindowClone *self, gfloat *borderX, gfloat *borderY)
{
    MetaRectangle outerRect, inputRect;

    // We need to adjust the position of the actor because of the
    // consequences of invisible borders -- in reality, the texture
    // has an extra set of "padding" around it that we need to trim
    // down.

    // The outer rect paradoxically is the smaller rectangle,
    // containing the positions of the visible frame. The input
    // rect contains everything, including the invisible border
    // padding.

    //let outerRect = this.metaWindow.get_outer_rect();
    meta_window_get_frame_rect ( self->metaWindow, &outerRect );

    //let inputRect = this.metaWindow.get_input_rect();
    meta_window_get_buffer_rect ( self->metaWindow, &inputRect );

    *borderX = outerRect.x - inputRect.x;
    *borderY = outerRect.y - inputRect.y;	
}

static void
onRealWindowSizeChanged (GObject *gobject, gpointer user_data)
{
	gfloat borderX, borderY;
	MetaRectangle outerRect;
	CdosWindowClone *self = CDOS_WINDOW_CLONE ( user_data );
	
	getInvisibleBorderPadding ( self, &borderX, &borderY );

	meta_window_get_buffer_rect (self->metaWindow, &outerRect );	

	clutter_actor_set_size ((ClutterActor *)self, outerRect.width, outerRect.height );

	clutter_actor_set_position (self->clone, -borderX, -borderY );
}

static void 
onWorkspaceChanged (MetaWindow *w, gint oldws, gpointer user_data )
{
	CdosWindowClone *self = CDOS_WINDOW_CLONE ( user_data );
	g_signal_emit (self, cdos_window_clone_signals[WORKSPACE_CHANGED], 0, oldws );
}

static void
disconnect_window_signals (MetaWindowActor *window, CdosWindowClone *self)
{
    if (self->workspaceChangedId)
        g_signal_handler_disconnect (self->metaWindow, self->workspaceChangedId);
    self->workspaceChangedId = 0L;

    if (self->sizeChangedId)
        g_signal_handler_disconnect (self->metaWindow, self->sizeChangedId);
    self->sizeChangedId = 0L;
    if (self->realWindowDestroyId)
        g_signal_handler_disconnect (self->realWindow, self->realWindowDestroyId);
    self->realWindowDestroyId = 0L;
}

static void 
on_real_window_destroyed (ClutterActor *actor, CdosWindowClone *self)
{
    clutter_actor_destroy ((ClutterActor *)self);
}

static gboolean
on_button_release_event (ClutterActor *actor, ClutterButtonEvent *button_event)
{
    CdosWindowClone *self = CDOS_WINDOW_CLONE (actor);
	guint32 current_time = cdos_global_get_current_time ( cdos_global_get () );	
    ClutterEvent *event = (ClutterEvent *)button_event;
	guint32 event_type = clutter_event_get_button (event);

    if (event_type != 2)
    {
        self->selected = TRUE;
        g_signal_emit_by_name ( self, "activated", current_time );
        return TRUE;
    }
    else if (event_type == 2)
    {
        g_signal_emit_by_name ( self, "closed", current_time );

        return TRUE;
    }
    return FALSE;
}

static gboolean
on_button_press_event (ClutterActor *actor, ClutterButtonEvent *event)
{
	guint32 current_time = cdos_global_get_current_time ( cdos_global_get () );

	g_signal_emit_by_name (actor, "selected", current_time);
    return TRUE;
}

static gboolean
on_leave_event (ClutterActor *actor, ClutterCrossingEvent *event)
{
    CdosWindowClone *self = CDOS_WINDOW_CLONE (actor);
    return FALSE;
}

static void
cdos_window_clone_dispose (GObject *gobject)
{
    CdosWindowClone *self = CDOS_WINDOW_CLONE (gobject);
    disconnect_window_signals (NULL, self);
    /* clutter_actor_dispose () 发出 destroy() 信号，销毁 window-overlay */
    G_OBJECT_CLASS (cdos_window_clone_parent_class)->dispose (gobject);
}

static void 
cdos_window_clone_get_preferred_height (ClutterActor *actor,
                                        gfloat        for_width,
                                        gfloat       *min_height,
                                        gfloat       *nat_height)
{
    CdosWindowClone *clone = CDOS_WINDOW_CLONE (actor);
    MetaWindow *window = clone->metaWindow;
    MetaRectangle rect;
    meta_window_get_frame_rect (window, &rect);
    if (min_height)
        *min_height = rect.height;
    if(nat_height)
        *nat_height = rect.height;
}

static void 
cdos_window_clone_get_preferred_width (ClutterActor *actor,
                                       gfloat        for_height,
                                       gfloat       *min_width,
                                       gfloat       *nat_width)
{
    CdosWindowClone *clone = CDOS_WINDOW_CLONE (actor);
    MetaWindow *window = clone->metaWindow;
    MetaRectangle rect;
    meta_window_get_frame_rect (window, &rect);
    if (min_width)
        *min_width = rect.width;
    if(nat_width)
        *nat_width = rect.width;
}

static void
cdos_window_clone_allocate (ClutterActor *actor,
                            const ClutterActorBox *box,
                            ClutterAllocationFlags flags)
{
    CLUTTER_ACTOR_CLASS (cdos_window_clone_parent_class)->allocate (actor, box, flags);

    CdosWindowClone *clone = CDOS_WINDOW_CLONE (actor);
    MetaWindow *window = clone->metaWindow;
    MetaRectangle frame_rect, buffer_rect;
    meta_window_get_frame_rect (window, &frame_rect);
    meta_window_get_buffer_rect (window, &buffer_rect);
    ClutterActorBox box1;
    box1.x1 = buffer_rect.x - frame_rect.x;
    box1.y1 = buffer_rect.y - frame_rect.y;
    box1.x2 = box1.x1 + buffer_rect.width;
    box1.y2 = box1.y1 + buffer_rect.height;
    clutter_actor_allocate (clone->clone, &box1, flags);
}

static void
cdos_window_clone_class_init ( CdosWindowCloneClass *klass )
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    gobject_class->dispose = cdos_window_clone_dispose;

    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);
    actor_class->get_preferred_width = cdos_window_clone_get_preferred_width;
    actor_class->get_preferred_height = cdos_window_clone_get_preferred_height;
    actor_class->allocate = cdos_window_clone_allocate;
    actor_class->button_press_event = on_button_press_event;
    actor_class->button_release_event = on_button_release_event;
    actor_class->leave_event = on_leave_event;

	cdos_window_clone_signals [WORKSPACE_CHANGED] =
                g_signal_new ("workspace-changed",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__POINTER,
                        G_TYPE_NONE, 1 ,
			            G_TYPE_INT );

    cdos_window_clone_signals [SELECTED] =
                g_signal_new ("selected",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__UINT,
                        G_TYPE_NONE, 1,
                        G_TYPE_UINT );

    cdos_window_clone_signals [ACTIVATED] =
                g_signal_new ("activated",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__UINT,
                        G_TYPE_NONE, 1,
                        G_TYPE_UINT );

    cdos_window_clone_signals [CLOSED] =
                g_signal_new ("closed",
                        G_TYPE_FROM_CLASS (klass),
                        G_SIGNAL_RUN_LAST,
                        0,
                        NULL, NULL,
                        g_cclosure_marshal_VOID__UINT,
                        G_TYPE_NONE, 1,
                        G_TYPE_UINT );
    /**
     * CdosWindowClone::drag-begin:
     * @window_clone: 正在拖动的CdosWindowClone对象
     * @time: 事件发生的时间
     *
     * 在开始拖动时发出 ::drag-begin 信号。
     *
     * Since: 1.0.0
     */
    cdos_window_clone_signals[DRAG_BEGIN] = 
        g_signal_new ("drag-begin",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__UINT,
                    G_TYPE_NONE, 1,
                    G_TYPE_UINT);
    /**
     * CdosWindowClone::drag-cancelled:
     * @window_clone: 正在拖动的CdosWindowClone对象
     * @draggable: 正在拖动的DND对象
     * @time: 事件发生的时间
     *
     * 在拖动取消时发出 ::drag-cancelled 信号。
     *
     * Since: 1.0.0
     */
    cdos_window_clone_signals[DRAG_CANCELLED] = 
        g_signal_new ("drag-cancelled",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    g_cclosure_marshal_VOID__UINT,
                    G_TYPE_NONE, 1,
                    G_TYPE_UINT);
    /**
     * CdosWindowClone::drag-end:
     * @window_clone: 正在拖动的CdosWindowClone对象
     * @time: 事件发生的时间
     * @snap: 拖动取消时为 %FALSE, 正常结束为 %TRUE。
     *
     * 在拖动结束时发出 ::drag-end 信号。
     *
     * Since: 1.0.0
     */
    cdos_window_clone_signals[DRAG_END] = 
        g_signal_new ("drag-end",
                    G_TYPE_FROM_CLASS (klass),
                    G_SIGNAL_RUN_LAST,
                    0,
                    NULL, NULL,
                    _st_marshal_VOID__UINT_UINT,
                    G_TYPE_NONE, 2,
                    G_TYPE_UINT, G_TYPE_UINT);

}

static void
cdos_window_clone_init (CdosWindowClone *self)
{
    self->overlay = NULL;
    self->stackAbove = NULL;
    self->selected = FALSE;
}

static void
cdos_window_clone_post_init (CdosWindowClone *self ,
                             MetaWindowActor *realWindow,
                             CdosWorkspaceMonitor *myContainer)
{
    if (realWindow == NULL)
        return;

    self->myContainer = myContainer;
    self->realWindow = realWindow;
    self->origX = clutter_actor_get_x ((ClutterActor *)realWindow);
    self->origY = clutter_actor_get_y ((ClutterActor *)realWindow);
    self->metaWindow = meta_window_actor_get_meta_window (self->realWindow);
    self->clone = clutter_clone_new (meta_window_actor_get_texture (realWindow));
	cdos_base_util_set_hidden_from_pick (self->clone, TRUE);
    clutter_actor_add_child ((ClutterActor *)self, self->clone);

	self->sizeChangedId = g_signal_connect (self->metaWindow , "size-changed", G_CALLBACK (onRealWindowSizeChanged), self);

    self->workspaceChangedId = g_signal_connect (self->metaWindow , "workspace-changed", G_CALLBACK (onWorkspaceChanged), self);	

	self->realWindowDestroyId = g_signal_connect (self->realWindow, "destroy", G_CALLBACK (on_real_window_destroyed), self);
}

/**
 * cdos_window_clone_new:
 * @real_window: #MetaWindowActor 对象，实际打开的窗口
 * @my_container: #CdosWorkspaceMonitor 对象
 *
 * 新建窗口克隆体
 *
 * Return value: #CdosWindowClone 对象
 */
CdosWindowClone* 
cdos_window_clone_new (MetaWindowActor *window, CdosWorkspaceMonitor *my_container)
{
    CdosWindowClone *instance = g_object_new (CDOS_TYPE_WINDOW_CLONE, "reactive", TRUE, NULL);

	cdos_window_clone_post_init (instance, window, my_container);

    return instance;
}

/**
 * cdos_window_clone_set_stack_above:
 * @self: #CdosWindowClone 对象
 * @actor: 需要置于顶层的对象，可空
 *
 * 将 actor 置于最顶层。 #CdosWindowClone 对象的 actor 成员用于处理所有的交互事件，例如鼠标点击等，
 * 之所以用 actor 而不用 clone［存放克隆窗口的组］，是因为 actor 才能对事件做出响应。
 */
void
cdos_window_clone_set_stack_above (CdosWindowClone *self, ClutterActor *actor)
{
    self->stackAbove = actor;

    if (self->stackAbove == NULL)
    {
        //this.actor.lower_bottom();
        ClutterActor *parent = clutter_actor_get_parent ((ClutterActor *)self);
        clutter_actor_set_child_below_sibling (parent, (ClutterActor *)self, NULL);
        //		clutter_actor_lower_bottom ( self->actor );                   
    }
    else
    {
        ClutterActor *parent = clutter_actor_get_parent (self->stackAbove);
        clutter_actor_set_child_above_sibling (parent, (ClutterActor *)self, self->stackAbove);
    }
}

