#include <stdarg.h>
#include <math.h>

#include "cdos-debug.h"
#include <meta/display.h>
#include <meta/keybindings.h>
#include <meta/common.h>

#include "cdos.h"
#include "cdos-wm.h"
#include "cdos-start.h"
#include "cdos-window-manager.h"
#include "cdos-overview.h"
#include "cdos-export-overview.h"
#include "cdos-coverflow-switcher.h"

/**
 * SECTION: cdos-window-manager
 * @Title: CdosWM
 * @short_description: 管理窗口及工作区切换映射等
 *
 * 创建一个 CdosWM, 管理窗口切换、映射、关闭和切换工作区等的动画及绑定快捷键。
 */

#define WINDOW_ANIMATION_TIME   250
#define TILE_HUD_ANIMATION_TIME 150
#define DIM_TIME                500
#define DIM_DESATURATION        0.600
#define DIM_BRIGHTNESS          -0.2
#define UNDIM_TIME              250
#define MAP_TIME                250
#define MAXIMIZE_TIME           250
#define MINIMIZE_TIME           150
#define DESTROY_TIME            150
#define SWITCH_TIME             300

// add by liting
static ClutterActor  *snap_osd = NULL; // StBoxLayout
static ClutterActor   *workspace_osd = NULL; //StLabel
// end

static int
get_top_invisible_border (MetaWindow *window)
{
    MetaRectangle outer, input;
    meta_window_get_frame_rect (window, &outer);
    meta_window_get_buffer_rect (window, &input);
    return outer.y - input.y;
}

static gboolean desktop_effects = FALSE;

static void 
_should_animate (GSettings *settings, gchar *key)
{
    if (g_strcmp0 (key, "desktop-effects") == 0)
    {
        desktop_effects = g_settings_get_boolean (settings, "desktop-effects");
    }
}

static void
should_animate ()
{
    GSettings *settings = cdos_global_get_settings (global);
    desktop_effects = g_settings_get_boolean (settings, "desktop-effects");
    g_signal_connect (settings, "changed", G_CALLBACK (_should_animate), NULL);
}

static gboolean
enable_desktop_effects ()
{

    if (cdos_get_modal_count () || cdos_is_software_rendering ())
    {
        return FALSE;
    }
    return desktop_effects;
}

/* 动画 */
typedef struct _DefaultTransData {
    ClutterActor *actor;
    const gchar *name;
}DefaultTransData;

static void 
default_transition_complete_callback (ClutterTimeline *tml, DefaultTransData *data)
{
    if (data->name)
        clutter_actor_remove_transition (data->actor, data->name);
    g_free (data);
}

static void 
default_transition_stop_callback (ClutterTimeline *tml, gboolean is_finished, DefaultTransData *data)
{
    if (!is_finished)
    {
        if (data->name)
            clutter_actor_remove_transition (data->actor, data->name);
        g_free (data);
    }
}

/**
 * 不需要回调函数的动画
 */
ClutterTransition*
cdos_transition_simple (ClutterActor *actor,
                        ClutterAnimationMode mode,
                        guint duration,
                        const char *first_property_name,
                        ...)
{
    g_return_val_if_fail (CLUTTER_IS_ACTOR (actor), NULL);
    g_return_val_if_fail (first_property_name != NULL, NULL);
    va_list args;

    clutter_actor_save_easing_state (actor);
    clutter_actor_set_easing_mode (actor, mode);
    clutter_actor_set_easing_duration (actor, duration);

    va_start (args, first_property_name);
    g_object_set_valist ((GObject *)actor, first_property_name, args);
    va_end (args);
    ClutterTransition *transition = clutter_actor_get_transition (actor, first_property_name);
    clutter_actor_restore_easing_state (actor);
    return transition;
}

/**
 * cdos_transition_to:
 * @actor: A #ClutterActor
 * @mode: 动画模式, #ClutterAnimationMode
 * @duration: 一次动画持续时间
 * @repeat_count: 动画重复次数。-1 无限重复。0 不重复。
 * @callback: 动画结束回调函数, 可以为 %NULL。
 * @data: @callback 参数
 * @name: 动画名称
 * @first_property_name: @actor 的属性，不能为 %NULL
 *
 * 例如：
 *     cdos_transition_to (actor,
 *                         CLUTTER_EASE_OUT_QUAD,
 *                         WINDOW_ANIMATION_TIME,
 *                         0,
 *                         NULL, NULL,
 *                         "switch-workspace",
 *                         "x", TRANS_PROP_FLOAT, 0.f,
 *                         "y", TRANS_PROP_FLOAT, 0.f,
 *                         NULL);
 *
 * Returns: #ClutterTransition 对象
 */
ClutterTransition *
cdos_transition_to (ClutterActor *actor, 
                     ClutterAnimationMode mode,
                     guint duration,
                     guint repeat_count,
                     TransCallback callback,
                     void *data,
                     const char *name,
                     const char *first_property_name,
                     ...)
{
    /* 可变参数格式:
     * prop_name, (x, y, signal::new-frame, ...)
     * prop_type, (TRANS_PROP_FLOAT, ...)
     * to value
     */
    g_return_val_if_fail (CLUTTER_IS_ACTOR (actor), NULL);
    g_return_val_if_fail (first_property_name != NULL, NULL);

    va_list args;

    clutter_actor_save_easing_state (actor);

    ClutterTransition *trans = clutter_transition_group_new ();
    const char *prop_name = first_property_name;
    va_start (args, first_property_name);
    while (1)
    {
        if (g_str_has_prefix (prop_name, "signal::"))
        {
            GCallback callback = va_arg (args, GCallback);
            gpointer user_data = va_arg (args, gpointer);
            g_signal_connect (trans, prop_name + 8, callback, user_data);
        }
        else {
            gint value_type = va_arg (args, gint);
            if (value_type < TRANS_PROP_CHAR || value_type > TRANS_PROP_DOUBLE)
            {
                break;
            }
            ClutterTransition *trans_p = NULL;
            trans_p = clutter_property_transition_new (prop_name);
            switch (value_type)
            {
                case TRANS_PROP_CHAR:
                    {
                        gchar p = va_arg (args, gint);
                        clutter_transition_set_to (trans_p, G_TYPE_CHAR, p);
                        break;
                    }
                case TRANS_PROP_UCHAR:
                    {
                        guchar p = va_arg (args, guint);
                        clutter_transition_set_to (trans_p, G_TYPE_UCHAR, p);
                        break;
                    }
                case TRANS_PROP_INT:
                    {
                        gint p = va_arg (args, gint);
                        clutter_transition_set_to (trans_p, G_TYPE_INT, p);
                        break;
                    }
                case TRANS_PROP_UINT:
                    {
                        guint p = va_arg (args, guint);
                        clutter_transition_set_to (trans_p, G_TYPE_UINT, p);
                        break;
                    }
                case TRANS_PROP_FLOAT:
                    {
                        gfloat p = va_arg (args, double);
                        clutter_transition_set_to (trans_p, G_TYPE_FLOAT, p);
                        break;
                    }
                case TRANS_PROP_DOUBLE:
                    {
                        gdouble p = va_arg (args, double);
                        clutter_transition_set_to (trans_p, G_TYPE_DOUBLE, p);
                        break;
                    }
            }
            clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_p);
        }
        prop_name = va_arg (args, char *);
        if (prop_name == NULL)
            break;
    }

    va_end (args);

    clutter_actor_set_easing_mode (actor, mode);
    clutter_timeline_set_duration (CLUTTER_TIMELINE (trans), duration);

    clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (trans), repeat_count);

    clutter_actor_add_transition (actor, name, trans);
    if (callback)
    {
        g_signal_connect (trans, "completed", 
                G_CALLBACK (callback), data);
    }
    else {
        DefaultTransData *data = g_new0 (DefaultTransData, 1);
        data->actor = actor;
        data->name = name;
        g_signal_connect (trans, "completed", G_CALLBACK (default_transition_complete_callback), data);
        g_signal_connect (trans, "stopped", G_CALLBACK (default_transition_stop_callback), data);
    }
    clutter_actor_restore_easing_state (actor);
    return trans;
}


/* Switch Workspace */

struct TransData {
    MetaWindowActor *actor;
    gboolean visible;
    ClutterPoint orig;
};

/* 是否正在切换工作区 */

static GMutex mutex;
static gint animation_count = 0;

static void
kill_window_effects (CdosBaseWM *cdos_wm, MetaWindowActor *actor)
{
    clutter_actor_remove_all_transitions ((ClutterActor *)actor);
}

#define TRANSITION_SWITCH_WKS "switch-workspace"

static void
kill_switch_workspace (CdosBaseWM *cdos_wm, CdosWM *wm)
{
    GList *windows = cdos_global_get_window_actors (global);
    GList *iter;
    for (iter = windows; iter; iter = iter->next)
    {
        MetaWindowActor *actor = iter->data;
        ClutterTransition *trans = clutter_actor_get_transition ((ClutterActor *)actor, TRANSITION_SWITCH_WKS);
        if (trans)
            clutter_actor_remove_transition ((ClutterActor *)actor, TRANSITION_SWITCH_WKS);
    }
}

static void 
on_switch_wks_complete (ClutterTimeline *tml, struct TransData *data)
{
    clutter_actor_remove_transition ((ClutterActor *)data->actor, TRANSITION_SWITCH_WKS);
    clutter_actor_set_position ((ClutterActor *)data->actor,
            data->orig.x, data->orig.y);
    MetaWindow *window = meta_window_actor_get_meta_window (data->actor);
    MetaRectangle rect;
    meta_window_get_buffer_rect (window, &rect);
    clutter_actor_set_position ((ClutterActor *)data->actor, rect.x, rect.y);
    if (data->visible == 0)
        clutter_actor_hide ((ClutterActor *)data->actor);

    g_free (data);

    g_mutex_lock (&mutex);
    animation_count --;
    g_mutex_unlock (&mutex);

    if (animation_count == 0)
    {
        CdosBaseWM *wm = cdos_global_get_window_manager (global);
        cdos_base_wm_completed_switch_workspace (wm);
        g_mutex_clear (&mutex);
    }
}

static void
on_switch_wks_stop (ClutterTimeline *tml, gboolean is_finished, struct TransData *data)
{
    if (!is_finished)
    {
        on_switch_wks_complete (tml, data);
    }
}

static void
switch_workspace (gpointer cdos_wm,
                  gint from,
                  gint to,
                  gint direction,
                  CdosWM *wm)
{
    cdos_debug ("from=%d, to=%d, direction=%d\n", from, to, direction);
    /* FIXME
     * 切换工作区动画可能会导致窗口不能正常最大/最小化等bug
     * 如果 expo、overview 正在显示，不显示动画
     */
    //if (!enable_desktop_effects ())
    {
        cdos_base_wm_completed_switch_workspace (CDOS_BASE_WM (cdos_wm));    
        return;
    }
    GList *windows = cdos_global_get_window_actors (global);
    gint width, height;
    cdos_global_get_screen_size (global, &width, &height);

    /* @direction 是相机移动方向。
     * 窗口 actor 应该反向移动
     */
    gint x_dest = 0, y_dest = 0;
    if (direction == META_MOTION_UP ||
            direction == META_MOTION_UP_LEFT ||
            direction == META_MOTION_UP_RIGHT)
        y_dest = height;
    else if (direction == META_MOTION_DOWN ||
            direction == META_MOTION_DOWN_LEFT ||
            direction == META_MOTION_DOWN_RIGHT)
        y_dest = -height;

    if (direction == META_MOTION_LEFT ||
            direction == META_MOTION_UP_LEFT ||
            direction == META_MOTION_DOWN_LEFT)
        x_dest = width;
    else if (direction == META_MOTION_RIGHT ||
            direction == META_MOTION_UP_RIGHT ||
            direction == META_MOTION_DOWN_RIGHT)
        x_dest = -width;

    GList *iter = NULL;
    MetaDisplay *display = cdos_global_get_display (global);
    MetaScreen *screen = cdos_global_get_screen (global);

    g_mutex_init (&mutex);

    for (iter = windows; iter != NULL; iter = iter->next)
    {
        MetaWindowActor *actor = iter->data;
        MetaWindow *window = meta_window_actor_get_meta_window (actor);

        if (!meta_window_showing_on_its_workspace (window))
            continue;
        if ((window == wm->moving_window) ||
               (meta_display_get_grab_op (display) == META_GRAB_OP_MOVING ||
                meta_display_get_grab_op (display) == META_GRAB_OP_KEYBOARD_MOVING)
               && (window == meta_display_get_focus_window (display)))
        {
            /* clutter_actor_show_all () */
            clutter_actor_show (CLUTTER_ACTOR (actor));
            wm->moving_window = NULL;
        }
        else if (meta_screen_get_active_workspace_index (screen) == from)
        {
            animation_count++;
            struct TransData *data = g_new0 (struct TransData, 1);
            data->actor = actor;
            data->visible = 0;
            clutter_actor_get_position ((ClutterActor *)actor,
                    &data->orig.x, &data->orig.y);
            cdos_transition_to ((ClutterActor *)actor,
                                CLUTTER_EASE_OUT_QUAD,
                                WINDOW_ANIMATION_TIME,
                                0,
                                (TransCallback)on_switch_wks_complete, data,
                                TRANSITION_SWITCH_WKS,
                                "x", TRANS_PROP_FLOAT, data->orig.x + x_dest,
                                "y", TRANS_PROP_FLOAT, data->orig.y + y_dest,
                                "signal::stopped", on_switch_wks_stop, data,
                                NULL);
        }
        else if (meta_screen_get_active_workspace_index (screen) == to)
        {
            animation_count++;
            struct TransData *data = g_new0 (struct TransData, 1);
            data->actor = actor;
            data->visible = 1;
            clutter_actor_get_position ((ClutterActor *)actor,
                    &data->orig.x, &data->orig.y);
            clutter_actor_set_position ((ClutterActor *)actor,
                    data->orig.x - x_dest, data->orig.y - y_dest);
            cdos_transition_to ((ClutterActor *)actor,
                                CLUTTER_EASE_OUT_QUAD,
                                WINDOW_ANIMATION_TIME,
                                0,
                                (TransCallback)on_switch_wks_complete, data,
                                TRANSITION_SWITCH_WKS,
                                "x", TRANS_PROP_FLOAT, data->orig.x,
                                "y", TRANS_PROP_FLOAT, data->orig.y,
                                "signal::stopped", on_switch_wks_stop, data,
                                NULL);
            clutter_actor_show ((ClutterActor *)actor);
        }

    }
}

typedef struct _WindowData {
    CdosBaseWM *wm;
    ClutterActor *actor;
    gdouble orig_scale_x;
    gdouble orig_scale_y;
    gfloat  orig_x;
    gfloat  orig_y;
    guint8  orig_opacity;
}WindowData;

typedef struct _EffectData {
#define MINIMIZING 1
#define MAPPING 2
#define DESTROYING 4
    guint state;
    guint elapsed_time;
    gdouble scale_x;
    gdouble scale_y;
    gfloat x;
    gfloat y;
}EffectData;

static EffectData *
get_effect_data (MetaWindowActor *actor)
{
#define EFFECT_DATA "effect-data"
    EffectData *data = NULL;
    data = g_object_get_data ((GObject *)actor, EFFECT_DATA);

    if (!data)
    {
        data = g_new0 (EffectData, 1);
        g_object_set_data_full ((GObject *)actor, EFFECT_DATA, data, g_free);
    }
    return data;
}

static void
on_minimize_effect_complete (ClutterTimeline *tml, WindowData *data)
{
    ClutterActor *actor = data->actor;

    clutter_actor_remove_transition (actor, "minimize-window-trans");
    clutter_actor_set_scale (actor, 1.0f, 1.0f);
    clutter_actor_set_pivot_point (actor, 0.0f, 0.0f);
    clutter_actor_set_opacity (actor, data->orig_opacity);

    cdos_base_wm_completed_minimize (data->wm, META_WINDOW_ACTOR (actor));
    
    /*
    gfloat x, y;
    clutter_actor_get_transformed_position (actor, &x, &y);
    cdos_debug ("%f %f", x, y);
    clutter_actor_get_position (actor, &x, &y);
    cdos_debug ("%f %f", x, y);
    */

    EffectData *effect = get_effect_data ((MetaWindowActor *)data->actor);
    effect->state = 0;

    g_free (data);
}

static void 
on_minimize_effect_stop (ClutterTimeline *tml, gboolean is_finished, WindowData *data)
{
    if (!is_finished)
    {
        EffectData *effect = get_effect_data ((MetaWindowActor *)data->actor);
        effect->state = MINIMIZING;
        effect->elapsed_time = clutter_timeline_get_elapsed_time (tml);
        clutter_actor_get_scale (data->actor, &effect->scale_x, &effect->scale_y);
        clutter_actor_get_position (data->actor, &effect->x, &effect->y);
        on_minimize_effect_complete (tml, data);
    }
}

static void
minimize_window (CdosBaseWM *cdos_wm,
                 gpointer actor)
{
    cdos_sound_manager_play ("minimize");
    MetaWindow *window = meta_window_actor_get_meta_window (actor);

    Window xid = meta_window_get_xwindow (window);
    printf("%s xwindow: 0x%lx\n", __func__, xid);

    MetaWindowType type = meta_window_get_window_type (window);

    if (!enable_desktop_effects ())
    {
        cdos_base_wm_completed_minimize (cdos_wm, actor);
        return;
    }
    clutter_actor_show (actor);

    WindowData *data = g_new0 (WindowData, 1);
    data->actor = actor;
    data->wm = cdos_wm;
    clutter_actor_get_scale (actor, &data->orig_scale_x, &data->orig_scale_y);
    data->orig_opacity = clutter_actor_get_opacity (actor);

    g_object_set_data ((GObject *)actor, "prev-state", GINT_TO_POINTER(TRUE));

    MetaRectangle rect;
    ClutterTransition *trans = NULL;
    if (meta_window_get_icon_geometry (window, &rect))
    {
        clutter_actor_set_pivot_point (actor, 0.0f, 0.0f);

        EffectData *effect = get_effect_data (actor);
        if (effect->state == MAPPING)
        {
            clutter_actor_set_scale (actor, effect->scale_x, effect->scale_y);
            clutter_actor_set_position (actor, effect->x, effect->y);
            effect->state = MINIMIZING;
        }
        trans = cdos_transition_to (actor,
                CLUTTER_EASE_OUT_QUAD,
                WINDOW_ANIMATION_TIME, 0,
                (TransCallback)on_minimize_effect_complete, data,
                "minimize-window-trans",
                "scale-x", TRANS_PROP_FLOAT, 0.0f,
                "scale-y", TRANS_PROP_FLOAT, 0.0f,
                "opacity", TRANS_PROP_UCHAR, 0,
                "x", TRANS_PROP_FLOAT, (gfloat)(rect.x + rect.width / 2.0),
                "y", TRANS_PROP_FLOAT, (gfloat)rect.y,
                "signal::stopped", on_minimize_effect_stop, data,
                NULL);
    }

    else {
        clutter_actor_set_pivot_point (actor, 0.5f, 0.5f);
        trans = cdos_transition_to (actor,
                CLUTTER_EASE_OUT_QUAD,
                WINDOW_ANIMATION_TIME, 0,
                (TransCallback)on_minimize_effect_complete, data,
                "minimize-window-trans",
                "scale-x", TRANS_PROP_FLOAT, 0.0f,
                "scale-y", TRANS_PROP_FLOAT, 0.0f,
                /*"z-position", TRANS_PROP_FLOAT, -3000.f,*/
                /*"rotation-angle-z", TRANS_PROP_DOUBLE, 70.0f,*/
                "opacity", TRANS_PROP_UCHAR, 0,
                "signal::stopped", on_minimize_effect_stop, data,
                NULL);
    }
    if (trans == NULL)
        cdos_base_wm_completed_minimize (cdos_wm, actor);
}

static void
on_maximize_effect_complete (ClutterTimeline *tml, WindowData *data)
{
    ClutterActor *actor = data->actor;
    clutter_actor_remove_transition (actor, "maximize-window-trans");
    clutter_actor_set_scale (actor, 1.0, 1.0);

    cdos_base_wm_completed_unminimize (data->wm, META_WINDOW_ACTOR (actor));

    g_free (data);
}

static void
size_change_window (CdosBaseWM *cdos_wm,
                    ClutterActor *actor,
                    MetaSizeChange which_change,
                    MetaRectangle *old_frame_rect,
                    MetaRectangle *old_buffer_rect)
{
    MetaWindow *window = meta_window_actor_get_meta_window ((MetaWindowActor *)actor);

    MetaMaximizeFlags max_flags =  meta_window_get_maximized (window);

    if (max_flags == META_MAXIMIZE_BOTH) {
        cdos_sound_manager_play ("maximize");        
    } else {
        cdos_sound_manager_play ("unmaximize");
    }

    MetaWindowType type = meta_window_get_window_type (window);

    //if (type != META_WINDOW_NORMAL || !enable_desktop_effects ())
    {
        cdos_base_wm_completed_size_change (cdos_wm, (MetaWindowActor *)actor);
        return;
    }
}

static void
unminimize_window (CdosBaseWM *cdos_wm,
                   MetaWindowActor *actor)
{
    cdos_sound_manager_play ("unminimize");
    MetaWindow *window = meta_window_actor_get_meta_window (actor);

    cdos_base_wm_completed_unminimize (cdos_wm, actor);
}

static void
tile_window (gpointer cdos_wm,
             gpointer actor,
             gint     x,
             gint     y,
             gint     width,
             gint     height)
{
    cdos_sound_manager_play ("tile");
    MetaWindow *window = meta_window_actor_get_meta_window (META_WINDOW_ACTOR (actor));

    Window xid = meta_window_get_xwindow (window);
    printf("%s xwindow: 0x%lx\n", __func__, xid);

    //cdos_base_wm_completed_tile (CDOS_BASE_WM (cdos_wm), META_WINDOW_ACTOR (actor));
} 

static void
show_tile_preview (gpointer cdos_wm,
                   MetaWindow  *window,
                   MetaRectangle *rectangle,
                   gint     a,
                   guint    b)
{

}

static void G_GNUC_UNUSED
hide_tile_preview (gpointer cdos_wm)
{

}

static void G_GNUC_UNUSED
show_hud_preview (gpointer cdos_wm,
                  gint a,
                  MetaRectangle *rectangle,
                  guint b)
{

}

static void G_GNUC_UNUSED
hide_hud_preview (gpointer cdos_wm) 
{

}

static void
on_map_effect_complete (/*ClutterAnimation *anim*/ ClutterTimeline *tml, ClutterActor *actor)
{
    clutter_actor_remove_transition (actor, "map-window-scale-trans");
    
    MetaWindowActor *window_actor = META_WINDOW_ACTOR (actor);
   
    clutter_actor_set_scale (actor, 1.0f, 1.0f);
    //clutter_actor_set_pivot_point (actor, 0.f, 0.f);
    cdos_base_wm_completed_map (cdos_global_get_window_manager (global), window_actor);

    EffectData *effect = get_effect_data (window_actor);
    effect->state = 0;
}

static void 
on_map_effect_stop (ClutterTimeline *tml, gboolean is_finished, ClutterActor *actor)
{
    if (!is_finished)
    {
        EffectData *effect = get_effect_data ((MetaWindowActor *)actor);
        effect->state = MAPPING;
        effect->elapsed_time = clutter_timeline_get_elapsed_time (tml);
        clutter_actor_get_scale (actor, &effect->scale_x, &effect->scale_y);
        clutter_actor_get_position (actor, &effect->x, &effect->y);
        on_map_effect_complete (tml, actor);
    }
}

static void
on_dialog_closed (GPid     pid,
                  gint     status,
                  gpointer user_data)
{
  MetaPlugin *plugin = user_data;
  gboolean ok;

  ok = g_spawn_check_exit_status (status, NULL);
  meta_plugin_complete_display_change (plugin, ok);
}

static gboolean
xunli_confirm_display_change (CdosBaseWM *cdos_wm){
  GPid pid;  
 //将倒计时对话框 改为永久不会自动消失
  MetaPlugin *plugin = cdos_base_wm_get_plugin (cdos_wm);
  pid = meta_show_dialog ("--question",
                          _("Does the display look OK?"),
                          "-1",
                          NULL,
                          _("_Keep This Configuration"),
                          _("_Restore Previous Configuration"),
                          "preferences-desktop-display",
                          0,
                          NULL, NULL);

  g_child_watch_add (pid, on_dialog_closed, plugin);	

}

/*
 * 在特殊模式下过滤掉其他所有快捷键，除了 overlay-key
 * overlay-key 的过滤在 menu applet 中实现。
 * 补充： overlay-key 就是 super-L 打开隐藏菜单的键
 * */
static gboolean
xunli_filter_keybinding (CdosBaseWM *cdos_wm, MetaKeyBinding *binding, gpointer data){
	//g_print ("key name is : %s\n", meta_key_binding_get_name (binding));
	gchar *screenshot = meta_key_binding_get_name (binding);
	if (g_str_has_prefix (screenshot, "settings-daemon")){
		//g_print ("hit\n");
		// 任何情况下都不控制截屏键
		return FALSE;
	}
	
	if (cdos_get_action_mode () == CDOS_ACTION_MODE_NONE){
	    if ( g_strcmp0 (meta_key_binding_get_name (binding) , "overlay-key") != 0 )
	    	return TRUE;
	    else 
		return FALSE;
	}
	if (cdos_get_action_mode () == CDOS_ACTION_MODE_NORMAL && meta_key_binding_is_builtin (binding)){
	    return FALSE;
	}
	if (cdos_get_action_mode () == CDOS_ACTION_MODE_NORMAL && meta_key_binding_is_customized (binding)){
	    return FALSE;
	}
	return TRUE;
}

static void
map_window (CdosBaseWM *cdos_wm, gpointer actor)
{
    MetaWindowActor *window_actor = META_WINDOW_ACTOR (actor);
    MetaWindow *window = meta_window_actor_get_meta_window(window_actor);

    Window xid = meta_window_get_xwindow (window);
    /*
    Display *dpy = meta_display_get_xdisplay (meta_window_get_display (window));
    XWindowAttributes attr;
    Status s G_GNUC_UNUSED = XGetWindowAttributes (dpy, xid, &attr);
    */
    //printf("%s xwindow: 0x%lx (%s)\n", 
    //        __func__, xid, meta_window_get_wm_class (window));

    MetaWindowType type = meta_window_get_window_type(window);
    ClutterActor *clutter_actor = CLUTTER_ACTOR(actor);

    if (!enable_desktop_effects ())
    {
        cdos_base_wm_completed_map (cdos_wm, window_actor);
        return;
    }
    /*
    ClutterAnimation *anim = NULL;
    if ((type == META_WINDOW_NORMAL) || (type == META_WINDOW_DIALOG))
    {
        clutter_actor_set_scale (clutter_actor, 0.0, 0.0);
        clutter_actor_show (actor);

        clutter_actor_move_anchor_point_from_gravity (clutter_actor,
                                                      CLUTTER_GRAVITY_CENTER);
        anim = clutter_actor_animate (clutter_actor,
                                        CLUTTER_EASE_IN_SINE,
                                        MAP_TIME,
                                        "scale-x", 1.0,
                                        "scale-y", 1.0,
                                        NULL);
        
    }
    else if ((type == META_WINDOW_MENU) || (type == META_WINDOW_POPUP_MENU))
    {
        clutter_actor_set_scale (clutter_actor, 0.0, 1.0);
        clutter_actor_show (actor);

        anim = clutter_actor_animate (clutter_actor,
                                        CLUTTER_EASE_IN_SINE,
                                        MAP_TIME / 2.0,
                                        "scale-x", 1.0,
                                        "scale-y", 1.0,
                                        NULL);
    }
    if (anim)
    {
        g_signal_connect (anim, "completed",
                G_CALLBACK (on_map_effect_complete),
                clutter_actor);
    }
    else
        cdos_base_wm_completed_map (CDOS_BASE_WM (cdos_wm), META_WINDOW_ACTOR (actor));

    */
    ClutterTransition *trans = NULL;

    if ((type == META_WINDOW_NORMAL) || (type == META_WINDOW_DIALOG))
    {
        ClutterPoint dst;
        //clutter_actor_get_transformed_position (actor, &dst.x , &dst.y);
        MetaRectangle input_rect;
        meta_window_get_buffer_rect (window, &input_rect);
        dst.x = input_rect.x;
        dst.y = input_rect.y;
        clutter_actor_set_scale (actor, 0.0, 0.0);    
        clutter_actor_show (actor);

        MetaRectangle rect;
        /* previous window state is minimized ? */
        gboolean prev_state = FALSE;
        prev_state = GPOINTER_TO_INT(g_object_get_data ((GObject *)actor, "prev-state"));

        if (!meta_window_is_attached_dialog (window) && meta_window_get_icon_geometry (window, &rect) && prev_state)
        {
            //clutter_actor_set_position (actor, (gfloat)(rect.x + rect.width/2), (gfloat)rect.y);
            clutter_actor_set_pivot_point (actor, 0.0f, 0.0f);

            clutter_actor_save_easing_state (actor);

            trans = clutter_transition_group_new ();

            ClutterTransition *trans_x = clutter_property_transition_new ("x");
            ClutterTransition *trans_y = clutter_property_transition_new ("y");
            ClutterTransition *trans_sx = clutter_property_transition_new ("scale-x");
            ClutterTransition *trans_sy = clutter_property_transition_new ("scale-y");

            clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_x);
            clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_y);
            clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_sx);
            clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_sy);

            clutter_actor_set_easing_mode (actor, CLUTTER_EASE_IN_SINE);

            EffectData *effect = get_effect_data (window_actor);
            if (effect->state == MINIMIZING)
            {
                clutter_timeline_set_duration ((ClutterTimeline *)trans, effect->elapsed_time);
                clutter_transition_set_from (trans_x, G_TYPE_FLOAT, effect->x);
                clutter_transition_set_from (trans_y, G_TYPE_FLOAT, effect->y);
                clutter_transition_set_from (trans_sx, G_TYPE_DOUBLE, effect->scale_x);
                clutter_transition_set_from (trans_sy, G_TYPE_DOUBLE, effect->scale_y);
                effect->state = MAPPING;
            }else {
                clutter_timeline_set_duration (CLUTTER_TIMELINE (trans), MAP_TIME);
                clutter_transition_set_from (trans_x, G_TYPE_FLOAT, (gfloat)(rect.x + rect.width/2));
                clutter_transition_set_from (trans_y, G_TYPE_FLOAT, (gfloat)rect.y);
                clutter_transition_set_from (trans_sx, G_TYPE_DOUBLE, 0.f);
                clutter_transition_set_from (trans_sy, G_TYPE_DOUBLE, 0.f);
            }

            clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (trans), 0);

            clutter_transition_set_to (trans_x, G_TYPE_FLOAT, dst.x);

            clutter_transition_set_to (trans_y, G_TYPE_FLOAT, dst.y);

            clutter_transition_set_to (trans_sx, G_TYPE_DOUBLE, 1.0f);

            clutter_transition_set_to (trans_sy, G_TYPE_DOUBLE, 1.0f);
        }
        else {
            cdos_sound_manager_play ("map");
            clutter_actor_set_pivot_point (actor, 0.5f, 0.5f);    
            clutter_actor_save_easing_state (clutter_actor);

            trans = clutter_transition_group_new ();

            ClutterTransition *trans_x = clutter_property_transition_new ("scale-x");
            ClutterTransition *trans_y = clutter_property_transition_new ("scale-y");

            clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_x);
            clutter_transition_group_add_transition ((ClutterTransitionGroup *)trans, trans_y);

            clutter_actor_set_easing_mode (clutter_actor, CLUTTER_EASE_OUT_QUAD);
            clutter_timeline_set_duration (CLUTTER_TIMELINE (trans), MAP_TIME);

            /**
             * 动画重复次数：
             * -1: 无限循环，直到 clutter_actor_remove_transition() 或 destroy/hide actor
             *  0: 只播放一次
             *  n > 0: 重复 n 次
             */
            clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (trans), 0);

            clutter_transition_set_from (trans_x, G_TYPE_FLOAT, 0.f);
            clutter_transition_set_to (trans_x, G_TYPE_FLOAT, 1.0f);

            clutter_transition_set_from (trans_y, G_TYPE_FLOAT, 0.f);
            clutter_transition_set_to (trans_y, G_TYPE_FLOAT, 1.0f);
        }
    }
    else if (/* other window type */ 0)
    {

    }

    if (trans != NULL)
    {
        /* 每个动画具有唯一字符串 identifier */
        clutter_actor_add_transition (clutter_actor, "map-window-scale-trans", trans);
        /* ClutterTransition 继承自 ClutterTimeline, 设置timeline 的 completed 信号 */
        g_signal_connect (trans, "completed", 
                G_CALLBACK (on_map_effect_complete), clutter_actor);
        g_signal_connect (trans, "stopped", 
                G_CALLBACK (on_map_effect_stop), clutter_actor);
        clutter_actor_restore_easing_state (clutter_actor);
    }
    else
        cdos_base_wm_completed_map (cdos_wm, window_actor); 
}

static void
on_destroy_effect_complete (/*ClutterAnimation *anim*/ ClutterTimeline *tml, ClutterActor *actor)
{
    clutter_actor_remove_transition (actor, "unmap-window-scale-trans");
    
    MetaWindowActor *window_actor = META_WINDOW_ACTOR (actor);
    cdos_base_wm_completed_destroy (cdos_global_get_window_manager (global), window_actor);
}

static void
destroy_window (CdosBaseWM *cdos_wm, ClutterActor *actor)
{
    cdos_sound_manager_play ("close");
    MetaWindowActor *window_actor = META_WINDOW_ACTOR (actor);
    MetaWindow *window = meta_window_actor_get_meta_window(window_actor);

    MetaWindowType type = meta_window_get_window_type(window);
    if (type != META_WINDOW_NORMAL || !enable_desktop_effects ())
    {
        cdos_base_wm_completed_destroy (cdos_wm, window_actor);
        return;
    }
    clutter_actor_show (actor);
    clutter_actor_set_pivot_point (actor, 0.5f, 0.5f);
    ClutterTransition *trans = cdos_transition_to (actor, CLUTTER_EASE_IN_OUT_SINE,
            MAP_TIME, 0,
            (TransCallback)on_destroy_effect_complete, actor,
            "unmap-window-scale-trans",
            "scale-x", TRANS_PROP_FLOAT, 0.0f,
            "scale-y", TRANS_PROP_FLOAT, 0.0f,
            /*"z-position", TRANS_PROP_FLOAT, -3000.f,*/
            /*"rotation-angle-z", TRANS_PROP_DOUBLE, 70.0f,*/
            "opacity", TRANS_PROP_UCHAR, 0,
            NULL);
    if (trans == NULL)
        cdos_base_wm_completed_destroy (CDOS_BASE_WM (cdos_wm), window_actor);
}


static void 
shift_window_to_workspace (CdosWM *wm,
                           MetaWindow *window,
                           MetaMotionDirection direction)
{
    if (meta_window_get_window_type (window) == META_WINDOW_DESKTOP)
        return;
    wm->moving_window = window;
    MetaScreen *screen = cdos_global_get_screen (global);
    MetaWorkspace *active = meta_screen_get_active_workspace (screen);
    MetaWorkspace *wks = meta_workspace_get_neighbor (active, direction);
    if (wks != active)
    {
        meta_window_change_workspace (window, wks);
        meta_workspace_activate_with_focus (wks, window, cdos_global_get_current_time (global));
        show_workspace_OSD ( screen );
    }
}
/*
 * typedef void (* MetaKeyHandleFunc) (MetaDisplay *display,
 *                                     MetaScreen  *screen,
 *                                     MetaWindow  *window,
 *                                     XEvent      *event,
 *                                     MetaKeyBinding *binding,
 *                                     gpointer    user_data);
 */
static void
move_window_to_workspace_left (MetaDisplay    *display,
                               MetaScreen     *screen,
                               MetaWindow     *window,
                               XEvent         *event,
                               MetaKeyBinding *binding,
                               gpointer        data)
{
    CdosWM *wm = data;
    shift_window_to_workspace (wm, window, META_MOTION_LEFT);
}

static void
move_window_to_workspace_right (MetaDisplay    *display,
                                MetaScreen     *screen,
                                MetaWindow     *window,
                                XEvent         *event,
                                MetaKeyBinding *binding,
                                gpointer        data)
{
    CdosWM *wm = data;
    shift_window_to_workspace (wm, window, META_MOTION_RIGHT);
}

/**
 * cdos_wm_action_move_workspace_left:
 *
 * 切换到左侧工作区
 *
 */
void 
cdos_wm_action_move_workspace_left ()
{
    MetaScreen *screen = cdos_global_get_screen (global);
    MetaWorkspace *active = meta_screen_get_active_workspace (screen);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, 
            META_MOTION_LEFT);
    if (active != neighbour)
    {
        cdos_sound_manager_play ("switch");
        meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
    }
}

/**
 * cdos_wm_action_move_workspace_right:
 *
 * 切换到右侧工作区
 *
 */
void 
cdos_wm_action_move_workspace_right ()
{
    MetaScreen *screen = cdos_global_get_screen (global);
    MetaWorkspace *active = meta_screen_get_active_workspace (screen);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, 
            META_MOTION_RIGHT);
    if (active != neighbour)
    {
        cdos_sound_manager_play ("switch");
        meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
    }
}

/**
 * cdos_wm_action_move_workspace_up
 *
 * 切换到上一个工作区
 */
void 
cdos_wm_action_move_workspace_up ()
{
    MetaScreen *screen = cdos_global_get_screen (global);
    MetaWorkspace *active = meta_screen_get_active_workspace (screen);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, 
            META_MOTION_UP);

    meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
}

/**
 * cdos_wm_action_move_workspace_down:
 *
 * 切换到下一个工作区
 */
void 
cdos_wm_action_move_workspace_down ()
{
    MetaScreen *screen = cdos_global_get_screen (global);
    MetaWorkspace *active = meta_screen_get_active_workspace (screen);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, 
            META_MOTION_DOWN);

    meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
}

/**
 * cdos_wm_action_flip_workspace_left:
 *
 * 切到左侧工作区，并重置光标位置
 */
void 
cdos_wm_action_flip_workspace_left ()
{
    MetaScreen *screen = cdos_global_get_screen (global);
    MetaWorkspace *active = meta_screen_get_active_workspace (screen);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, 
            META_MOTION_LEFT);
    if (active == neighbour)
        return;
    meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
    gint x, y;
    ClutterModifierType mods;
    cdos_global_get_pointer (global, &x, &y, &mods);
    gint w, h;
    cdos_global_get_screen_size (global, &w, &h);
    cdos_global_set_pointer (global, w - 10, y);
}

/**
 * cdos_wm_action_flip_workspace_right:
 *
 * 切到右侧工作区，并重置光标位置
 */
void 
cdos_wm_action_flip_workspace_right ()
{
    MetaScreen *screen = cdos_global_get_screen (global);
    MetaWorkspace *active = meta_screen_get_active_workspace (screen);
    MetaWorkspace *neighbour = meta_workspace_get_neighbor (active, 
            META_MOTION_RIGHT);
    if (active == neighbour)
        return;
    meta_workspace_activate (neighbour, cdos_global_get_current_time (global));
    gint x, y;
    ClutterModifierType mods;
    cdos_global_get_pointer (global, &x, &y, &mods);
    gint w, h;
    cdos_global_get_screen_size (global, &w, &h);
    cdos_global_set_pointer (global, 10, y);
}

// add by liting 
static void
hide_snap_OSD () {
    if (snap_osd != NULL){
        clutter_actor_hide (CLUTTER_ACTOR (snap_osd));
        CdosLayoutManager *layout = cdos_monkey_plugin_get_layout_manager (monkey_plugin_get ());
        cdos_layout_manager_remove_chrome (layout, snap_osd);
        clutter_actor_destroy (snap_osd);
        snap_osd = NULL;
    }
}

static void
hide_workspace_OSD (ClutterTimeline *tml, void *data) {
    if (workspace_osd != NULL){
        clutter_actor_hide (CLUTTER_ACTOR (workspace_osd));
        CdosLayoutManager *layout = cdos_monkey_plugin_get_layout_manager (monkey_plugin_get ());
        clutter_actor_destroy (workspace_osd);
        workspace_osd = NULL;
    }
}

void
show_workspace_OSD (MetaScreen *screen) {
    hide_snap_OSD ();
    hide_workspace_OSD (NULL, NULL);
    GSettings *settings = cdos_global_get_settings (global);
    gboolean visible = g_settings_get_boolean (settings, "workspace-osd-visible");
    CdosLayoutManager *layout = cdos_monkey_plugin_get_layout_manager (monkey_plugin_get ());
    if (visible){
        guint current_workspace_index = meta_screen_get_active_workspace_index (screen);

        MetaRectangle *monitor = cdos_layout_manager_get_primary_monitor (layout);
        if (workspace_osd == NULL){
            workspace_osd = g_object_new (ST_TYPE_LABEL, "style-class", "workspace-osd", NULL);
        }
        st_label_set_text ((StLabel *)workspace_osd, cdos_get_workspace_name (current_workspace_index));
        clutter_actor_set_opacity (CLUTTER_ACTOR (workspace_osd), 0);
        cdos_layout_manager_add_chrome (layout, workspace_osd, 0);

        guint workspace_osd_x = g_settings_get_int (settings, "workspace-osd-x");
        guint workspace_osd_y = g_settings_get_int (settings, "workspace-osd-y");

        guint minX = 5;
        guint maxX = 95;
        guint minY = 5;
        guint maxY = 95;

        gfloat label_width = clutter_actor_get_width (workspace_osd);
        gfloat label_height = clutter_actor_get_height (workspace_osd);
        guint delta = (workspace_osd_x - minX) / (maxX - minX);
        guint x = round ((monitor->width *workspace_osd_x / 100) - (label_width * delta)); 

        delta = (workspace_osd_y - minY) / (maxY - minY);

        guint y = round ((monitor->height * workspace_osd_y / 100) - (label_height * delta));

        clutter_actor_set_position (workspace_osd, x, y);
        // 开始动画
        guint duration = g_settings_get_int (settings, "workspace-osd-duration");

        clutter_actor_save_easing_state (workspace_osd);
        ClutterTransition *transition = clutter_property_transition_new ("opacity");
        clutter_actor_set_easing_mode (workspace_osd, CLUTTER_EASE_IN_SINE);
        clutter_timeline_set_duration (CLUTTER_TIMELINE (transition), duration);
        clutter_timeline_set_repeat_count (CLUTTER_TIMELINE (transition), 0);
        clutter_transition_set_from (transition, G_TYPE_UCHAR, 0);
        clutter_transition_set_to (transition, G_TYPE_UCHAR, 255);
        clutter_actor_add_transition (workspace_osd, "show-workspace-osd", transition);
        g_signal_connect (transition, "completed", G_CALLBACK (hide_workspace_OSD), screen);
        clutter_actor_restore_easing_state (workspace_osd);
    }

}


/* 切换工作区时，先执行 show_workspace_switcher(),
 * 切换完成后，执行 switch_workspace ();
 */
static void
show_workspace_switcher (MetaDisplay    *display,
                         MetaScreen     *screen,
                         MetaWindow     *window,
                         XEvent         *event,
                         MetaKeyBinding *binding,
                         gpointer       user_data)
{
    const char *name = meta_key_binding_get_name (binding);
    if (name == NULL)
        return;
    if (strcmp (name, "switch-to-workspace-up") == 0)
    {
        CdosOverview * overview = cdos_overview_get ();
        cdos_overview_toggle (overview);        
        return;
    }
    if (strcmp (name, "switch-to-workspace-down") == 0)
    {
        CdosOverview * overview = cdos_overview_get ();
        cdos_overview_toggle (overview);
        return;
    }
    if (meta_screen_get_n_workspaces (screen) == 1)
        return;

    MetaScreen *scr = cdos_global_get_screen (global);
    gint current_wks_idx = meta_screen_get_active_workspace_index (scr);
    if (strcmp (name, "switch-to-workspace-left") == 0)
    {
        cdos_wm_action_move_workspace_left ();
        if (current_wks_idx != meta_screen_get_active_workspace_index (scr))
            show_workspace_OSD (scr);
    }
    else if (strcmp (name, "switch-to-workspace-right") == 0)
    {
        cdos_wm_action_move_workspace_right ();
        if (current_wks_idx != meta_screen_get_active_workspace_index (scr))
            show_workspace_OSD (scr);
    }
}

static void
start_appswitcher (MetaDisplay      *display,
                    MetaScreen      *screen,
                    MetaWindow      *window,
                    ClutterKeyEvent *event,
                    MetaKeyBinding  *binding,
                    gpointer         user_data)
{
    if (cdos_get_windows_for_binding (binding) == NULL)
        return;
    cdos_coverflow_switcher_new (binding);
}


static void
start_multi_switcher (MetaDisplay      *display,
                    MetaScreen      *screen,
                    MetaWindow      *window,
                    ClutterKeyEvent *event,
                    MetaKeyBinding  *binding,
                    gpointer         user_data)
{
    //g_print ("multi monitor switcher\n");
    //if (cdos_get_windows_for_binding (binding) == NULL)
        //return;
    
    cdos_util_spawn_command_line ("cdos-multimonitor-switcher");
}


/**
 * cdos_wm_new:
 * @cdos_wm: #CdosBaseWM
 *
 * 创建一个 #CdosWM, 管理窗口切换、映射、关闭和切换工作区等的动画及绑定快捷键。
 *
 * Returns: #CdosWM
 *
 */
CdosWM *
cdos_wm_new (CdosBaseWM *cdos_wm)
{
    CdosWM *wm = g_new0 (CdosWM, 1);
    wm->should_animate = 1;
    wm->cdos_wm = cdos_wm;

    should_animate ();

    g_signal_connect (cdos_wm, "kill-window-effects", G_CALLBACK (kill_window_effects), NULL);
    g_signal_connect (cdos_wm, "kill-switch-workspace", G_CALLBACK (kill_switch_workspace), wm);
    g_signal_connect (cdos_wm, "switch-workspace", G_CALLBACK (switch_workspace), wm);
    g_signal_connect (cdos_wm, "minimize", G_CALLBACK (minimize_window), NULL);
    g_signal_connect (cdos_wm, "unminimize", G_CALLBACK (unminimize_window), NULL);
    g_signal_connect (cdos_wm, "size-change", G_CALLBACK (size_change_window), NULL);
    g_signal_connect (cdos_wm, "map", G_CALLBACK (map_window), NULL);
    g_signal_connect (cdos_wm, "filter-keybinding", G_CALLBACK (xunli_filter_keybinding), NULL);
    g_signal_connect (cdos_wm, "confirm-display-change", G_CALLBACK (xunli_confirm_display_change), NULL);
    g_signal_connect (cdos_wm, "destroy", G_CALLBACK (destroy_window), NULL);

    meta_keybindings_set_custom_handler ("move-to-workspace-left", (MetaKeyHandlerFunc)move_window_to_workspace_left, wm, NULL);
    meta_keybindings_set_custom_handler ("move-to-workspace-right", (MetaKeyHandlerFunc)move_window_to_workspace_right, wm, NULL);
    meta_keybindings_set_custom_handler ("switch-to-workspace-left", (MetaKeyHandlerFunc)show_workspace_switcher, wm, NULL);
    meta_keybindings_set_custom_handler ("switch-to-workspace-right", (MetaKeyHandlerFunc)show_workspace_switcher, wm, NULL);
    meta_keybindings_set_custom_handler ("switch-to-workspace-up", (MetaKeyHandlerFunc)show_workspace_switcher, wm, NULL);
    meta_keybindings_set_custom_handler ("switch-to-workspace-down", (MetaKeyHandlerFunc)show_workspace_switcher, wm, NULL);

    meta_keybindings_set_custom_handler ("switch-applications", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-group", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-windows-backward", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("cycle-windows", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-group-backward", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-panels", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("switch-panels-backward", (MetaKeyHandlerFunc)start_appswitcher, NULL, NULL);
    meta_keybindings_set_custom_handler ("multimonitor-switcher", (MetaKeyHandlerFunc)start_multi_switcher, NULL, NULL);

    //wm->settings = g_settings_new ("org.cdos.mutter");

    return wm;
}
