#include <meta/window.h>
#include <meta/meta-cursor-tracker.h>
#include "cdos-window-overlay.h"
#include "cdos-window-clone.h"

#include "cdos-window-tracker.h"
#include "cdos-app.h"
#include "cdos-global.h"
#include "cdos-debug.h"

#include <math.h>

/**
 * SECTION: cdos-window-overlay
 * @Title: CdosWindowOverlay
 * @Short_description: 在 CdosWindowClone 上显示的 overlay
 *
 * 在 #CdosOverview 的 #CdosWindowClone 的窗口缩略图周围显示关闭按钮、图标及标题。
 */

#define WINDOWOVERLAY_ICON_SIZE    32
#define DEMANDS_ATTENTION_CLASS_NAME "window-list-item-demands-attention"

#define CLOSE_BUTTON_FADE_TIME 100


G_DEFINE_TYPE (CdosWindowOverlay, cdos_window_overlay, G_TYPE_OBJECT );

/* signal */
enum
{
    ACTIVATED,        //activated
    SHOW_CLOSE_BUTTON,     //show-close-button    
    LAST_SIGNAL
};

static guint cdos_window_overlay_signals [LAST_SIGNAL] = { 0 };

static void cdos_window_overlay_close_window (ClutterActor *button,
                                  ClutterEvent *e,
                                  CdosWindowOverlay *self);

static void cdos_window_overlay_disconnect_window_added ( CdosWindowOverlay *self );

static void cdos_window_overlay_disconnect_attention_signals ( CdosWindowOverlay *self );

static void cdos_window_overlay_refresh_title (CdosWindowOverlay *self, const gchar *titleText);

static void
onStyleChanged (StWidget *widget, gpointer user_data )
{
    CdosWindowOverlay *self = CDOS_WINDOW_OVERLAY(user_data);
    
    StThemeNode *titleNode = st_widget_get_theme_node ( ST_WIDGET(self->title) );
    StThemeNode *closeNode = st_widget_get_theme_node ( ST_WIDGET(self->closeButton) );

    self->title_spacing = st_theme_node_get_length ( titleNode, "-cdos-caption-spacing"); 
    self->closeButton_overlap = st_theme_node_get_length ( closeNode, "-cdos-close-overlap");

    clutter_actor_queue_relayout (self->parentActor);
}

static void
onNotifyTitle (MetaWindow *window, GParamSpec *spec, gpointer user_data )
{
    CdosWindowOverlay *self = CDOS_WINDOW_OVERLAY(user_data);    

    cdos_window_overlay_refresh_title ( self, meta_window_get_title (window));
}

static void
onPointerMotion (ClutterActor *actor, ClutterEvent *event, gpointer user_data )
{
    CdosWindowOverlay *self = (CdosWindowOverlay *)user_data;
    if (!cdos_pointer_tracker_has_moved (self->pointerTracker) )
        return;

    // We might get motion events on the clone while the overlay is
    // hidden, e.g. during animations, we ignore these events,
    // as the close button will be shown as needed when the overlays
    // are shown again
    if (self->hidden) 
        return;
    if (self->hovering) 
        return;
    self->hovering = TRUE;
    cdos_window_overlay_show_close_button (self);
}

static gboolean
idleToggleCloseButton (CdosWindowOverlay *self)
{
    self->idleToggleCloseId = 0ul;

    if (!clutter_actor_has_pointer ((ClutterActor *)self->windowClone) && !clutter_actor_has_pointer ((ClutterActor *)self->closeButton))
        clutter_actor_hide ((ClutterActor *)self->closeButton);

    return FALSE;
}

static void
idleHideCloseButton ( CdosWindowOverlay *self )
{
    //if ( self->idleToggleCloseId == 0ul )
    //    self->idleToggleCloseId = g_timeout_add ( 750, (GSourceFunc) idleToggleCloseButton, self ); 
    idleToggleCloseButton (self);
}

static gboolean
onPointerLeave (ClutterActor *gobject, ClutterEvent *event, gpointer user_data )
{
    CdosWindowOverlay *self = CDOS_WINDOW_OVERLAY(user_data);

    //if (!cdos_pointer_tracker_has_moved (self->pointerTracker) )
    //    return;

    self->hovering = FALSE;
    idleHideCloseButton (self);
    return FALSE;
}

static void
onDestroy (ClutterActor *gobject , gpointer user_data)
{
    CdosWindowOverlay *self = CDOS_WINDOW_OVERLAY(user_data);

    if (self->disconnectWindowAdded)
        self->disconnectWindowAdded (self);
    
    if (self->idleToggleCloseId > 0){
        g_source_remove ( self->idleToggleCloseId );
        self->idleToggleCloseId = 0;
    }

    self->disconnectAttentionSignals (self);

    g_signal_handler_disconnect ( self->windowClone->metaWindow, self->updateCaptionId );    

    clutter_actor_destroy ((ClutterActor *)self->title);
    clutter_actor_destroy ((ClutterActor *)self->closeButton);
    clutter_actor_destroy (self->icon);
    clutter_actor_destroy ((ClutterActor *)self->applicationIconBox);

    g_free (self->pointerTracker);
}

static void
onWindowDemandsAttention (MetaDisplay *display, MetaWindow *metaWindow, gpointer user_data)
{
    CdosWindowOverlay *self = CDOS_WINDOW_OVERLAY(user_data);

    if ( metaWindow != self->windowClone->metaWindow ) 
        return;

    if (!st_widget_has_style_class_name ( ST_WIDGET(self->title), DEMANDS_ATTENTION_CLASS_NAME ))
    {
        st_widget_add_style_class_name ( ST_WIDGET(self->title), DEMANDS_ATTENTION_CLASS_NAME );
        st_widget_add_style_class_name ( ST_WIDGET(self->applicationIconBox), DEMANDS_ATTENTION_CLASS_NAME );
    }
}

static gboolean
onIdleFunc (gpointer user_data)
{
    CdosWindowOverlay *self = CDOS_WINDOW_OVERLAY(user_data);

    g_signal_emit_by_name ( self->windowClone, "activated" );
    
    return FALSE;
}

static void
onWindowAdded (GObject *gobject, MetaWindow *win, gpointer user_data)
{
    CdosWindowOverlay *self = CDOS_WINDOW_OVERLAY(user_data);

    if ( self->disconnectWindowAdded )
        self->disconnectWindowAdded (self);

    if ( meta_window_get_transient_for (win) == self->windowClone->metaWindow ) {
        // use an idle handler to avoid mapping problems -
        // see comment in Workspace._windowAdded
        g_idle_add ( onIdleFunc, self );
    }
}

static void 
cdos_window_overlay_class_init (CdosWindowOverlayClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);

    /**
     * CdosWindowOverlay::activated:
     * @overlay: #CdosWindowOverlay
     *
     *
     */
    cdos_window_overlay_signals [ACTIVATED] =
        g_signal_new ("activated",
                G_TYPE_FROM_CLASS (klass),
                G_SIGNAL_RUN_LAST,
                0,
                NULL, NULL,
                g_cclosure_marshal_VOID__VOID,
                G_TYPE_NONE, 0 );

    /**
     * CdosWindowOverlay::show-close-button:
     * @overlay: #CdosWindowOverlay
     *
     * 显示关闭按钮时发出该信号。
     *
     */
    cdos_window_overlay_signals [SHOW_CLOSE_BUTTON] =
        g_signal_new ("show-close-button",
                G_TYPE_FROM_CLASS (klass),
                G_SIGNAL_RUN_LAST,
                0,
                NULL, NULL,
                g_cclosure_marshal_VOID__VOID,
                G_TYPE_NONE, 0 );

}

static void
cdos_window_overlay_init (CdosWindowOverlay *self)
{
    self->title = NULL;
}

static void
cdos_window_overlay_post_init (CdosWindowOverlay *self, 
                               CdosWindowClone *windowClone,
                               ClutterActor *parentActor)
{

    MetaWindow *metaWindow = windowClone->metaWindow;
    MetaDisplay *display = cdos_global_get_display (global);

    self->windowClone = windowClone;
    self->parentActor = parentActor;
    self->hidden = FALSE;
    self->hovering = FALSE;
    self->isSelected = FALSE;
//clutter_group_new ();
    CdosWindowTracker *tracker = cdos_window_tracker_get_default ();
    CdosApp *app = cdos_window_tracker_get_window_app (tracker, metaWindow);

    ClutterActor *icon = NULL;
    if (app)
        icon = cdos_app_create_icon_texture ( app, WINDOWOVERLAY_ICON_SIZE );

    if (!icon) {

        icon = g_object_new (ST_TYPE_ICON, "icon-name", "application-default-icon",
                "icon-type", ST_ICON_FULLCOLOR,
                "icon-size", WINDOWOVERLAY_ICON_SIZE, NULL);
    }

    self->icon = icon;
    clutter_actor_set_width (icon, WINDOWOVERLAY_ICON_SIZE );
    clutter_actor_set_height (icon, WINDOWOVERLAY_ICON_SIZE );

    self->applicationIconBox = ST_BIN(st_bin_new ());
    st_widget_set_style_class_name ((StWidget *)self->applicationIconBox, "window-iconbox" );
    
    clutter_actor_set_opacity ( CLUTTER_ACTOR(self->applicationIconBox), 255 );        
    
    st_bin_set_child (self->applicationIconBox, icon);
    clutter_actor_add_child ( parentActor, CLUTTER_ACTOR(self->applicationIconBox) );

    StButton *button = ST_BUTTON( st_button_new () );
    st_widget_set_style_class_name ((StWidget *)button, "window-close" );
    self->closeButton = button;
        
    self->closeButton_overlap = 0;
    
    clutter_actor_hide ((ClutterActor *)button);
      
    clutter_actor_add_child ( parentActor, (ClutterActor *)button);

    g_signal_connect ( button , "style-changed", G_CALLBACK ( onStyleChanged ), self );
    g_signal_connect ( button , "clicked", G_CALLBACK ( cdos_window_overlay_close_window ), self );

    cdos_window_overlay_refresh_title ( self, meta_window_get_title (metaWindow) );    

    self->updateCaptionId = g_signal_connect ( metaWindow, "notify::title", G_CALLBACK ( onNotifyTitle ), self );

    self->pointerTracker = cdos_pointer_tracker_new ();

    g_signal_connect (windowClone, "motion-event", G_CALLBACK ( onPointerMotion ), self );
    g_signal_connect (windowClone, "leave-event", G_CALLBACK ( onPointerLeave ), self );

    self->idleToggleCloseId = 0;

    g_signal_connect (windowClone, "destroy", G_CALLBACK (onDestroy), self );

    self->attentionId = g_signal_connect ( display, "window-demands-attention", G_CALLBACK ( onWindowDemandsAttention ), self );

    self->urgentId = g_signal_connect ( display, "window-marked-urgent", G_CALLBACK ( onWindowDemandsAttention ), self ); 

    self->disconnectAttentionSignals = cdos_window_overlay_disconnect_attention_signals;
        
        // force a style change if we are already on a stage - otherwise
        // the signal will be emitted normally when we are added
    if ( clutter_actor_get_stage ( parentActor ) ) {
        onStyleChanged ( NULL, self );  //unknow
    }
}

/**
 * cdos_window_overlay_new:
 * @windowClone: 对 #CdosWindowClone 创建一个 #CdosWindowOverlay.
 * @parentActor: #CdosWorkspaceMonitor 的 #windowOverlaysGroup (#ClutterActor).
 *
 * 创建一个 #CdosWindowOverlay.
 *
 * Returns: a #CdosWindowOverlay.
 */
CdosWindowOverlay *
cdos_window_overlay_new (CdosWindowClone *windowClone, ClutterActor *parentActor)
{
    CdosWindowOverlay *instance = g_object_new ( CDOS_TYPE_WINDOW_OVERLAY, NULL );

    cdos_window_overlay_post_init ( instance, windowClone, parentActor );

    return instance;
}

static void
cdos_window_overlay_refresh_title (CdosWindowOverlay *self, const gchar *titleText)
{
    gchar *name = NULL;
    MetaWindow *mw;

    if (self->title) {
        name = g_strdup ( clutter_actor_get_name ( CLUTTER_ACTOR(self->title) ) );
        //clutter_container_remove_actor ( CLUTTER_CONTAINER(self->parentActor), CLUTTER_ACTOR(self->title) );
        //clutter_actor_remove_child ( self->parentActor, CLUTTER_ACTOR(self->title) );
        //clutter_actor_destroy ( CLUTTER_ACTOR(self->title) );
        st_label_set_text ((StLabel *)self->title, titleText);
    }
    else {
        self->title = st_label_new (titleText);
        clutter_actor_add_child ( self->parentActor, (ClutterActor *)self->title);
    }
    //StWidget *title = st_label_new ( titleText );
    st_widget_set_style_class_name (self->title, "window-caption" ) ;
    
    clutter_actor_set_name ( CLUTTER_ACTOR(self->title), name );

    clutter_text_set_ellipsize ((ClutterText *)st_label_get_clutter_text ((StLabel *)self->title),  PANGO_ELLIPSIZE_END );
    self->title_spacing = 0;


    mw = self->windowClone->metaWindow;

    if ( meta_window_is_above (mw))
    { 
        st_widget_add_style_class_name (self->title, DEMANDS_ATTENTION_CLASS_NAME );
        st_widget_add_style_class_name (ST_WIDGET(self->applicationIconBox), DEMANDS_ATTENTION_CLASS_NAME );

    }
    
    g_signal_connect ( self->title, "style-changed", G_CALLBACK ( onStyleChanged ), self );

    if ( clutter_actor_get_stage ( self->parentActor ) )
        onStyleChanged (self->title, self );
}

/**
 * cdos_window_overlay_set_selected:
 * @self: #CdosWindowOverlay
 * @selected: 是否被选中。
 *
 * 如果 #CdosWindowClone 被选中，则设置与它对应的 #CdosWindowOverlay 的 selected 为 %TRUE.
 */
void
cdos_window_overlay_set_selected (CdosWindowOverlay *self, gboolean selected)
{
    if (self->isSelected == selected)
                return;

    self->isSelected = selected;

    //g_strdup ( clutter_actor_get_name ( CLUTTER_ACTOR(self->title) ) );

    if ( selected )
        clutter_actor_set_name ( CLUTTER_ACTOR(self->title), "selected" );
    else
        clutter_actor_set_name ( CLUTTER_ACTOR(self->title), "" );

    cdos_window_overlay_refresh_title ( self, st_label_get_text ((StLabel *)self->title) );

    if (selected)
        cdos_window_overlay_show_close_button (self);
    else 
        cdos_window_overlay_hide_close_button (self);
}

/**
 * cdos_window_overlay_hide:
 * @clone: #CdosWindowClone.
 * @self: #CdosWindowOverlay.
 *
 * 隐藏 #CdosWindowClone 的 #CdosWindowOverlay.
 */
void
cdos_window_overlay_hide (CdosWindowClone *clone, CdosWindowOverlay *self)
{
    self->hidden = TRUE;
    clutter_actor_hide ((ClutterActor *)self->closeButton);
    clutter_actor_hide ((ClutterActor *)self->title);
    clutter_actor_hide ((ClutterActor *)self->applicationIconBox);
}

/**
 * cdos_window_overlay_show:
 * @clone: #CdosWindowClone.
 * @self: #CdosWindowOverlay.
 *
 * 显示 #CdosWindowClone 的 #CdosWindowOverlay.
 */
void
cdos_window_overlay_show (CdosWindowClone *clone, CdosWindowOverlay *self)
{
    self->hidden = FALSE;
    self->hovering = FALSE;
    clutter_actor_show ((ClutterActor *)self->title);
    clutter_actor_show ((ClutterActor *)self->applicationIconBox);
}

/* *
 * cdos_window_overlay_fadein:
 * @self: #CdosWindowOverlay.
 *
 * #CdosWindowOverlay 渐入。
 */
void
cdos_window_overlay_fadein (CdosWindowOverlay *self)
{
    cdos_window_overlay_show (NULL, self);
    ClutterActor *title = (ClutterActor *)self->title;
    clutter_actor_set_opacity (title, 0);

    ClutterActor *parent = clutter_actor_get_parent (self->parentActor);
    /*
    if (ST_IS_CONTAINER (parent))
        st_container_raise ((ClutterContainer *)parent, self->parentActor, NULL);
    else
    */
    clutter_actor_set_child_above_sibling (parent, self->parentActor, NULL);
    clutter_actor_remove_transition (title, "wo-fadein");
    cdos_transition_to (title, CLUTTER_EASE_OUT_QUAD, CLOSE_BUTTON_FADE_TIME,
            0,
            NULL, NULL,
            "wo-fadein",
            "opacity", TRANS_PROP_UCHAR, 255,
            NULL);
}


gfloat
cdos_window_overlay_chrome_width (CdosWindowOverlay *self)
{
    return clutter_actor_get_width ( CLUTTER_ACTOR(self->closeButton) ) - self->closeButton_overlap;
}


void
cdos_window_overlay_chrome_heights (CdosWindowOverlay *self, gfloat *h_overlap , gfloat *h_space)
{
    *h_overlap = clutter_actor_get_height ((ClutterActor *)self->closeButton ) - self->closeButton_overlap;
    *h_space = clutter_actor_get_height ((ClutterActor *)self->applicationIconBox) - self->title_spacing;
}

static void
cdos_window_overlay_update_icon_caption_width ( CdosWindowOverlay *self, 
                                                gfloat cloneX, 
                                                gfloat cloneY, 
                                                gfloat cloneWidth, 
                                                gfloat cloneHeight, 
                                                gfloat maxWidth )
{
    ClutterActor *title = (ClutterActor *)self->title;
    StBin *icon;
    gfloat titleX, titleY;
    gfloat iconX, iconY;
    gfloat iconWidth; 

    iconWidth = clutter_actor_get_width ((ClutterActor *)self->applicationIconBox) + self->title_spacing;

    gfloat twidth = fmin ( maxWidth - iconWidth , clutter_actor_get_width (title));
    clutter_actor_set_width (title, twidth);

    titleX = cloneX + ( iconWidth + cloneWidth - twidth ) / 2;

    titleY = cloneY + cloneHeight + self->title_spacing + (WINDOWOVERLAY_ICON_SIZE/2) - ( clutter_actor_get_height (title)/2 ) ;

    clutter_actor_set_position (title, floor(titleX), floor(titleY) );

    icon = self->applicationIconBox;

    iconX = titleX - iconWidth;
    iconY = cloneY + cloneHeight + self->title_spacing;

    clutter_actor_set_position ((ClutterActor *)icon, floor(iconX), floor(iconY) );
}

/**
 * cdos_window_overlay_update_positions:
 * @self: #CdosWindowOverlay.
 * @cloneX: #CdosWindowClone 的 x
 * @cloneY: #CdosWindowClone 的 y
 * @cloneWidth: #CdosWindowClone 的 width
 * @cloneHeight: #CdosWindowClone 的 height
 * @maxWidth: 最大宽度。
 *
 * 设置 #CdosWindowOverlay 的位置。
 */
void
cdos_window_overlay_update_positions ( CdosWindowOverlay *self, 
                        gfloat cloneX, 
                        gfloat cloneY, 
                        gfloat cloneWidth, 
                        gfloat cloneHeight, 
                        gfloat maxWidth )
{


    StButton *button = self->closeButton;
    StSide side = ST_SIDE_RIGHT;
    gfloat buttonX, buttonY;
    buttonY = cloneY - ( clutter_actor_get_height ((ClutterActor *)button) - self->closeButton_overlap );

    if (side == ST_SIDE_LEFT)
        buttonX = cloneX - ( clutter_actor_get_width ((ClutterActor *)button) - self->closeButton_overlap );
    else
        buttonX = cloneX + ( cloneWidth - self->closeButton_overlap );


    clutter_actor_set_position ((ClutterActor *)button, buttonX, buttonY);

    cdos_window_overlay_update_icon_caption_width ( self, cloneX, cloneY, cloneWidth, cloneHeight, maxWidth );
}


static void
cdos_window_overlay_close_window (ClutterActor *button,
                                  ClutterEvent *e,
                                  CdosWindowOverlay *self)
{
    
    MetaWindow *metaWindow = self->windowClone->metaWindow;
    MetaWorkspace *workspace = meta_window_get_workspace (metaWindow);
    //gint windowAddedId;

    if ( self->disconnectWindowAdded ) { 
        self->disconnectWindowAdded (self);
    }
    
    self->windowAddedId = g_signal_connect ( workspace, "window-added", G_CALLBACK ( onWindowAdded ), self );

    self->disconnectWindowAdded = cdos_window_overlay_disconnect_window_added;    

    meta_window_delete ( metaWindow, cdos_global_get_current_time ( cdos_global_get() ) );
}

/**
 * cdos_window_overlay_show_close_button:
 * @self: #CdosWindowOverlay.
 *
 * 显示 #CdosWindowOverlay 的关闭按钮。
 */
void cdos_window_overlay_show_close_button ( CdosWindowOverlay *self ) {

    ClutterActor *parent = clutter_actor_get_parent ( self->parentActor );
    /*
    if (CLUTTER_IS_GROUP (parent) || ST_IS_CONTAINER (parent))
        clutter_container_raise_child (parent, self->parentActor, NULL);
    else
    */
        clutter_actor_set_child_above_sibling (parent , self->parentActor, NULL );

    clutter_actor_show ((ClutterActor *)self->closeButton);
    g_signal_emit (self, cdos_window_overlay_signals[SHOW_CLOSE_BUTTON], 0);
}

/**
 * cdos_window_overlay_hide_close_button:
 * @self: #CdosWindowOverlay.
 *
 * 隐藏 #CdosWindowOverlay 的关闭按钮。
 */
void cdos_window_overlay_hide_close_button ( CdosWindowOverlay *self )
{
    if ( self->idleToggleCloseId > 0 ) {
        g_source_remove ( self->idleToggleCloseId );
        self->idleToggleCloseId = 0;
    }
    clutter_actor_hide ( CLUTTER_ACTOR(self->closeButton) );
    self->hovering = FALSE;
}


static void 
cdos_window_overlay_disconnect_attention_signals ( CdosWindowOverlay *self ) {

    MetaDisplay *display = cdos_global_get_display (global);

    g_signal_handler_disconnect ( display, self->attentionId );
    g_signal_handler_disconnect ( display, self->urgentId );
}


static void
cdos_window_overlay_disconnect_window_added ( CdosWindowOverlay *self ) {

    MetaWindow *metaWindow = self->windowClone->metaWindow;
    MetaWorkspace *workspace = meta_window_get_workspace (metaWindow);

    g_signal_handler_disconnect ( workspace, self->windowAddedId );

    self->disconnectWindowAdded = NULL;
}

/**
 * cdos_pointer_tracker_new:
 *
 * 记录光标位置，用于判断两次光标位置是否相同。
 *
 * Returns: a #CdosPointerTracker
 */
CdosPointerTracker *
cdos_pointer_tracker_new ()
{
    CdosPointerTracker *self = g_new0 (CdosPointerTracker, 1);
    MetaScreen *screen = cdos_global_get_screen (global);
    self->pointer = meta_cursor_tracker_get_for_screen (screen);
    meta_cursor_tracker_get_pointer (self->pointer, &self->lastPointerX, &self->lastPointerY, NULL);
    return self;
}

/**
 * cdos_pointer_tracker_has_moved:
 * @self: #CdosPointerTracker
 *
 * 判断光标相对上次位置是否已经移动
 *
 * Returns: %TRUE / %FALSE
 *
 */
gboolean
cdos_pointer_tracker_has_moved (CdosPointerTracker *self)
{
    gint pointerX, pointerY;
    gboolean has_moved = FALSE;

    meta_cursor_tracker_get_pointer (self->pointer, &pointerX, &pointerY, NULL);

    if (pointerX != self->lastPointerX || pointerY != self->lastPointerY )
        has_moved = TRUE;
    
    self->lastPointerX = pointerX;
    self->lastPointerY = pointerY; 

    return has_moved;
}

