#include <stdlib.h>
#include "cdos-workspace.h"
#include "cdos-workspaces-view.h"
#include "cdos-window-overlay.h"
#include "cdos-overview.h"
//#include "cdos-export-overview.h"
#include "cdos-overview-workspace-thumbnail.h"
/**
 * SECTION:cdos-workspaces-view
 * @Title: CdosWorkspacesView
 * @Short_description: 工作区视图
 * 
 * CdosWorkspacesView 的成员 actor 保存所有 workspace，cdos 桌面环境默认有四个工作区，不提供增加删除功能，
 * 作为 CdosOverview 和 CdosWorkspace 之间的桥梁， CdosWorkspacesView 在初始化时会先获取当前所有的工作
 * 区，保存到链表中，并隐藏非 active 的工作区。
 */

G_DEFINE_TYPE (CdosWorkspacesView, cdos_workspaces_view, ST_TYPE_GROUP);

/* 下面这个宏可以用来访问新建对象的私有结构*/
#define CDOS_WORKSPACES_VIEW_GET_PRIVATE(obj) \
    (G_TYPE_INSTANCE_GET_PRIVATE ((obj), CDOS_TYPE_WORKSPACES_VIEW, CdosWorkspacesViewPrivate))

struct _CdosWorkspacesViewPrivate
{
    guint width;
    guint height;
    guint x;
    guint y;
    guint workspaceRatioSpacing;
    guint spacing;

    gboolean animating;
    gboolean animatingScroll;

    GList *workspaces;
    
    gulong nWorkspacesChangedId;
    gulong overviewShowingId;
    gulong switchWorkspaceNotifyId;
};

static void
cdos_workspaces_view_update_visibility (CdosWorkspacesView *self)
{
    CdosWorkspacesViewPrivate *priv = self->priv;

    MetaScreen *screen = cdos_global_get_screen (global);
    MetaWorkspace *active = meta_screen_get_active_workspace (screen);
    GList *workspaces = priv->workspaces;
    GList *iter;

    for (iter = workspaces; iter; iter = iter->next)
    {
        CdosWorkspace *workspace = iter->data;

        if (priv->animating){
            cdos_workspace_hide_windows_overlays (workspace);
            clutter_actor_show ((ClutterActor *)workspace);
        }else {
            cdos_workspace_show_windows_overlays (workspace);
            if (active == workspace->metaWorkspace)
                clutter_actor_show ((ClutterActor *)workspace);
            else
                clutter_actor_hide ((ClutterActor *)workspace);
        }
    }  
}

static void
on_workspaces_changed (MetaScreen *screen, GParamSpec *spec, gpointer data){
    CdosWorkspacesView *self = (CdosWorkspacesView *)data;
    CdosWorkspacesViewPrivate *priv = self->priv;

    guint removedCount = 0;
    GList *workspaces = priv->workspaces;
    GList *recorder = priv->workspaces;
    guint i = 0;

    while (workspaces != NULL){
        GList *next = workspaces->next;
        CdosWorkspace *workspace = workspaces->data;
        MetaWorkspace *mworkspace = meta_screen_get_workspace_by_index (screen, i - removedCount);

        if (workspace->metaWorkspace != mworkspace ){
            // remove tween from cdos workspace's actor
            recorder = g_list_remove (recorder, workspace);
            clutter_actor_destroy ((ClutterActor *)workspace);
            ++removedCount;
        }
        ++i;
        workspaces = next;
    }
    g_list_free (priv->workspaces);
    priv->workspaces = NULL;

    GList *wks = meta_screen_get_workspaces (screen);
    GList *iter;

    for (iter = wks; iter; iter = iter->next)
    {
        MetaWorkspace *lastWs = iter->data;
        CdosWorkspace *newWs= cdos_workspace_new (lastWs, self);
        priv->workspaces = g_list_append (priv->workspaces, newWs);
        st_container_add_actor ((ClutterContainer *)self, (ClutterActor *)newWs);
    }

    priv->animating = FALSE;
    cdos_workspaces_view_update_visibility (self);
}

static gboolean
cdos_stage_key_press_event (ClutterActor *actor, ClutterKeyEvent *ek)
{
    CdosWorkspacesView *self = (CdosWorkspacesView *)actor;
    CdosWorkspacesViewPrivate *priv = self->priv;

    ClutterEvent *event = (ClutterEvent *)ek;
    //ClutterModifierType modifiers = clutter_event_get_state (event);

    guint symbol = clutter_event_get_key_symbol (event);

    if (symbol == CLUTTER_KEY_Escape ){
        // FIXME overview.hide
        CdosOverview *overview = cdos_overview_get ();
        //cdos_overview_hide (overview);
        clutter_actor_hide ((ClutterActor *)overview);
        return TRUE;
    }

    MetaScreen *screen = cdos_global_get_screen (global);
    guint activeWorkspaceIndex = meta_screen_get_active_workspace_index (screen);
    CdosWorkspace *activeWorkspace = g_list_nth_data (priv->workspaces, activeWorkspaceIndex);
    return cdos_workspace_on_key_press (activeWorkspace, actor, event);
}

static void
cdos_workspaces_view_set_geometry (CdosWorkspacesView *self, guint x, guint y,
                                                guint width, guint height, guint spacing){
    CdosWorkspacesViewPrivate *priv = self->priv;
    priv->width = width;
    priv->height = height;
    priv->x = x;
    priv->y = y;
    priv->workspaceRatioSpacing = spacing;
}

static void
cdos_workspaces_view_hide (ClutterActor *actor){
    CdosWorkspacesView *self = (gpointer)actor;
    CdosWorkspacesViewPrivate *priv = self->priv;

    MetaScreen *screen = cdos_global_get_screen (global);
    guint activeWorkspaceIndex = meta_screen_get_active_workspace_index (screen);
    CdosWorkspace *activeWorkspace = g_list_nth_data (priv->workspaces, activeWorkspaceIndex);

    ClutterActor *parent = clutter_actor_get_parent ( CLUTTER_ACTOR (activeWorkspace));
    st_container_raise ((ClutterContainer *)parent, (ClutterActor *)activeWorkspace, NULL);
    cdos_workspace_zoom_from_overview (activeWorkspace);
    //CLUTTER_ACTOR_CLASS (cdos_workspaces_view_parent_class)->hide (actor);
}

static void
on_workspace_transitions_completed (ClutterTimeline *tml, CdosWorkspacesView *self)
{
    self->priv->animating = FALSE;
    cdos_workspaces_view_update_visibility (self);
}

static void
cdos_workspaces_view_update_workspace_actors (CdosWorkspacesView *self, gboolean animate)
{
    CdosWorkspacesViewPrivate *priv = self->priv;

    MetaScreen *screen = cdos_global_get_screen (global);
    guint active = meta_screen_get_active_workspace_index (screen);

    CdosLayoutManager *layout = cdos_monkey_plugin_get_layout_manager (monkey_plugin_get ());
    GPtrArray *monitors = cdos_layout_manager_get_monitors (layout);
    MetaRectangle *primary = cdos_layout_manager_get_primary_monitor (layout);
    gboolean animating = priv->animating;
    priv->animating = (monitors->len < 2) && animate;

    guint i = 0;
    GList *workspaces = priv->workspaces;
    GList *iter;

    for (iter = workspaces; iter; iter = iter->next, i++)
    {
        CdosWorkspace *workspace = iter->data;
        
        gfloat x = (i-active) * (priv->width + priv->spacing + priv->workspaceRatioSpacing);
        gfloat w = primary->width * thumbnail_box_scale / 2.0;
        x += w;

        if (!animating && priv->animating)
        {
            clutter_actor_show ((ClutterActor *)workspace);
            ClutterTransition *trans = cdos_transition_to ((ClutterActor *)workspace,
                                CLUTTER_EASE_IN_ELASTIC,
                                300,
                                0,
                                NULL, NULL,
                                "update-workspace",
                                "x", TRANS_PROP_FLOAT, x,
                                NULL);
            if (i == 0){
                g_signal_connect (trans, "completed", 
                        G_CALLBACK (on_workspace_transitions_completed), self);
            }
        }else {
            clutter_actor_set_position ((ClutterActor *)workspace, x, 0);
            if ( i == 0)
                cdos_workspaces_view_update_visibility (self);
        }

    }
}

static void
on_switch_workspace (MetaScreen *screen,
                     gint from,
                     gint to,
                     gint direction,
                     CdosWorkspacesView *self)
{
    cdos_workspaces_view_update_workspace_actors (self, TRUE);
}

/*
static void
cdos_workspaces_view_map (ClutterActor *actor)
{
    CLUTTER_ACTOR_CLASS (cdos_workspaces_view_parent_class)->map (actor);

    CdosWorkspacesView *self = (CdosWorkspacesView *)actor;
    CdosWorkspacesViewPrivate *priv = CDOS_WORKSPACES_VIEW_GET_PRIVATE (self);

    gboolean ismapped;
    ismapped = clutter_actor_is_mapped (actor);

    CdosOverview *overview = cdos_overview_get ();
    if (ismapped){
        guint direction = SWIPE_DIRECTION_HORIZONTAL;
        cdos_overview_set_scroll_adjustment (overview, priv->scrollAdjustment, direction);

        priv->swipeScrollBeginId = g_signal_connect (overview, "swipe-scroll-begin",
                                        G_CALLBACK (on_swipe_scroll_begin), self); 

        priv->swipeScrollEndId = g_signal_connect (overview, "swipe-scroll-end",
                                        G_CALLBACK (on_swipe_scroll_end), self);
    }else {
        g_signal_handler_disconnect ( overview, priv->swipeScrollBeginId);
        g_signal_handler_disconnect ( overview, priv->swipeScrollEndId);
    } 
}
*/

static void
cdos_workspaces_view_on_destroy (CdosWorkspacesView *self)
{
    CdosWorkspacesViewPrivate *priv = self->priv;
    if (priv->switchWorkspaceNotifyId)
    {
        CdosBaseWM *window_manager = cdos_global_get_window_manager (global);
        g_signal_handler_disconnect (window_manager, priv->switchWorkspaceNotifyId);
        priv->switchWorkspaceNotifyId = 0L;
    }
    if (priv->nWorkspacesChangedId)
    {
        MetaScreen *screen = cdos_global_get_screen (global);
        g_signal_handler_disconnect (screen, priv->nWorkspacesChangedId);
        priv->nWorkspacesChangedId = 0L;
    }
    if (priv->overviewShowingId)
    {
        CdosOverview *overview = cdos_overview_get ();
        g_signal_handler_disconnect (overview, priv->overviewShowingId);
        priv->overviewShowingId = 0L;
    }
}

static void
cdos_workspaces_view_dispose (GObject *gobject)
{
    CdosWorkspacesView *self = CDOS_WORKSPACES_VIEW (gobject);

    cdos_workspaces_view_on_destroy (self);

    G_OBJECT_CLASS (cdos_workspaces_view_parent_class)->dispose (gobject);
}

static void
cdos_workspaces_view_finalize (GObject *gobject)
{
    CdosWorkspacesView *self = CDOS_WORKSPACES_VIEW (gobject);
    CdosWorkspacesViewPrivate *priv = self->priv;
    g_list_free (priv->workspaces);
    priv->workspaces = NULL;
}

static void
on_overview_showing (CdosOverview *overview, gpointer data)
{
    CdosWorkspacesView *self = (CdosWorkspacesView *)data;
    CdosWorkspacesViewPrivate *priv = CDOS_WORKSPACES_VIEW_GET_PRIVATE (self);
//    CdosOverview *overview = cdos_overview_get ();
    MetaScreen *screen = cdos_global_get_screen (global);

    g_signal_handler_disconnect (overview, priv->overviewShowingId);
    priv->overviewShowingId = 0L;
    guint activeWorkspaceIndex = meta_screen_get_active_workspace_index (screen);

    CdosWorkspace *current_workspace = g_list_nth_data (priv->workspaces, activeWorkspaceIndex);
    cdos_workspace_zoom_to_overview (current_workspace);
}

static void
cdos_workspaces_view_init (CdosWorkspacesView *self){
    CdosWorkspacesViewPrivate *priv = CDOS_WORKSPACES_VIEW_GET_PRIVATE (self);
    self->priv = priv;
    MetaScreen *screen = cdos_global_get_screen (global);

    priv->width = 0;
    priv->height =0;
    priv->x = 0;
    priv->y = 0;

    priv->workspaceRatioSpacing = 0;
    priv->spacing = 0;
    priv->animating = FALSE;
    priv->animatingScroll = FALSE;

    CdosOverview *overview = cdos_overview_get ();
    priv->overviewShowingId = g_signal_connect (overview, "showing", G_CALLBACK (on_overview_showing), self);

    CdosBaseWM *wm = cdos_global_get_window_manager (global);
    priv->switchWorkspaceNotifyId = g_signal_connect (wm, "switch-workspace", G_CALLBACK (on_switch_workspace), self);

    priv->nWorkspacesChangedId = g_signal_connect (screen, "notify::n-workspaces",
                                    G_CALLBACK (on_workspaces_changed), self);
}

static void
cdos_workspaces_view_class_init (CdosWorkspacesViewClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    ClutterActorClass *actor_class = CLUTTER_ACTOR_CLASS (klass);

    gobject_class->dispose = cdos_workspaces_view_dispose;
    gobject_class->finalize = cdos_workspaces_view_finalize;

    //actor_class->map = cdos_workspaces_view_map;
    actor_class->key_press_event = cdos_stage_key_press_event;
    actor_class->hide = cdos_workspaces_view_hide;

    g_type_class_add_private (klass, sizeof (CdosWorkspacesViewPrivate));
}

/**
 * cdos_workspaces_view_new:
 * 
 * 新建 #CdosWorkspacesView 对象
 *
 * Return value: 工作区视图对象
 */
CdosWorkspacesView *
cdos_workspaces_view_new ()
{
    CdosWorkspacesView *self;
    self = g_object_new (CDOS_TYPE_WORKSPACES_VIEW, 
                         "style-class", "workspaces-view",
                         NULL);

    CdosWorkspacesViewPrivate *priv = CDOS_WORKSPACES_VIEW_GET_PRIVATE (self);

    CdosLayoutManager *layout = cdos_layout_manager_get_default ();
    MetaRectangle *primary = cdos_layout_manager_get_primary_monitor (layout);
    cdos_workspaces_view_set_geometry (self, primary->x, primary->y, primary->width, 
                                                                primary->height, priv->spacing);

    MetaScreen *screen = cdos_global_get_screen (global);

    MetaWorkspace *active = meta_screen_get_active_workspace (screen);
    GList *workspaces = meta_screen_get_workspaces (screen);
    GList *iter;
    gint active_index = g_list_index (workspaces, active);
    gint i = 0;
    CdosWorkspace *cactive = NULL;
    for (iter = workspaces; iter; iter = iter->next, i++)
    {
        MetaWorkspace *workspace = iter->data;
        CdosWorkspace *cworkspace = cdos_workspace_new (workspace, self);
        priv->workspaces = g_list_append (priv->workspaces, cworkspace);
        st_container_add_actor ((ClutterContainer *)self, (ClutterActor *)cworkspace);
        gfloat x = (i-active_index) * (priv->width + priv->spacing + priv->workspaceRatioSpacing);
        gfloat w = primary->width * thumbnail_box_scale / 2.0;
        x += w;
        clutter_actor_set_position ((ClutterActor *)cworkspace, x, 0);
        if (workspace != active)
            clutter_actor_hide ((ClutterActor *)cworkspace);
        else {
            cactive = cworkspace;
        }
    }
    st_container_raise ((ClutterContainer *)self, (ClutterActor *)cactive, NULL);

    ClutterStage *stage = cdos_global_get_stage (global);
    clutter_stage_set_key_focus (stage, (ClutterActor *)self);

    return self;
}
